#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Classes and methods for laying the data location foundation for various types of CASA processing
"""
import json
import logging
import os
import re
import shutil
import sys
import tarfile
from glob import glob
from pathlib import Path
from typing import Optional

from casa_envoy import RESTORE_ORIG_CAL_PPR_NAME
from casa_envoy.interfaces import FoundationIF
from casa_envoy.schema import AbstractTextFile
from casa_envoy.utilities import get_abs_file, get_value_for, get_xml_content


def set_permissions(path: Path):
    """
    reusable function for setting file permissions

    :param path: path to set permissions on
    :return:
    """
    for root, dirs, files in os.walk(path):
        for d in dirs:
            os.chmod(os.path.join(root, d), 0o755)
        for f in files:
            os.chmod(os.path.join(root, f), 0o755)


class GeneralFoundation(FoundationIF):
    """
    Class to set up the general data foundation for anything running CASA.
    - Special attention to user provided file locations, must be placed in working directory.
    """

    def __init__(self, parameters: dict, metadata: AbstractTextFile):
        self.logger = logging.getLogger("casa_envoy")
        self.parent_path = parameters["parent_path"]
        self.rawdata_dir = self.parent_path / "rawdata"
        self.working_dir = self.parent_path / "working"
        self.content = json.loads(metadata.content)

    def data_foundation(self):
        """
        All user provided files should be placed in the 'working' directory inorder for CASA to pick them up.

        :return:
        """
        self.logger.info("Laying general data foundation...")

        if "providedFiles" in self.content:
            provided_files = self.content["providedFiles"].replace("[", "").replace("'", "").replace("]", "").split(",")

            # move provided files from parent directory into working directory for CASA
            for file in provided_files:
                stripped_file = file.strip()
                if stripped_file:
                    shutil.copy(stripped_file, self.working_dir)

        self.copy_into_rawdata()

        self.logger.info("General data foundation complete!")

    def copy_into_rawdata(self):
        """
        CASA expects some files from the 'working' directory to also exist next to the data in the 'rawdata' directory.

        :return:
        """
        # Copy files that need to exist next to the data into the rawdata directory
        for file in Path(self.working_dir).glob("*flagtargetstemplate.txt"):
            if Path(file).exists():
                shutil.copy(file, self.rawdata_dir)


class RestoreFoundation(FoundationIF):
    """
    Class to set up the data foundation for restores
    """

    def __init__(self, parameters: dict, metadata: AbstractTextFile):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters
        self.parent_path = parameters["parent_path"]
        self.content = json.loads(metadata.content)
        self._piperestorescript_path: Optional[Path] = None

    def data_foundation(self):
        """
        CMS Restore requires two inputs: An EB and a Calibration for that EB
        After download, all data is in rawdata and the calibrations tables are contained
        in a tar file. We need to extract all the calibration tables to the /products directory
        for CASA processing.

        :return:
        """

        self.logger.info("Laying restore data foundation...")
        # ensure we're starting from the parent directory
        os.chdir(self.parent_path)
        self.extract_cal()
        path = Path("./products")
        self.logger.info("Ensuring correct file permissions....")
        set_permissions(path)
        self.find_piperestore(path)

        self.logger.info("Restore data foundation complete!")

    @property
    def piperestorescript_path(self) -> Optional[Path]:
        """Absolute path to the casa_piperestorescript.py for this restore, or None if one doesn't exist"""
        return self._piperestorescript_path

    def is_alma(self) -> bool:
        """Returns true if the telescope in the metadata is ALMA"""
        return (
            "projectMetadata" in self.content
            and "telescope" in self.content["projectMetadata"]
            and self.content["projectMetadata"]["telescope"]
            and self.content["projectMetadata"]["telescope"].lower() == "alma"
        )

    def extract_cal(self):
        self.logger.info("Extracting calibration tar file to products directory...")
        cal_name = self.content["fileSetIds"][1]
        # ALMA calibrations do not need extracted
        cal_path = "./rawdata/" + cal_name
        if not self.is_alma():
            cal_path += "/" + cal_name
        if Path(cal_path).exists():
            chosen_src = None
            if self.is_alma():
                chosen_src = cal_path
            else:
                calibration = tarfile.open(cal_path)
                # extract to products
                calibration.extractall("./rawdata")
                calibration.close()
                src = "./rawdata/products"
                src2: str = "./rawdata/" + cal_name[:-4] + "/products"
                if Path(src).exists():
                    chosen_src = src
                elif Path(src2).exists():
                    chosen_src = src2

                cal_weblog_path = Path(chosen_src + "/weblog.tgz")
                # If there is a calibration weblog present in chosen_src, extract it, will be delivered to user
                if cal_weblog_path.exists():
                    cal_weblog = tarfile.open(cal_weblog_path)
                    cal_weblog.extractall(chosen_src)

            if not chosen_src:
                self.logger.error(f"ERROR: calibration tar file {cal_name} failed to extract correctly!")
                sys.exit(1)
            shutil.copytree(chosen_src, "./products", dirs_exist_ok=True)
            # Search for a cal PPR or a PPR produced by CASA
            cal_ppr = Path("./products/PPR.xml")
            pprequests = glob("./products/*pprequest.xml")
            if not cal_ppr.exists() and len(pprequests) == 1:
                cal_ppr = Path(pprequests[0])
            if cal_ppr.exists():
                # Rename the PPR.xml so that it's not clobbered by the restore's one in auditor.correct_for_condor
                cal_ppr.rename(f"./products/{RESTORE_ORIG_CAL_PPR_NAME}")
        else:
            self.logger.error(f"ERROR: calibration tar file {cal_name} not found in rawdata!")
            # abort since required data not found
            sys.exit(1)

    def edit_piperestore(self, path: Path):
        """
        Ensures the given piperestore script is compatible with the desired CASA version.

        :param path: Path object to a piperestore script
        :return:
        """
        # Grab the first digit in the user specified CASA home or that in CAPO
        if self.content.get("casaHome") is not None:
            casa_home = Path(self.content["casaHome"])
        else:
            casa_home = Path(self.parameters["homeForReprocessing"])
        match = re.search(r"\d+", casa_home.name)
        if match is not None:
            casa_home_major = int(match.group(0))
        else:
            self.logger.info("CASA home not specified or defined in parameters, using piperestore as is")
            return

        # Correct import statements to work with CASA > 5
        recipes = "from recipes"
        casarecipes = "from casarecipes"
        # Correct recipes made with CASA < 5
        hif_deprecated = "hif_restoredata"
        hif_corrected = "hifa_restoredata" if self.is_alma() else "hifv_restoredata"
        with open(path, "r") as f:
            content = f.read()
            content = (
                content.replace(recipes, casarecipes) if casa_home_major > 5 else content.replace(casarecipes, recipes)
            )
            content = (
                content.replace(hif_deprecated, hif_corrected)
                if casa_home_major >= 5
                else content.replace(hif_corrected, hif_deprecated)
            )
        with open(path, "w") as f:
            f.write(content)
        self.logger.info(f"Finished editing piperestore to work with CASA {casa_home_major}")

    def find_piperestore(self, path: Path):
        """
        Searches for a piperestore script in the given path and sets the script's path in self._piperestorescript_path.

        :param path: Path object used to find the piperestore script
        :return:
        """
        # Find *casa_piperestorescript.py and copy it to the working directory
        CASAPIPERESTORESCRIPT = "casa_piperestorescript.py"
        self.logger.info(f"Checking for {CASAPIPERESTORESCRIPT}...")
        for file in path.glob(f"*{CASAPIPERESTORESCRIPT}"):
            if Path(file).exists():
                dest = Path(f"working/{file.name}")
                shutil.copyfile(file, dest)
                self.edit_piperestore(dest)
                self._piperestorescript_path = dest.resolve()
                self.logger.info(f"Found {file.name}, copied it to {dest}")
        if not self._piperestorescript_path:
            self.logger.info(f"Unable to locate a piperestore script, calling CASA with the PPR")


class CubeFoundation(FoundationIF):
    """
    Class to set up the data foundation for VLASS Cubes
    """

    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters
        self.parent_path = self.parameters["parent_path"]
        self.working_dir = self.parent_path / "working"
        self.products_dir = self.parent_path / "products"

    def data_foundation(self):
        """
        VLASS Cubes require two inputs: the reimaging resources of an SECI and the associated SEIP_parameters.list
        After download, all data is in rawdata and the reimaging resources are contained
        in a tar file. We need to extract all the reimaging resources from the tar and move the parameters.list file
        to the working directory for CASA processing.

        :return:
        """
        self.logger.info("Laying cube data foundation...")
        os.chdir(self.parent_path)
        # self.extract_reimaging_resources()
        self.move_ancillaries()

        self.logger.info("Ensuring correct file permissions....")
        set_permissions(self.working_dir)
        self.logger.info("Cube data foundation complete!")

    def extract_reimaging_resources(self):
        """
        Extract resources to working directory

        :return:
        """
        self.logger.info("Extracting reimaging resources to working directory...")
        resource_path = self.parent_path / "rawdata" / "reimaging_resources.tgz"
        if Path(resource_path).exists():
            resources = tarfile.open(resource_path)
            # extract to working
            resources.extractall(self.working_dir)
            resources.close()
        else:
            self.logger.error("ERROR: reimaging_resources.tgz not found in rawdata!")

    def move_ancillaries(self):
        """
        Move ancillary products to correct locations for processing

        :return:
        """
        list_file_path = Path("./rawdata/SEIP_parameter.list")
        resources_path = Path("./rawdata/reimaging_resources.tgz")
        component_path = list(Path("./rawdata").glob("*catalog.csv"))[0]

        if Path(list_file_path).exists():
            shutil.copy(list_file_path, self.working_dir)
        else:
            self.logger.error("ERROR: Could not find SEIP_parameter.list in rawdata!")

        if Path(resources_path).exists():
            shutil.copy(resources_path, self.working_dir)
        else:
            self.logger.error("ERROR: Could not find reimaging_resources.tgz in rawdata!")

        if Path(component_path).exists():
            shutil.copy(component_path, self.products_dir)
        else:
            self.logger.error("ERROR: Could not find component list in rawdata!")
