#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for casa_envoy.palaver
"""
import argparse
from pathlib import Path
from unittest.mock import MagicMock, patch

import casa_envoy.palaver as palaver
import pytest
from pycapo import CapoConfig


@pytest.fixture
def arg_parser() -> argparse.ArgumentParser:
    return palaver.arg_parser()


class TestPalaver:
    def test_get_settings(self, arg_parser, cal_settings, resource_path_root: Path):
        """
        verifies correct determination of needed settings

        :return:
        """
        expected_settings = cal_settings
        arg_parser.parse_args(
            [
                "--standard-cal",
                f"{resource_path_root}/input_files/test.json",
                f"{resource_path_root}/input_files/PPR.xml",
            ]
        )

        with patch(
            "casa_envoy.palaver.CapoConfig",
            MagicMock(return_value=CapoConfig(profile="test-capo", path=f"{resource_path_root}/input_files/")),
        ):
            with patch(
                "pathlib.Path.cwd",
                MagicMock(return_value="/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmpo1ca1pp_"),
            ) as cwd:
                settings = palaver._get_settings(
                    cwd, [f"{resource_path_root}/input_files/test.json", f"{resource_path_root}/input_files/PPR.xml"]
                )
                assert settings["useCasa"] == expected_settings["useCasa"]
                assert settings["homeForReprocessing"] == expected_settings["homeForReprocessing"]
                assert settings["metadata"] == expected_settings["metadata"]
                assert settings["ppr"] == expected_settings["ppr"]

    def test_parser(self):
        """
        Verifies parser setup

        :return:
        """
        parser = palaver.arg_parser()
        assert isinstance(parser, argparse.ArgumentParser)

    @patch("json.loads")
    @patch("os.chdir")
    @patch("os.getcwd")
    def test_main_cal(self, mock_cwd, mock_chdir, mock_json, arg_parser, resource_path_root: Path):
        """
        verifies use of correct launcher for calibrations

        :param mock_cwd:
        :param mock_chdir:
        :param mock_json:
        :return:
        """
        args = arg_parser.parse_args(
            [
                "--standard-cal",
                f"{resource_path_root}/input_files/test.json",
                f"{resource_path_root}/input_files/PPR.xml",
            ]
        )

        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args)) as mock_args:
            with patch("casa_envoy.launchers.CalibrationLauncher.launch_casa") as cal_launcher:
                palaver.main()
                assert cal_launcher.call_count == 1

    @patch("json.loads")
    @patch("os.chdir")
    @patch("os.getcwd")
    def test_main_img(self, mock_cwd, mock_chdir, mock_json, arg_parser, resource_path_root: Path):
        """
        verifies use of correct launcher for images

        :param mock_cwd:
        :param mock_chdir:
        :param mock_json:
        :return:
        """
        args = arg_parser.parse_args(
            [
                "--standard-img",
                f"{resource_path_root}/input_files/image-metadata.json",
                f"{resource_path_root}/input_files/cmsimage-PPR.xml",
            ]
        )

        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args)) as mock_args:
            with patch("casa_envoy.launchers.ImagingLauncher.launch_casa") as img_launcher:
                palaver.main()
                assert img_launcher.call_count == 1

    @patch("json.loads")
    @patch("os.chdir")
    @patch("os.getcwd")
    def test_main_restore(self, mock_cwd, mock_chdir, mock_json, arg_parser, resource_path_root: Path):
        """
        verifies use of correct launcher for restores

        :param mock_cwd:
        :param mock_chdir:
        :param mock_json:
        :return:
        """
        args = arg_parser.parse_args(
            ["--restore", f"{resource_path_root}/input_files/restore.json", f"{resource_path_root}/input_files/PPR.xml"]
        )

        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args)) as mock_args:
            with patch("casa_envoy.launchers.CalibrationLauncher.launch_casa") as cal_launcher:
                palaver.main()
                assert cal_launcher.call_count == 1
