#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Tests for casa_envoy Configure
"""
import json
import os
from pathlib import Path
from test import make_abstract_text_file
from unittest.mock import patch

import pytest
from casa_envoy.configure import (
    CasaConfigure,
    ParallelRunner,
    PiperestorescriptRunner,
    StandardRunner,
)
from casa_envoy.schema import AbstractTextFile


@pytest.fixture
def parallel_parameters(resource_path_root: Path) -> dict:
    return {
        "useCasa": False,
        "homeForReprocessing": "/home/casa/packages/pipeline/current",
        "rootDirectory": "/tmp/workspaces_tmp/",
        "processingDirectory": "tmpiox5trbp",
        "metadata": f"{resource_path_root}/input_files/test.json",
        "ppr": f"{resource_path_root}/input_files/PPR.xml",
        "product_type": "standard-cal",
        "parent_path": Path("I/am/a/path"),
        "requested_parallel": "4",
    }


@pytest.fixture
def test_cal_metadata(resource_path_root: Path) -> AbstractTextFile:
    return make_abstract_text_file(resource_path_root / "input_files/test.json")


class TestCasaConfigure:
    """
    Test methods for CASA environment set up
    """

    def test_setup_environment(self, cal_settings, test_cal_metadata: AbstractTextFile):
        content = json.loads(test_cal_metadata.content)
        CasaConfigure(parameters=cal_settings, content=content).setup_environment()
        assert os.environ.get("SCIPIPE_ROOTDIR") == cal_settings["rootDirectory"]
        assert os.environ.get("CASA_HOME") == cal_settings["homeForReprocessing"]
        assert os.environ.get("PPR_FILENAME") == cal_settings["ppr"]


class TestStandardRunner:
    """
    Test methods for Standard CASA calls
    """

    @patch("os.chdir")
    @patch("subprocess.Popen")
    def test_std_call_casa(self, mock_subprocess, mock_chdir, cal_settings):
        runner = StandardRunner(parameters=cal_settings)
        runner.call_casa()
        assert mock_subprocess.called_with(runner.build_command)


class TestParallelRunner:
    """
    Test methods for Parallel CASA calls
    """

    @patch("os.chdir")
    @patch("subprocess.Popen")
    def test_parallel_call_casa(self, mock_subprocess, mock_chdir, parallel_parameters: dict):
        runner = ParallelRunner(parameters=parallel_parameters)
        runner.call_casa()
        assert mock_subprocess.called_with(runner.build_command)


class TestPiperestorescriptRunner:
    """
    Test methods for Piperestorescript CASA calls
    """

    @patch("os.chdir")
    @patch("subprocess.Popen")
    def test_piperestorescript_call_casa(self, mock_subprocess, mock_chdir, restore_settings_sans_ppr):
        assert "piperestorescript_path" in restore_settings_sans_ppr
        runner = PiperestorescriptRunner(parameters=restore_settings_sans_ppr)
        runner.call_casa()
        assert mock_subprocess.called_with(runner.build_command) @ patch("os.chdir")

    @patch("os.chdir")
    @patch("subprocess.Popen")
    def test_piperestorescript_call_casa_no_piperestorescript(self, mock_subprocess, mock_chdir, cal_settings):
        assert "piperestorescript_path" not in cal_settings
        runner = PiperestorescriptRunner(parameters=cal_settings)
        with pytest.raises(ValueError, match="casa_piperestorescript"):
            runner.call_casa()
        assert not mock_subprocess.called
