#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
# -------------------------------------------------------------------------
#
#        D E S T I N A T I O N   S Y S T E M
#
# -------------------------------------------------------------------------
import pathlib

from .destinations.checksum import ChecksumDecorator
from .destinations.fetchfile import FetchFileGenerator
from .destinations.interfaces import DeliveryContextIF, Destination
from .destinations.local import LocalDestination
from .destinations.sharedweb import SharedWebDestination
from .destinations.tar import TarArchiver


class DestinationBuilder:
    """
    Facilitates building a stack of destination and its decorators.
    """

    def __init__(self, context: DeliveryContextIF):
        self.context = context
        self._destination = None

    def local(self, path: pathlib.Path, prefix: str = ""):
        """Base the delivery on a local destination with the given path"""
        self._destination = LocalDestination(self.context, path / prefix)
        return self

    def web(self, prefix: str = ""):
        """Base the delivery on a shared web destination"""
        self._destination = SharedWebDestination(self.context, prefix)
        return self

    def checksums(self):
        "Add a SHA1SUMS file to the delivery"
        self._destination = ChecksumDecorator(self._destination)
        return self

    def fetchfile(self):
        """Add the fetch-all.sh script to the delivery."""
        self._destination = FetchFileGenerator(self._destination)
        return self

    def tar(self, projects):
        """Add the tar decorator to the destination"""
        self._destination = TarArchiver(self.context, self._destination, projects)
        return self

    def build(self) -> Destination:
        """Create the destination"""
        return self._destination
