#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import getpass
import pathlib
from typing import Dict

from pycapo import CapoConfig

from .interfaces import DeliveryContextIF, DestinationDecorator
from .local import LocalDestination


class SharedWebDestination(DestinationDecorator):
    """
    Users who do not specify a local destination receive the standard shared web destination automatically.
    This is a location in a shared filesystem where a web service is configured to allow access files.
    Otherwise, it behaves like a local destination.
    """

    def __init__(self, context: DeliveryContextIF, prefix: str = ""):
        # The implementation trick here is basically that the constructor makes it look like a final destination,
        # but internally it is a decorator for a local destination with some slightly different
        # initialization parameters, which come from Capo
        capo = CapoConfig().settings("edu.nrao.workspaces.DeliverySettings")

        # get the username, or "anonymous" (handled by the parser)
        username = context.username

        # determine the destination directory
        self.destination_dir = pathlib.Path(capo.downloadDirectory) / username / prefix / context.token()

        # we're ready to build the local destination and thread it through our super constructor
        super().__init__(LocalDestination(context, self.destination_dir))

        # Open permissions on the username dir for the group so both vlapipe and almapipe can make new delivery dirs
        user_path = pathlib.Path(capo.downloadDirectory) / username
        if user_path.owner() == getpass.getuser():
            user_path.chmod(0o2775)
            for p in user_path.rglob("**"):
                if p.owner() == getpass.getuser():
                    p.chmod(0o2775)

        # generate the download URL for later usage
        self.download_url = f"{capo.downloadUrl}/{username}{'/'+ prefix if prefix else ''}/{context.token()}"

        # retain the token for later
        self.token = context.token()

    def close(self):
        """
        Closes the underlying destination and updates the permissions to be readable and writable by the group.
        """
        self.underlying.close()

        # Set permissions for the group so vlapipe and almapipe can make delivery directories
        if self.destination_dir.owner() == getpass.getuser():
            self.destination_dir.chmod(0o2775)

        # Set permissions on delivered files for the group so vlapipe and almapipe can modify them
        for p in self.destination_dir.rglob("*"):
            if p.owner() == getpass.getuser():
                p.chmod(0o2775)

    def results(self) -> Dict:
        return {"delivered_to": str(self.destination_dir), "url": self.download_url, "token": self.token}

    def __str__(self):
        return f"shared web root at {self.destination_dir}"
