#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import pathlib
from test import assert_directories_are_same

import pytest
from delivery.context import DeliveryContext
from delivery.deliverer import LocalDestination


@pytest.fixture
def dest_dir(tmpdir) -> pathlib.Path:
    """
    Generate a destination directory.
    :param tmpdir:  the tmpdir fixture we depend on
    :return:  path to the destination directory
    """
    # generate the destination directory
    dest_dir = pathlib.Path(tmpdir) / "after"
    dest_dir.mkdir()
    return dest_dir


@pytest.fixture
def file_to_deliver(tmpdir) -> pathlib.Path:
    """
    Generate some files in a test directory for delivery.
    :param tmpdir: pytest fixture for the temporary directory
    :return: a file in the test directory
    """
    # rewrap the tmpdir as a pathlib.Path instance
    tmpdir = pathlib.Path(tmpdir)

    # generate the test file in prior
    (tmpdir / "prior").mkdir()
    file_name = "test.txt"

    temp_file = tmpdir / "prior" / file_name  # place a file into the source
    temp_file.write_text("content", encoding=None)  # write something into that file

    # secretly generate a second file
    temp_file2 = tmpdir / "prior" / "test2.txt"  # make another file in source
    temp_file2.write_text("more content", encoding=None)  # add some content to the second file

    return temp_file


def test_local_delivery_add_file(resource_path_root, file_to_deliver: pathlib.Path, dest_dir: pathlib.Path):
    """
    Ensure that local delivery works
    """
    test_data_path = resource_path_root / "724126739"
    prefix = "1234"
    test_context = DeliveryContext.parse_commandline(["-r", str(test_data_path), "--prefix", prefix])

    # generate the destination
    local_dest = LocalDestination(test_context, dest_dir)

    # add the file to it
    local_dest.add_file(file_to_deliver, file_to_deliver.name)

    # see if the source file ended up in the destination
    assert (dest_dir / file_to_deliver.name).exists()

    # see if the content of the file is intact
    assert (dest_dir / file_to_deliver.name).read_text(encoding=None) == "content"

    # see if the results are alright
    results = local_dest.results()
    assert len(results.keys()) == 3
    assert results["delivered_to"] == str(dest_dir)


@pytest.mark.parametrize("flatten", [True, False])
def test_local_delivery_add_directory(
    resource_path_root: pathlib.Path, file_to_deliver: pathlib.Path, dest_dir: pathlib.Path, flatten: bool
):
    """Ensure that local delivery does something"""
    test_data_path = resource_path_root / "724126739"
    test_data_dir_to_add = test_data_path / "b2f536ca-bdce-4073-b9ee-96d8266109e7"
    prefix = "1234"
    args = ["-r", str(test_data_path), "--prefix", prefix]
    if flatten is True:
        assert any(entry.is_dir() for entry in test_data_path.iterdir())  # Needs to be something to flatten
        args.append("--flatten")
        args.append("execblock")
    test_context = DeliveryContext.parse_commandline(args)

    local_dest = LocalDestination(test_context, dest_dir)
    local_dest.add_directory(test_data_dir_to_add, test_data_dir_to_add.name, flatten)

    if flatten:
        assert all(entry.is_file() for entry in dest_dir.iterdir())
        # compare the source and destination -- destination should have all the files at the top level
        test_data_filenames = sorted(entry.name for entry in test_data_dir_to_add.rglob("*") if entry.is_file())
        temp_directory_entries = sorted(entry.name for entry in dest_dir.iterdir())
        assert test_data_filenames == temp_directory_entries
    else:
        assert_directories_are_same(dest_dir / test_data_dir_to_add.name, test_data_dir_to_add)

    # see if the results are alright
    results = local_dest.results()
    assert len(results.keys()) == 3
    assert results["delivered_to"] == str(dest_dir)


def test_subdirectories(file_to_deliver: pathlib.Path, dest_dir: pathlib.Path):
    """
    Ensure that subdirectories work the way we expect them to.
    """
    subdirectory = "subdirectory"

    local_dest = LocalDestination(DeliveryContext(file_to_deliver.parent), dest_dir)
    subdir = local_dest / subdirectory
    subdir.add_directory(file_to_deliver.parent, file_to_deliver.parent.name, False)

    # ensure that we only got one file in the tmpdir called "subdirectory"
    assert list(dest_dir.iterdir()) == [dest_dir / subdirectory]

    # compare the subdirectories
    assert_directories_are_same(file_to_deliver.parent, dest_dir / subdirectory / file_to_deliver.parent)
