#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Various 'n' sundry utilities for our tests """
# pylint: disable=E0401, R1721
import json
from pathlib import Path
from typing import Generator, List, Tuple
from unittest.mock import patch

import pytest
from ingest_envoy.ingestion_manifest import IngestionManifest, IngestionManifestBuilder
from ingest_envoy.manifest_components import (
    INIT_WEBLOG_FILENAME,
    TARFILE_EXT,
    WEBLOG_FILENAME,
    AncillaryProduct,
    OutputScienceProduct,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType, Telescope
from pycapo import CapoConfig

CAPO_CONFIG = CapoConfig()

# --------------------------------
#    EVLA CAL manifest test data
# --------------------------------
EVLA_CAL_INPUT_FILENAMES = ["20A-346_2021_07_23_T13_37_08.376.tar", WEBLOG_FILENAME]
UNWANTED = ["ignore_me.fits", "just_a_lotta_nothing", "uninteresting_metadata.xml"]
SECI_SCIENCE_PRODUCT = {
    "external_name": "VLASS3.1.se.SqDeg2.J232156+603000.06.2048",
    "science_product_locator": "uid://evla/fits_image/b79afe31-81ab-436e-9029-170fb93482ce",
}

EVLA_EB_FILENAME = "19A-001.sb1234567890.eb233423545632.54321.894327984569"


def ordered(obj: dict | list) -> dict | list:
    """
    Ripped this off SO, recursively sorts the given object and all nested lists.

    :param obj: An object to sort
    :return: The given object sorted all the way down
    """
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    return obj


def evla_eb_information() -> list[str]:
    """
    :return: The EVLA EB file name, directory, telescope, and project code used for testing.
    """
    return [EVLA_EB_FILENAME, "/fake/directory", "EVLA", "19A-001"]


def expected_settings_evla_eb() -> dict:
    """
    :return: The expected return value for _get_settings.
    """
    test_filename, _, telescope, project_code = evla_eb_information()
    return {
        "filename": test_filename,
        "telescope": telescope,
        "project": project_code,
        "workflowDir": "directory",
        "sourcePath": Path("/fake"),
        "staging_area": "/fake/staging/directory",
        "storage_area": "/fake/storage/directory",
        "useIngest": True,
        "script_location": "/fake/script/location",
        "workflowUrl": "https://fake-service-url.com",
    }


@pytest.fixture(name="evla_eb_staging_dir", scope="session")
def get_evla_eb_staging_dir() -> Path:
    """
    :return: The path to the fake EVLA EB staging directory.
    """
    staging_dir = Path(CAPO_CONFIG.getstring("edu.nrao.workspaces.IngestionSettings.stagingDirectory"))
    return staging_dir / EVLA_EB_FILENAME


@pytest.fixture
def mock_evla_eb_solicitor_solicit_contents():
    """
    Yields a MagicMock that mocks Solicitor.solicit_contents for the tested EVLA EB.
    """
    with patch("ingest_envoy.solicitor.Solicitor.solicit_contents") as mock_solicit_contents:
        mock_solicit_contents.return_value = {"projectMetadata": {"telescope": "EVLA", "projectCode": "19A-001"}}
        yield mock_solicit_contents  # This allows the test to use the mock


@pytest.fixture(scope="function")
def ingest_path(tmp_path: Path) -> Path:
    """
    Make a directory to use as the ingestion staging dir, or curation source

    :param tmp_path: built-in pytest fixture, Pytest cleans them up periodically
    :return: Path to new directory
    """

    # cast is necessary because otherwise we get a LocalPath, which doesn't work
    fake_ingest_path = tmp_path / "ingestion"
    fake_ingest_path.mkdir()
    return fake_ingest_path


@pytest.fixture(scope="function")
def spectral_cube_path(tmp_path: Path) -> Path:
    """
    Make a directory to use as the ingestion staging dir, or curation source

    :param tmp_path: built-in pytest fixture, Pytest cleans them up periodically
    :return: Path to new directory
    """

    # cast is necessary because otherwise we get a LocalPath, which doesn't work
    fake_ingest_path = tmp_path / "ingestion"
    fake_ingest_path.mkdir()
    populate_fake_path(
        fake_ingest_path, SPECTRAL_CUBE_SCIENCE_PRODUCTS, SPECTRAL_CUBE_ANCILLARIES, [AUX_IMAGE_METADATA_FILENAME]
    )

    yield fake_ingest_path


@pytest.fixture(scope="function")
def vlass_coarse_path() -> Path:
    """
    Make a directory to use as a VLASS spool directory.

    :return: Path to new directory
    """
    fake_path = (
        Path(CAPO_CONFIG.getstring("OodtSettings.spoolDir"))
        / "quicklook"
        / "VLASS3.1_T10t10.J063800-023000_P106057v1_2023_03_07T22_58_33.300"
    )
    fake_path.mkdir(parents=True)
    (fake_path / "products").mkdir()
    (fake_path / "working").mkdir()

    populate_fake_path(
        (fake_path / "products"),
        VLASS_COARSE_CUBE_SCIENCE_PRODUCTS,
        VLASS_COARSE_CUBE_ANCILLARIES,
        VLASS_COARSE_CUBE_CACHE["products"],
    )
    populate_fake_path((fake_path / "working"), [], [], VLASS_COARSE_CUBE_CACHE["working"])

    yield fake_path


@pytest.fixture
def aux_data() -> dict:
    aux_file = "test/input_files/" + AUX_IMAGE_METADATA_FILENAME
    with open(aux_file) as f:
        aux_data = json.load(f)
    return aux_data


@pytest.fixture
def alternate_manifest_destination(tmp_path: Path) -> Path:
    """
    Make an alternative directory to ingest_path for tests to put their manifests in

    :param tmp_path: built-in pytest fixture, Pytest cleans them up periodically
    :return: Path to new directory
    """
    alternate_manifest_destination = tmp_path / "manifest_destination"
    alternate_manifest_destination.mkdir()
    return alternate_manifest_destination


@pytest.fixture
def image_manifest(ingest_path: Path) -> Tuple[IngestionManifest, Path]:
    """
    Build an image ingestion manifest that "should" match image_manifest_tmpx_ratuqh.json

    :return: the ingestion manifest and the manifest file in which it lives
    """

    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)
    additional_metadata = AbstractTextFile(filename=AUX_IMAGE_METADATA_FILENAME, content="")

    return IngestionManifestBuilder(
        manifest_source_dir=ingest_path,
        additional_metadata=additional_metadata,
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator="uid://evla/calibration/ea93dae5-3495-47fa-887d-4be2852f5f14",
    ).build()


def find_example_manifest(manifest_name: str) -> Path:
    """
    Get this example manifest for comparison with one we've generated in a test.

    :param manifest_name: unique file identifier
    :return: full path to the manifest file
    """
    filename = manifest_name + ".json"
    for file in Path.cwd().rglob(filename):
        return file

    raise FileNotFoundError(filename)


def populate_fake_evla_cal_ingest_path(staging_dir: Path) -> List[Path]:
    """
    Create a directory containing fake calibration products, plus other stuff
    that we -don't- want to ingest.

    :param staging_dir: our temporary dir
    :return:
    """

    files = []
    filenames = [filename for filename in EVLA_CAL_INPUT_FILENAMES]
    for filename in UNWANTED:
        filenames.append(filename)

    for filename in filenames:
        path = staging_dir / filename
        path.touch()
        files.append(path)

    return files


def populate_fake_final_evla_cal_ingest_path(staging_dir: Path) -> List[Path]:
    """
    Create a directory containing fake calibration products, plus other stuff
    that we -don't- want to ingest, PLUS an initial weblog

    :param staging_dir: our temporary dir
    :return:
    """
    files = populate_fake_evla_cal_ingest_path(staging_dir)
    init_weblog = staging_dir / "initial_weblog.tgz"
    init_weblog.touch()
    files.append(init_weblog)
    return files


def mocked_archive_service_get(*args, **kwargs):
    """
    Can mock various archive service calls and returns a response with a code and JSON.
    """

    class MockResponse:
        def __init__(self, json_data, status_code):
            self.json_data = json_data
            self.status_code = status_code

        def json(self):
            return self.json_data

    if "archive-service" in args[0] and SECI_SCIENCE_PRODUCT["external_name"] in kwargs["params"]["identifier"]:
        return MockResponse({"locator": SECI_SCIENCE_PRODUCT["science_product_locator"]}, 200)

    return MockResponse(None, 404)


# -----------------------------
#    Image manifest test data
# -----------------------------
IMAGE_PRODUCT_GROUP = 1
IMAGE_LOCATOR = "uid://evla/calibration/ea93dae5-3495-47fa-887d-4be2852f5f14"
AUX_IMAGE_METADATA_FILENAME = "aux_image_metadata.json"

PRIMARY_BEAM_ANCILLARY_K = AncillaryProduct(
    type=AncillaryProductType.PB_FITS, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.K_band.cont.I.pb.tt0.fits"
)
CLEAN_MASK_ANCILLARY_K = AncillaryProduct(
    type=AncillaryProductType.FITS_MASK, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.K_band.cont.I.mask.fits"
)
ALPHA_ERROR_K = AncillaryProduct(
    type=AncillaryProductType.FITS,
    filename="16B-069.MJD57713.51329133102.J1522+3934_sci.K_band.cont.I.alpha.error.fits",
)
ALPHA_K = AncillaryProduct(
    type=AncillaryProductType.FITS, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.K_band.cont.I.alpha.fits"
)
TT0_K = AncillaryProduct(
    type=AncillaryProductType.FITS, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.K_band.cont.I.tt0.fits"
)
TT1_K = AncillaryProduct(
    type=AncillaryProductType.FITS, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.K_band.cont.I.tt1.fits"
)
OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES_K = [
    PRIMARY_BEAM_ANCILLARY_K,
    CLEAN_MASK_ANCILLARY_K,
    ALPHA_ERROR_K,
    ALPHA_K,
    TT0_K,
    TT1_K,
]

OUTPUT_SCIENCE_PRODUCT_K = OutputScienceProduct(
    type=AncillaryProductType.FITS,
    filename="16B-069.MJD57713.51329133102.J1522+3934_sci.K_band.cont.I.pbcor.tt0.fits",
    ancillary_products=OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES_K,
)

PRIMARY_BEAM_ANCILLARY_X = AncillaryProduct(
    type=AncillaryProductType.PB_FITS, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.X_band.cont.I.pb.tt0.fits"
)
CLEAN_MASK_ANCILLARY_X = AncillaryProduct(
    type=AncillaryProductType.FITS_MASK, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.X_band.cont.I.mask.fits"
)
ALPHA_ERROR_X = AncillaryProduct(
    type=AncillaryProductType.FITS,
    filename="16B-069.MJD57713.51329133102.J1522+3934_sci.X_band.cont.I.alpha.error.fits",
)
ALPHA_X = AncillaryProduct(
    type=AncillaryProductType.FITS, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.X_band.cont.I.alpha.fits"
)
TT0_X = AncillaryProduct(
    type=AncillaryProductType.FITS, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.X_band.cont.I.tt0.fits"
)
TT1_K = AncillaryProduct(
    type=AncillaryProductType.FITS, filename="16B-069.MJD57713.51329133102.J1522+3934_sci.X_band.cont.I.tt1.fits"
)
OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES_X = [
    PRIMARY_BEAM_ANCILLARY_X,
    CLEAN_MASK_ANCILLARY_X,
    ALPHA_ERROR_X,
    ALPHA_X,
    TT0_X,
    TT1_K,
]

OUTPUT_SCIENCE_PRODUCT_X = OutputScienceProduct(
    type=AncillaryProductType.FITS,
    filename="16B-069.MJD57713.51329133102.J1522+3934_sci.X_band.cont.I.pbcor.tt0.fits",
    ancillary_products=OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES_X,
)

# input files
WEBLOG_ANCILLARY = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG, filename=WEBLOG_FILENAME)
INIT_WEBLOG_ANCILLARY = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG, filename=INIT_WEBLOG_FILENAME)
PIPELINE_AF_ANCILLARY = AncillaryProduct(
    type=AncillaryProductType.PIPELINE_ARTIFACTS,
    filename="pipeline_artifacts_2021_08_04T15_46_02.tar",
)
# an output file
INGESTION_AF_ANCILLARY = AncillaryProduct(
    type=AncillaryProductType.INGESTION_ARTIFACTS,
    filename="ingestion_artifacts_2021_08_04T01_57_08.564.tar",
)
ANCILLARY_PRODUCTS = [WEBLOG_ANCILLARY, INGESTION_AF_ANCILLARY, PIPELINE_AF_ANCILLARY]

STAGING_DIR_FILES = [
    "aux_image_metadata.json",
    "oussid.J1522+3934_sci.K_band.cont.I.pbcor.tt0.fits",
    "oussid.J1522+3934_sci.X_band.cont.I.pbcor.tt0.fits",
]


def populate_fake_tmpx_ratuqh_ingest_path(
    staging_source_dir: Path, is_final: bool = False, is_for_curation_test: bool = False
) -> List[Path]:
    """
    Make a bunch of fake files that should result in the example manifest.
    If this is version 2 or later of a standard calibration, include the initial weblog.

    :return:
    """
    fake_files_to_create = []

    if not is_for_curation_test:
        # Curator doesn't care about the additional metadata file for images
        fake_files_to_create.append(AUX_IMAGE_METADATA_FILENAME)
    fake_files_to_create.append(PIPELINE_AF_ANCILLARY.filename)

    for product in OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES_K:
        fake_files_to_create.append(product.filename)

    for product in OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES_X:
        fake_files_to_create.append(product.filename)

    fake_files_to_create.append(WEBLOG_ANCILLARY.filename)
    if is_final:
        fake_files_to_create.append(INIT_WEBLOG_ANCILLARY.filename)

    fake_files_to_create.append(OUTPUT_SCIENCE_PRODUCT_K.filename)
    fake_files_to_create.append(OUTPUT_SCIENCE_PRODUCT_X.filename)

    if is_for_curation_test:
        fake_files_to_create.append(INGESTION_AF_ANCILLARY.filename)

    files = []
    for filename in fake_files_to_create:
        file = staging_source_dir / filename
        file.touch()
        files.append(file)

    assert len(files) == len(fake_files_to_create)

    return files


# -----------------------------
# SDM manifest test data
# -----------------------------
SDM_FILE_LIST = [
    "Antenna.xml",
    "ASDM.xml",
    "CalData.xml",
    "CalDevice.xml",
    "CalPointing.xml",
    "CalReduction.xml",
    "ConfigDescription.xml",
    "CorrelatorMode.xml",
    "DataDescription.xml",
    "Doppler.xml",
    "Ephemeris.xml",
    "ExecBlock.xml",
    "Feed.xml",
    "Field.xml",
    "Flag.xml",
    "Main.xml",
    "Pointing.xml",
    "PointingModel.xml",
    "Polarization.xml",
    "Processor.xml",
    "Receiver.xml",
    "SBSummary.xml",
    "Scan.xml",
    "Source.xml",
    "SpectralWindow.xml",
    "State.xml",
    "Station.xml",
    "Subscan.xml",
    "SwitchCycle.xml",
    "SysCal.xml",
    "SysPower.bin",
    "Weather.xml",
]
# From file ./examples/full_curation_evla_eb_manifest.json
EVLA_EB_NAME = "19A-001.sb1234567890.eb233423545632.54321.894327984569"
EVLA_EB_LOCATOR = "uid://I/am/a/locator"
EVLA_EB_PRODUCT_GROUP = "2"


def populate_fake_evla_eb_curator_source_path(staging_dir: Path) -> list[Path]:
    eb_dir = staging_dir / EVLA_EB_NAME
    eb_dir.mkdir()
    for sdm_filename in SDM_FILE_LIST:
        sdm_file_path = eb_dir / sdm_filename
        sdm_file_path.touch()
    # Only really care about the directory for the manifest
    return [eb_dir]


# -----------------------
# Spectral Cube test data
# -----------------------

SPECTRAL_CUBE_SCIENCE_PRODUCTS = [
    OutputScienceProduct(
        ScienceProductType.IMAGE,
        "oussid.L1473maser_sci.spw11.cube.regcal.I.pbcor.fits",
        [
            AncillaryProduct(AncillaryProductType.FITS_MASK, "oussid.L1473maser_sci.spw11.cube.regcal.I.mask.fits"),
            AncillaryProduct(AncillaryProductType.PB_FITS, "oussid.L1473maser_sci.spw11.cube.regcal.I.pb.fits"),
        ],
    ),
    OutputScienceProduct(
        ScienceProductType.IMAGE,
        "oussid.L1473maser_sci.spw12.cube.regcal.I.pbcor.fits",
        [
            AncillaryProduct(AncillaryProductType.FITS_MASK, "oussid.L1473maser_sci.spw12.cube.regcal.I.mask.fits"),
            AncillaryProduct(AncillaryProductType.PB_FITS, "oussid.L1473maser_sci.spw12.cube.regcal.I.pb.fits"),
        ],
    ),
    OutputScienceProduct(
        ScienceProductType.IMAGE,
        "oussid.L1473maser_sci.spw13.cube.regcal.I.pbcor.fits",
        [
            AncillaryProduct(AncillaryProductType.FITS_MASK, "oussid.L1473maser_sci.spw13.cube.regcal.I.mask.fits"),
            AncillaryProduct(AncillaryProductType.PB_FITS, "oussid.L1473maser_sci.spw13.cube.regcal.I.pb.fits"),
        ],
    ),
    OutputScienceProduct(
        ScienceProductType.IMAGE,
        "oussid.L1482maser_sci.spw11.cube.regcal.I.pbcor.fits",
        [
            AncillaryProduct(AncillaryProductType.FITS_MASK, "oussid.L1482maser_sci.spw11.cube.regcal.I.mask.fits"),
            AncillaryProduct(AncillaryProductType.PB_FITS, "oussid.L1482maser_sci.spw11.cube.regcal.I.pb.fits"),
        ],
    ),
    OutputScienceProduct(
        ScienceProductType.IMAGE,
        "oussid.L1482maser_sci.spw12.cube.regcal.I.pbcor.fits",
        [
            AncillaryProduct(AncillaryProductType.FITS_MASK, "oussid.L1482maser_sci.spw12.cube.regcal.I.mask.fits"),
            AncillaryProduct(AncillaryProductType.PB_FITS, "oussid.L1482maser_sci.spw12.cube.regcal.I.pb.fits"),
        ],
    ),
    OutputScienceProduct(
        ScienceProductType.IMAGE,
        "oussid.L1482maser_sci.spw13.cube.regcal.I.pbcor.fits",
        [
            AncillaryProduct(AncillaryProductType.FITS_MASK, "oussid.L1482maser_sci.spw13.cube.regcal.I.mask.fits"),
            AncillaryProduct(AncillaryProductType.PB_FITS, "oussid.L1482maser_sci.spw13.cube.regcal.I.pb.fits"),
        ],
    ),
]


SPECTRAL_CUBE_ANCILLARIES = [
    AncillaryProduct(AncillaryProductType.AUX_PRODUCTS, "unknown.hifv_calimage_cont_cube_selfcal.auxproducts.tgz"),
    AncillaryProduct(AncillaryProductType.PIPELINE_WEBLOG, "weblog.tgz"),
    AncillaryProduct(AncillaryProductType.INGESTION_ARTIFACTS, "ingestion_artifacts_2024_10_15T17_38_02.293.tar"),
    AncillaryProduct(AncillaryProductType.PIPELINE_ARTIFACTS, "pipeline_artifacts_2024_10_15T10_00_02.tar"),
]


SPECTRAL_CUBE_CALIBRATION_SPL = "uid://evla/calibration/not-the-real-calibration"


# ---------------------------
# VLASS Coarse Cube test data
# ---------------------------

VLASS_COARSE_CUBE_SCIENCE_PRODUCTS = [
    OutputScienceProduct(
        ScienceProductType.VLASS_COARSE,
        "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw10.IQU.iter3.image.pbcor.tt0.subim.fits",
        [
            AncillaryProduct(
                AncillaryProductType.RMS_IMAGE,
                "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw10.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            ),
        ],
    ),
    OutputScienceProduct(
        ScienceProductType.VLASS_COARSE,
        "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw11.IQU.iter3.image.pbcor.tt0.subim.fits",
        [
            AncillaryProduct(
                AncillaryProductType.RMS_IMAGE,
                "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw11.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            ),
        ],
    ),
    OutputScienceProduct(
        ScienceProductType.VLASS_COARSE,
        "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw12.IQU.iter3.image.pbcor.tt0.subim.fits",
        [
            AncillaryProduct(
                AncillaryProductType.RMS_IMAGE,
                "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw12.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            ),
        ],
    ),
    OutputScienceProduct(
        ScienceProductType.VLASS_COARSE,
        "VLASS2.1.cc.T10t02.J005000-023000.06.2048.V_mean_minmaxrej.subim.fits",
        [],
    ),
]


VLASS_COARSE_CUBE_ANCILLARIES = [
    AncillaryProduct(AncillaryProductType.PIPELINE_WEBLOG, "weblog.tgz"),
]


VLASS_COARSE_CUBE_CACHE = {
    "products": [
        "casa_commands.log",
        "casa_pipescript.py",
        "pipeline_manifest.xml",
        "placeholder.catalog.csv",
        "unknown.hifv_vlassSEIP.pprequest.xml",
        "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw10.IQU.iter3.image.pbcor.tt0.subim.com.fits",
        "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw11.IQU.iter3.image.pbcor.tt0.subim.com.fits",
        "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw12.IQU.iter3.image.pbcor.tt0.subim.com.fits",
    ],
    "working": [
        "CCIP_parameter.list",
        "SEIP_parameter.list",
    ],
}


def populate_fake_path(
    source_dir: Path,
    science_products: list[OutputScienceProduct],
    ancillary_products: list[AncillaryProduct],
    additional_files: list[str],
) -> set[Path]:
    """
    Make a bunch of fake files that should result in the example manifest.

    :param source_dir: Our temporary dir
    :param science_products: A list of science products with any ancillaries
    :param ancillary_products: A list of group ancillaries
    :param additional_files: Any additional files in a list
    :return:
    """
    source_dir.mkdir(parents=True, exist_ok=True)

    fake_files_to_create: set[str] = set()

    # Make science products with their ancillaries
    for product in science_products:
        fake_files_to_create.add(product.filename)
        if product.ancillary_products:
            for ancillary in product.ancillary_products:
                fake_files_to_create.add(ancillary.filename)

    # Make group ancillary products
    for product in ancillary_products:
        if product.type != AncillaryProductType.INGESTION_ARTIFACTS:
            # Ingestion artifacts are created during manifest creation
            fake_files_to_create.add(product.filename)

    # Make any additional files passed
    for f in additional_files:
        fake_files_to_create.add(f)

    files = set()
    for filename in fake_files_to_create:
        file = source_dir / filename
        file.touch()
        files.add(file)

    assert len(files) == len(fake_files_to_create)

    return files


def find_ingestion_artifacts_tar(staging_source_dir: Path):
    """
    There should be an ingestion artifacts tar after manifest creation.

    :param staging_source_dir:
    :return:
    """
    ing_artifacts_tars = [
        file
        for file in staging_source_dir.iterdir()
        if file.name.startswith(AncillaryProductType.INGESTION_ARTIFACTS.value) and file.name.endswith(TARFILE_EXT)
    ]
    if len(ing_artifacts_tars) == 0:
        # we're testing curation, no new tar in directory
        return None
    # otherwise ensure there is only 1 tar file present
    assert len(ing_artifacts_tars) == 1
    return ing_artifacts_tars[0]


# ---------------------------
# VLASS coarse cube content
# ---------------------------
def populate_directory(dir_path: Path, filenames: list[str]):
    """
    Creates the specified files in the given directory.

    :param dir_path: Path. The path to the directory which is to be populated.
    :param filenames: list[str]. The list of files to be created in the specified directory.
    """
    for filename in filenames:
        (dir_path / filename).touch()


# ---------
# Test data
# ---------
@pytest.fixture(scope="session")
def vlass_coarse_cube_staging_dir() -> Path:
    """
    :return: Path. The path to the fake VLASS coarse cube staging directory.
    """
    staging_dir = Path(CAPO_CONFIG.getstring("edu.nrao.workspaces.IngestionSettings.stagingDirectory"))
    return staging_dir / "tmp_coarse_cube_staging"


def get_vlass_coarse_cube_cache_products_dict() -> dict:
    """
    :return: dict. The VLASS coarse cube cache directory products in a dictionary (separated out into groups).
    """
    # As for SPW file names, they were created based on the information in the below Confluence document.
    # For the other files, the names were pulled from the list in product_collectors.py.
    # Confluence document: https://open-confluence.nrao.edu/pages/viewpage.action?pageId=166396955
    SPW_RANGE = range(2, 18)
    return {
        "science_products": {
            "per_channel_fits_files": [
                f"VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw{i}.IQU.iter3.image.pbcor.tt0.subim.fits"
                for i in SPW_RANGE
            ],
            "mean_minmaxrej_fits_file": "VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.V_mean_minmaxrej.subim.fits",
        },
        "ancillary_products": {
            "rms_iqu_cubes": [
                f"VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw{i}.IQU.iter3.image.pbcor.tt0.rms.subim.fits"
                for i in SPW_RANGE
            ],
            "others": [
                "weblog.tgz",
                "placeholder.catalog.csv",
                "casa_pipescript.py",
                "casa_commands.log",
                "pipeline_manifest.xml",
                "unknown.hifv_vlassSEIP.pprequest.xml",
                "CCIP_parameter.list",
                "SEIP_parameter.list",
            ],
        },
        "other_products": [
            # These are .subim.com.fits files that are in cache but not staging
            f"VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw{i}.V.iter3.image.pbcor.tt0.subim.com.fits"
            for i in SPW_RANGE
        ],
    }


@pytest.fixture
def expected_num_ancillary_products() -> int:
    """
    :return: int. The expected number of ancillary products.
    """
    cache_products = get_vlass_coarse_cube_cache_products_dict()
    return len(cache_products["ancillary_products"]["rms_iqu_cubes"]) + len(
        cache_products["ancillary_products"]["others"]
    )


@pytest.fixture(scope="module", name="vlass_coarse_cache_products")
def get_vlass_coarse_cube_cache_products_list() -> list[str]:
    """
    :return: list[str]. The list of products that should be in the cache directory for VLASS coarse cubes.
    """
    CACHE_PRODUCTS = get_vlass_coarse_cube_cache_products_dict()
    return (
        CACHE_PRODUCTS["science_products"]["per_channel_fits_files"]
        + CACHE_PRODUCTS["ancillary_products"]["rms_iqu_cubes"]
        + CACHE_PRODUCTS["ancillary_products"]["others"]
        + CACHE_PRODUCTS["other_products"]
        + [CACHE_PRODUCTS["science_products"]["mean_minmaxrej_fits_file"]]
    )


@pytest.fixture(scope="module", name="vlass_coarse_staging_products")
def get_vlass_coarse_cube_staging_products_list() -> list[str]:
    """
    :return: list[str]. The list of products that should be in the staging directory for VLASS coarse cubes.
    """
    STAGING_PRODUCTS = get_vlass_coarse_cube_cache_products_dict()
    del STAGING_PRODUCTS["other_products"]  # Drop the .subim.com.fits files because they shouldn't be ingested

    # Add the pipeline artifacts tar made in the product collector
    STAGING_PRODUCTS["ancillary_products"]["others"].append("pipeline_artifacts.tar")

    return (
        STAGING_PRODUCTS["science_products"]["per_channel_fits_files"]
        + STAGING_PRODUCTS["ancillary_products"]["rms_iqu_cubes"]
        + STAGING_PRODUCTS["ancillary_products"]["others"]
        + [STAGING_PRODUCTS["science_products"]["mean_minmaxrej_fits_file"]]
    )
