#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import glob
import logging
import os
import shutil
import subprocess
import sys
from pathlib import Path

import arrow
from ingest_envoy.interfaces import CollectorIF
from ingest_envoy.product_collectors import (
    ImageProductCollector,
    ObservationProductCollector,
    VlassCacheProductCollector,
    VlassCoarseProductCollector,
)
from ingest_envoy.schema import AbstractTextFile


def collect_image_metadata() -> AbstractTextFile:
    """
    Create an AbstractTextFile object from the auxiliary metadata file. Needed by manifest generator

    :return: AbstractTextFile
    """
    file = open("./aux_image_metadata.json", "r")
    return AbstractTextFile(file.name, file.read())


def aux_file_to_staging(staging_dir: Path):
    """
    Stage auxiliary metadata file for ingestion

    :return:
    """
    shutil.copy2(
        "./aux_image_metadata.json",
        staging_dir,
    )


def catalog_file_to_staging(staging_dir: Path):
    """
    Stage SECI components list file for ingestion

    :return:
    """
    for i in os.listdir("."):
        if os.path.isfile(os.path.join(".", i)) and "catalog.csv" in i:
            shutil.copy2(
                i,
                staging_dir,
            )


class ImageCollector(CollectorIF):
    """Collect Image Products for Ingestion"""

    def __init__(self, parameters):
        self.logger = logging.getLogger("ingest_envoy")
        self.parameters = parameters
        self.staging_source_dir = Path(self.parameters["staging_area"]) / self.parameters["workflowDir"]

    def collect_products(self):
        """
        Run image product collection

        :return:
        """
        self.logger.info("Collecting image products for staging...")
        workflow_dir = self.parameters["workflowDir"]
        staging_dir = Path(self.parameters["staging_area"])
        tar_name = self.create_artifacts_name()

        # rename image files to allow for duplicate ingestion
        self.rename_images()

        # Run product collection
        collector = ImageProductCollector(workflow_dir, staging_dir, tar_name)
        collector()

        aux_file_to_staging(self.staging_source_dir)

    def create_artifacts_name(self) -> str:
        """
        Determine name for pipeline artifacts tar file

        :return: string
        """
        name = "pipeline_artifacts_"
        date = self.parameters["processingStart"]
        return name + date

    def rename_images(self):
        """
        Rename images to allow for duplicate ingestion

        :return:
        """
        self.logger.info("Renaming images...")
        sdm = self.parameters["sdmId"]
        ebMjd = sdm.split(".")[-2] + "." + sdm.split(".")[-1]

        prefix = self.parameters["project"] + ".MJD" + ebMjd

        path = self.parameters["spoolDir"] + "/products"

        images = glob.glob(path + "/*.fits")

        for image in images:
            newname = image.replace("oussid", prefix)
            self.logger.info(f"Renaming {image} to {newname}")
            os.rename(image, newname)


class CoarseCollector(CollectorIF):
    """Collect Coarse Cube Products for Ingestion"""

    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.parameters = parameters

    @staticmethod
    def create_artifacts_name() -> str:
        """
        Determine name for pipeline artifacts tar file
        * Using current ingestion time for filename since processing time isn't
        available for coarse cube ingestion

        :return: string
        """
        name = "pipeline_artifacts_"
        date = arrow.now().format("YYYY_MM_DDThh_mm_ss")
        return name + date

    def collect_products(self):
        """
        Run coarse cube product collection

        :return:
        """
        self.logger.info("Collecting coarse cube products for staging...")
        workflow_dir = self.parameters["workflowDir"]
        staging_dir = Path(self.parameters["staging_area"]) / workflow_dir
        tar_name = self.create_artifacts_name()

        source_dir = Path(self.parameters["sourcePath"]) / workflow_dir
        qa_status = self.parameters["qa_status"]

        # Run product collection to both cache and staging
        self.logger.info("Running VLASS cache collector...")
        cache_collector = VlassCacheProductCollector(source_dir, qa_status)
        cache_collector()

        if qa_status == "PASS":
            cache_path = cache_collector.get_cache_dir()
            collector = VlassCoarseProductCollector(cache_path, staging_dir, tar_name)
            collector()

            aux_file_to_staging(staging_dir)
            catalog_file_to_staging(staging_dir)


class SECICollector(CollectorIF):
    """Collect SECI Image Products for Ingestion"""

    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.parameters = parameters
        self.staging_source_dir = Path(self.parameters["staging_area"]) / self.parameters["workflowDir"]

    @staticmethod
    def create_artifacts_name() -> str:
        """
        Determine name for pipeline artifacts tar file
        * Using current ingestion time for filename since processing time isn't
        available for SECI ingestion

        :return: string
        """
        name = "pipeline_artifacts_"
        date = arrow.now().format("YYYY_MM_DDThh_mm_ss")
        return name + date

    def collect_products(self):
        """
        Run SECI product collection script

        :return:
        """
        self.logger.info("Collecting SECI products for staging...")
        cache_path = self.parameters["sourcePath"]
        workflow_dir = self.parameters["workflowDir"]
        staging_dir = self.parameters["staging_area"]
        sbin_dir = self.parameters["script_location"]
        tar_name = self.create_artifacts_name()

        collector = subprocess.run(
            [f"{sbin_dir}/seci-product-collector.sh", cache_path, workflow_dir, staging_dir, tar_name],
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        if collector.returncode != 0:
            self.logger.error("ERROR: SECI product collection failed!")
            exit(1)

        aux_file_to_staging(self.staging_source_dir)
        catalog_file_to_staging(self.staging_source_dir)


class ObservationCollector(CollectorIF):
    """Collect Observation Products for Ingestion"""

    def __init__(self, parameters):
        self.logger = logging.getLogger("ingest_envoy")
        self.parameters = parameters

    def collect_products(self):
        """
        Run observation product collection script
        """
        self.logger.info("Collecting observation products for staging...")
        source_path = Path(self.parameters["sourcePath"]) / self.parameters["workflowDir"]
        staging_dir = Path(self.parameters["staging_area"])

        # Run product collection
        collector = ObservationProductCollector(source_path, staging_dir)
        collector()
