#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" This is the entrypoint for ingestion launching """
import logging
import sys
import tarfile
from pathlib import Path

from ingest_envoy.spectral_cube_manifest_utils import (
    SpectralCubeIngestionProductsFinder,
)
from marshmallow import fields
from pycapo import CapoConfig

# pylint: disable=C0301, E0401, R0903, R1721
from typing import Callable, Optional

import arrow
from ingest_envoy.manifest_components import (
    CURATOR_MANIFEST_FILENAME,
    INGESTION_ARTIFACTS_NAME,
    INIT_WEBLOG_FILENAME,
    MANIFEST_FILENAME,
    TARFILE_EXT,
    AncillaryProduct,
    BaseSchema,
    IngestionManifestKey,
    InputGroup,
    InputGroupSchema,
    InputScienceProduct,
    ManifestParameters,
    ManifestParametersSchema,
    OutputGroup,
    OutputGroupSchema,
    OutputScienceProduct,
    ReingestGroup,
    ReingestGroupSchema,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.std_img_manifest_utils import ImageIngestionProductsFinder
from ingest_envoy.std_obs_manifest_utils import ObservationIngestionProductsFinder
from ingest_envoy.utilities import (
    AncillaryProductType,
    CuratorType,
    IngestionManifestException,
    NoScienceProductException,
    ScienceProductType,
    Telescope,
    find_output_tars,
    find_weblogs,
)
from ingest_envoy.vlass_coarse_cube_manifest_utils import (
    VlassCoarseCubeIngestionProductsFinder,
)

logger = logging.getLogger("ingest_envoy")

# pylint: disable=R0902, R0913


class IngestionManifest:
    """Encapsulates an ingestion manifest"""

    def __init__(
        self,
        telescope: Telescope,
        sp_type: ScienceProductType,
        staging_source_dir: Path | None,  # curation has no staging directory
        locator: str | None,  # initial observation ingestion has no locator
        # if running curator, use ReingestGroup, otherwise none
        reingest_group: ReingestGroup | None,
        # all except Partial Curation, EVLA_EB and VLASS catalog manifest have input group
        input_group: InputGroup | None,
        # every non-curator manifest has at least one output group
        output_group: OutputGroup | None,
        # image manifest has this:
        additional_metadata: Optional[AbstractTextFile] = None,
        filename: Optional[str] = None,
        destination_dir: Path | None = None,
        # Alternative to staging_source_dir as a way of getting files
        file_list: Optional[list[str]] = None,
    ):
        self.staging_source_dir = staging_source_dir
        self.sp_type = sp_type
        self.locator = locator
        self.reingest = reingest_group
        self.input_group = input_group
        self.output_group = output_group
        self.telescope = telescope
        self.filename = filename
        self.destination_dir = destination_dir
        if destination_dir is None:
            self.destination_dir = self.staging_source_dir

        # Check if destination_dir is a valid path
        try:
            Path(self.destination_dir).resolve(strict=True)
        except FileNotFoundError:
            raise ValueError("Destination directory or staging source directory must be an existing path")

        # Check if additional_metadata is a valid path to an AbstractTextFile
        if additional_metadata and not (Path(self.staging_source_dir) / additional_metadata.filename).is_file():
            raise ValueError("Additional metadata must be a valid text file")

        # Check if the file list contains valid paths
        if file_list:
            for file in file_list:
                try:
                    Path(file).resolve(strict=True)
                except FileNotFoundError:
                    raise ValueError("Destination directory must be an existing path")

        # Check if NGAS ingestion should be enabled for all manifests in this environment
        self.ngas_ingest = self.get_ngas_flag(self.sp_type)

        self.parameters = self.build_ingest_parameters(additional_metadata)
        if file_list:
            self.files_found = file_list
        elif staging_source_dir is not None:
            # we are not running curation, find files
            self.files_found = [file for file in staging_source_dir.iterdir()]

    def get_ngas_flag(self, sp_type: ScienceProductType) -> bool:
        """Determine and retrieve the correct ingestNGAS flag for this type of ingestion/curation"""

        match sp_type:
            case ScienceProductType.EXEC_BLOCK:
                flag = CapoConfig().getboolean("archive-ingestion.ingestNGAS.observation")
            case ScienceProductType.EVLA_CAL:
                flag = CapoConfig().getboolean("archive-ingestion.ingestNGAS.calibration")
            case ScienceProductType.IMAGE | ScienceProductType.VLASS_SECI | ScienceProductType.VLASS_COARSE:
                flag = CapoConfig().getboolean("archive-ingestion.ingestNGAS.imaging")
            case _:
                flag = True

        return flag

    def build_ingest_parameters(self, additional_metadata: Optional[AbstractTextFile]):
        """
        Make the "parameters" section of the manifest

        :return:
        """
        if self.sp_type not in [
            ScienceProductType.EVLA_CAL,
            ScienceProductType.IMAGE,
            ScienceProductType.VLASS_SECI,
            ScienceProductType.EXEC_BLOCK,
            ScienceProductType.VLASS_COARSE,
        ]:
            raise NotImplementedError()

        staging_source_dir = self.staging_source_dir

        # Change the staging_source_dir to the parent if it's not curation and if it's an EVLA EB.
        # This is because SDM ingestion is expecting to work upon a directory given as the product name
        # with an expected set of files (XML files and others) underneath.
        # We need to pass that directory's parent. for example, We should pass /lustre/.../workspaces/sdm-staging
        # instead of /lustre/.../workspaces/sdm-staging/19A-001.sb1234567890.eb233423545632.54321.894327984569.
        if all(
            [
                not self.reingest,  # Not curation
                self.telescope == Telescope.EVLA,
                self.sp_type == ScienceProductType.EXEC_BLOCK,
            ]
        ):
            staging_source_dir = self.staging_source_dir.parent

        params = ManifestParameters(
            telescope=self.telescope,
            ngas_ingest=self.ngas_ingest,
            staging_source_dir=staging_source_dir,
            additional_metadata=additional_metadata,
        )

        if self.reingest is not None:
            params.ngas_ingest = False
        return params

    def __eq__(self, other):
        if isinstance(other, IngestionManifest):
            return (
                other.parameters == self.parameters
                and other.input_group == self.input_group
                and other.output_group == self.output_group
            )

        return False


class IngestionManifestSchema(BaseSchema):
    """
    Schema to validate and generate JSON from an IngestionManifest object.
    This is done by making an instance of this class and then calling .dumps() on it with an IngestionManifest object.
    """

    reingest = fields.Nested(ReingestGroupSchema(), data_key=IngestionManifestKey.REINGEST.value, allow_none=True)
    input_group = fields.Nested(InputGroupSchema(), data_key=IngestionManifestKey.INPUT_GROUP.value, allow_none=True)
    output_group = fields.Nested(OutputGroupSchema(), data_key=IngestionManifestKey.OUTPUT_GROUP.value, allow_none=True)
    parameters = fields.Nested(ManifestParametersSchema(), data_key=IngestionManifestKey.PARAMETERS.value)


def write(manifest: IngestionManifest, schema: IngestionManifestSchema) -> Path:
    """
    Write the manifest .json file.

    :return:
    """
    if manifest.reingest is not None:
        output_path = manifest.destination_dir / CURATOR_MANIFEST_FILENAME
    else:
        output_path = manifest.destination_dir / MANIFEST_FILENAME

    to_write = schema.dumps(manifest, indent=4)
    with open(output_path, "w") as out:
        out.write(to_write)

    return output_path


class IngestionManifestBuilder:
    """Builds ingestion manifest and associated files from files in ingestion_path"""

    def __init__(
        self,
        manifest_source_dir: Path | None,  # partial curation has no source path
        sp_type: str,
        telescope: str,
        locator: Optional[str] = None,  # initial observation ingestion, and full curation, have no locator
        product_group: Optional[int] = None,  # used for full curation
        additional_metadata: Optional[AbstractTextFile] = None,
        filename: Optional[str] = None,
        # (curator_type, target_list)
        curate: Optional[tuple[CuratorType, Optional[list[str]]]] = None,
        # If None, self.manifest_destination_dir = manifest_source_dir
        manifest_destination_dir: Optional[Path] = None,
        # alternative to staging_source_dir and curation_source for specifying source files
        file_list: Optional[list[str]] = None,
        input_group_locator: Optional[str] = None,
    ):
        # get the telescope
        self.telescope = Telescope(telescope)
        if self.telescope is None:
            raise ValueError("telescope is required.")

        # get the directory containing the files to be ingested
        self.manifest_source_dir = manifest_source_dir

        # get the additional metadata, if any (EVLA CAL and EB ingestion will have none)
        self.additional_metadata = additional_metadata

        # are we running curator?
        self.curation_type = curate[0] if curate is not None else None

        # determine the science product type
        self.sp_type = ScienceProductType(sp_type)
        if self.sp_type is None or not isinstance(self.sp_type, ScienceProductType):
            raise ValueError(f"Invalid/unknown science product type: {sp_type}")
        if self.sp_type not in [
            ScienceProductType.EVLA_CAL,
            ScienceProductType.IMAGE,
            ScienceProductType.VLASS_SECI,
            ScienceProductType.VLASS_COARSE,
            ScienceProductType.EXEC_BLOCK,
        ]:
            raise NotImplementedError(f"Don't know yet how to build a {self.sp_type.value} manifest")

        if self.curation_type is not None:
            if self.curation_type not in [CuratorType.PARTIAL, CuratorType.FULL]:
                raise NotImplementedError(
                    f"Don't know how to build a {self.curation_type.value} curation {self.sp_type} manifest"
                )
            self.target_list = curate[1]
            if self.curation_type == CuratorType.FULL and self.target_list:
                raise IngestionManifestException(
                    f"Expected targets=None for full curation, got targets={self.target_list}"
                )

        # we are not running observation ingestion, use a locator
        self.locator = locator
        self.product_group = product_group

        self.input_group_locator = input_group_locator
        if self.input_group_locator is None:
            self.input_group_locator = locator

        # directory to put the manifest in
        self.manifest_destination_dir = manifest_destination_dir
        if self.manifest_destination_dir is None:
            self.manifest_destination_dir = manifest_source_dir
        if self.manifest_destination_dir is None:
            raise IngestionManifestException(
                f"IngestionManifestBuilder: Expected a directory to house the manifest, found instead staging_source_dir={self.manifest_source_dir}, manifest_destination_dir={self.manifest_destination_dir}"
            )

        if file_list:
            self.files_found = [Path(file) for file in file_list]
        elif self.manifest_source_dir is not None:
            # we are not running partial curation, look for files
            self.files_found = [file for file in self.manifest_source_dir.iterdir()]
            if len(self.files_found) == 0:
                raise IngestionManifestException(f"No ingestion files found at {manifest_source_dir}")
        # Partial curation doesn't require source files, it can fetch them from productfetcher via their SPLs

        self.filename = filename

    def build(self) -> tuple[IngestionManifest, Path]:
        """
        Using only -relevant- files in ingestion_path, write the manifest
        and produce other files required for ingestion.

        :return: the ingestion manifest and the file containing its JSON
        """
        manifest, manifest_path = (None, None)
        match (self.curation_type, self.sp_type):
            case (CuratorType.PARTIAL, _):
                manifest, manifest_path = self._build_partial_curation_manifest()
            case (None | CuratorType.FULL, ScienceProductType.IMAGE | ScienceProductType.VLASS_SECI):
                manifest, manifest_path = self._build_manifest_with_input_group_and_locator(
                    self._build_imaging_output_group, self.curation_type == CuratorType.FULL
                )
            case (None, ScienceProductType.EVLA_CAL):
                manifest, manifest_path = self._build_manifest_with_input_group_and_locator(
                    self._build_evla_cal_output_group
                )
            case (None | CuratorType.FULL, ScienceProductType.EXEC_BLOCK):
                manifest, manifest_path = self._build_observation_manifest(self.curation_type == CuratorType.FULL)
            case (None, ScienceProductType.VLASS_COARSE):
                manifest, manifest_path = self._build_manifest_with_input_group_and_locator(
                    self._build_vlass_coarse_output_group
                )
            case (a, b):
                raise NotImplementedError(
                    f"Don't know yet how to build a CuratorType={a.value if a is not None else a}, ScienceProductType={b.value} manifest"
                )
        return manifest, manifest_path

    def _build_partial_curation_manifest(self) -> tuple[IngestionManifest, Path]:
        """
        Build a manifest for partial curator
        Partial curation is simple, only parameters and reingest groups are needed

        :return: the manifest file object and path
        """
        manifest = IngestionManifest(
            telescope=self.telescope,
            staging_source_dir=self.manifest_source_dir,
            locator=self.locator,
            sp_type=self.sp_type,
            reingest_group=self._build_reingest_group(),
            input_group=None,
            output_group=None,
            destination_dir=self.manifest_destination_dir,
        )
        schema = IngestionManifestSchema()
        manifest_file = write(manifest, schema)
        return manifest, manifest_file

    def _build_observation_manifest(self, is_full_curation: bool = False) -> tuple[IngestionManifest, Path]:
        reingest_group = self._build_reingest_group() if is_full_curation else None

        # create the manifest
        manifest = IngestionManifest(
            telescope=self.telescope,
            locator=None,
            sp_type=self.sp_type,
            staging_source_dir=self.manifest_source_dir,
            reingest_group=reingest_group,
            input_group=None,
            output_group=self._build_observation_output_group(),
            filename=self.filename,
            destination_dir=self.manifest_destination_dir,
        )

        if not manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products = []
        artifacts_ap = self._build_artifacts_product()
        if artifacts_ap not in manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products.append(artifacts_ap)

        schema = IngestionManifestSchema()
        manifest_file = write(manifest, schema)

        self.write_ingestion_artifacts_tar(self.manifest_destination_dir / artifacts_ap.filename)

        return manifest, manifest_file

    def _find_init_weblog_if_any(self):
        """
        Is there an initial weblog in the staging source dir?
        (If so, this calibration is v2 or higher)

        :return: initial weblog, if any
        """
        for file in [file for file in self.files_found]:
            if file.name == INIT_WEBLOG_FILENAME:
                return file

        return None

    def _build_manifest_with_input_group_and_locator(
        self, output_group_builder: Callable[[], OutputGroup], is_full_curation: bool = False
    ) -> tuple[IngestionManifest, Path]:
        """
        Builds a manifest with the provided callable for building the output group along with any additional_metadata.

        :param output_group_builder: Builder for this manifest's OutputGroup.
        :return: A manifest and the path to where it's been written to storage
        """
        reingest_group = self._build_reingest_group() if is_full_curation else None

        # create the manifest
        manifest = IngestionManifest(
            telescope=self.telescope,
            locator=self.locator,
            additional_metadata=self.additional_metadata,
            sp_type=self.sp_type,
            staging_source_dir=self.manifest_source_dir,
            reingest_group=reingest_group,
            input_group=self._build_input_group(),
            output_group=output_group_builder(),
            destination_dir=self.manifest_destination_dir,
        )

        if is_full_curation:
            # find the existing artifacts tar file name
            artifacts_ap = self._find_existing_record(self.files_found)
        else:
            # build new artifacts tar
            artifacts_ap = self._build_artifacts_product()

        if artifacts_ap not in manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products.append(artifacts_ap)
        schema = IngestionManifestSchema()
        manifest_file = write(manifest, schema)
        if not is_full_curation:
            # we're running initial ingestion, make a new artifacts tar
            self.write_ingestion_artifacts_tar(self.manifest_destination_dir / artifacts_ap.filename)

        return manifest, manifest_file

    def _build_reingest_group(self) -> ReingestGroup:
        """
        Create the curation reingest group

        :return:
        """
        return ReingestGroup(
            # make type consistent for later validation
            locator=str(self.product_group) if self.product_group else self.locator,
            product_type=self.sp_type,
            targets=self.target_list,
            # product_group=self.product_group,
        )

    def _build_input_group(self) -> InputGroup:
        """
        Create the input group using the parameters.

        :return:
        """

        # N.B. this is sufficient for most types of ingestion,
        # but ALMA CALs will have multiple EB SPs, identified only by locator,
        # and VLBAs have no input group at all.
        sp_in = InputScienceProduct(locator=self.input_group_locator)

        return InputGroup([sp_in])

    def _build_evla_cal_output_group(self) -> OutputGroup:
        """
        Create EVLA standard calibration manifest output group using the parameters
        and the contents of the staging dir.

        :return: an output group, if a science product is found
        """

        # find science product (we expect just one for this SP type)
        tars_found = find_output_tars(self.files_found, self.manifest_source_dir)

        sci_prod = None
        for file in tars_found:
            sci_prod = OutputScienceProduct(type=self.sp_type, filename=file.name)
            break

        if sci_prod:
            weblog_files = find_weblogs(self.files_found, self.manifest_source_dir)
            weblogs = []
            for file in weblog_files:
                ap = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG, filename=file.name)
                if ap not in weblogs:
                    weblogs.append(ap)
        else:
            raise NoScienceProductException(f">>> NO SCIENCE PRODUCT FOUND in {self.manifest_source_dir}")

        return OutputGroup(science_products=[sci_prod], ancillary_products=weblogs)

    def _build_vlass_coarse_output_group(self) -> OutputGroup:
        """
        Create VLASS coarse manifest output group using the parameters
        and the contents of the staging dir.

        :return: OutputGroup. AN output group with the retrieved science and ancillary products passed in.
        """
        products_finder = VlassCoarseCubeIngestionProductsFinder(self.files_found)
        science_products = products_finder.output_science_products
        ancillary_products = products_finder.ancillary_products

        return OutputGroup(science_products=science_products, ancillary_products=ancillary_products)

    def _build_imaging_output_group(self) -> OutputGroup:
        """
        Create imaging manifest output group using the parameters
        and the contents of the staging dir.

        This also includes any spectral-cube images.

        :return:
        """

        products_finder = ImageIngestionProductsFinder(self.files_found, self.sp_type)
        science_products = products_finder.output_science_products
        ancillary_products = products_finder.ancillary_products

        # Search for spectral cube science products + ancillaries and add them to the output group lists
        spectral_cube_products_finder = SpectralCubeIngestionProductsFinder(self.files_found, self.sp_type)
        science_products.extend(spectral_cube_products_finder.output_science_products)

        return OutputGroup(science_products=science_products, ancillary_products=ancillary_products)

    def _build_observation_output_group(self) -> OutputGroup:
        """
        Create observation manifest output group using the parameters
        and the contents of the staging dir.

        :return:
        """

        products_finder = ObservationIngestionProductsFinder(self.files_found, self.sp_type)
        science_products = products_finder.output_science_products
        ancillary_products = products_finder.ancillary_products

        return OutputGroup(science_products=science_products, ancillary_products=ancillary_products)

    @staticmethod
    def _build_artifacts_filename() -> str:
        """
        Build unique manifest filename in standard format.

        :return: the filename
        """
        current_time = arrow.now()
        timestamp = format_timestamp(current_time)
        return f"{INGESTION_ARTIFACTS_NAME}{timestamp}{TARFILE_EXT}"

    @staticmethod
    def _build_artifacts_product() -> AncillaryProduct:
        return AncillaryProduct(
            AncillaryProductType.INGESTION_ARTIFACTS, IngestionManifestBuilder._build_artifacts_filename()
        )

    @staticmethod
    def _find_existing_record(available_files: list) -> AncillaryProduct:
        """
        Find an existing artifacts tar for curation

        :param available_files: existing files list
        :return: Ancillary Product of existing tar
        """
        for file in available_files:
            if INGESTION_ARTIFACTS_NAME in file.name:
                return AncillaryProduct(AncillaryProductType.INGESTION_ARTIFACTS, file.name)

    def write_ingestion_artifacts_tar(self, artifacts_path: Path) -> tarfile.TarFile:
        """
        Take the list of files and write a tar file for inclusion in the archive.
        This happens in the staging area for ingestion.
        The EVLA CAL tar will contain just the manifest.

        :param artifacts_path: Path to create the resulting tar file at
        :return: tar file of the ingestion artifacts
        """

        addl_md_file = None
        if self.additional_metadata:
            # find the additional metadata
            addl_md_filename = self.additional_metadata.filename
            addl_md_file = self.manifest_destination_dir / addl_md_filename

        with tarfile.open(artifacts_path, "w") as ingestion_artifacts_tar:
            if addl_md_file:
                ingestion_artifacts_tar.add(addl_md_file)

            # The manifest file itself is considered an ingestion artifact.
            # (It's turtles all the way down.)
            manifest_file = self.manifest_destination_dir / MANIFEST_FILENAME
            if not manifest_file.exists():
                manifest_file = self.manifest_destination_dir / CURATOR_MANIFEST_FILENAME
            if not manifest_file.exists():
                raise FileNotFoundError(
                    f"No manifest (i.e. {MANIFEST_FILENAME} or {CURATOR_MANIFEST_FILENAME}) found in {self.manifest_destination_dir}"
                )
            ingestion_artifacts_tar.add(manifest_file)

        return ingestion_artifacts_tar


def format_timestamp(datetime: arrow.Arrow) -> str:
    """
    Format the current time as
        2021_07_01T13_49_17.237

    :param datetime: current timestamp
    :return: timestamp suitable for ingestion manifest filename
    """

    return datetime.format("YYYY_MM_DDThh_mm_ss.SSS")


def find_manifest(ingestion_path: Path) -> Path:
    """
    Find the ingestion manifest at this ingestion path.

    :param ingestion_path: home of ingestion files
    :return:
    """
    for json_file in ingestion_path.glob(MANIFEST_FILENAME):
        return json_file

    raise FileNotFoundError(f"No ingestion manifest found at {ingestion_path}")
