#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Interfaces for ingest_envoy
"""
import abc
from abc import ABC

from ingest_envoy.schema import AbstractTextFile


class LauncherIF(ABC):
    """
    Generic Ingestion Launcher methods.
    Should be implemented for any type of ingestion launcher
    """

    # launch all ingest steps, should be called in ingest.main
    @abc.abstractmethod
    def launch(self) -> (int, bool):
        """
        Prepare for and run ingest

        :return: (return code from ingest, whether to run the indexer)
        """
        pass

    # setup workflow results for ingestion, ensure placement in staging area
    @abc.abstractmethod
    def prepare_for_launch(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        pass

    # create ingestion manifest
    @abc.abstractmethod
    def create_manifest(self, additional_file: AbstractTextFile = None):
        """
        Create ingestion manifest

        :param additional_file: AbstractTextFile containing any addition metadata required for ingestion
        :return:
        """
        pass


class CollectorIF(ABC):
    """
    Generic Collection methods.
    """

    @abc.abstractmethod
    def collect_products(self):
        """
        Generic method for running product collection scripts

        :return:
        """
        pass
