#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import logging
import re
import subprocess
import sys
from pathlib import Path
from subprocess import PIPE, STDOUT, CalledProcessError, CompletedProcess, Popen
from typing import Any, Union

from ingest_envoy.collectors import (
    CoarseCollector,
    ImageCollector,
    ObservationCollector,
    SECICollector,
    collect_image_metadata,
)
from ingest_envoy.ingestion_manifest import IngestionManifestBuilder
from ingest_envoy.interfaces import LauncherIF
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import CuratorType, IngestType, VLASSIngestType

logger = logging.getLogger("ingest_envoy")


def get_ingest_spls(subprocess_stdout: str) -> list[str]:
    """
    :param subprocess_stdout: The standard output of the ingestion pex.
    :return: The product locators found in the passed-in standard output.
    """
    # OSPs: output science products
    # First, find the line that starts with "OSPs: "
    osps_line_match = re.search(r"^OSPs: \[.*\]$", subprocess_stdout, re.MULTILINE)

    if not osps_line_match:
        return []

    # Get the full OSPs line
    osps_line = osps_line_match.group(0)

    # Extract all SPLs from that line
    locators = re.findall(r"uid://[a-z]+/[a-z_]+/[a-z0-9-]+", osps_line)

    return locators


def run_process_with_realtime_logging_and_search(
    args: list[str],
    search_method: Any | None,
    stdout_handler=logging.info,
    check=True,
    text=True,
    stdout=PIPE,
    stderr=STDOUT,
    **kwargs,
) -> (CompletedProcess, list):
    """
    Run process like subprocess.run() but with realtime logging and log parsing.
    This solution is based on the following stackoverflow answer: https://stackoverflow.com/a/76626021

    :param args: argument for the process to run.
    :param search_method: method to use for log searching. Optional
    :param stdout_handler: where to send the log output. Defaults to logging.info.
    :param check: whether to raise an exception if the process returns a non-zero return code. Defaults to True.
    :param text: whether to decode the output as text. Defaults to True.
    :param stdout: how to capture the stdout. Defaults to PIPE.
    :param stderr: how to capture the stderr. Defaults to STDOUT.
    :param kwargs: additional keyword arguments to pass to subprocess.Popen().
    :return: CompletedProcess object and search result (if any)
    """

    search_result = []

    # Mimic subprocess.run while processing command output in realtime with log searching if needed.
    with subprocess.Popen(args, stdout=stdout, stderr=stderr, text=text, **kwargs) as process:
        for line in process.stdout:
            search_result.extend(search_method(line)) if search_method else None
            stdout_handler(line.rstrip())
    return_code = process.wait()
    if check and return_code:
        raise CalledProcessError(return_code, process.args)
    return CompletedProcess(process.args, return_code), search_result


def trigger_ingest(
    real_ingest: bool, staging_dir: str, bin_dir: str = ".", logger: logging.Logger = logger
) -> (int, list[str]):
    """
    Run ingest

    :param real_ingest: real ingestion vs. testing flag
    :param staging_dir: staging directory to ingest from
    :param bin_dir: directory containing the ingestion utility
    :param logger: logger to use for logging messages. Defaults to the base logger for this module.
    :return: return code, list of SPLs generated by ingest.
    """
    if real_ingest:
        ingest_args = [f"{bin_dir}/ingest", "--json", "-p", staging_dir]
        logger.info(f"Calling the ingest pex with arguments: {' '.join(ingest_args)}")

        ingest_process, generated_spls = run_process_with_realtime_logging_and_search(
            ingest_args,
            search_method=get_ingest_spls,
            stdout_handler=logger.info,
        )

        return ingest_process.returncode, generated_spls
    else:
        # For local testing
        return 0, []


def trigger_curator(real_ingest: bool, manifest_dir: Path, bin_dir: str = ".", logger: logging.Logger = logger) -> int:
    """
    Run curator

    :param real_ingest: real curation or testing only?
    :param manifest_dir: The directory containing the manifest
    :param bin_dir: directory containing the curator utility
    :param logger: logger to use for logging messages. Defaults to the base logger for this module.
    :return: return code
    """

    if real_ingest:
        curator_args = [f"{bin_dir}/curator", "-m", f"{manifest_dir / 'manifest.json'}"]
        logger.info(f"Calling curator with arguments: {' '.join(curator_args)}")
        curator_process = subprocess.run(
            curator_args,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        return curator_process.returncode
    else:
        # for local testing
        return 0


class IngestCalibrationLauncher(LauncherIF):
    """Setup and Launch Calibration Ingestion"""

    def __init__(self, arg_type: IngestType, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = arg_type.value.__str__()
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]

    def launch(self) -> (int, bool):
        """
        Prepare for and run ingest

        :return: (return code from ingest, whether to run the indexer)
        """
        self.logger.info("RUNNING CALIBRATION INGESTION!")
        self.prepare_for_launch()

        self.logger.info("Running ingest!")
        ingest_returncode, _ = trigger_ingest(
            self.parameters["useIngest"], self.staging_source_dir, self.parameters["script_location"], self.logger
        )
        return (
            ingest_returncode,
            True,
        )

    def prepare_for_launch(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. run collection script to create calibration tarfile
        self.run_collection_script()

        # 2. create ingestion manifest
        self.create_manifest()

    def run_collection_script(self):
        """
        Run calibration collection script, creates collection tars
        ex. caltables, weblogs, pipeline-artifacts, etc

        If this was a multi-version request, also package up the initial version weblog

        :return:
        """
        self.logger.info("Collecting calibration tables for staging...")
        workflow_dir = self.parameters["workflowDir"]
        sbin_dir = self.parameters["script_location"]
        sdm_id = self.parameters["sdmId"]
        cal_processing_datetime = self.parameters["processingStart"]

        # run script
        if self.parameters["multiVersion"] is False:
            collector = subprocess.run(
                [f"{sbin_dir}/calibration-table-collector.sh", workflow_dir, sdm_id, cal_processing_datetime],
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
        else:
            initial_version_dir = self.parameters["initialVersionDir"]

            collector = subprocess.run(
                [
                    f"{sbin_dir}/calibration-table-collector.sh",
                    workflow_dir,
                    sdm_id,
                    cal_processing_datetime,
                    initial_version_dir,
                ],
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
        if collector.returncode != 0:
            self.logger.error("ERROR: Calibration product collection failed!")
            exit(1)

    def create_manifest(self, additional_file=None):
        """
        Create the calibration ingestion manifest

        :param additional_file: Unneeded for calibration ingestion
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        spl = self.parameters["spl"]
        telescope = self.parameters["telescope"]

        IngestionManifestBuilder(
            manifest_source_dir=Path(self.staging_source_dir),
            sp_type=self.sci_product_type,
            locator=spl,
            telescope=telescope,
        ).build()


class IngestVLASSLauncher(LauncherIF):
    """Setup and Launch VLASS Caching and Ingestion"""

    def __init__(self, arg_type: VLASSIngestType, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = arg_type.value.__str__()
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]
        self.qa_passed = parameters["qa_status"] == "PASS"

        match arg_type:
            case VLASSIngestType.COARSE:
                self.collector = CoarseCollector(self.parameters)
            case _:
                raise NotImplementedError(f"Error: {arg_type.value} is not currently supported.")

    def launch(self) -> (int, bool):
        """
        Prepare and run ingestion script

        :return: Return code from ingestion script process
        """
        self.logger.info(f"RUNNING {self.sci_product_type} CACHING AND INGESTION!")
        self.prepare_for_launch()

        if self.qa_passed:
            self.logger.info("Running ingest!")
            ingest_returncode, _ = trigger_ingest(
                self.parameters["useIngest"], self.staging_source_dir, self.parameters["script_location"], self.logger
            )
            return (
                ingest_returncode,
                True,
            )

        # In the case that we didn't run ingest for a coarse cube due to QA fail
        # Return 0 because this is expected behavior
        self.logger.info("Not an error; ingest not run for a coarse cube due to QA fail.")
        return 0, False

    def prepare_for_launch(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. collect products and auxiliary image metadata
        aux_file = self.run_collector()
        # 2. create manifest
        if self.qa_passed:
            self.create_manifest(additional_file=aux_file)

    def run_collector(self) -> AbstractTextFile:
        """
        Run ImageCollector

        :return: AbstractTextFile containing additional metadata needed for manifest creation
        """
        # 1. collect products to staging area
        self.collector.collect_products()
        # 2. create auxiliary metadata file and return
        return collect_image_metadata()

    def create_manifest(self, additional_file: AbstractTextFile = None):
        """
        Create VLASS cube ingestion manifest

        :param additional_file: image ingestion requires additional metadata
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        # Input product locator
        precursor_spl = self.parameters["precursorSpl"]
        telescope = self.parameters["telescope"]
        additional_metadata = additional_file

        IngestionManifestBuilder(
            manifest_source_dir=Path(self.staging_source_dir),
            sp_type=self.sci_product_type,
            locator=precursor_spl,
            telescope=telescope,
            additional_metadata=additional_metadata,
        ).build()


class IngestImageLauncher(LauncherIF):
    """Setup and Launch Image Ingestion"""

    def __init__(self, arg_type: Union[IngestType, VLASSIngestType], parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = arg_type.value.__str__()
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]

        match arg_type:
            case VLASSIngestType.SECI:
                self.collector = SECICollector(self.parameters)
            case IngestType.IMG:
                self.collector = ImageCollector(self.parameters)
            case _:
                raise NotImplementedError(f"Error: {arg_type.value} is not currently supported for an image launcher.")

    def launch(self) -> (int, bool):
        """
        Prepare and run ingestion script

        :return: Return code from ingestion script process
        """
        self.logger.info(f"RUNNING {self.sci_product_type} INGESTION!")
        self.prepare_for_launch()

        self.logger.info("Running ingest!")
        ingest_returncode, _ = trigger_ingest(
            self.parameters["useIngest"], self.staging_source_dir, self.parameters["script_location"], self.logger
        )
        return (
            ingest_returncode,
            True,
        )

    def prepare_for_launch(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. collect products and auxiliary image metadata
        aux_file = self.run_collector()
        # 2. create manifest
        self.create_manifest(additional_file=aux_file)

    def run_collector(self) -> AbstractTextFile:
        """
        Run ImageCollector

        :return: AbstractTextFile containing additional metadata needed for manifest creation
        """
        # 1. collect products to staging area
        self.collector.collect_products()
        # 2. create auxiliary metadata file and return
        return collect_image_metadata()

    def create_manifest(self, additional_file: AbstractTextFile = None):
        """
        Create image ingestion manifest

        :param additional_file: image ingestion requires additional metadata
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        # needs to be the CAL locator!!
        spl = self.parameters["calSpl"]
        telescope = self.parameters["telescope"]
        additional_metadata = additional_file

        IngestionManifestBuilder(
            manifest_source_dir=Path(self.staging_source_dir),
            sp_type=self.sci_product_type,
            locator=spl,
            telescope=telescope,
            additional_metadata=additional_metadata,
        ).build()


class IngestObservationLauncher(LauncherIF):
    """Setup and Launch Observation Ingestion"""

    def __init__(self, arg_type: IngestType, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = arg_type.value.__str__()
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]
        self.collector = ObservationCollector(self.parameters)

    def launch_curator_for_spls(self, spls: list[str]) -> dict[str, int]:
        """
        :param spls: The locators for which we are going to run curator.
        :return: The dictionary mapping SPLs where curator failed to the return values from curator.
        """
        parameters = {
            **self.parameters,
            "product_type": "execution_block",
            "input_group_locator": None,
            "curation_source": Path(self.staging_source_dir),
            "file_list": None,
            "product_group_id": None,
            "target_list": [],
        }
        launcher = CuratorLauncher(CuratorType.PARTIAL, parameters)
        curator_results = {}

        for spl in spls:
            launcher.parameters["spl"] = spl
            trigger_curator_returncode, _ = launcher.launch()
            if trigger_curator_returncode != 0:
                curator_results[spl] = trigger_curator_returncode

        return curator_results

    def launch(self) -> (int, bool):
        """
        Prepare for and run ingest

        :return: (return code from ingest, whether to run the indexer)
        """
        self.logger.info("RUNNING OBSERVATION INGESTION!")
        self.prepare_for_launch()

        self.logger.info("Running ingest!")
        ingest_returncode, generated_spls = trigger_ingest(
            self.parameters["useIngest"], self.staging_source_dir, self.parameters["script_location"], self.logger
        )

        if ingest_returncode == 0:
            curator_results = self.launch_curator_for_spls(generated_spls)
            if curator_results:
                self.logger.error("Curator failed: %s", curator_results)

        return (
            ingest_returncode,
            True,
        )

    def prepare_for_launch(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. run collection script to move observation data to the staging area
        self.run_collector()

        # 2. create ingestion manifest
        self.create_manifest()

    def run_collector(self):
        """
        Run ObservationCollector
        """
        # 1. collect products to staging area
        self.collector.collect_products()

    def create_manifest(self, additional_file=None):
        """
        Create the observation ingestion manifest

        :param additional_file: Unneeded for observation ingestion
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        telescope = self.parameters["telescope"]
        filename = self.parameters["filename"]
        IngestionManifestBuilder(
            manifest_source_dir=Path(self.staging_source_dir),
            sp_type=self.sci_product_type,
            locator=None,
            telescope=telescope,
            filename=filename,
        ).build()


class CuratorLauncher(LauncherIF):
    """Setup and Launch Curator"""

    def __init__(self, arg_type: CuratorType, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.curator_type = arg_type
        self.sci_product_type = parameters["product_type"]
        self.curation_source = parameters["curation_source"]
        self.manifest_destination_dir = Path.cwd()
        self.parameters = parameters

    def prepare_for_launch(self):
        """
        Prepare for curation, manifest building

        :return:
        """
        self.logger.info("Preparing for curator...")
        #  no collection needed for curation, manifest building only
        self.create_manifest()

    def create_manifest(self, additional_file: AbstractTextFile = None):
        """
        Create the curation manifest

        :param additional_file: Unneeded for curation
        :return:
        """
        self.logger.info("Creating curation manifest...")
        spl = self.parameters["spl"]
        telescope = self.parameters["telescope"]
        curation_source = self.parameters["curation_source"]
        target_list = self.parameters["target_list"]
        file_list = self.parameters["file_list"]
        product_group = self.parameters["product_group_id"]
        input_group_locator = self.parameters["input_group_locator"]

        IngestionManifestBuilder(
            manifest_source_dir=curation_source,
            sp_type=self.sci_product_type,
            locator=spl,
            telescope=telescope,
            curate=(self.curator_type, target_list),
            manifest_destination_dir=self.manifest_destination_dir,
            file_list=file_list,
            product_group=product_group,
            input_group_locator=input_group_locator,
        ).build()

    def launch(self) -> (int, bool):
        """
        Prepare for and run ingest

        :return: (return code from ingest, whether to run the indexer)
        """
        self.logger.info(f"RUNNING {self.sci_product_type.upper()} CURATION!")
        self.prepare_for_launch()

        self.logger.info("Running curator!")
        return (
            trigger_curator(
                self.parameters["useIngest"], self.manifest_destination_dir, self.parameters["script_location"]
            ),
            True,
        )
