#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Pieces of any ingestion manifest """

import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional, Union

from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType, Telescope
from marshmallow import Schema, ValidationError, fields
from marshmallow.decorators import post_dump, validates

MANIFEST_FILENAME = "ingestion-manifest.json"
CURATOR_MANIFEST_FILENAME = "manifest.json"
INGESTION_ARTIFACTS_NAME = "ingestion_artifacts_"
TARFILE_EXT = ".tar"
WEBLOG_FILENAME = "weblog.tgz"
INIT_WEBLOG_FILENAME = "initial_weblog.tgz"
REIMAGING_FILENAME = "reimaging_resources.tgz"
PARAMETER_FILENAME = "SEIP_parameter.list"
SELFCAL_AUX_FILENAME = "selfcal.auxproducts.tgz"
SCIENCE_PRODUCT_PATTERN = re.compile("[a-zA-Z0-9._\\-+]*\\.tar")
JSON = Union[int, float, str, list["JSON"], dict[str, "JSON"]]


class IngestionManifestKey(Enum):
    """Sections we expect to see in a manifest"""

    PARAMETERS = "parameters"
    INPUT_GROUP = "input_group"
    OUTPUT_GROUP = "output_group"
    SCIENCE_PRODUCTS = "science_products"
    PRIMARY_PRODUCTS = "primary_products"
    ANCILLARY_PRODUCTS = "ancillary_products"
    REINGEST = "reingest"


class ParamsKey(Enum):
    """Keys in ManifestParameters"""

    PARAMETERS = "parameters"
    TELESCOPE = "telescope"
    REINGEST = "reingest"
    NGAS_INGEST = "ngas_ingest"
    INGESTION_PATH = "ingestion_path"
    ADDITIONAL_METADATA = "additional_metadata"
    COLLECTION_METADATA = "additional_metadata"  # needed for VLASS, realfast, elwa, alfalfa. Not yet implemented


class ReingestKey(Enum):
    """Keys in ReingestGroup"""

    TARGETS = "targets"
    LOCATOR = "locator"
    TYPE = "type"
    PRODUCT_GROUP = "product_group"


class ProductTypeField(fields.Field):
    """A generic product type field that allows either a ScienceProductType or AncillaryProductType"""

    def _serialize(self, value, attr, data, **kwargs):
        if not isinstance(value, ScienceProductType) and not isinstance(value, AncillaryProductType):
            raise ValidationError("Serializing product, not a ScienceProductType or AncillaryProductType")
        return value.value

    def _deserialize(self, value, attr, data, **kwargs):
        if value not in [sp.value for sp in ScienceProductType] and value not in [
            ap.value for ap in AncillaryProductType
        ]:
            raise ValidationError("Deserializing product, not a ScienceProductType or AncillaryProductType")
        return value


class BaseSchema(Schema):
    @post_dump
    def remove_none_values(self, data, **kwargs):
        return {key: value for key, value in data.items() if value is not None}

    @post_dump
    def post_dump_validate(self, data, **kwargs):
        # Validation only happens automatically on load, let's also do it on dump
        errors = self.validate(data)
        if errors:
            raise ValidationError(errors)

        return data


@dataclass
class ManifestParameters:
    """Represents "parameters" section of ingestion manifest.
    ASSUMPTIONS:
    * "ngas_ingest" is always True (per our testing examples)
    """

    telescope: Telescope
    ngas_ingest: bool
    staging_source_dir: Optional[Path]
    additional_metadata: Optional[AbstractTextFile] = None


class ManifestParametersSchema(BaseSchema):
    """Schema to generate and validate a ManifestParameters object"""

    ngas_ingest = fields.Bool(data_key=ParamsKey.NGAS_INGEST.value, required=True)
    telescope = fields.Enum(Telescope, by_value=True, data_key=ParamsKey.TELESCOPE.value, required=True)
    staging_source_dir = fields.Str(data_key=ParamsKey.INGESTION_PATH.value, allow_none=True)
    additional_metadata = fields.Str(data_key=ParamsKey.ADDITIONAL_METADATA.value, allow_none=True)

    @validates("staging_source_dir")
    def validate_staging_source_dir(self, staging_source_dir):
        # Check if the string is a valid path
        if staging_source_dir:
            try:
                Path(staging_source_dir).resolve(strict=True)
            except FileNotFoundError:
                raise ValidationError("Staging source directory must be an existing path")


@dataclass
class AncillaryProduct:
    """Represents an ancillary or primary product in an ingestion manifest"""

    type: AncillaryProductType
    filename: str
    science_associate: Optional[str] = None
    group_with: Optional[str] = None


class AncillaryProductSchema(BaseSchema):
    """Schema to generate and validate an AncillaryProductSchema object"""

    type = fields.Enum(AncillaryProductType, by_value=True, required=True)
    filename = fields.Str(required=True)
    science_associate = fields.Str(allow_none=True)
    group_with = fields.Str(allow_none=True)


@dataclass
class InputScienceProduct:
    """Simplest type of science product: has only a locator"""

    locator: str


class InputScienceProductSchema(BaseSchema):
    """Schema to generate and validate an InputScienceProduct object"""

    locator = fields.Str(required=True)


@dataclass
class InputGroup:
    """Generic ingestion manifest input group"""

    science_products: list[InputScienceProduct]


class InputGroupSchema(BaseSchema):
    """Schema to generate and validate an InputGroup object"""

    science_products = fields.List(
        fields.Nested(InputScienceProductSchema()),
        data_key=IngestionManifestKey.SCIENCE_PRODUCTS.value,
        allow_none=True,
    )

    @validates("science_products")
    def validate_science_products(self, science_products):
        # Ingestion manifests with empty input groups can cause errors on ingest
        if science_products and len(science_products) <= 0:
            raise ValidationError("The input group can be None but cannot be empty")


@dataclass
class OutputScienceProduct:
    """Generic science product contained in manifest output group"""

    type: Union[ScienceProductType, AncillaryProductType]
    filename: str
    ancillary_products: Optional[list[AncillaryProduct]] = None
    primary_products: Optional[list[AncillaryProduct]] = None


class OutputScienceProductSchema(BaseSchema):
    """Schema to generate and validate an OutputScienceProduct object"""

    type = ProductTypeField(required=True)
    filename = fields.Str(required=True)
    ancillary_products = fields.List(
        fields.Nested(AncillaryProductSchema()), data_key=IngestionManifestKey.ANCILLARY_PRODUCTS.value, allow_none=True
    )
    primary_products = fields.List(
        fields.Nested(AncillaryProductSchema()), data_key=IngestionManifestKey.PRIMARY_PRODUCTS.value, allow_none=True
    )


@dataclass
class OutputGroup:
    """Generic ingestion manifest output group"""

    science_products: list[OutputScienceProduct]
    ancillary_products: Optional[list[AncillaryProduct]] = None


class OutputGroupSchema(BaseSchema):
    """Schema to generate and validate an OutputGroup object"""

    science_products = fields.List(
        fields.Nested(OutputScienceProductSchema()), data_key=IngestionManifestKey.SCIENCE_PRODUCTS.value, required=True
    )
    ancillary_products = fields.List(
        fields.Nested(AncillaryProductSchema()), data_key=IngestionManifestKey.ANCILLARY_PRODUCTS.value, allow_none=True
    )


@dataclass
class ReingestGroup:
    """Generic manifest reingest group"""

    product_type: ScienceProductType
    locator: Optional[str] = None  # Replaced by product_group for full curation
    product_group: Optional[int] = None  # Replaced by locator for partial curation
    targets: Optional[list[str]] = None  # optional: presence determines Partial vs Full curation


class ReingestGroupSchema(BaseSchema):
    """Schema to generate and validate a ReingestGroup object"""

    product_type = fields.Enum(ScienceProductType, by_value=True, data_key=ReingestKey.TYPE.value, required=True)
    locator = fields.Str(data_key=ReingestKey.LOCATOR.value, allow_none=True)
    product_group = fields.Int(data_key=ReingestKey.PRODUCT_GROUP.value, allow_none=True)
    targets = fields.List(fields.Str(), data_key=ReingestKey.TARGETS.value, allow_none=True)
