#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Contains helper methods and product collector classes.
"""
import logging
import os
import re
import shutil
import tarfile
from pathlib import Path
from tarfile import TarInfo

from pycapo import CapoConfig

logger = logging.getLogger("ingest_envoy")

SELFCAL_REGEX = r"^.*selfcal\.auxproducts\.tgz"
WEBLOG_DIR_REGEX = r"^pipeline-.*"
WEBLOG_TAR_REGEX = r"^.*weblog\.tgz"
IMAGE_REGEX = r"^.*\.fits"


def find_in_dir(search_dir: Path, pattern: str) -> [Path]:
    """
    Finds files matching the given regex in the given directory and returns their paths in a list.

    :param search_dir: Directory to search in
    :param pattern: Pattern to search for
    :return: List of paths found
    """
    return [f for f in search_dir.iterdir() if re.match(pattern, f.name)]


def find_newest_in_dir(search_dir: Path, pattern: str) -> Path:
    """
    Finds the newest file/directory in the given directory matching the given pattern.

    :param search_dir: Directory to search in
    :param pattern: Pattern to search for
    :return: Path of the newest file
    """
    return max(find_in_dir(search_dir, pattern), key=lambda item: item.stat().st_ctime)


def copy_to_dir(dest_dir: Path, item: Path):
    """
    Copies the given file to the given directory.
    Will make a hard link when possible to save space, otherwise performs a copy.

    :param dest_dir: The path to copy to
    :param item: A path to copy/link into the given directory
    """
    # Check if we're cross-device to see if we can hard link (trying to hard link cross-devices will throw an OSError)
    if os.stat(item).st_dev == os.stat(dest_dir).st_dev:
        logger.info(f"Hard linking {item.name} to {dest_dir}")
        (dest_dir / item.name).hardlink_to(item)
    else:
        logger.info(f"Copying {item.name} to {dest_dir}")
        shutil.copy2(item, (dest_dir / item.name))


def exclude_artifacts(tarinfo: TarInfo) -> TarInfo | None:
    """
    Filter function for tarring artifacts that can be passed as an argument to tar.add().

    :param tarinfo: A TarInfo object provided by tar.add()
    :return: The same TarInfo object if it's to be included in the tarball, otherwise None.
    """
    exclude_artifacts_regexes = [SELFCAL_REGEX, WEBLOG_DIR_REGEX, WEBLOG_TAR_REGEX, IMAGE_REGEX]
    for pattern in exclude_artifacts_regexes:
        if re.match(pattern, tarinfo.name):
            return None
    return tarinfo


def tar_artifacts(source_dir: Path, artifacts_tar: Path, dest_dir: Path):
    """
    Tars the artifacts into the destination directory using the filename set in the constructor.
    """
    with tarfile.open(artifacts_tar, "w") as tar:
        for f in source_dir.iterdir():
            tar.add(f, filter=exclude_artifacts)
    copy_to_dir(dest_dir, artifacts_tar)


class StagingProductCollector:
    """
    Abstract product collector class to prepare a staging directory for metadata + NGAS ingestion.
    """

    SELFCAL_REGEX = r"^.*selfcal\.auxproducts\.tgz"
    WEBLOG_DIR_REGEX = r"^pipeline-.*"
    WEBLOG_TAR_REGEX = r"^.*weblog\.tgz"

    def __init__(self, workflow_dir: str, stage_dir: Path, artifacts_tar: str):
        """
        Class to link files from the given workflow's products directory to the staging directory.

        :param workflow_dir: The workflow directory name to append to spool and staging paths
        :param stage_dir: The full staging directory path as a string
        :param artifacts_tar: A name for the artifacts tarfile
        """
        self.logger = logging.getLogger("ingest_envoy")
        self.workflow_dir = workflow_dir
        self.spool_dir = Path(CapoConfig().getstring("edu.nrao.workspaces.ProcessingSettings.rootDirectory"))
        self.source_dir = self.spool_dir / self.workflow_dir / "products"
        self.stage_dir = stage_dir / self.workflow_dir
        self.artifacts_tar = self.source_dir / (artifacts_tar + ".tar")

    def __call__(self):
        """
        Takes the steps to produce the prepared staging directory for ingestion.
        """
        logger.info(f"Collecting products from {self.source_dir} into {self.stage_dir}")

        self.prepare()
        self.collect()
        self.cleanup()

    def prepare(self):
        """
        Prepares for product collection into the staging directory.
        """
        self.remove_staging()
        self.create_staging()

    def collect(self):
        """
        Links files from the products directory to the staging directory.
        """
        self.copy_weblog()
        self.copy_selfcal()
        tar_artifacts(self.source_dir, self.artifacts_tar, self.stage_dir)

    def cleanup(self):
        """
        Performs any cleanup after collection if necessary.
        """
        pass

    def find_in_source(self, pattern: str) -> [Path]:
        """
        Finds files matching the given regex in the source directory and returns their paths in a list.

        :param pattern: Pattern to search for
        :return: List of paths found
        """
        return find_in_dir(self.source_dir, pattern)

    def find_newest(self, pattern: str) -> Path:
        """
        Finds the newest file/directory in the source directory matching the given pattern.

        :param pattern: Pattern to search for
        :return: Path of the newest file
        """
        return find_newest_in_dir(self.source_dir, pattern)

    def remove_staging(self):
        """
        Deletes the staging directory if it exists.
        """
        if self.stage_dir.exists() and self.stage_dir.is_dir():
            logger.info("Staging directory existed, removing")
            shutil.rmtree(self.stage_dir)

    def create_staging(self):
        """
        Creates the staging directory (assumes it didn't exist).
        """
        try:
            self.stage_dir.mkdir(parents=True)
        except OSError:
            logger.error(f"Unable to make the staging directory at {self.stage_dir}, exiting")
            exit(1)

    def copy_to_staging(self, item: Path):
        """
        Makes a hard link in the staging directory pointing to the given item.

        :param item: A path to link in the staging directory
        """
        copy_to_dir(self.stage_dir, item)

    def copy_weblog(self):
        """
        Copies the most recent version of the weblog to the staging directory. This is done separate from the other
        artifacts since weblogs are considered a separate ancillary product.
        """
        # Check if an untarred weblog exists. If so, prefer it in case it's been updated by the DAs
        if self.find_in_source(WEBLOG_DIR_REGEX):
            logger.info("Replacing weblog tar")
            # Remove any old tarred weblogs
            for f in self.find_in_source(WEBLOG_TAR_REGEX):
                f.unlink()

            # Tar and zip the newest weblog
            with tarfile.open(self.source_dir / "weblog.tgz", "w:gz") as tar:
                tar.add(self.find_newest(WEBLOG_DIR_REGEX))

        # Copy the newest weblog tarfile
        if self.find_in_source(WEBLOG_TAR_REGEX):
            self.copy_to_staging(self.find_newest(WEBLOG_TAR_REGEX))
        else:
            # Ingest falls over if there's not a weblog.tgz in staging,
            #     so let's fall over here instead if there's not even one to copy over
            raise FileNotFoundError(
                f"Expected to find tarred weblog (*weblog.tgz) in source directory '{self.source_dir}', found none"
            )

    def copy_selfcal(self):
        """
        Copies the selfcal auxproducts to the staging directory if it exists.
        """
        if self.find_in_source(SELFCAL_REGEX):
            self.copy_to_staging(self.find_newest(SELFCAL_REGEX))


class ImageProductCollector(StagingProductCollector):
    """
    Image product collector class to link files from the products directory to the staging directory.
    """

    def collect(self):
        super().collect()
        self.copy_images()

    def copy_images(self):
        """
        Copies fits files into the staging directory.
        """
        for f in self.find_in_source(IMAGE_REGEX):
            self.copy_to_staging(f)


class ObservationProductCollector(StagingProductCollector):
    """
    Observation product collector to copy everything from the given source directory to the staging directory.
    """

    def __init__(self, source_dir: Path, stage_dir: Path):
        # The staging dir name will be the same as the source
        # There won't be artifacts, but providing a name just in case
        super().__init__(source_dir.name, stage_dir, "artifacts")

        # Want to use the given source directory as the true source to the files, typically located outside the spool
        self.source_dir = source_dir

    def collect(self):
        """
        Copies everything from the source directory to the staging directory.
        """
        for f in self.source_dir.iterdir():
            self.copy_to_staging(f)


class VlassCoarseProductCollector:
    _ = """
    Files to ingest.

    Science products
    - VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.
        - Cube entity name (for display in the archive)
    - VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw[?].IQU.iter3.image.pbcor.tt0.subim.fits
        - Per-channel FITS files
        - IQU image cube, one per spectral window.
    - VLASS2.1.cc.T10t02.J005000-023000.06.2048.V_mean_minmaxrej.subim.fits
        - Circular polarization image:
        - mean (with min/max reject) Stokes V image made from the common beam Stokes V images, one per cube.

    Ancillary products
    - VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw[?].IQU.iter3.image.pbcor.tt0.rms.subim.fits
        - Per channel
        - RMS IQU cube,one per spectral window.
    - weblog.tgz
        - Per cube
        - tarred and gzipped weblog
    - pipeline-<YYYMMDDTXXXXXX>
        - untarred weblog
    - placeholder.catalog.csv
        - empty file, will eventually contain source list
    - casa_pipescript.py
        - pipeline command script
    - casa_commands.log
        - list of equivalent CASA commands.
    - pipeline_manifest.xml
        - XML-formatted list of files
    - unknown.hifv_vlassSEIP.pprequest.xml
        - XML pipeline processing request
    - CCIP_parameter.list, SEIP_parameter.list
        - parameters used for running the cube and continuum pipelines, respectively.
    """
    STAGING_PERMISSIONS = 0o755

    def __init__(self, cache_path: Path, staging_dir: Path, tar_name: str):
        self.source_dir = cache_path
        self.staging_dir = staging_dir
        self.artifacts_tar = self.source_dir / (tar_name + ".tar")
        logger.info(f"Ready to collect products from {self.source_dir} and put them in {self.staging_dir}")

    def __call__(self):
        """
        Takes the steps to populate the staging directory.
        """
        self.prepare()
        self.collect()

    def prepare(self):
        """
        Creates the staging directory and sets permissions (if needed).
        """
        if self.staging_dir.exists() and self.staging_dir.is_dir():
            logger.info("Staging directory existed, removing")
            shutil.rmtree(self.staging_dir)

        self.staging_dir.mkdir()
        self.staging_dir.chmod(self.STAGING_PERMISSIONS)

    def collect(self):
        """
        Copies everything into the staging directory.
        """
        # .subim.com.fits files are to be cached, not ingested,
        # so we exclude them from what is copied into the staging directory
        FILES_TO_INGEST = r"^((?!\.subim\.com\.fits).)*$"

        # Copy everything in the file dict to the staging directory
        for file in find_in_dir(self.source_dir, FILES_TO_INGEST):
            copy_to_dir(self.staging_dir, file)

        # Set permissions on everything in the staging dir
        for f in self.staging_dir.iterdir():
            f.chmod(self.STAGING_PERMISSIONS)

        # Create the artifacts tar
        tar_artifacts(self.source_dir, self.artifacts_tar, self.staging_dir)


class VlassCacheProductCollector:
    """
    Product collector class to move products into a VLASS cache directory.
    """

    CACHE_PERMISSIONS = 0o755

    ACCEPTED_PLANES_REGEX = r"^accepted_planes\.txt"
    IMAGE_REGEX = r"^.*\.fits"
    IMAGE_BASE_REGEX = r"^(.*)\.spw[0-9]+"
    MEAN_IMAGE_REGEX = r"^.*.V_mean_minmaxrej.subim.fits"
    PROD_NUM_REGEX = r"^.*P([0-9]+)v[0-9]+.*"

    def __init__(self, source_dir: Path, status: str):
        # List of file regexes to collect
        self.FILE_REGEXES = {
            "products": [
                # Original set of desired files from Quicklook caching
                "casa_commands\\.log",
                "casa_pipescript\\.py",
                "pipeline_manifest\\.xml",
                ".*pprequest\\.xml",
                "weblog\\.tgz",
                # SECI specific files
                ".*catalog\\.csv",
                ".*reimaging_resources\\.tgz",
            ],
            "working": [
                ".*parameter.*list",
            ],
        }

        # source_dir = CapoConfig().getstring("OodtSettings.spoolDir") / queue / working
        # Example: /lustre/aoc/cluster/pipeline/vlass_dev/spool/se_coarse_cube_imaging/VLASS2.1_T10t32.J210201-033000_P106057v1_2023_03_07T22_58_33.300
        # Needed to clean up other versions of this product
        self.queue_dir = source_dir.parent

        # Needed to construct the cache path
        queue = self.queue_dir.name
        working = source_dir.name

        self.products_dir = self.queue_dir / working / "products"

        # Get what we need to generate the cache directory name
        # code = VLASS#.#
        code, tile = working.split("_", 1)
        # tile = T#t#
        tile = "QA_REJECTED" if status != "PASS" else tile.split(".")[0]

        # Any image base name (before 'spw#')
        img = find_in_dir(self.products_dir, self.IMAGE_BASE_REGEX)[0].name
        img = re.match(self.IMAGE_BASE_REGEX, img).group(1)

        # This comes from the VLASS CAPO file from the same environment (e.g. dsoc-dev -> vlass.dev)
        cache_prefix = CapoConfig(profile=self.get_vlass_profile()).getstring("OodtSettings.cacheDir")
        # Example: /lustre/aoc/cluster/pipeline/vlass_dev/cache/se_coarse_cube_imaging/VLASS2.1/T10t32/VLASS2.1.cc.T10t32.J210201-033000.06.2048.v1
        self.cache_dir = Path(cache_prefix) / queue / code / tile / img

        # Parse the product number for clean up
        self.prod_num = re.match(self.PROD_NUM_REGEX, working).group(1)
        logger.info(f"Ready to collect products from {self.products_dir} into {self.cache_dir}")

    def __call__(self):
        """
        Takes the steps to populate the cache directory.
        """
        self.prepare()
        self.collect()
        self.cleanup()

    def get_cache_dir(self):
        return self.cache_dir

    @staticmethod
    def get_vlass_profile() -> str:
        """
        Determines the correct VLASS CAPO profile from the current one.

        :return: The VLASS CAPO profile as a string
        """
        profile = CapoConfig().profile

        # Determine the VLASS location from the current profile if not running locally
        return profile if profile == "docker" else "vlass." + profile.split("-")[1]

    def prepare(self):
        """
        Creates the cache directory and sets permissions (if needed).
        Any previous cache directory will be deleted first.
        """
        if self.cache_dir.exists():
            logger.info("Cache directory existed, removing")
            shutil.rmtree(self.cache_dir)

        self.cache_dir.mkdir(parents=True)

        # Set permissions on both the image and tile directories
        self.cache_dir.parent.chmod(self.CACHE_PERMISSIONS)
        self.cache_dir.chmod(self.CACHE_PERMISSIONS)

    def collect(self):
        """
        Finds image files and copies everything into the cache directory.
        """
        self.find_images()
        self.tar_weblog()

        # Copy everything in the file dict to the cache directory, they aren't necessarily in the products_dir
        for subdir, regexes in self.FILE_REGEXES.items():
            search_dir = self.products_dir.parent / subdir
            for regex in regexes:
                for f in find_in_dir(search_dir, regex):
                    copy_to_dir(self.cache_dir, f)

        # Set permissions on everything in the cache dir
        for f in self.cache_dir.iterdir():
            f.chmod(self.CACHE_PERMISSIONS)

    def cleanup(self):
        """
        Removes any remaining left over directories, both the specified run and
        those generated by previous attempts to produce the same product ID.
        """
        logger.info(f"Removing all directories from {self.queue_dir} for product # {self.prod_num}")
        for f in self.queue_dir.iterdir():
            if self.prod_num in f.name:
                logger.info(f"Removing {f.resolve()}")
                shutil.rmtree(f)

    def find_images(self):
        """
        Adds all desired images to the file list to collect.
        """
        accepted_planes = find_in_dir(self.products_dir, self.ACCEPTED_PLANES_REGEX)
        if accepted_planes:
            # First, steal the accepted planes file from tomcat to be able to set permissions on it and gain its power
            accepted_planes_file = accepted_planes[0]
            accepted_planes_bak = Path(accepted_planes_file.parent, accepted_planes_file.name + ".bak")
            shutil.copy2(accepted_planes_file, accepted_planes_bak)
            accepted_planes_file.unlink()
            accepted_planes_bak.rename(accepted_planes_file)

            # For coarse cubes, identify correct plane images to cache from accepted_planes.txt
            self.FILE_REGEXES["products"].append(self.ACCEPTED_PLANES_REGEX)
            with open(accepted_planes_file, "r") as f:
                # For each spw, add its primary and ancillary images (each line in the file looks like 'spw#')
                for spw in f:
                    self.FILE_REGEXES["products"].append("^.*" + spw.strip() + "\\..*\\.fits")
            self.FILE_REGEXES["products"].append(self.MEAN_IMAGE_REGEX)
        else:
            # Otherwise add all primary images
            self.FILE_REGEXES["products"].append(self.IMAGE_REGEX)

    def tar_weblog(self):
        """
        Tars the most recent version of the weblog if it exists.
        """
        # Check if an untarred weblog exists. If so, prefer it in case it's been updated by the DAs
        if find_in_dir(self.products_dir, WEBLOG_DIR_REGEX):
            logger.info("Replacing weblog tar")
            # Remove any old tarred weblogs
            for f in find_in_dir(self.products_dir, WEBLOG_TAR_REGEX):
                f.unlink()

            # Tar and zip the newest weblog
            with tarfile.open(self.products_dir / "weblog.tgz", "w:gz") as tar:
                tar.add(find_newest_in_dir(self.products_dir, WEBLOG_DIR_REGEX))
