#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Solicitor: uses metadata.json to determine input parameters for ingestion
"""
import http
import json
import logging
import pathlib
from typing import Union

import requests
from ingest_envoy.utilities import (
    CuratorType,
    IngestType,
    InvalidCalLocatorException,
    VLASSIngestType,
)

INVALID_INITIAL_VERSION = "Initial version not valid for ingest"


class Solicitor:
    def __init__(self, arg_type: Union[IngestType, VLASSIngestType, CuratorType], urls: list, filename: str = None):
        self.metadata = {}
        if filename is not None:
            self.filename = filename
            self.metadata = self.solicit_contents(self.filename)

        self.logger = logging.getLogger("ingest_envoy")
        self.argument = arg_type
        self.capability_url = urls[0]
        self.workflow_url = urls[1]
        self.archive_url = urls[2]

    @staticmethod
    def solicit_contents(filename: str) -> dict:
        """
        Read provided metadata file

        :return: dict of parameters
        """
        with open(filename) as file:
            metadata = json.loads(file.read())
            return metadata

    def solicit_aux_metadata(self) -> dict:
        """
        Read aux_image_metadata file

        :return: dict of parameters
        """
        return self.solicit_contents(filename="aux_image_metadata.json")

    def solicit_workflow_directory_name(self) -> str:
        """
        Determine the original workflow processing directory's name without the path

        :return: directory name string
        """
        destination_path = pathlib.Path(self.metadata["destinationDirectory"])
        return destination_path.name

    def solicit_initial_directory_name(self, wf_id: int) -> str:
        """
        Determine the initial directory if more than one version was needed

        :param wf_id: the workflow id of the workflow being ingested
        :return: directory name string
        """
        current_execution_url = f"{self.capability_url}/capability/request/execution/{wf_id}"
        current_execution = requests.get(current_execution_url)

        capability_req_id = current_execution.json()["capability_request_id"]
        current_version_number = current_execution.json()["version_number"]

        current_version_url = (
            f"{self.capability_url}/capability/request/{capability_req_id}/version/{current_version_number}"
        )
        current_version = requests.get(current_version_url)
        # UI only updates the *current* version parameters, not the workflow arguments or other versions.
        # pull srdp status from current version
        srdp_status = current_version.json()["parameters"]["metadata"]["is_srdp"]

        initial_execution_url = f"{self.capability_url}/capability/request/{capability_req_id}/version/1/execution"
        initial_execution = requests.get(initial_execution_url)

        wf_request_url = (
            f"{self.workflow_url}/workflows/std_calibration/requests/"
            f"{initial_execution.json()['current_workflow_request_id']}"
        )

        wf_request = requests.get(wf_request_url)

        if self.verify_valid_initial_version(wf_request.json(), srdp_status):
            initial_path = pathlib.Path(wf_request.json()["results_dir"])
            return initial_path.name
        else:
            self.logger.error(
                "Error: Invalid initial version! No weblog found or reqeust is not SRDP. "
                "Running Single Version Ingestion..."
            )
            return INVALID_INITIAL_VERSION

    def verify_valid_initial_version(self, wf_request: dict, srdp_flag: bool) -> bool:
        """
        Ping the weblog display endpoint to verify the existence of an initial version weblog

        **Ingestion fails if this is unavailable**

        :param wf_request: the initial workflow request to verify
        :param srdp_flag: Is this request SRDP?
        :return: True if is SRDP and weblog exists, False otherwise
        """

        # only run ingestion with initial version weblog if this is SRDP
        if srdp_flag:
            # if this is SRDP, verify the initial weblog exists
            get_weblog_url = (
                f"{self.workflow_url}/workflows/std_calibration/requests/{wf_request['workflow_request_id']}/weblog"
            )
            response = requests.get(get_weblog_url)

            if response.status_code == http.HTTPStatus.OK:
                return True
        else:
            self.logger.info("Request is not SRDP. Skipping initial weblog ingestion.")

        # if weblog response is not 200 or request is not SRDP
        return False

    def solicit_processing_creation_time(self) -> str:
        """
        Transform a datetime into and NGAS friendly representation

        Example:
            Incoming datetime: 2021-07-06T21:50:48
            Transformation for NGAS friendliness: 2021_07_06T21_50_48

        :return: string
        """
        creation_time = self.metadata["creationTime"].replace("-", "_").replace(":", "_")
        return creation_time

    def solicit_calibration_product_locator(self, sdm_id: str) -> str:
        """
        Returns the calibration product locator from the metadata or requests it from the archive service if needed.

        :param sdm_id: SDM ID of the exec block
        :return: Calibration product locator or None if not found
        """
        if "calProductLocator" in self.metadata:
            return self.metadata["calProductLocator"]
        else:
            # Make a call to the archive service to get the cal SPL if it's not in the metadata
            self.logger.info("Cal SPL not in the metadata, requesting it from the archive service")
            return self.lookup_calibration_product_locator(sdm_id)

    def lookup_calibration_product_locator(self, sdm_id: str) -> str:
        """
        Requests the calibration product locator from the archive service using the given exec block SDM ID.

        :param sdm_id: SDM ID of the exec block
        :return: Calibration product locator
        """
        response = requests.get(self.archive_url, params={"identifier": sdm_id + ".%"})

        if response.status_code == http.HTTPStatus.OK and "locator" in response.json():
            self.logger.info(f"Obtained cal SPL from the archive service: {response.json()['locator']}")
            return response.json()["locator"]
        else:
            raise InvalidCalLocatorException("Failed to fetch calibration SPL from the archive service")

    def get_general_params(self) -> dict:
        """
        Determine general ingestion parameters (WS only)
        :return: dict
        """

        gen = {
            "sdmId": self.metadata["fileSetIds"],
            "requestId": self.metadata["systemId"],
            "telescope": self.metadata["projectMetadata"]["telescope"],  # all, needed by manifest generator
            "project": self.metadata["projectMetadata"]["projectCode"],  # needed for post ingestion messaging
            "workflowName": self.metadata["workflowName"],  # all, needed by ingestion type check
            "processingStart": self.solicit_processing_creation_time(),  # cal/image
            "workflowDir": self.solicit_workflow_directory_name(),  # cal/image
            "initialVersionDir": self.solicit_initial_directory_name(self.metadata["systemId"]),  # currently, only cal
        }

        if gen["workflowDir"] == gen["initialVersionDir"] or gen["initialVersionDir"] == INVALID_INITIAL_VERSION:
            gen["multiVersion"] = False
        else:
            gen["multiVersion"] = True

        return gen

    def solicit_calibration_params(self) -> dict:
        """
        Determine calibration specific parameters (WS Only)

        :return: dict
        """

        general = self.get_general_params()

        cal = {
            "spl": self.metadata["productLocator"],  # calibration only
        }

        return {**general, **cal}

    def solicit_image_params(self) -> dict:
        """
        Determine image specific parameters (WS Only)

        :return: dict
        """
        general = self.get_general_params()

        img = {
            "calSpl": self.solicit_calibration_product_locator(general["sdmId"]),  # image only
            "spoolDir": self.metadata["destinationDirectory"],
        }

        return {**general, **img}

    def solicit_observation_params(self) -> dict:
        """
        Determine observation specific parameters (WS Only)

        :return: dict
        """

        obs = {
            "filename": self.filename,
            "telescope": self.metadata["projectMetadata"]["telescope"],  # all, needed by manifest generator
            "project": self.metadata["projectMetadata"]["projectCode"],  # needed for post ingestion messaging
        }

        return {**obs}

    def solicit_curation_params(self) -> dict:
        """
        Determine curation specific parameters

        :return: dict
        """

        if self.argument == CuratorType.PARTIAL and "target_list" not in self.metadata:
            targets = []  # empty list is untargeted partial curation
        elif self.argument == CuratorType.FULL:
            targets = None
        else:
            targets = list(filter(None, [target.strip() for target in self.metadata.get("target_list").split(",")]))

        params = {
            "telescope": self.metadata["projectMetadata"]["telescope"],  # all, needed by manifest generator
            "project": self.metadata["projectMetadata"]["projectCode"],  # needed for post ingestion messaging
            "spl": self.metadata.get("product_locator"),
            "input_group_locator": self.metadata.get("input_group_locator"),
            "product_type": self.metadata["product_type"],
            "curation_source": self.metadata.get("data_location"),  # not required for curation
            "file_list": list(filter(None, self.metadata.get("file_list").split(","))),
            "product_group_id": self.metadata.get("product_group_id"),
            "target_list": targets,
        }
        return params

    def solicit_seci_params(self) -> dict:
        """
        Determine SECI image specific parameters

        :return: dict
        """
        aux = self.solicit_aux_metadata()

        aux2 = {"project": aux["project_code"]}

        seci = {
            "telescope": "EVLA",
            "workflowName": "vlass_seci",
            "multiVersion": False,
        }

        return {**aux2, **seci}

    def solicit_coarse_params(self) -> dict:
        """
        Determine coarse cube image specific parameters

        :return: dict
        """
        aux = self.solicit_aux_metadata()

        aux2 = {"project": aux["project_code"]}

        coarse = {
            "telescope": "EVLA",
            "workflowName": "vlass_coarse",
            "multiVersion": False,
        }

        return {**aux2, **coarse}

    def solicit_parameters(self) -> dict:
        """
        Create a dict of parameters needed for ingestion

        :return: parameters dict
        """

        match self.argument:
            case IngestType.CAL:
                return self.solicit_calibration_params()
            case IngestType.IMG:
                return self.solicit_image_params()
            case IngestType.OBS:
                return self.solicit_observation_params()
            case VLASSIngestType.SECI:
                return self.solicit_seci_params()
            case VLASSIngestType.COARSE:
                return self.solicit_coarse_params()
            case CuratorType.PARTIAL | CuratorType.FULL:
                return self.solicit_curation_params()
