# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
import logging
import re
from pathlib import Path

from ingest_envoy.manifest_components import AncillaryProduct, OutputScienceProduct
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType

CUBE_SCIENCE_PRODUCT_REGEX = r"^.*cube\.(regcal|selfcal)\.I\.pbcor\.fits$"
CUBE_ANCILLARY_MASK_REGEX = r"^.*cube\.(regcal|selfcal)\.I\.mask\.fits$"
CUBE_ANCILLARY_PB_REGEX = r"^.*cube\.(regcal|selfcal)\.I\.pb\.fits$"


class SpectralCubeIngestionProductsFinder:
    """Finds science products and ancillary products needed for spectral-cube ingestion"""

    def __init__(self, files_found: list[Path], sp_type: ScienceProductType):
        self.logger = logging.getLogger("ingest_envoy")
        self.sp_type = sp_type
        self.files_found = files_found
        self.output_science_products = self._find_output_science_products()

    def _find_output_science_products(self) -> list[OutputScienceProduct]:
        """
        Find the ancillary products belonging to the science product*
        in the staging dir.

        N.B. ingestion artifacts file will not be there until manifest has been created!

        :return: The output science products as a list
        """

        sp_files = [
            file for file in self.files_found if re.fullmatch(CUBE_SCIENCE_PRODUCT_REGEX, file.name) is not None
        ]
        self.logger.info(f"Spectral Cube Science Product(s) to ingest: {sp_files}")
        sp_ancillary_files = [
            file
            for file in self.files_found
            if re.fullmatch(CUBE_ANCILLARY_MASK_REGEX, file.name) or re.fullmatch(CUBE_ANCILLARY_PB_REGEX, file.name)
        ]
        self.logger.info(f"Spectral Cube Science Product ancillaries are: {sp_ancillary_files}")

        # Build a mapping of science products -> ancillaries
        sp_aps = {
            sp_file: [
                self._build_ancillary_science_product(sp_ancillary_file)
                for sp_ancillary_file in sp_ancillary_files
                if sp_ancillary_file.name.startswith(sp_file.name.split("pbcor")[0])
            ]
            for sp_file in sp_files
        }
        self.logger.info(f"Spectral Cube Science Product -> Ancillary mapping: {sp_aps}")

        science_products = []
        for file in sp_files:
            # Add all science products and their ancillaries
            science_products.append(
                OutputScienceProduct(
                    type=ScienceProductType.IMAGE,
                    filename=file.name,
                    ancillary_products=sp_aps.get(file),
                )
            )

        return science_products

    @staticmethod
    def _build_ancillary_science_product(file: Path) -> AncillaryProduct:
        """
        Spectral-cube science products will have ancillary products of their very own,
        distinct from other ancillary products that might be in the staging dir.

        :param file: a possible ancillary science product
        :return: the corresponding AncillaryProduct
        """
        filename = file.name
        if re.fullmatch(CUBE_ANCILLARY_MASK_REGEX, filename):
            return AncillaryProduct(type=AncillaryProductType.FITS_MASK, filename=filename)
        if re.fullmatch(CUBE_ANCILLARY_PB_REGEX, filename):
            return AncillaryProduct(type=AncillaryProductType.PB_FITS, filename=filename)
        raise ValueError(f"Cannot determine what type of spectral-cube ancillary '{filename}' is.")
