#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for ingest_envoy.collectors
"""
import logging
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from conftest import expected_settings_evla_eb
from ingest_envoy.collectors import (
    CoarseCollector,
    ImageCollector,
    ObservationCollector,
    aux_file_to_staging,
    collect_image_metadata,
)
from ingest_envoy.schema import AbstractTextFile

image_parameters = {
    "sdmId": "brain_000.58099.67095825232",
    "telescope": "EVLA",
    "workflowName": "std_cms_imaging",
    "workflowRequestId": "12",
    "spl": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",
    "processingStart": "2021_07_06T21_50_48",
    "destinationDir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmprb1se376",
    "workflowDir": "tmprb1se376",
    "calSpl": "uid://evla/calibration/c78ccfd6-fe4e-43c6-a5c5-70e5bcfde16b",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
    "script_location": "/lustre/aoc/cluster/pipeline/docker/workspaces/sbin",
    "project": "Operations",
    "spoolDir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmprb1se376",
}

coarse_parameters = {
    "telescope": "EVLA",
    "workflowName": "vlass_coarse",
    "project": "TEST000",
    "multiVersion": False,
    "calSpl": "uid://evla/calibration/ba6c6404-b196-4d5d-9ab8-8e2b5bce62c1",
    "workflowDir": "VLASS3.1_T10t10.J063800-023000_P106057v1_2023_03_07T22_58_33.300",
    "sourcePath": "/lustre/aoc/cluster/pipeline/vlass_dev/spool/quicklook/",
    "qa_status": "PASS",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
    "script_location": "/lustre/aoc/cluster/pipeline/docker/workspaces/sbin",
}

image_collector = ImageCollector(parameters=image_parameters)
coarse_collector = CoarseCollector(parameters=coarse_parameters)


@patch("builtins.open")
def test_collect_image_metadata(mock_open):
    aux = collect_image_metadata()
    assert mock_open.call_count == 1
    assert isinstance(aux, AbstractTextFile)


def test_aux_file_to_staging():
    with patch("shutil.copy2") as mock_copy2:
        aux_file_to_staging(Path(image_parameters["staging_area"]) / image_parameters["workflowDir"])
        assert mock_copy2.call_count == 1


class TestImageCollector:
    @patch("ingest_envoy.product_collectors.ImageProductCollector.__call__")
    @patch("shutil.copy2")
    @patch("os.rename")
    def test_collect_image_products(self, mock_rename, mock_copy2, mock_product_collector):
        image_collector.collect_products()
        assert mock_product_collector.call_count == 1

    def test_create_artifacts_name(self):
        expected_name = "pipeline_artifacts_" + image_parameters["processingStart"]

        name = image_collector.create_artifacts_name()
        assert name == expected_name


class TestCoarseCollector:
    @patch("ingest_envoy.product_collectors.VlassCacheProductCollector.__init__", return_value=None)
    @patch("ingest_envoy.product_collectors.VlassCacheProductCollector.__call__")
    @patch("ingest_envoy.product_collectors.VlassCacheProductCollector.get_cache_dir", return_value=Path("."))
    @patch("ingest_envoy.product_collectors.VlassCoarseProductCollector.__call__")
    @patch("shutil.copy2")
    @patch("os.rename")
    def test_collect_image_products(
        self,
        mock_rename,
        mock_copy2,
        mock_staging_collector_call,
        mock_cache_collector_dir,
        mock_cache_collector_call,
        mock_cache_collector_init,
    ):
        coarse_collector.collect_products()
        assert mock_cache_collector_call.call_count == 1
        assert mock_staging_collector_call.call_count == 1


class TestObservationCollector:
    """
    Tests for the ObservationCollector class.
    """

    @staticmethod
    @pytest.fixture
    def collector():
        """
        Yields an ObservationCollector instance.
        """
        yield ObservationCollector(expected_settings_evla_eb())

    def test_init(self, collector):
        """
        Test the class' __init__ method.
        """
        params = expected_settings_evla_eb()
        assert collector.parameters == params

    @patch("ingest_envoy.product_collectors.ObservationProductCollector.__call__")
    def test_collect_products(self, mock_product_collector, collector, caplog):
        """
        Tests the collect_products method when the product collector succeeds.
        """
        caplog.set_level(logging.INFO)
        collector.collect_products()
        assert "Collecting observation products for staging..." in caplog.text
        assert mock_product_collector.call_count == 1
