# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Runs the tests for EVLA EB manifest generation.
"""
import json
import re
import shutil
import tarfile
from pathlib import Path
from typing import Any, Generator

import pytest
from conftest import (
    EVLA_EB_FILENAME,
    find_example_manifest,
    find_ingestion_artifacts_tar,
    ordered,
    populate_directory,
)
from ingest_envoy.ingestion_manifest import (
    IngestionManifest,
    IngestionManifestBuilder,
    IngestionManifestSchema,
)
from ingest_envoy.manifest_components import MANIFEST_FILENAME
from ingest_envoy.utilities import IngestType, Telescope


@pytest.fixture(name="evla_eb_manifest")
def build_manifest(evla_eb_staging_dir: Path) -> Generator[tuple[IngestionManifest, Path], Any, None]:
    """
    Builds the ingestion manifest and returns it as well as a Path of the file it came from.
    :param evla_eb_staging_dir: the path to the temporary staging directory.
    :return: The manifest and the file it came from.
    """
    evla_eb_staging_dir.mkdir(parents=True)
    evla_eb_sdm_files = [
        "Annotation.xml",
        "CalData.xml",
        "CalReduction.xml",
        "DelayModelFixedParameters.xml",
        "Ephemeris.xml",
        "Flag.xml",
        "Polarization.xml",
        "SBSummary.xml",
        "State.xml",
        "SysCal.xml",
        "Antenna.xml",
        "CalDelay.xml",
        "ConfigDescription.xml",
        "DelayModelVariableParameters.xml",
        "ExecBlock.xml",
        "Main.xml",
        "Processor.xml",
        "Scan.xml",
        "Station.xml",
        "SysPower.bin",
        "ASDM.xml",
        "CalDevice.xml",
        "CorrelatorMode.xml",
        "DelayModel.xml",
        "Feed.xml",
        "Pointing.bin",
        "Pulsar.xml",
        "Source.xml",
        "Subscan.xml",
        "VLAWVR.xml",
        "CalAntennaSolutions.xml",
        "CalPointing.xml",
        "DataDescription.xml",
        "Doppler.xml",
        "Field.xml",
        "PointingModel.xml",
        "Receiver.xml",
        "SpectralWindow.xml",
        "SwitchCycle.xml",
        "Weather.xml",
    ]
    populate_directory(evla_eb_staging_dir, [EVLA_EB_FILENAME, *evla_eb_sdm_files])

    yield IngestionManifestBuilder(
        manifest_source_dir=evla_eb_staging_dir,
        sp_type=IngestType.OBS.value,
        locator=None,
        telescope=Telescope.EVLA.value,
        filename=EVLA_EB_FILENAME,
    ).build()

    # Remove staging directory
    shutil.rmtree(evla_eb_staging_dir)


@pytest.fixture(name="actual_manifest_json")
def get_actual_manifest_json(evla_eb_manifest) -> dict:
    """
    :param evla_eb_manifest:  A Pytest fixture.
    :return: The dict representation of the actual (generated) manifest JSON.
    """
    _, actual_manifest_file = evla_eb_manifest
    with open(actual_manifest_file, encoding="utf8") as f:
        manifest_json = json.load(f)
    return manifest_json


@pytest.fixture(name="expected_manifest_json")
def get_expected_manifest_json() -> dict:
    """
    :return: dict. The dict representation of the expected manifest JSON.
    """
    expected_manifest_file = find_example_manifest("evla_eb_manifest")
    with open(expected_manifest_file, encoding="utf8") as f:
        manifest_json = json.load(f)
    return manifest_json


def test_manifest_ingestion_artifacts_tar_contents(
    evla_eb_manifest: tuple[IngestionManifest, Path], evla_eb_staging_dir: Path
):
    """
    Tests for the correct contents of the manifest's ingestion artifacts tar.
    """
    ingestion_artifacts_tar = find_ingestion_artifacts_tar(evla_eb_staging_dir)
    with tarfile.open(ingestion_artifacts_tar, "r") as tar:
        # We need to get just the file name, not the full path to the file
        members = [member.name.split("/")[-1] for member in tar.getmembers()]
        assert members == [MANIFEST_FILENAME]


def test_manifest_json(actual_manifest_json: dict, expected_manifest_json: dict):
    """
    Tests that the generated manifest JSON is what is expected.
    """
    assert ordered(actual_manifest_json["parameters"]) == ordered(expected_manifest_json["parameters"])

    # No input group is expected; test for this
    with pytest.raises(KeyError):
        _ = actual_manifest_json["input_group"]
    with pytest.raises(KeyError):
        _ = expected_manifest_json["input_group"]

    assert (
        actual_manifest_json["output_group"]["science_products"]
        == expected_manifest_json["output_group"]["science_products"]
    )

    # Test output group ancillary products
    actual_og_aps = actual_manifest_json["output_group"]["ancillary_products"]
    expected_og_aps = expected_manifest_json["output_group"]["ancillary_products"]

    # Because the name is based on the time it was generated, we have to compare it differently
    actual_ingestion_artifacts_tar = [x for x in actual_og_aps if x["type"] == "ingestion_artifacts"][0]
    expected_ingestion_artifacts_tar = [x for x in expected_og_aps if x["type"] == "ingestion_artifacts"][0]

    assert re.match(
        r"ingestion_artifacts_\d{4}_\d{2}_\d{2}T\d{2}_\d{2}_\d{2}\.\d{3}\.tar",
        actual_ingestion_artifacts_tar["filename"],
    )

    # Drop ingestion artifacts tars from the lists
    actual_og_aps.remove(actual_ingestion_artifacts_tar)
    expected_og_aps.remove(expected_ingestion_artifacts_tar)

    # Check that there are no other output group ancillary products
    assert len(actual_og_aps) == 0
    assert len(expected_og_aps) == 0


def test_manifest_object_vs_dumped_json(evla_eb_manifest: tuple[IngestionManifest, Path], actual_manifest_json: dict):
    """
    Tests that the correct content was dumped to the manifest file.
    """
    actual_manifest, _ = evla_eb_manifest
    actual_manifest_schema = IngestionManifestSchema()
    # These are sorted because ancillary products are ordered differently between the two dumps for some reason
    assert ordered(actual_manifest_schema.dump(actual_manifest)) == ordered(actual_manifest_json)


def test_manifest_parameters(
    evla_eb_manifest: tuple[IngestionManifest, Path],
    evla_eb_staging_dir: Path,
):
    """
    Tests for the correct parameters on the IngestionManifest instance.
    """
    actual_manifest, _ = evla_eb_manifest

    actual_manifest_params = actual_manifest.parameters
    assert actual_manifest_params.telescope == Telescope.EVLA
    assert actual_manifest_params.staging_source_dir == evla_eb_staging_dir.parent
    assert actual_manifest_params.ngas_ingest is True
