#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for ingest_envoy.ingest
"""
import argparse
import http
from typing import Sequence
from unittest.mock import MagicMock, patch

import pytest
from conftest import (
    SECI_SCIENCE_PRODUCT,
    evla_eb_information,
    expected_settings_evla_eb,
    mocked_archive_service_get,
)
from ingest_envoy.ingest import (
    _determine_ingestion_info,
    _get_settings,
    arg_parser,
    check_ingest_type,
    main,
    send_indexer_message,
)
from ingest_envoy.launchers import IngestObservationLauncher
from ingest_envoy.utilities import IngestType

expected_settings = {
    "sdmId": "16B-069_sb32814386_1_001.57685.66193635417",
    "telescope": "EVLA",
    "workflowName": "std_calibration",
    "workflowRequestId": "2",
    "spl": "uid://evla/execblock/48ba4c9d-d7c7-4a8f-9803-1115cd52459b",
    "processingStart": "2021_07_06T21_50_48",
    "destinationDir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmp576dknjp",
    "workflowDir": "tmp576dknjp",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
}
expected_settings_image = {
    "sdmId": "brain_000.58099.67095825232",
    "telescope": "EVLA",
    "workflowName": "std_cms_imaging",
    "workflowRequestId": "12",
    "spl": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",
    "processingStart": "2021_07_06T21_50_48",
    "destinationDir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmprb1se376",
    "workflowDir": "tmprb1se376",
    "calSpl": "uid://evla/calibration/c78ccfd6-fe4e-43c6-a5c5-70e5bcfde16b",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
}

args = argparse.Namespace()
FILENAME = "test/input_files/test-metadata.json"
FILENAME2 = "test/input_files/test-image-metadata.json"
coarse_args = [
    SECI_SCIENCE_PRODUCT["external_name"],
    "/lustre/aoc/cluster/pipeline/vlass_dev/spool/quicklook/"
    "VLASS3.1_T10t10.J063800-023000_P106057v1_2023_03_07T22_58_33.300",
    "PASS",
]


class TestIngest:
    """
    Tests for ingest.
    """

    @staticmethod
    def get_parsed_args(arguments: Sequence[str] | None = None) -> argparse.Namespace:
        """
        :param arguments: a list of arguments to parse.
        :return: The parsed arguments.
        """
        return arg_parser().parse_args(arguments)

    @staticmethod
    def evla_eb_parsed_args() -> argparse.Namespace:
        """
        :return: The parsed arguments for the test EVLA EB.
        """
        return TestIngest.get_parsed_args(TestIngest.evla_eb_observation_args(True))

    def test_determine_ingestion_info(self):
        """
        Tests the _determine_ingestion_info function.
        """
        # For EVLA EBs
        with patch("ingest_envoy.ingest._get_settings", return_value=expected_settings_evla_eb()):
            arg_type, parameters, launcher = _determine_ingestion_info(TestIngest.evla_eb_parsed_args())

        assert arg_type == IngestType.OBS
        assert parameters == expected_settings_evla_eb()
        assert isinstance(launcher, IngestObservationLauncher)

    def test_get_settings(self, mock_evla_eb_solicitor_solicit_contents):
        """
        Tests _get_settings function.
        """
        with (
            patch("pycapo.CapoConfig.settings") as mock_capo_config_settings,
            patch("pycapo.CapoConfig.getboolean") as mock_capo_config_boolean,
        ):
            mock_capo_config_settings.side_effect = TestIngest.mock_capo_config_get_settings_evla_eb_side_effect
            mock_capo_config_boolean.return_value = True
            test_filename, source_dir, *_ = evla_eb_information()

            actual_parameters = _get_settings(IngestType.OBS, filename=test_filename, source_dir=source_dir)

        assert actual_parameters == expected_settings_evla_eb()

    def test_check_ingest_type(self):
        """
        Tests the check_ingest_type function.
        """
        assert check_ingest_type(args_type=IngestType.CAL, parameters=expected_settings)
        assert check_ingest_type(args_type=IngestType.IMG, parameters=expected_settings_image)
        assert check_ingest_type(args_type=IngestType.OBS, parameters=expected_settings_evla_eb())

    def test_parser(self):
        """
        Tests the arg_parser function.
        """
        parser = arg_parser()
        assert isinstance(parser, argparse.ArgumentParser)

        # Test the correct behavior for an observation
        actual_observation_args = TestIngest.evla_eb_parsed_args().observation
        assert actual_observation_args == TestIngest.evla_eb_observation_args(False)

    def test_send_indexer_message(self, caplog):
        """
        Tests the send_indexer_message function.
        """

        # EVLA EB
        with patch("requests.post") as mock_post:
            parameters = expected_settings_evla_eb()

            # Successfully sent message
            mock_post.return_value = MagicMock(status_code=http.HTTPStatus.OK)
            send_indexer_message(parameters, IngestType.OBS)
            assert mock_post.call_count == 1
            mock_post.assert_called_with(
                f"{parameters['workflowUrl']}/workflows/requests/message-archive/obs-ingestion-complete",
                json={"project_code": parameters["project"], "telescope": parameters["telescope"]},
            )
            assert caplog.text == ""  # No warning was logged

            # Message that failed to send
            mock_post.return_value = MagicMock(status_code=http.HTTPStatus.REQUEST_TIMEOUT)
            send_indexer_message(parameters, IngestType.OBS)
            assert (
                "WARNING: Failed to send observation ingestion complete "
                f"to archive for observation {parameters['filename']}. "
                "Please set off the indexer manually"
            ) in caplog.text

    @pytest.mark.skip("Broken by multiversion changes")
    def test_main_calibration(self):
        """
        Tests the main function for a calibration.
        """
        args.calibration = [FILENAME]

        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args)):
            with patch("ingest_envoy.launchers.IngestCalibrationLauncher.launch", return_value=0) as cal_ingest:
                main()
                assert cal_ingest.call_count == 1
                assert cal_ingest.return_value == 0

        # reset for other testing
        args.calibration = None

    @pytest.mark.skip("Broken by multiversion changes")
    def test_main_image(self):
        """
        Tests the main function for an image.
        """
        args.calibration = None
        args.image = [FILENAME2]

        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args)):
            with patch("ingest_envoy.launchers.IngestImageLauncher.launch", return_value=0) as img_ingest:
                main()
                assert img_ingest.call_count == 1
                assert img_ingest.return_value == 0

        # reset for other testing
        args.image = None

    def test_main_coarse(self, aux_data: dict):
        """
        Tests the main function for a VLASS coarse cube.
        """
        args.calibration = None
        args.image = None
        args.seci = None
        args.coarse = coarse_args

        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args)):
            with patch("ingest_envoy.solicitor.Solicitor.solicit_aux_metadata", return_value=aux_data):
                with patch("requests.get", side_effect=mocked_archive_service_get):
                    with patch(
                        "ingest_envoy.launchers.IngestVLASSLauncher.launch", return_value=(0, False)
                    ) as coarse_ingest:
                        with patch("requests.post", MagicMock(status_code=http.HTTPStatus.OK)):
                            main()
                            assert coarse_ingest.call_count == 1

        # reset for other testing
        args.coarse = None

    @staticmethod
    def evla_eb_observation_args(including_flag: bool) -> list[str]:
        """
        :param including_flag: Whether to als return the "--observation" flag.
        :return: Sample EVLA EB observation arguments.
        """
        file_name, source_directory, *_ = evla_eb_information()
        arguments = [file_name, source_directory]

        if including_flag:
            arguments.insert(0, "--observation")
        return arguments

    @staticmethod
    def mock_capo_config_get_settings_evla_eb_side_effect(setting: str) -> MagicMock:
        """
        :param setting: which settings to get.
        :raises ValueError: when setting is invalid.
        :return: MagicMock. The appropriate settings.
        """
        match setting:
            case "edu.nrao.workspaces.IngestionSettings":
                return MagicMock(
                    findLocatorServiceUrlPrefix="https://fake-locator-service.com/fake-service/",
                    stagingDirectory="/fake/staging/directory",
                    storageDirectory="/fake/storage/directory",
                )
            case "edu.nrao.workspaces.ProcessingSettings":
                return MagicMock(scriptLocation="/fake/script/location")
            case "edu.nrao.workspaces.CapabilitySettings":
                return MagicMock(externalServiceUrl="https://fake-external-service.com")
            case "edu.nrao.workspaces.WorkflowSettings":
                return MagicMock(serviceUrl="https://fake-service-url.com")
            case _:
                raise ValueError(f"Error: unexpected setting {setting}.")

    def test_main_evla_eb(self, caplog):
        """
        Tests that the main method correctly processes an observation (at a high level).
        """
        args_observation = TestIngest.evla_eb_parsed_args()
        with patch("argparse.ArgumentParser.parse_args", MagicMock(return_value=args_observation)), patch(
            "ingest_envoy.ingest._determine_ingestion_info"
        ) as mock_determine_ingestion_info, patch(
            "ingest_envoy.launchers.IngestObservationLauncher"
        ) as observation_ingest, patch(
            "ingest_envoy.ingest.send_indexer_message"
        ) as mock_send_indexer_message:
            mock_determine_ingestion_info.return_value = (IngestType.OBS, {}, observation_ingest)
            observation_ingest.launch.return_value = (0, True)

            main()

            assert mock_determine_ingestion_info.call_count == 1
            assert observation_ingest.launch.call_count == 1
            assert mock_send_indexer_message.call_count == 1
            assert "Ingestion finished successfully! Sending status message to archive..." in caplog.text
