#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for ingest_envoy.launchers
"""
import logging
import sys
import unittest
from io import StringIO
from pathlib import Path
from subprocess import CalledProcessError, CompletedProcess
from typing import Any, Generator
from unittest.mock import MagicMock, call, patch

import pytest
from conftest import expected_settings_evla_eb
from ingest_envoy import launchers
from ingest_envoy.launchers import (
    CuratorLauncher,
    IngestCalibrationLauncher,
    IngestImageLauncher,
    IngestObservationLauncher,
    IngestVLASSLauncher,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import CuratorType, IngestType, VLASSIngestType

parameters = {
    "sdmId": "16B-069_sb32814386_1_001.57685.66193635417",
    "telescope": "EVLA",
    "workflowName": "std_calibration",
    "spl": "uid://evla/execblock/48ba4c9d-d7c7-4a8f-9803-1115cd52459b",
    "processingStart": "2021_07_06T21_50_48",
    "destinationDir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmp576dknjp",
    "workflowDir": "tmp576dknjp",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
    "script_location": "/lustre/aoc/cluster/pipeline/docker/workspaces/sbin",
    "useIngest": False,
    "multiVersion": False,
}

image_parameters = {
    "sdmId": "brain_000.58099.67095825232",
    "telescope": "EVLA",
    "workflowName": "std_cms_imaging",
    "spl": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",
    "processingStart": "2021_07_06T21_50_48",
    "destinationDir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmprb1se376",
    "workflowDir": "tmprb1se376",
    "calSpl": "uid://evla/calibration/c78ccfd6-fe4e-43c6-a5c5-70e5bcfde16b",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
    "script_location": "/lustre/aoc/cluster/pipeline/docker/workspaces/sbin",
    "useIngest": False,
}

coarse_parameters = {
    "telescope": "EVLA",
    "workflowName": "vlass_coarse",
    "project": "TEST000",
    "multiVersion": False,
    "calSpl": "uid://evla/calibration/ba6c6404-b196-4d5d-9ab8-8e2b5bce62c1",
    "workflowDir": "VLASS3.1_T10t10.J063800-023000_P106057v1_2023_03_07T22_58_33.300",
    "sourcePath": "/lustre/aoc/cluster/pipeline/vlass_dev/spool/quicklook/",
    "qa_status": "PASS",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
    "script_location": "/lustre/aoc/cluster/pipeline/docker/workspaces/sbin",
}

expected_spls = [
    "uid://vlba/correlation/ebfac8c5-6b86-489f-84f8-e77c046d29a8",
    "uid://vlba/correlation/cd61f65d-e1ee-4972-9c67-973fccb0dd58",
    "uid://vlba/correlation/a2042206-dc72-4351-b662-b5d08008307e",
    "uid://vlba/correlation/299dc994-c168-4d77-9155-92f5d3516b3a",
    "uid://vlba/correlation/3f619a60-5790-4d05-8e1f-346e5ecf150f",
    "uid://vlba/correlation/3124cf11-5793-42d4-b909-9cd4b17501ab",
    "uid://vlba/correlation/c0808e39-fb9c-42dd-bb13-a5ace0024752",
    "uid://vlba/correlation/1834cd4a-9390-4a03-874e-1acf652d717d",
    "uid://vlba/correlation/70d70b22-d1bd-4ca5-a426-2d316ea5cf29",
    "uid://vlba/correlation/8bc3d9c3-b85d-4758-9931-5ffd5885f1e7",
]
sample_input = """
*Input science products-2: []
*Output science products-2: [<ScienceProduct#uid://vlba/correlation/ebfac8c5-6b86-489f-84f8-e77c046d29a8 type=Execution Block>, <ScienceProduct#uid://vlba/correlation/cd61f65d-e1ee-4972-9c67-973fccb0dd58 type=Execution Block>, <ScienceProduct#uid://vlba/correlation/a2042206-dc72-4351-b662-b5d08008307e type=Execution Block>, <ScienceProduct#uid://vlba/correlation/299dc994-c168-4d77-9155-92f5d3516b3a type=Execution Block>, <ScienceProduct#uid://vlba/correlation/3f619a60-5790-4d05-8e1f-346e5ecf150f type=Execution Block>, <ScienceProduct#uid://vlba/correlation/3124cf11-5793-42d4-b909-9cd4b17501ab type=Execution Block>, <ScienceProduct#uid://vlba/correlation/c0808e39-fb9c-42dd-bb13-a5ace0024752 type=Execution Block>, <ScienceProduct#uid://vlba/correlation/1834cd4a-9390-4a03-874e-1acf652d717d type=Execution Block>, <ScienceProduct#uid://vlba/correlation/70d70b22-d1bd-4ca5-a426-2d316ea5cf29 type=Execution Block>, <ScienceProduct#uid://vlba/correlation/8bc3d9c3-b85d-4758-9931-5ffd5885f1e7 type=Execution Block>]
*Output group ancillary products: [{'type': 'observation_log', 'filename': 'VLBA_TR039HD_tr039hd_250321T142044.difx.log.gz'}, {'type': 'observation_details', 'filename': 'TR039HD.tr039hd.sum.txt'}, {'type': 'ingestion_artifacts', 'filename': 'ingestion_artifacts_2025_03_21T09_08_50.708.tar'}]
ISPs: []
OSPs: [<ScienceProduct#uid://vlba/correlation/ebfac8c5-6b86-489f-84f8-e77c046d29a8 type=Execution Block>, <ScienceProduct#uid://vlba/correlation/cd61f65d-e1ee-4972-9c67-973fccb0dd58 type=Execution Block>, <ScienceProduct#uid://vlba/correlation/a2042206-dc72-4351-b662-b5d08008307e type=Execution Block>, <ScienceProduct#uid://vlba/correlation/299dc994-c168-4d77-9155-92f5d3516b3a type=Execution Block>, <ScienceProduct#uid://vlba/correlation/3f619a60-5790-4d05-8e1f-346e5ecf150f type=Execution Block>, <ScienceProduct#uid://vlba/correlation/3124cf11-5793-42d4-b909-9cd4b17501ab type=Execution Block>, <ScienceProduct#uid://vlba/correlation/c0808e39-fb9c-42dd-bb13-a5ace0024752 type=Execution Block>, <ScienceProduct#uid://vlba/correlation/1834cd4a-9390-4a03-874e-1acf652d717d type=Execution Block>, <ScienceProduct#uid://vlba/correlation/70d70b22-d1bd-4ca5-a426-2d316ea5cf29 type=Execution Block>, <ScienceProduct#uid://vlba/correlation/8bc3d9c3-b85d-4758-9931-5ffd5885f1e7 type=Execution Block>]
OGAPs: [{'type': 'observation_log', 'filename': 'VLBA_TR039HD_tr039hd_250321T142044.difx.log.gz'}, {'type': 'observation_details', 'filename': 'TR039HD.tr039hd.sum.txt'}, {'type': 'ingestion_artifacts', 'filename': 'ingestion_artifacts_2025_03_21T09_08_50.708.tar'}]
mous: None
"""


@pytest.fixture
def expected_curator_params() -> Generator[dict[str, None | list[Any] | str], Any, None]:
    """
    Yields the expected parameters for the curator launcher.
    """
    obs_launcher_settings = expected_settings_evla_eb()
    yield {
        **obs_launcher_settings,
        "product_type": "execution_block",
        "input_group_locator": None,
        "curation_source": f"{obs_launcher_settings['staging_area']}/{obs_launcher_settings['workflowDir']}",
        "file_list": None,
        "product_group_id": None,
        "target_list": [],
        "spl": expected_spls[0],
    }


def test_get_ingest_product_locators():
    """
    Tests the get_ingest_product_locators function.
    """
    assert launchers.get_ingest_spls(sample_input) == expected_spls


@patch("ingest_envoy.launchers.run_process_with_realtime_logging_and_search")
def test_trigger_ingest(mock_run):
    """
    Tests the trigger_ingest function.
    """
    # Create a mock return value - CompletedProcess object and search result list
    mock_process = MagicMock()
    mock_process.stdout = sample_input
    mock_process.returncode = 0
    mock_search_results = ["spl1", "spl2"]  # Example SPL results

    # Set the return value as a tuple (CompletedProcess, list)
    mock_run.return_value = (mock_process, mock_search_results)

    # Don't actually run ingest
    result = launchers.trigger_ingest(
        parameters["useIngest"], parameters["destinationDir"], parameters["script_location"]
    )
    assert mock_run.call_count == 0
    assert result == (0, [])

    # "Actually" run ingest (we are mocking subprocess.Popen)
    result = launchers.trigger_ingest(True, parameters["destinationDir"], parameters["script_location"])
    assert mock_run.call_count == 1
    assert result == (mock_process.returncode, mock_search_results)

    # Verify the call arguments
    mock_run.assert_called_with(
        [f"{parameters['script_location']}/ingest", "--json", "-p", parameters["destinationDir"]],
        search_method=launchers.get_ingest_spls,  # Should be the function reference, not the called function
        stdout_handler=launchers.logger.info,
    )


@patch("subprocess.run")
def test_trigger_curator(mock_run, expected_curator_params):
    """
    Tests the trigger_curator function.
    """
    # Create a mock return value (a CompletedProcess[str] object)
    mock_process = MagicMock()
    mock_process.stdout = sample_input
    mock_process.returncode = 0

    # Set this as the return value of subprocess.run
    mock_run.return_value = mock_process

    # Don't actually run curator
    launchers.trigger_curator(
        False, Path(expected_curator_params["curation_source"]), expected_curator_params["script_location"]
    )
    assert mock_run.call_count == 0

    # "Actually" run curator (we are mocking subprocess.run)
    launchers.trigger_curator(
        True, Path(expected_curator_params["curation_source"]), expected_curator_params["script_location"]
    )
    assert mock_run.call_count == 1
    mock_run.assert_called_with(
        [
            f"{expected_curator_params['script_location']}/curator",
            "-m",
            f"{expected_curator_params['curation_source']}/manifest.json",
        ],
        stdout=sys.stdout,
        stderr=sys.stderr,
    )


class TestRunProcessWithRealtimeLoggingAndSearch(unittest.TestCase):
    @patch("subprocess.Popen")
    def test_successful_process_without_search(self, mock_popen):
        # Setup mock process
        mock_process = MagicMock()
        mock_process.stdout = ["output line 1\n", "output line 2\n"]
        mock_process.wait.return_value = 0
        mock_process.args = ["command", "arg1"]

        # Configure the mock to return the mock_process
        mock_popen.return_value.__enter__.return_value = mock_process

        # Mock logger
        mock_logger = MagicMock()

        # Execute function
        completed_process, search_results = launchers.run_process_with_realtime_logging_and_search(
            args=["command", "arg1"], search_method=None, stdout_handler=mock_logger
        )

        # Verify results
        self.assertEqual(completed_process.returncode, 0)
        self.assertEqual(completed_process.args, ["command", "arg1"])
        self.assertEqual(search_results, [])

        # Verify logger was called
        mock_logger.assert_has_calls([call("output line 1"), call("output line 2")])

    @patch("subprocess.Popen")
    def test_with_search_method(self, mock_popen):
        # Setup mock process
        mock_process = MagicMock()
        mock_process.stdout = ["error: found a bug\n", "nothing to see here", "warning: check this\n"]
        mock_process.wait.return_value = 0
        mock_process.args = ["command", "arg1"]

        # Configure the mock to return the mock_process
        mock_popen.return_value.__enter__.return_value = mock_process

        # Create a search method that finds "error" and "warning" keywords
        def mock_search(line):
            if "error" in line:
                return ["error found"]
            elif "warning" in line:
                return ["warning found"]
            return []

        # Execute function
        completed_process, search_results = launchers.run_process_with_realtime_logging_and_search(
            args=["command", "arg1"], search_method=mock_search, stdout_handler=logging.info
        )

        # Verify results
        self.assertEqual(completed_process.returncode, 0)
        self.assertEqual(search_results, ["error found", "warning found"])

    @patch("subprocess.Popen")
    def test_process_with_error(self, mock_popen):
        # Setup mock process
        mock_process = MagicMock()
        mock_process.stdout = ["error output\n"]
        mock_process.wait.return_value = 1  # Non-zero return code
        mock_process.args = ["command", "arg1"]

        # Configure the mock to return the mock_process
        mock_popen.return_value.__enter__.return_value = mock_process

        # Execute function and expect CalledProcessError
        with self.assertRaises(CalledProcessError):
            launchers.run_process_with_realtime_logging_and_search(
                args=["command", "arg1"], search_method=None, stdout_handler=logging.info
            )

    @patch("subprocess.Popen")
    def test_process_with_error_check_false(self, mock_popen):
        # Setup mock process
        mock_process = MagicMock()
        mock_process.stdout = ["error output\n"]
        mock_process.wait.return_value = 1  # Non-zero return code
        mock_process.args = ["command", "arg1"]

        # Configure the mock to return the mock_process
        mock_popen.return_value.__enter__.return_value = mock_process

        # Execute function with check=False
        completed_process, search_results = launchers.run_process_with_realtime_logging_and_search(
            args=["command", "arg1"], search_method=None, stdout_handler=logging.info, check=False
        )

        # Verify results - should return with error code but not raise exception
        self.assertEqual(completed_process.returncode, 1)
        self.assertEqual(search_results, [])


class TestCuratorLauncher:
    """
    Tests for the CuratorLauncher class.
    """

    @staticmethod
    @pytest.fixture
    def curator_launcher(expected_curator_params):
        """
        Yields the CuratorLauncher instance used in testing.
        """
        yield CuratorLauncher(CuratorType.PARTIAL, expected_curator_params)

    def test_init(self, curator_launcher, expected_curator_params):
        """
        Tests that the class initializes correctly.
        """
        assert curator_launcher.curator_type == CuratorType.PARTIAL
        assert curator_launcher.curation_source == expected_curator_params["curation_source"]
        assert curator_launcher.sci_product_type == IngestType.OBS.value
        assert curator_launcher.manifest_destination_dir == Path.cwd()
        assert curator_launcher.parameters == expected_curator_params

    def test_prepare_for_launch(self, curator_launcher):
        """
        Tests the prepare_for_launch method.
        """
        with patch("ingest_envoy.launchers.CuratorLauncher.create_manifest") as mock_create_manifest:
            curator_launcher.prepare_for_launch()
            assert mock_create_manifest.call_count == 1

    def test_create_manifest(self, curator_launcher, expected_curator_params):
        """
        Tests the create_manifest method.
        """

        with patch("ingest_envoy.ingestion_manifest.IngestionManifestBuilder.build") as mock_manifest_build, patch(
            "ingest_envoy.ingestion_manifest.IngestionManifestBuilder.__init__"
        ) as mock_init:
            mock_init.return_value = None  # Otherwise, we get a TypeError

            curator_launcher.create_manifest()

            assert mock_init.call_count == 1
            mock_init.assert_called_with(
                manifest_source_dir=expected_curator_params["curation_source"],
                sp_type=expected_curator_params["product_type"],
                locator=expected_curator_params["spl"],
                telescope=expected_curator_params["telescope"],
                curate=(CuratorType.PARTIAL, expected_curator_params["target_list"]),
                manifest_destination_dir=Path.cwd(),
                file_list=expected_curator_params["file_list"],
                product_group=expected_curator_params["product_group_id"],
                input_group_locator=expected_curator_params["input_group_locator"],
            )

            assert mock_manifest_build.call_count == 1

    def test_launch(self, curator_launcher):
        """
        Tests the launch method.
        """
        with patch("ingest_envoy.launchers.CuratorLauncher.prepare_for_launch") as mock_prepare, patch(
            "ingest_envoy.launchers.trigger_curator"
        ) as mock_trigger_curator:
            mock_trigger_curator.return_value = 0
            curator_returncode, run_indexer = curator_launcher.launch()

            assert mock_prepare.call_count == 1
            assert mock_trigger_curator.call_count == 1
            assert curator_returncode == 0
            assert run_indexer is True

    def test_launch_trigger_curator_failure(self, curator_launcher):
        """
        Tests that the launch method returns the trigger_curator return code,
        even when it is a nonzero value.
        """
        with patch("ingest_envoy.launchers.CuratorLauncher.prepare_for_launch"), patch(
            "ingest_envoy.launchers.trigger_curator"
        ) as mock_trigger_curator:
            mock_trigger_curator.return_value = 1
            curator_returncode, run_indexer = curator_launcher.launch()

            assert curator_returncode == 1
            assert run_indexer is True


class TestIngestCalibrationLauncher:
    def test_launch_ingestion(self):
        with patch("ingest_envoy.launchers.IngestCalibrationLauncher.prepare_for_launch") as prepare:
            with patch("ingest_envoy.launchers.trigger_ingest") as ingest:
                ingest.return_value = (0, [expected_spls[0]])
                ingest_returncode, run_indexer = IngestCalibrationLauncher(IngestType.CAL, parameters).launch()
                assert ingest_returncode == 0
                assert run_indexer is True
                assert prepare.call_count == 1
                assert ingest.call_count == 1

    @patch("subprocess.run", return_value=0)
    def test_prepare_for_ingest(self, mock_run):
        with patch("ingest_envoy.launchers.IngestCalibrationLauncher.run_collection_script") as collector:
            with patch("ingest_envoy.launchers.IngestCalibrationLauncher.create_manifest") as manifest:
                IngestCalibrationLauncher(IngestType.CAL, parameters).prepare_for_launch()
                assert collector.call_count == 1
                assert manifest.call_count == 1

    @patch("subprocess.run")
    def test_run_collection_script(self, mock_run):
        mock_run.return_value.returncode = 0
        IngestCalibrationLauncher(IngestType.CAL, parameters).run_collection_script()
        assert mock_run.call_count == 1

    @pytest.mark.skip("Skip. Ignores manifest builder mock")
    def test_create_manifest(self):
        with patch("ingest_envoy.ingestion_manifest.IngestionManifestBuilder.build") as manifest:
            IngestCalibrationLauncher(IngestType.CAL, parameters).create_manifest()
            assert manifest.call_count == 1


class TestIngestImageLauncher:
    def test_launch_ingestion(self):
        with patch("ingest_envoy.launchers.IngestImageLauncher.prepare_for_launch") as prepare:
            with patch("ingest_envoy.launchers.trigger_ingest") as ingest:
                ingest.return_value = (0, [expected_spls[0]])
                ingest_returncode, run_indexer = IngestImageLauncher(IngestType.IMG, image_parameters).launch()

                assert ingest_returncode == 0
                assert run_indexer is True
                assert prepare.call_count == 1
                assert ingest.call_count == 1

    def test_prepare_for_ingest(self):
        with patch("ingest_envoy.launchers.IngestImageLauncher.run_collector") as collector:
            with patch("ingest_envoy.launchers.IngestImageLauncher.create_manifest") as manifest:
                IngestImageLauncher(IngestType.IMG, image_parameters).prepare_for_launch()
                assert collector.call_count == 1
                assert manifest.call_count == 1

    @patch("builtins.open")
    @patch("subprocess.run")
    def test_run_collector(self, mock_run, mock_open):
        with patch("ingest_envoy.collectors.ImageCollector.collect_products") as products:
            mock_run.return_value.returncode = 0
            aux_file = IngestImageLauncher(IngestType.IMG, image_parameters).run_collector()
            assert products.call_count == 1
            assert isinstance(aux_file, AbstractTextFile)

    @patch("builtins.open")
    @patch("subprocess.run")
    def test_run_coarse_collector(self, mock_run, mock_open):
        with patch("ingest_envoy.collectors.CoarseCollector.collect_products") as products:
            mock_run.return_value.returncode = 0
            aux_file = IngestVLASSLauncher(VLASSIngestType.COARSE, coarse_parameters).run_collector()
            assert products.call_count == 1
            assert isinstance(aux_file, AbstractTextFile)

    @pytest.mark.skip("Skip. Ignores manifest builder mock")
    def test_create_manifest(self):
        with patch("ingest_envoy.ingestion_manifest.IngestionManifestBuilder.build") as manifest:
            IngestImageLauncher(IngestType.IMG, image_parameters).create_manifest()
            assert manifest.call_count == 1


class TestIngestObservationLauncher:
    """
    Tests for the IngestObservationLauncher class.
    """

    @staticmethod
    @pytest.fixture
    def observation_launcher():
        """
        Yields the IngestObservationLauncher instance used in testing.
        """
        yield IngestObservationLauncher(IngestType.OBS, expected_settings_evla_eb())

    def test_init(self, observation_launcher):
        """
        Tests that the class initializes correctly.
        """
        expected_parameters = expected_settings_evla_eb()
        assert observation_launcher.sci_product_type == IngestType.OBS.value
        assert observation_launcher.parameters == expected_parameters

        expected_staging_source_dir = f"{expected_parameters['staging_area']}/{expected_parameters['workflowDir']}"
        assert observation_launcher.staging_source_dir == expected_staging_source_dir

    def test_launch_curator_for_spls(self, observation_launcher):
        """
        Tests the launch_curator_for_spls method for normal execution (nothing in the returned dictionary).
        """
        with patch("ingest_envoy.launchers.CuratorLauncher.launch") as mock_curator_launch:
            mock_curator_launch.return_value = (0, True)
            assert observation_launcher.launch_curator_for_spls(expected_spls) == {}

    def test_launch_curator_for_spls_curator_fails(self, observation_launcher):
        """
        Tests the launch_curator_for_spls method for when curator fails.
        """
        with patch("ingest_envoy.launchers.CuratorLauncher.launch") as mock_curator_launch:
            mock_trigger_curator_returncode = 2
            mock_curator_launch.return_value = (mock_trigger_curator_returncode, True)
            expected_results = {spl: mock_trigger_curator_returncode for spl in expected_spls}

            actual_results = observation_launcher.launch_curator_for_spls(expected_spls)

            assert actual_results == expected_results

    def test_launch(self, observation_launcher):
        """
        Tests the launch method.
        """
        with patch("ingest_envoy.launchers.IngestObservationLauncher.prepare_for_launch") as mock_prepare, patch(
            "ingest_envoy.launchers.trigger_ingest"
        ) as mock_trigger_ingest, patch(
            "ingest_envoy.launchers.IngestObservationLauncher.launch_curator_for_spls"
        ) as mock_launch_curator_for_spls:
            mock_trigger_ingest.return_value = (0, [expected_spls[0]])
            mock_launch_curator_for_spls.return_value = {}  # Simulate all SPLs having curator launched successfully

            ingest_returncode, run_indexer = observation_launcher.launch()

            assert mock_prepare.call_count == 1
            assert mock_trigger_ingest.call_count == 1
            assert ingest_returncode == 0
            assert run_indexer is True
            assert mock_launch_curator_for_spls.call_count == 1

    def test_launch_trigger_ingest_failure_no_curator(self, observation_launcher):
        """
        Test that when trigger_ingest fails, launch_curator_for_spls isn't run.
        """
        with patch("ingest_envoy.launchers.IngestObservationLauncher.prepare_for_launch"), patch(
            "ingest_envoy.launchers.trigger_ingest"
        ) as mock_trigger_ingest, patch(
            "ingest_envoy.launchers.IngestObservationLauncher.launch_curator_for_spls"
        ) as mock_launch_curator_for_spls:
            mock_trigger_ingest.return_value = (2, [])

            ingest_returncode, run_indexer = observation_launcher.launch()

            assert ingest_returncode == 2
            assert run_indexer is True
            assert mock_launch_curator_for_spls.call_count == 0

    def test_launch_curator_launch_fails(self, observation_launcher, caplog):
        """
        Test that when trigger_ingest succeeds but launch_curator_for_spls fails for at least one SPL,
         the failed SPLs are logged out.
        """
        with patch("ingest_envoy.launchers.IngestObservationLauncher.prepare_for_launch"), patch(
            "ingest_envoy.launchers.trigger_ingest"
        ) as mock_trigger_ingest, patch(
            "ingest_envoy.launchers.IngestObservationLauncher.launch_curator_for_spls"
        ) as mock_launch_curator_for_spls:
            mock_trigger_ingest.return_value = (0, [expected_spls[0]])
            mock_launch_curator_for_spls.return_value = {expected_spls[0]: 1}  # Simulate curator failing for an SPL
            caplog.set_level(logging.ERROR)

            ingest_returncode, run_indexer = observation_launcher.launch()

            assert ingest_returncode == 0
            assert run_indexer is True
            assert mock_launch_curator_for_spls.call_count == 1
            assert caplog.record_tuples == [
                ("ingest_envoy", logging.ERROR, f"Curator failed: {{'{expected_spls[0]}': 1}}")
            ]

    def test_prepare_for_launch(self, observation_launcher):
        """
        Tests the prepare_for_launch method.
        """
        with patch("ingest_envoy.launchers.IngestObservationLauncher.run_collector") as mock_run_collector, patch(
            "ingest_envoy.launchers.IngestObservationLauncher.create_manifest"
        ) as mock_create_manifest:
            observation_launcher.prepare_for_launch()

            assert mock_run_collector.call_count == 1
            assert mock_create_manifest.call_count == 1

    def test_run_collector(self, observation_launcher):
        """
        Tests the run_collector method.
        """
        with patch("ingest_envoy.collectors.ObservationCollector.collect_products") as mock_collect_products:
            observation_launcher.run_collector()
            assert mock_collect_products.call_count == 1

    def test_create_manifest(self, observation_launcher):
        """
        Tests the create_manifest method.
        """
        expected_parameters = expected_settings_evla_eb()

        with patch("ingest_envoy.ingestion_manifest.IngestionManifestBuilder.build") as mock_manifest_build, patch(
            "ingest_envoy.ingestion_manifest.IngestionManifestBuilder.__init__"
        ) as mock_init:
            mock_init.return_value = None  # Otherwise, we get a TypeError

            observation_launcher.create_manifest()

            assert mock_init.call_count == 1
            mock_init.assert_called_with(
                manifest_source_dir=Path(observation_launcher.staging_source_dir),
                sp_type=IngestType.OBS.value,
                locator=None,
                telescope=expected_parameters["telescope"],
                filename=expected_parameters["filename"],
            )

            assert mock_manifest_build.call_count == 1
