#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import shutil
from pathlib import Path
from typing import Any, Generator

import pytest
from ingest_envoy.product_collectors import ObservationProductCollector
from pycapo import CapoConfig

SDM_DIR = "19A-001.sb1234567890.eb233423545632.54321.894327984569"
VLBA_DIR = "BD276G3.bd276g3"
GMVA_DIR = "GMVA_C202"

STAGING = Path(CapoConfig().getstring("edu.nrao.workspaces.IngestionSettings.stagingDirectory"))
STORAGE = Path(CapoConfig().getstring("edu.nrao.workspaces.IngestionSettings.storageDirectory"))

SDM_FILES = [
    "Antenna.xml",
    "ASDM.xml",
    "CalData.xml",
    "CalDevice.xml",
    "CalPointing.xml",
    "CalReduction.xml",
    "ConfigDescription.xml",
    "CorrelatorMode.xml",
    "DataDescription.xml",
    "Doppler.xml",
    "Ephemeris.xml",
    "ExecBlock.xml",
    "Feed.xml",
    "Field.xml",
    "Flag.xml",
    "Main.xml",
    "PointingModel.xml",
    "Pointing.xml",
    "Polarization.xml",
    "Processor.xml",
    "Receiver.xml",
    "SBSummary.xml",
    "Scan.xml",
    "Source.xml",
    "SpectralWindow.xml",
    "State.xml",
    "Station.xml",
    "Subscan.xml",
    "SwitchCycle.xml",
    "SysCal.xml",
    "Weather.xml",
]

VLBA_FILES = [
    "BD276G3.bd276g3.sum.txt",
    "VLBA_BD276G3_bd276g3_250408T141334.difx.log.gz",
    "VLBA_BD276G3_bd276g3_BIN0_SRC0_0_250408T141334.idifits",
    "VLBA_BD276G3_bd276g3_BIN0_SRC1_0_250408T141334.idifits",
]

GMVA_FILES = [
    "GMVA_c202cPart1.idifits",
]


def stub_area(src_path: Path, dir_name: str, files: list) -> Generator[Path, Any, None]:
    """
    Fixture that creates a temporary mock ingest input directory.

    :return: Path to the base of the directory
    """
    # Make spool, staging, and storage directories
    for d in [src_path / dir_name, STAGING / dir_name, STORAGE / dir_name]:
        d.mkdir(parents=True, exist_ok=True)

    # Populate dir
    workflow_path = src_path / dir_name
    for f in files:
        (workflow_path / f).touch()

    yield workflow_path

    shutil.rmtree(src_path / dir_name)
    shutil.rmtree(STAGING / dir_name)
    shutil.rmtree(STORAGE / dir_name)

    assert not (src_path / dir_name).exists()
    assert not (STAGING / dir_name).exists()
    assert not (STORAGE / dir_name).exists()


@pytest.fixture(scope="function")
def stub_sdm_area(tmp_path) -> Generator[Path, Any, None]:
    """
    Fixture that creates a temporary mimic of an SDM directory.

    :return: Path to the base of the directory
    """
    yield from stub_area(tmp_path, SDM_DIR, SDM_FILES)


@pytest.fixture(scope="function")
def stub_vlba_area(tmp_path) -> Generator[Path, Any, None]:
    """
    Fixture that creates a temporary mimic of a VLBA directory.

    :return: Path to the base of the directory
    """
    yield from stub_area(tmp_path, VLBA_DIR, VLBA_FILES)


@pytest.fixture(scope="function")
def stub_gmva_area(tmp_path) -> Generator[Path, Any, None]:
    """
    Fixture that creates a temporary mimic of a GMVA directory.

    :return: Path to the base of the directory
    """
    yield from stub_area(tmp_path, GMVA_DIR, GMVA_FILES)


def test_obs_product_collector_sdm(stub_sdm_area: Path):
    """
    Test that everything from an SDM source directory was copied into staging.
    """
    collector = ObservationProductCollector(stub_sdm_area, STAGING)
    collector()
    for f in SDM_FILES:
        assert (STAGING / SDM_DIR / f).exists()


def test_obs_product_collector_vlba(stub_vlba_area: Path):
    """
    Test that everything from a VLBA source directory was copied into staging.
    """
    collector = ObservationProductCollector(stub_vlba_area, STAGING)
    collector()
    for f in VLBA_FILES:
        assert (STAGING / VLBA_DIR / f).exists()


def test_obs_product_collector_gmva(stub_gmva_area: Path):
    """
    Test that everything from a GMVA source directory was copied into staging.
    """
    collector = ObservationProductCollector(stub_gmva_area, STAGING)
    collector()
    for f in GMVA_FILES:
        assert (STAGING / GMVA_DIR / f).exists()
