#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for solicitor.py
"""

# pylint: disable=E0401, C0301, C0116, R0201, W0621
from unittest.mock import patch

import pytest
from conftest import (
    SECI_SCIENCE_PRODUCT,
    expected_settings_evla_eb,
    mocked_archive_service_get,
)
from ingest_envoy.solicitor import Solicitor
from ingest_envoy.utilities import CuratorType, IngestType, VLASSIngestType

SOLICIT_WORKFLOWS_PATCH = "ingest_envoy.solicitor.Solicitor.solicit_contents"
SOLICITOR_URLS = ["http://capability:3457", "http://workflow:3456", "fake-archive-service"]


@pytest.fixture(scope="function")
def solicitor() -> Solicitor:
    filename = "test/input_files/test-metadata.json"
    return Solicitor(IngestType.CAL, SOLICITOR_URLS, filename=filename)


@pytest.fixture(scope="function")
def image_solicitor() -> Solicitor:
    image_filename = "test/input_files/test-image-metadata.json"
    return Solicitor(IngestType.IMG, SOLICITOR_URLS, filename=image_filename)


@pytest.fixture(scope="function")
def image_solicitor_nocal() -> Solicitor:
    image_filename = "test/input_files/test-image-nocal-metadata.json"
    return Solicitor(IngestType.IMG, SOLICITOR_URLS, filename=image_filename)


@pytest.fixture(scope="function")
def coarse_solicitor() -> Solicitor:
    return Solicitor(VLASSIngestType.COARSE, SOLICITOR_URLS, filename=None)


expected_metadata = {
    "fileSetIds": "16B-069_sb32814386_1_001.57685.66193635417",
    "workflowName": "std_calibration",
    "systemId": "2",
    "creationTime": "2021-07-06T21:50:48",
    "productLocator": "uid://evla/execblock/48ba4c9d-d7c7-4a8f-9803-1115cd52459b",
    "projectMetadata": {
        "projectCode": "16B-069",
        "title": "JVLA monitoring of an in-progress extreme scattering event",
        "telescope": "EVLA",
        "startTime": "57685.661952546296",
        "observer": "Keith Bannister",
    },
    "destinationDirectory": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmp576dknjp",
}

expected_image_metadata = {
    "sdmId": "brain_000.58099.67095825232",
    "telescope": "EVLA",
    "workflowName": "std_cms_imaging",
    "processingStart": "2021_07_29T14_26_31",
    "workflowDir": "tmprb1se376",
    "initialVersionDir": "iamthefirst",
    "multiVersion": True,
    "calSpl": "uid://evla/calibration/c78ccfd6-fe4e-43c6-a5c5-70e5bcfde16b",
    "project": "Operations",
    "requestId": "12",
    "spoolDir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmprb1se376",
}

expected_coarse_metadata = {
    "telescope": "EVLA",
    "workflowName": "vlass_coarse",
    "project": "TEST000",
    "multiVersion": False,
}


class TestSolicitor:
    def test_solicit_contents(self, solicitor: Solicitor):
        metadata = solicitor.solicit_contents(solicitor.filename)
        assert metadata == expected_metadata

    def test_solicit_workflow_directory_name(self, solicitor: Solicitor):
        expected_workflow_dir_name = "tmp576dknjp"

        name = solicitor.solicit_workflow_directory_name()
        assert name == expected_workflow_dir_name

    def test_solicit_processing_creation_time(self, solicitor: Solicitor):
        expected_corrected_time = "2021_07_06T21_50_48"

        time_string = solicitor.solicit_processing_creation_time()
        assert time_string == expected_corrected_time

    def test_solicit_parameters_std_cal(self, solicitor: Solicitor):
        expected_parameters = {
            "sdmId": "16B-069_sb32814386_1_001.57685.66193635417",
            "telescope": "EVLA",
            "workflowName": "std_calibration",
            "spl": "uid://evla/execblock/48ba4c9d-d7c7-4a8f-9803-1115cd52459b",
            "processingStart": "2021_07_06T21_50_48",
            "workflowDir": "tmp576dknjp",
            "initialVersionDir": "tmp24i4o1yl",
            "multiVersion": True,
            "project": "16B-069",
            "requestId": "2",
        }
        with patch("ingest_envoy.solicitor.Solicitor.solicit_initial_directory_name", return_value="tmp24i4o1yl"):
            parameters = solicitor.solicit_parameters()
            assert parameters == expected_parameters

    def test_solicit_parameters_image(self, image_solicitor: Solicitor):
        initial_version_dir = "iamthefirst"
        with patch("ingest_envoy.solicitor.Solicitor.solicit_initial_directory_name", return_value=initial_version_dir):
            parameters = image_solicitor.solicit_parameters()
            assert parameters == expected_image_metadata

    def test_solicit_parameters_image_without_cal_spl(self, image_solicitor_nocal: Solicitor):
        initial_version_dir = "iamthefirst"
        with (
            patch("ingest_envoy.solicitor.Solicitor.solicit_initial_directory_name", return_value=initial_version_dir),
            patch(
                "ingest_envoy.solicitor.Solicitor.lookup_calibration_product_locator",
                return_value="uid://evla/calibration/c78ccfd6-fe4e-43c6-a5c5-70e5bcfde16b",
            ),
        ):
            parameters = image_solicitor_nocal.solicit_parameters()
            assert parameters == expected_image_metadata

    def test_solicit_parameters_coarse_aux_metadata(self, coarse_solicitor: Solicitor, aux_data: dict):
        with patch("ingest_envoy.solicitor.Solicitor.solicit_aux_metadata", return_value=aux_data):
            parameters = coarse_solicitor.solicit_parameters()
            assert parameters == expected_coarse_metadata

    def test_solicit_parameters_coarse_precursor_spl(self, coarse_solicitor: Solicitor):
        with patch("requests.get", side_effect=mocked_archive_service_get):
            locator = coarse_solicitor.solicit_calibration_product_locator(SECI_SCIENCE_PRODUCT["external_name"])
            assert locator == SECI_SCIENCE_PRODUCT["science_product_locator"]

    def test_solicit_observation_params(self, mock_evla_eb_solicitor_solicit_contents):
        """
        Tests the solicit_observation_params method.
        """
        evla_eb_settings = expected_settings_evla_eb()

        # Drop all but the key/values produced by solicit
        evla_eb_settings = {
            "filename": evla_eb_settings["filename"],
            "telescope": evla_eb_settings["telescope"],
            "project": evla_eb_settings["project"],
        }

        solicitor = Solicitor(IngestType.OBS, SOLICITOR_URLS, evla_eb_settings["filename"])
        assert solicitor.solicit_observation_params() == evla_eb_settings

    def test_solicit_parameters_full_curation_image_file_list(self):
        filename = "test/input_files/test-full-curation-image-evla-metadata.json"
        expected_parameters = {
            "telescope": "EVLA",
            "project": "20B-377",
            "product_group_id": 328918,
            "product_type": "image",
            "curation_source": None,
            "target_list": None,
            "spl": None,
            "input_group_locator": "uid://evla/calibration/620195ad-1d11-49f4-be1f-3532092565c1",
            "file_list": [
                "20B-377.MJD59909.70629903935.AT2019teq_sci.X_band.cont.I.mask.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.X_band.cont.I.tt1.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.X_band.cont.I.pbcor.tt0.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.X_band.cont.I.pb.tt0.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.X_band.cont.I.alpha.error.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.X_band.cont.I.alpha.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.X_band.cont.I.tt0.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.C_band.cont.I.pb.tt0.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.C_band.cont.I.alpha.error.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.C_band.cont.I.pbcor.tt0.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.C_band.cont.I.tt0.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.C_band.cont.I.alpha.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.C_band.cont.I.tt1.fits",
                "20B-377.MJD59909.70629903935.AT2019teq_sci.C_band.cont.I.mask.fits",
                "ingestion_artifacts_2022_12_14T06_22_03.874.tar",
                "weblog.tgz",
                "pipeline_artifacts_2022_12_09T19_08_32.tar",
            ],
        }
        full_curation_image_files_list_solicitor = Solicitor(CuratorType.FULL, SOLICITOR_URLS, filename)
        actual_parameters = full_curation_image_files_list_solicitor.solicit_parameters()
        assert len(actual_parameters.keys()) == len(expected_parameters.keys())
        assert set(actual_parameters.keys()) == set(expected_parameters.keys())
        for key, value in expected_parameters.items():
            if key != "file_list":
                assert value == actual_parameters[key]
            else:
                assert len(value) == len(actual_parameters[key])
                assert set(value) == set(actual_parameters[key])

    def test_solicit_parameters_targeted_curation(self):
        filename = "test/input_files/test-targeted-curation-metadata.json"
        expected_parameters = {
            "telescope": "EVLA",
            "project": "16B-069",
            "product_group_id": None,
            "product_type": "execution_block",
            "curation_source": None,
            "target_list": ["subscans.ra", "subscans.dec"],
            "spl": "uid://evla/execblock/9b60aa94-d037-4edb-bb1a-d642aa1fa441",
            "input_group_locator": None,
            "file_list": [],
        }
        targeted_curation_solicitor = Solicitor(CuratorType.PARTIAL, SOLICITOR_URLS, filename)
        actual_parameters = targeted_curation_solicitor.solicit_parameters()
        assert len(actual_parameters.keys()) == len(expected_parameters.keys())
        assert set(actual_parameters.keys()) == set(expected_parameters.keys())
        for key, value in expected_parameters.items():
            if key != "target_list":
                assert value == actual_parameters[key]
            else:
                assert len(value) == len(actual_parameters[key]) == 2
                assert set(value) == set(actual_parameters[key])
