# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
from pathlib import Path

from conftest import (
    VLASS_COARSE_CUBE_ANCILLARIES,
    VLASS_COARSE_CUBE_CACHE,
    VLASS_COARSE_CUBE_SCIENCE_PRODUCTS,
)
from ingest_envoy.product_collectors import VlassCacheProductCollector


def test_coarse_cube_cache(vlass_coarse_path: Path):
    """Test cache transfer for a coarse cube"""
    # Call the collector
    collector = VlassCacheProductCollector(vlass_coarse_path, "PASS")
    collector()
    cache_path = collector.get_cache_dir()

    # Verify science products are cached
    for sp in VLASS_COARSE_CUBE_SCIENCE_PRODUCTS:
        assert (cache_path / sp.filename).exists()
        for ap in sp.ancillary_products:
            assert (cache_path / ap.filename).exists()

    # Verify group ancillaries are cached
    for f in VLASS_COARSE_CUBE_ANCILLARIES:
        assert (cache_path / f.filename).exists()

    # Verify extra files are cached
    for f in VLASS_COARSE_CUBE_CACHE["products"]:
        assert (cache_path / f).exists()
    for f in VLASS_COARSE_CUBE_CACHE["working"]:
        assert (cache_path / f).exists()


def test_coarse_cube_cache_planes(vlass_coarse_path: Path):
    """Test cache transfer for a coarse cube filtering out some planes"""
    # Accepted list is for filtering images in the cache, rejected list is for verification
    accepted_planes = ["spw10", "spw12"]
    rejected_planes = ["spw11"]

    # Add an accepted_planes.txt file to filter out some images
    with open(vlass_coarse_path / "products" / "accepted_planes.txt", "w") as f:
        for plane in accepted_planes:
            f.write(f"{plane}\n")

    # Call the collector
    collector = VlassCacheProductCollector(vlass_coarse_path, "PASS")
    collector()
    cache_path = collector.get_cache_dir()

    # Verify only accepted science products are cached
    for sp in VLASS_COARSE_CUBE_SCIENCE_PRODUCTS:
        # Check if this plane is accepted
        should_exist = False
        for plane in accepted_planes:
            if plane in sp.filename:
                should_exist = True

        # Mean Stokes V image should always exist
        if "V_mean_minmaxrej" in sp.filename:
            should_exist = True

        if should_exist:
            # Verify in cache
            assert (cache_path / sp.filename).exists()
            for ap in sp.ancillary_products:
                assert (cache_path / ap.filename).exists()
        else:
            # Verify not in cache
            assert not (cache_path / sp.filename).exists()
            for ap in sp.ancillary_products:
                assert not (cache_path / ap.filename).exists()

    # Verify group ancillaries are cached
    for f in VLASS_COARSE_CUBE_ANCILLARIES:
        assert (cache_path / f.filename).exists()

    # Verify extra files are cached
    for f in VLASS_COARSE_CUBE_CACHE["products"]:
        # This list contains plane 'com' files, check if this plane is rejected (since other files don't contain spw)
        should_exist = True
        for plane in rejected_planes:
            if plane in f:
                should_exist = False

        if should_exist:
            # Verify in cache
            assert (cache_path / f).exists()
        else:
            # Verify not in cache
            assert not (cache_path / f).exists()
    for f in VLASS_COARSE_CUBE_CACHE["working"]:
        assert (cache_path / f).exists()
