# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Runs the tests for the VLASS coarse cube manifest generation.
"""
import json
import re
import shutil
import tarfile
from pathlib import Path

import pytest
from conftest import (
    AUX_IMAGE_METADATA_FILENAME,
    find_example_manifest,
    find_ingestion_artifacts_tar,
    ordered,
    populate_directory,
)
from ingest_envoy.ingestion_manifest import (
    IngestionManifest,
    IngestionManifestBuilder,
    IngestionManifestSchema,
)
from ingest_envoy.manifest_components import MANIFEST_FILENAME
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import ScienceProductType, Telescope


@pytest.fixture(name="vlass_coarse_cube_manifest")
def build_manifest(
    vlass_coarse_cube_staging_dir: Path, vlass_coarse_staging_products: list[str]
) -> tuple[IngestionManifest, Path]:
    """
    Builds the ingestion manifest and returns it as well as a Path of the file it came from.
    :param vlass_coarse_cube_staging_dir: Path. The path to the temporary staging directory.
    :param vlass_coarse_staging_products: list[str]. The list of products that should be in the staging directory.
    :return: tuple[IngestionManifest, Path]. The manifest and the file it came from.
    """
    vlass_coarse_cube_staging_dir.mkdir(parents=True)
    populate_directory(vlass_coarse_cube_staging_dir, vlass_coarse_staging_products + [AUX_IMAGE_METADATA_FILENAME])

    yield IngestionManifestBuilder(
        manifest_source_dir=vlass_coarse_cube_staging_dir,
        telescope=Telescope.EVLA.value,
        additional_metadata=AbstractTextFile(filename=AUX_IMAGE_METADATA_FILENAME, content=""),
        sp_type=ScienceProductType.VLASS_COARSE.value,
        locator="uid://evla/calibration/ba6c6404-b196-4d5d-9ab8-8e2b5bce62c1",
    ).build()

    # Remove staging directory
    shutil.rmtree(vlass_coarse_cube_staging_dir)


@pytest.fixture(name="actual_manifest_json")
def get_actual_manifest_json(vlass_coarse_cube_manifest) -> dict:
    """
    :param vlass_coarse_cube_manifest: tuple[IngestionManifest, Path]. A Pytest fixture.
    :return: dict. The dict representation of the actual (generated) manifest JSON.
    """
    _, actual_manifest_file = vlass_coarse_cube_manifest
    with open(actual_manifest_file, encoding="utf8") as f:
        manifest_json = json.load(f)
    return manifest_json


@pytest.fixture(name="expected_manifest_json")
def get_expected_manifest_json() -> dict:
    """
    :return: dict. The dict representation of the expected manifest JSON.
    """
    expected_manifest_file = find_example_manifest("vlass_coarse_cube_manifest")
    with open(expected_manifest_file, encoding="utf8") as f:
        manifest_json = json.load(f)
    return manifest_json


@pytest.fixture(scope="module", name="output_groups")
def get_output_groups(actual_manifest_json: dict, expected_manifest_json: dict) -> tuple[dict, dict]:
    """
    :param actual_manifest_json: dict. A Pytest fixture.
    :param expected_manifest_json: dict. A Pytest fixture.
    :return:  tuple[dict, dict]. The actual and expected manifests' output groups.
    """
    return expected_manifest_json["output_group"], actual_manifest_json["output_group"]


def test_manifest_json(actual_manifest_json: dict, expected_manifest_json: dict):
    """
    Checks that the generated manifest JSON is what is expected.

    :param actual_manifest_json: dict. A Pytest fixture.
    :param expected_manifest_json: dict. A Pytest fixture.
    """
    assert ordered(actual_manifest_json["parameters"]) == ordered(expected_manifest_json["parameters"])
    assert ordered(actual_manifest_json["input_group"]) == ordered(expected_manifest_json["input_group"])
    assert ordered(actual_manifest_json["output_group"]["science_products"]) == ordered(
        expected_manifest_json["output_group"]["science_products"]
    )

    # Test output group ancillary products
    actual_og_aps = actual_manifest_json["output_group"]["ancillary_products"]
    expected_og_aps = expected_manifest_json["output_group"]["ancillary_products"]

    # Because the name is based on the time it was generated, we have to compare it differently
    actual_ingestion_artifacts_tar = [x for x in actual_og_aps if x["type"] == "ingestion_artifacts"][0]
    expected_ingestion_artifacts_tar = [x for x in expected_og_aps if x["type"] == "ingestion_artifacts"][0]

    assert re.match(
        r"ingestion_artifacts_\d{4}_\d{2}_\d{2}T\d{2}_\d{2}_\d{2}\.\d{3}\.tar",
        actual_ingestion_artifacts_tar["filename"],
    )

    # Drop ingestion artifacts tars from the lists
    actual_og_aps.remove(actual_ingestion_artifacts_tar)
    expected_og_aps.remove(expected_ingestion_artifacts_tar)

    # Check the rest of the output group ancillary products
    assert ordered(actual_og_aps) == ordered(expected_og_aps)


def test_manifest_object_vs_dumped_json(
    vlass_coarse_cube_manifest: tuple[IngestionManifest, Path], actual_manifest_json: dict
):
    """
    Checks that the correct content was dumped to the file.

    :param vlass_coarse_cube_manifest: tuple[IngestionManifest, Path]. A Pytest fixture.
    :param actual_manifest_json: dict. A Pytest fixture.
    """
    actual_manifest, _ = vlass_coarse_cube_manifest
    actual_manifest_schema = IngestionManifestSchema()
    # These are sorted because ancillary products are ordered differently between the two dumps for some reason
    assert ordered(actual_manifest_schema.dump(actual_manifest)) == ordered(actual_manifest_json)


def test_manifest_parameters(
    vlass_coarse_cube_manifest: tuple[IngestionManifest, Path],
    vlass_coarse_cube_staging_dir: Path,
):
    """
    Tests the parameters of the IngestionManifest instance.
    :param vlass_coarse_cube_manifest: tuple[IngestionManifest, Path]. A Pytest fixture.
    :param vlass_coarse_cube_staging_dir: Path. A Pytest fixture.
    """
    actual_manifest, _ = vlass_coarse_cube_manifest

    actual_manifest_params = actual_manifest.parameters
    assert actual_manifest_params.ngas_ingest is False
    assert actual_manifest_params.telescope == Telescope.EVLA
    assert actual_manifest_params.staging_source_dir == vlass_coarse_cube_staging_dir
    assert isinstance(actual_manifest_params.additional_metadata, AbstractTextFile)
    assert actual_manifest_params.additional_metadata.filename == "aux_image_metadata.json"
    with pytest.raises(AttributeError):
        _ = actual_manifest_params.reingest


def test_manifest_ingestion_artifacts_tar_contents(
    vlass_coarse_cube_manifest: tuple[IngestionManifest, Path], vlass_coarse_cube_staging_dir: Path
):
    """
    Checks for the correct contents of the manifest's ingestion artifacts tar.

    :param vlass_coarse_cube_manifest: tuple[IngestionManifest, Path]. A Pytest fixture.
    :param vlass_coarse_cube_staging_dir: Path. A Pytest fixture.
    """
    actual_manifest, _ = vlass_coarse_cube_manifest
    # It should contain ONLY the ingestion manifest and the additional metadata file
    addl_md = actual_manifest.parameters.additional_metadata

    assert addl_md
    ingestion_artifacts_tar = find_ingestion_artifacts_tar(vlass_coarse_cube_staging_dir)
    with tarfile.open(ingestion_artifacts_tar, "r") as tar:
        # We need to get just the file name, not the full path to the file
        members = [member.name.split("/")[-1] for member in tar.getmembers()]
        assert sorted(members) == sorted([addl_md.filename, MANIFEST_FILENAME])
