#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Allows for manual intervention into running jobs.
Current primary use case: Manually force a workflow to fail
"""
import argparse
import logging
import sys
from argparse import ArgumentParser
from collections import deque

from pycapo import CapoConfig
from .arbitrator import Arbiter

logger = logging.getLogger("mediator")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


class CapabilityAction(argparse.Action):
    """
    Define an Action class for capability actions
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings, dest, nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        service_url = CapoConfig().settings("edu.nrao.workspaces.CapabilitySettings").externalServiceUrl
        values = deque(values)
        values.appendleft(service_url)
        setattr(namespace, self.dest, list(values))
        logger.debug("%r %r %r" % (namespace, values, option_string))


class WorkflowAction(argparse.Action):
    """
    Define an Action class for workflow actions
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings, dest, nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        service_url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl
        values = deque(values)
        values.appendleft(service_url)
        setattr(namespace, self.dest, list(values))
        logger.debug("%r %r %r" % (namespace, values, option_string))


def parser() -> ArgumentParser:
    """
    Initialize argument parser for CLI

    :return: Argument Parser
    """
    arg_parser = ArgumentParser(
        description="System Mediator allows interventions in currently running Workspaces requests",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    arg_parser.add_argument(
        "-ff",
        "--force-fail",
        nargs="+",
        action=WorkflowAction,
        required=False,
        help="Force a specified workflow into the failed state. Requires a workflow request id.",
    )
    arg_parser.add_argument(
        "-s",
        "--submit",
        nargs="+",
        action=WorkflowAction,
        required=False,
        help="Submit a specified workflow. Requires a workflow request id.",
    )

    arg_parser.add_argument(
        "-c",
        "--cancel",
        nargs="+",
        action=CapabilityAction,
        required=False,
        help="Cancel a given list of capability requests. Requires at least one Capability Request Id.",
    )
    return arg_parser


def main():
    """
    Launch the workspaces system mediator

    :return:
    """
    logger.info("**************************************")
    logger.info("* WELCOME TO THE WS SYSTEM MEDIATOR! *")
    logger.info("**************************************")

    args = parser().parse_args()

    Arbiter(args).intercede()
