#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Extracts workflow relevant metadata from the NRAO archive based on Science Product Locator.
"""
import json
import logging
from typing import List

import arrow
from aat_wrest.utilities import PENDULUM_FORMAT, TIME_ZONE, MDDBConnector


class WrestWorkflowMetadata:
    """
    Class for extracting metadata required for workflow execution
    """

    def __init__(
        self,
        connection: MDDBConnector,
        spl: List[str] | None = None,
        sdm_id: str | None = None,
        mous_id: str | None = None,
    ):
        self.logger = logging.getLogger("aat_wrest")
        self.conn = connection
        self.sdm_id = sdm_id
        self.spl = spl
        self.mous_id = mous_id

    def wrest_standard_cal_info(self) -> str:
        """
        Given an execution block science product locator, returns the required metadata to run
        the standard calibration workflow
        :return:
        """

        sql = """
        SELECT ngas_fileset_id as filesetId,
         e.project_code as projectCode,
         p.title as title,
         e.starttime as startTime,
         (a.firstname || ' ' || a.lastname) as observer,
         telescope as telescope,
         p.observing_types as observingTypes
         FROM execution_blocks e
            JOIN projects p on e.project_code = p.project_code
            JOIN authors a on p.project_code = a.project_code
         WHERE science_product_locator = %(spl)s AND a.is_pi = true
        """
        make_json = ""
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql, {"spl": self.spl})
            data = cursor.fetchall()

            if data:
                make_json = json.dumps(
                    {
                        "sdmId": data[0][0],
                        "projectCode": data[0][1],
                        "title": data[0][2],
                        "startTime": data[0][3],
                        "observer": data[0][4],
                        "telescope": data[0][5],
                        "created_at": str(arrow.now().to(TIME_ZONE).format(PENDULUM_FORMAT)),
                        "observingTypes": data[0][6],
                    }
                )
            else:
                self.logger.error(
                    "ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for SPL {self.spl}!"
                )
        finally:
            self.conn.close()
        return make_json

    def wrest_standard_image_info(self) -> str:
        """
        Given an execution block SDM ID, returns the required metadata to run
        the standard imaging workflow
        :return:
        """

        sql = """
        SELECT e.science_product_locator as spl,
         e.project_code as projectCode,
         p.title as title,
         e.starttime as startTime,
         (a.firstname || ' ' || a.lastname) as observer,
         telescope as telescope,
         c.science_product_locator as cal_locator,
         c.casa_version as casa_version
         FROM execution_blocks e
            JOIN projects p on e.project_code = p.project_code
            JOIN authors a on p.project_code = a.project_code
            LEFT JOIN calibrations c on e.execution_block_id = c.execution_block_id
         WHERE ngas_fileset_id = %(sdmId)s AND a.is_pi = true
        """
        make_json = ""
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql, {"sdmId": self.sdm_id})
            data = cursor.fetchall()
            if data:
                make_json = json.dumps(
                    {
                        "product_locator": data[0][0],
                        "projectCode": data[0][1],
                        "title": data[0][2],
                        "startTime": data[0][3],
                        "observer": data[0][4],
                        "telescope": data[0][5],
                        "cal_locator": data[0][6],
                        "casa_version": data[0][7],
                        "created_at": str(arrow.now().to(TIME_ZONE).format(PENDULUM_FORMAT)),
                    }
                )
            else:
                self.logger.error(
                    "ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for sdm id: {self.sdm_id}!"
                )
        finally:
            self.conn.close()
        return make_json

    def wrest_restore_info_via_spl(self) -> str:
        """
        Given an execution block science product locator and calibration science product locator, returns
        the required metadata to run the restore CMS workflow
        :return:
        """
        # Fetch the EBs from the SPL
        eb_sql = """
            SELECT e.ngas_fileset_id as filesetId,
             e.project_code as projectCode,
             p.title as title,
             e.starttime as startTime,
             (a.firstname || ' ' || a.lastname) as observer,
             telescope as telescope
             FROM execution_blocks e
                JOIN projects p on e.project_code = p.project_code
                JOIN authors a on p.project_code = a.project_code
             WHERE science_product_locator = %(spl)s AND a.is_pi = true
        """

        # Fetch the cal from the cal SPL
        cal_sql = """
            SELECT sp.external_name as calSdmId,
                c.casa_version as casa_version
            FROM science_products sp
                LEFT JOIN calibrations c on sp.science_product_locator = c.science_product_locator
            WHERE sp.science_product_locator = %(cal_spl)s
            ORDER BY c.calibration_id DESC
            FETCH FIRST ROW ONLY
        """

        make_json = ""
        try:
            cursor = self.conn.cursor()
            cursor.execute(eb_sql, {"spl": self.spl[0]})
            data = cursor.fetchall()
            cursor.execute(cal_sql, {"cal_spl": self.spl[1]})
            cal_data = cursor.fetchall()
            if data and cal_data:
                make_json = json.dumps(
                    {
                        "sdmId": data[0][0],
                        "projectCode": data[0][1],
                        "title": data[0][2],
                        "startTime": data[0][3],
                        "created_at": str(arrow.now().to(TIME_ZONE).format(PENDULUM_FORMAT)),
                        "observer": data[0][4],
                        "telescope": data[0][5],
                        "remote-cv": True if data[0][5] == "ALMA" else False,
                        "calSdmId": cal_data[0][0].replace(":", "_").replace("/", "_"),
                        "casa_version": cal_data[0][1],
                    },
                    indent=4,
                    default=str,
                )
            else:
                self.logger.error(
                    "ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for SPL {self.spl[0]} or Cal SPL {self.spl[1]}!"
                )
        finally:
            self.conn.close()
        return make_json

    def wrest_restore_info_via_mous(self) -> str:
        """
        Given a MOUS ID, returns the required metadata to run the restore CMS workflow
        :return: JSON object of metadata
        """
        # Fetch the EBs from the MOUS ID
        eb_sql = """
            SELECT e.science_product_locator as product_locator,
             e.ngas_fileset_id as filesetId,
             e.project_code as projectCode,
             p.title as title,
             e.starttime as startTime,
             (a.firstname || ' ' || a.lastname) as observer,
             telescope as telescope
             FROM execution_blocks e
                JOIN projects p on e.project_code = p.project_code
                JOIN authors a on p.project_code = a.project_code
             WHERE alma_ous_id = %(mous)s AND a.is_pi = true
        """

        # Fetch the cal from the MOUS ID using only the newest cal
        cal_sql = """
            SELECT c.science_product_locator as cal_locator,
                sp.external_name as calSdmId,
                c.casa_version as casa_version
            FROM science_products sp
                LEFT JOIN calibrations c on sp.science_product_locator = c.science_product_locator
            WHERE c.alma_ous_id = %(mous)s
            ORDER BY c.calibration_id DESC
            FETCH FIRST ROW ONLY
        """

        make_json = ""
        try:
            cursor = self.conn.cursor()
            cursor.execute(eb_sql, {"mous": self.mous_id})
            data = cursor.fetchall()
            cursor.execute(cal_sql, {"mous": self.mous_id})
            cal_data = cursor.fetchall()
            if data and cal_data:
                make_json = json.dumps(
                    {
                        "product_locator": ",".join([d[0] for d in data]),
                        "sdmId": [d[1] for d in data],
                        "projectCode": data[0][2],
                        "title": data[0][3],
                        "startTime": data[0][4],
                        "created_at": str(arrow.now().to(TIME_ZONE).format(PENDULUM_FORMAT)),
                        "observer": data[0][5],
                        "telescope": data[0][6],
                        "remote-cv": True if data[0][6] == "ALMA" else False,
                        "cal_locator": cal_data[0][0],
                        "calSdmId": cal_data[0][1].replace(":", "_").replace("/", "_"),
                        "casa_version": cal_data[0][2],
                    },
                    indent=4,
                    default=str,
                )
            else:
                self.logger.error(
                    "ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for MOUS ID {self.mous_id}!"
                )
        finally:
            self.conn.close()
        return make_json

    def wrest_aux_imaging_info(self) -> str:
        """
        Given a parent workflow request id for an image ingestion, returns the required metadata to run
        the image ingestion
        :return:
        """

        if self.spl and "calibration" in self.spl:
            sql = """
            SELECT e.project_code as projectCode,
                band_code as bandCodes,
                configuration
            FROM execution_blocks e
            JOIN calibrations c on e.execution_block_id = c.execution_block_id
            """

            condition = f"WHERE c.science_product_locator = '{self.spl}'"

        else:
            sql = """
            SELECT project_code as projectCode,
                band_code as bandCodes,
                configuration
            FROM execution_blocks e
            """

            condition = (
                f"WHERE science_product_locator = '{self.spl}'"
                if self.spl is not None
                else f"WHERE ngas_fileset_id = '{self.sdm_id}'"
            )

        make_json = ""
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql + condition)
            data = cursor.fetchall()
            if data:
                make_json = json.dumps(
                    {
                        "projectCode": data[0][0],
                        "bands": data[0][1],
                        "configurations": data[0][2],
                    }
                )
            else:
                self.logger.error(
                    f"ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for sdm id {self.sdm_id} or spl id {self.spl}!"
                )
        finally:
            self.conn.close()
        return make_json

    def wrest_image_name(self) -> str:
        """
        Given an image SPL, returns the image file name
        :return:
        """

        sql = """
        SELECT external_name as image_name
        FROM science_products
        WHERE science_product_locator = %(spl)s
            """

        make_json = ""
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql, {"spl": self.spl[0]})
            data = cursor.fetchall()
            if data:
                make_json = json.dumps(
                    {
                        "imageName": data[0][0],
                    }
                )
            else:
                self.logger.error(
                    f"ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for spl id {self.spl[0]}!"
                )
        finally:
            self.conn.close()
        return make_json

    def wrest_ancillary_locators(self, ancillary_types: tuple[str, str, str]) -> list[tuple[str]]:
        """
        Given the partial external name of a product, retrieve the locator for a specific
        ancillary product related to that product.

        :param ancillary_types:
        :return: list of locators
        """
        sql = """
        SELECT ancillary_product_locator FROM ancillary_products
        WHERE (product_group_id IN ( Select product_group_id
                                    FROM science_products_product_groups
                                    Join public.science_products sp on sp.science_product_locator = science_products_product_groups.science_product_locator
                                    WHERE sp.external_name LIKE %(partial_name)s
        ) OR science_product_locator IN ( select science_product_locator
                                   FROM science_products
                                   WHERE external_name like %(partial_name)s
        )) AND ancillary_product_type in %(ancillary_types)s
        """

        try:
            cursor = self.conn.cursor()
            name_like = self.sdm_id + "%"
            self.logger.debug(name_like)
            self.logger.debug(ancillary_types)
            cursor.execute(sql, {"partial_name": name_like, "ancillary_types": ancillary_types})
            data = cursor.fetchall()
            if not data:
                self.logger.error(
                    f"ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for sdm id: {self.sdm_id}!"
                )
        finally:
            self.conn.close()
        return data

    def wrest_product_info(self) -> str:
        """
        Given a list of locators, returns the basic product info for each

        :return:
        """
        query = """
        SELECT science_product_locator as locator,
        science_product_type as locator_type,
        external_name as name
        FROM archive.public.science_products
        WHERE science_product_locator IN %(locator_list)s
        """
        make_json = ""
        try:
            cursor = self.conn.cursor()
            cursor.execute(query, {"locator_list": tuple(self.spl[0].split(","))})
            data = cursor.fetchall()
            if data:
                make_json = json.dumps(
                    {
                        "products": [
                            {
                                "product_locator": product[0],
                                "product_type": product[1],
                                "external_name": product[2],
                            }
                            for product in data
                        ]
                    }
                )
            else:
                self.logger.error(
                    f"ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for spl(s): {self.spl}!"
                )
        finally:
            self.conn.close()

        return make_json

    def wrest_curator(self) -> dict:
        """
        Given a locator, returns the product information necessary to run curator on it.

        :return: JSON containing product information.
        """
        query = """
        SELECT sp.science_product_type,
        COALESCE(eb.telescope,im.telescope) as telescope,
        spp.project_code
        FROM science_products sp
        JOIN science_products_projects spp ON sp.science_product_locator = spp.science_product_locator
        LEFT JOIN execution_blocks eb ON sp.science_product_locator = eb.science_product_locator
        LEFT JOIN images im ON sp.science_product_locator = im.science_product_locator
        WHERE sp.science_product_locator = %(spl)s;
        """
        make_json = {}
        try:
            cursor = self.conn.cursor()
            spl = self.spl[0]
            if self.spl[0].isdigit():
                spl_query = """
                SELECT science_product_locator
                FROM science_products_product_groups
                WHERE product_group_id = %(pg_id)s;
                """
                cursor.execute(spl_query, {"pg_id": self.spl[0]})
                data = cursor.fetchall()
                if data:
                    spl = data[0][0]
                else:
                    self.logger.error(f"ERROR: Failed to determine SPL from product group id, got {data}")
                    return make_json

            cursor.execute(query, {"spl": spl})
            data = cursor.fetchall()
            if data:
                product_type = str(data[0][0]).lower().replace(" ", "_")

                if product_type != "calibration":
                    telescope = data[0][1]
                else:
                    # Need to query on the eb / alma_ous_id to get a telescope for calibrations
                    scope_query = """
                    SELECT telescope
                    FROM execution_blocks
                    JOIN calibrations c_eb ON execution_blocks.execution_block_id = c_eb.execution_block_id
                    WHERE c_eb.science_product_locator = %(spl)s
                    UNION
                    SELECT telescope
                    FROM execution_blocks
                    JOIN calibrations c_ous ON execution_blocks.alma_ous_id = c_ous.alma_ous_id
                    WHERE c_ous.science_product_locator = %(spl)s
                    """
                    cursor.execute(scope_query, {"spl": spl})
                    scope_data = cursor.fetchall()
                    if scope_data:
                        telescope = scope_data[0][0]
                    else:
                        self.logger.error(f"ERROR: Failed to determine telescope from calibration's EB, got {data}")
                        return make_json

                make_json = {
                    "product_type": product_type,
                    "telescope": telescope,
                    "projectCode": data[0][2],
                }

            else:
                self.logger.error(
                    f"ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for spl id {spl}!"
                )
        finally:
            self.conn.close()
        return make_json

    def wrest_curator_products(self) -> dict:
        """
        Given a locator or product group ID, returns the product filenames associated with it

        :return: JSON containing the product list
        """
        make_json = {}
        product_group_id = ""
        try:
            cursor = self.conn.cursor()
            if self.spl[0].isdigit():
                product_group_id = self.spl[0]
            else:
                # We were given an SPL, need to get the associated product group ID
                prod_id_query = """
                SELECT product_group_id
                FROM science_products_product_groups
                WHERE science_product_locator = %(spl)s;
                """
                cursor.execute(prod_id_query, {"spl": self.spl[0]})
                data = cursor.fetchall()
                if data:
                    product_group_id = data[0][0]
                else:
                    self.logger.error(f"ERROR: Failed to fetch product group id from SPL, got {data}")

            files_query = """
            WITH product_locators AS (
            SELECT sp.science_product_locator,
                   atspl.ancillary_product_locator
            FROM   science_products_product_groups sp
            JOIN   ancillary_to_science_product_locators atspl on sp.science_product_locator = atspl.science_product_locator
            WHERE  sp.product_group_id = %(pg_id)s
            ), filegroups AS (
                SELECT ap2.filegroup_id
                FROM ancillary_products ap2
                JOIN product_locators pl ON ap2.ancillary_product_locator = pl.ancillary_product_locator
            )
            SELECT external_name
            FROM science_products sp2
            JOIN product_locators pl ON pl.science_product_locator = sp2.science_product_locator
            UNION
            SELECT filename
            FROM files f
            JOIN filegroups fg ON f.filegroup = fg.filegroup_id
            """
            if product_group_id != "":
                cursor.execute(files_query, {"pg_id": product_group_id})
                data = cursor.fetchall()
                if data:
                    # This should have all the ancillary and science product names
                    # data is expected to be a list of strings
                    make_json = {"file_list": data}
                else:
                    self.logger.error(f"ERROR: Failed to fetch products from product ID {product_group_id}")
            else:
                self.logger.error("ERROR: Failed to get product group ID")

            input_group_query = """
            SELECT DISTINCT s.science_product_locator
            FROM product_groups pg2
            JOIN science_products_product_groups sppg on pg2.product_group_id = sppg.product_group_id
            JOIN science_products_product_groups s on pg2.parent_product_group_id = s.product_group_id
            WHERE sppg.product_group_id = %(pg_id)s;
            """
            cursor.execute(input_group_query, {"pg_id": product_group_id})
            input_data = cursor.fetchall()
            if input_data:
                make_json = {**make_json, "input_group_locator": input_data[0][0]}
        finally:
            self.conn.close()
        return make_json
