#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Wrester for obtaining metadata for a given spl observation for UI display.
"""
import json

# pylint: disable=E0401, E0402, R0903
import logging

import arrow
from aat_wrest.utilities import PENDULUM_FORMAT, MDDBConnector


def format_timestamp(timestamp: arrow.Arrow) -> str:
    """
    mjd_to_timestamp returns timestamps in the following form: 2019-01-18 05:21:07.550008 +00:00
    Convert this to YYY-MM-DDThh:mm:ss

    :param timestamp: the timestamp string returned from mjd_to_timestamp
    :return: correctly formatted datetime string
    """

    return timestamp.format(PENDULUM_FORMAT)


class WrestObservationMetadata:
    """
    Class for extracting observation metadata
    """

    def __init__(
        self, connection: MDDBConnector, sdm_id: str | None = None, spl: str | None = None, mous_id: str | None = None
    ):
        self.logger = logging.getLogger("aat_wrest")
        self.conn = connection
        self.sdm_id = sdm_id
        self.spl = spl
        self.mous_id = mous_id

    def wrest_observation_info(self) -> str:
        """
        Given a sdm_id, query the Metadata DB for observation data related to an ingestion-complete event

        :param sdm_id:
        :return science_product_locator:
        """
        metadata = ""
        sql = f"""
        SELECT e.science_product_locator,
         e.band_code as band_code,
         e.configuration as configuration,
         mjd_to_timestamp(e.starttime) as starttime,
         mjd_to_timestamp(e.endtime) as endtime,
         sp.is_srdp as is_srdp
         FROM execution_blocks e
            JOIN science_products sp on e.science_product_locator = sp.science_product_locator
         WHERE ngas_fileset_id = %(sdm_id)s
        """
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql, {"sdm_id": self.sdm_id})
            data = cursor.fetchall()
            if data:
                metadata = json.dumps(
                    {
                        "spl": data[0][0],
                        "bands": data[0][1],
                        "array_config": data[0][2],
                        "obs_start_time": format_timestamp(arrow.get(data[0][3].isoformat())),
                        "obs_end_time": format_timestamp(arrow.get(data[0][4].isoformat())),
                        "is_srdp": data[0][5],
                    },
                    indent=4,
                    default=str,
                )
            else:
                self.logger.error(
                    f"ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for SDM id {self.sdm_id}"
                )
        finally:
            self.conn.close()
        return metadata

    def wrest_observation_info_via_spl(self) -> str:
        """
        Given a science_product_locator, query the Metadata DB for observation data related to an ingestion-complete event

        :param science_product_locator:
        :return metadata:
        """
        metadata = ""
        sql = f"""
            SELECT e.science_product_locator,
            e.band_code as band_code,
            e.configuration as configuration,
            mjd_to_timestamp(e.starttime) as starttime,
            mjd_to_timestamp(e.endtime) as endtime,
            e.ngas_fileset_id as ngas_fileset_id,
            sp.is_srdp as is_srdp
            FROM execution_blocks e
                JOIN science_products sp on e.science_product_locator = sp.science_product_locator
            WHERE e.science_product_locator = %(spl)s
            """
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql, {"spl": self.spl})
            data = cursor.fetchall()
            if data:
                metadata = json.dumps(
                    {
                        "spl": data[0][0],
                        "bands": data[0][1],
                        "array_config": data[0][2],
                        "obs_start_time": format_timestamp(arrow.get(data[0][3].isoformat())),
                        "obs_end_time": format_timestamp(arrow.get(data[0][4].isoformat())),
                        "sdm_id": data[0][5],
                        "is_srdp": data[0][6],
                    },
                    indent=4,
                    default=str,
                )
            else:
                self.logger.error(
                    f"ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for SPL {self.spl}"
                )
        finally:
            self.conn.close()
        return metadata

    def wrest_observation_info_via_mous(self) -> str:
        """
        Given a MOUS ID, query the Metadata DB for observation data related to an ingestion-complete event.

        :return metadata:
        """
        metadata = ""
        sql = f"""
            SELECT e.science_product_locator,
            e.alma_ous_id,
            e.band_code as band_code,
            e.configuration as configuration,
            mjd_to_timestamp(e.starttime) as starttime,
            mjd_to_timestamp(e.endtime) as endtime,
            sp.is_srdp as is_srdp
            FROM execution_blocks e
                JOIN science_products sp on e.science_product_locator = sp.science_product_locator
            WHERE e.alma_ous_id = %(mous)s
            """
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql, {"mous": self.mous_id})
            data = cursor.fetchall()
            if data:
                metadata = json.dumps(
                    {
                        "spl": ",".join([d[0] for d in data]),
                        "mous_id": data[0][1],
                        "bands": data[0][2],
                        "array_config": data[0][3],
                        "obs_start_time": format_timestamp(arrow.get(data[0][4].isoformat())),
                        "obs_end_time": format_timestamp(arrow.get(data[0][5].isoformat())),
                        "is_srdp": data[0][6],
                    },
                    indent=4,
                    default=str,
                )
            else:
                self.logger.error(
                    f"ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for MOUS ID {self.mous_id}"
                )
        finally:
            self.conn.close()
        return metadata
