#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import argparse
import json
import logging
import sys

from aat_wrest.metadata_wrester import WrestWorkflowMetadata
from aat_wrest.observation_wrester import WrestObservationMetadata
from aat_wrest.utilities import MDDBConnector

"""
AAT Wrest extract information from the NRAO Archive metadata database
"""

logger = logging.getLogger("aat_wrest")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def parser() -> argparse.ArgumentParser:
    arg_parser = argparse.ArgumentParser(
        description="Archive Database Access Utility",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    arg_parser.add_argument(
        "-sc",
        "--stdcals",
        "--stdcalandimg",
        nargs=1,
        action="store",
        required=False,
        help="Find workflow metadata for standard calibrations or calibration+imaging with provided product locator",
    )
    arg_parser.add_argument(
        "-si",
        "--stdimg",
        nargs=1,
        action="store",
        required=False,
        help="Find workflow metadata for standard CMS imaging with provided SDM id",
    )
    arg_parser.add_argument(
        "-r",
        "--restore",
        nargs="+",
        default=[],
        required=False,
        help="Find workflow metadata for Restores with provided EB product locator and Cal product locator",
    )
    arg_parser.add_argument(
        "-obs",
        "--observation",
        nargs=1,
        action="store",
        required=False,
        help="Find display metadata for observations by SDM id",
    )
    arg_parser.add_argument(
        "-obspl",
        "--observation-via-spl",
        nargs=1,
        action="store",
        required=False,
        help="Find display metadata for observations by SPL",
    )
    arg_parser.add_argument(
        "-obsmous",
        "--observation-via-mous",
        nargs=1,
        action="store",
        required=False,
        help="Find display metadata for observations by MOUS ID",
    )
    arg_parser.add_argument(
        "-aux",
        "--auxiliary-ingestion",
        nargs=1,
        action="store",
        required=False,
        help="Find auxiliary imaging metadata for ingestion",
    )
    arg_parser.add_argument(
        "--carta",
        nargs=1,
        action="store",
        required=False,
        help="Find image names for single image carta requests",
    )
    arg_parser.add_argument(
        "--cube",
        nargs=1,
        action="store",
        required=False,
        help="Find the ancillary product locator for the cube reimaging resources",
    )
    arg_parser.add_argument(
        "--product",
        nargs="+",
        default=[],
        required=False,
        help="Find the basic product information for the provided product locator(s)",
    )
    arg_parser.add_argument(
        "--curator",
        nargs=1,
        action="store",
        required=False,
        help="Find the product information necessary to run curator on the provided product locator",
    )
    arg_parser.add_argument(
        "--curator_products",
        nargs=1,
        action="store",
        required=False,
        help=(
            "Find the product information necessary to run curator or ingest on the provided product locator or "
            "product group id"
        ),
    )
    return arg_parser


def determine_wrester(connection: MDDBConnector, args: argparse.Namespace):
    if args.stdcals:
        data = WrestWorkflowMetadata(connection, spl=args.stdcals[0]).wrest_standard_cal_info()
    elif args.stdimg:
        data = WrestWorkflowMetadata(connection, sdm_id=args.stdimg[0]).wrest_standard_image_info()
    elif args.restore:
        if "execblock" in args.restore[0] or "alma" in args.restore[0]:
            data = WrestWorkflowMetadata(connection, spl=args.restore).wrest_restore_info_via_spl()
        else:
            data = WrestWorkflowMetadata(connection, mous_id=args.restore[0]).wrest_restore_info_via_mous()
    elif args.carta:
        data = WrestWorkflowMetadata(connection, spl=args.carta).wrest_image_name()
    elif args.observation:
        data = WrestObservationMetadata(connection, sdm_id=args.observation[0]).wrest_observation_info()
    elif args.observation_via_spl:
        data = WrestObservationMetadata(connection, spl=args.observation_via_spl[0]).wrest_observation_info_via_spl()
    elif args.observation_via_mous:
        data = WrestObservationMetadata(
            connection, mous_id=args.observation_via_mous[0]
        ).wrest_observation_info_via_mous()
    elif args.auxiliary_ingestion:
        if str(args.auxiliary_ingestion[0]).startswith("uid"):
            data = WrestWorkflowMetadata(connection, spl=args.auxiliary_ingestion[0]).wrest_aux_imaging_info()
        else:
            data = WrestWorkflowMetadata(connection, sdm_id=args.auxiliary_ingestion[0]).wrest_aux_imaging_info()
    elif args.cube:
        data = WrestWorkflowMetadata(connection, sdm_id=args.cube[0]).wrest_ancillary_locators(
            ("reimaging_resources", "parameter_list", "component_list")
        )
        data = json.dumps(
            {
                "reimaging_locator": data[0][0],
                "parameter_locator": data[1][0],
                "component_locator": data[2][0],
            }
        )
    elif args.product:
        data = WrestWorkflowMetadata(connection, spl=args.product).wrest_product_info()
    elif args.curator:
        data_dict = WrestWorkflowMetadata(connection, spl=args.curator).wrest_curator()
        data = json.dumps(data_dict)
    elif args.curator_products:
        general_data = WrestWorkflowMetadata(connection, spl=args.curator_products).wrest_curator()
        connection2 = MDDBConnector()
        product_data = WrestWorkflowMetadata(connection2, spl=args.curator_products).wrest_curator_products()
        data = json.dumps({**general_data, **product_data})
    else:
        data = None

    print(data)


def main():
    args = parser().parse_args()
    connection = MDDBConnector()

    determine_wrester(connection, args)
