#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Testing suite for aat_wrest
"""

import argparse
import logging
import re
import sys
from unittest.mock import MagicMock, Mock, patch

import arrow
from aat_wrest.metadata_wrester import WrestWorkflowMetadata

logger = logging.getLogger("test_aat_wrest")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


args_cal = argparse.Namespace(stdcals="uid://evla/execblock/125dac14-c53b-4489-9c43-7d6e2739ec32", stdimg=None)
result_cal = (
    '{"sdmId": "17B-197.sb34663512.eb34806505.58108.78427738426", "projectCode": "17B-197", '
    '"title": "Radio-Active CoWS: Extended FIRST Sources in MaDCoWS Clusters", '
    '"startTime": 58108.7842916667, "observer": "Anthony Gonzalez", "telescope": "EVLA",'
    ' "created_at": 2021-01-01T00:00:00, "observing_types": "Type A, Type B"}'
)

args_image = argparse.Namespace(stdcals=None, stdimg="brain_000.58099.67095825232")
result_img = (
    '{"product_locator": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",'
    ' "projectCode": "Operations", "title": "", "startTime": 58099.6710792824, '
    '"observer": "VLA Operations", "telescope": "EVLA", "created_at": "2021-06-30T20:18:21"}'
)

args_carta = argparse.Namespace(
    carta="uid://evla/image/3d3db489-9331-4e61-aa80-002bc2989b1e", stdcals=None, stdimg=None
)
result_carta = '{"imageName": "VLASS1.1.ql.T29t04.J094850+743000.10.2048.v5"}'

args_curator = argparse.Namespace(
    curator="uid://evla/calibration/71d1f00c-a381-4688-9d5c-f528f43452f7", carta=None, stdcals=None, stdimg=None
)


def mock_wrester(args: argparse.Namespace) -> WrestWorkflowMetadata:
    with patch("psycopg2.connect") as mock_connect:
        if args.stdcals is not None:
            return WrestWorkflowMetadata(connection=mock_connect, spl=args.stdcals)
        if args.stdimg is not None:
            return WrestWorkflowMetadata(connection=mock_connect, sdm_id=args.stdimg)
        if args.carta is not None:
            return WrestWorkflowMetadata(connection=mock_connect, spl=args.carta)


def create_mock_connection_object(mock_cursor) -> Mock:
    """
    Creates a mock connection object to be used in the below tests.
    """
    # Testing utility method to create a reusable mock object to handle the functions of the MDDBConnector class
    mock_db_connection = Mock()

    mock_db_connection.cursor.return_value = mock_cursor
    return mock_db_connection


def wrest_curator_products_cursor_fetchall_side_effect():
    """
    A side effect method for mocking the cursor.fetchall() response.
    This is used in testing of the WrestWorkflowMetadata class' wrest_curator_products method.
    """
    # For consistency; some queries have upper-case and lower-case SQL syntax
    query_lower = mock_test_wrest_curator_prods_cursor.execute.call_args[0][0].lower()
    return_val = None

    if re.search("select filename", query_lower):
        return_val = [
            ("weblog.tgz"),
            ("18B-265_2019_12_10_T01_24_58.690.tar"),
            ("ingestion_artifacts_2019_12_13_T23_31_42.382.tar"),
        ]
    if re.search("select distinct s.science_product_locator", query_lower):
        return_val = [tuple(["uid://fake-spl"])]
    if re.search("select product_group_id\n +from science_products_product_groups", query_lower):
        return_val = [tuple([101])]

    if return_val:
        return return_val

    query_and_parameters = mock_test_wrest_curator_prods_cursor.execute.call_args
    value_error_msg = f"You have tried a query that is not currently handled.\n{query_and_parameters=}"
    raise ValueError(value_error_msg)


cal_wrester = mock_wrester(args_cal)
img_wrester = mock_wrester(args_image)
carta_wrester = mock_wrester(args_carta)

mock_test_wrest_curator_prods_cursor = Mock()


class TestAatWrest:
    @patch("json.dumps", MagicMock(return_value=result_cal))
    def test_wrest_standard_cal_info(self):
        cal_wrester.conn.cursor.return_value.fetchall.return_value = [
            "17B-197.sb34663512.eb34806505.58108.78427738426",
            "17B-197",
            "Radio-Active CoWS: Extended FIRST Sources in MaDCoWS Clusters",
            58108.7842916667,
            "Anthony Gonzalez",
            "EVLA",
            str(arrow.now().format("YYYY-MM-DDTHH:mm:ss")),
            "Type A, Type B",
        ]
        assert args_cal.stdcals == "uid://evla/execblock/125dac14-c53b-4489-9c43-7d6e2739ec32"
        value = cal_wrester.wrest_standard_cal_info()
        assert (
            value == '{"sdmId": "17B-197.sb34663512.eb34806505.58108.78427738426", "projectCode": "17B-197", '
            '"title": "Radio-Active CoWS: Extended FIRST Sources in MaDCoWS Clusters", '
            '"startTime": 58108.7842916667, "observer": "Anthony Gonzalez", "telescope": "EVLA",'
            ' "created_at": 2021-01-01T00:00:00, "observing_types": "Type A, Type B"}'
        )

    @patch("json.dumps", MagicMock(return_value=result_img))
    def test_wrest_standard_image_info(self):
        img_wrester.conn.cursor.return_value.fetchall.return_value = [
            "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",
            "Operations",
            "",
            58099.6710792824,
            "VLA Operations",
            "EVLA",
            "2021-06-30T20:18:21",
        ]
        assert args_image.stdimg == "brain_000.58099.67095825232"
        value = img_wrester.wrest_standard_image_info()
        assert (
            value == '{"product_locator": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",'
            ' "projectCode": "Operations", "title": "", "startTime": 58099.6710792824, '
            '"observer": "VLA Operations", "telescope": "EVLA", "created_at": "2021-06-30T20:18:21"}'
        )

    @patch("json.dumps", MagicMock(return_value=result_carta))
    def test_wrest_image_name(self):
        carta_wrester.conn.cursor.return_value.fetchall.return_value = ["VLASS1.1.ql.T29t04.J094850+743000.10.2048.v5"]
        assert args_carta.carta == "uid://evla/image/3d3db489-9331-4e61-aa80-002bc2989b1e"
        value = carta_wrester.wrest_image_name()
        assert value == '{"imageName": "VLASS1.1.ql.T29t04.J094850+743000.10.2048.v5"}'

    def test_wrest_curator_products(self):
        # Setup
        mock_test_wrest_curator_prods_cursor.fetchall.side_effect = wrest_curator_products_cursor_fetchall_side_effect
        mock_connection_object = create_mock_connection_object(mock_test_wrest_curator_prods_cursor)

        curator_wrester = WrestWorkflowMetadata(connection=mock_connection_object, spl="uid://fake-spl")

        # Run method
        value = curator_wrester.wrest_curator_products()

        # Verify results
        assert value == {
            "file_list": [
                "weblog.tgz",
                "18B-265_2019_12_10_T01_24_58.690.tar",
                "ingestion_artifacts_2019_12_13_T23_31_42.382.tar",
            ],
            "input_group_locator": "uid://fake-spl",
        }
