#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for aat_wrest.observation_wrester
"""

import json
import re
from datetime import datetime
from functools import partial
from unittest.mock import Mock

from aat_wrest.observation_wrester import WrestObservationMetadata


class TestWrestObservationInfo:
    """
    Tests the methods of the WrestObservationMetadata class.
    """

    def create_mock_connection_object(self, mock_cursor) -> Mock:
        """
        Creates a mock connection object to be used in the below tests.
        """
        # Testing utility method to create a reusable mock object to handle the functions of the MDDBConnector class
        mock_db_connection = Mock()

        mock_db_connection.cursor.return_value = mock_cursor
        return mock_db_connection

    def wrest_observation_info_fetchall_side_effect(self, cursor):
        """
        :param cursor: The mocked database connection cursor.
        :return: The appropriate mocked data based on the database query.
        """
        query_lower = cursor.execute.call_args[0][0].lower()
        return_val = None

        if re.search("select e.science_product_locator,\n +e.band_code as band_code", query_lower):
            return_val = [
                [
                    "uid://fake-spl",
                    "fake-band",
                    "fake-array-config",
                    datetime.strptime("2016-11-21 12:19:09.600000+00:00", "%Y-%m-%d %H:%M:%S.%f%z"),
                    datetime.strptime("2016-11-21 12:20:37.431992+00:00", "%Y-%m-%d %H:%M:%S.%f%z"),
                    "True",
                ]
            ]

        if return_val:
            return return_val

        query_and_parameters = cursor.execute.call_args
        value_error_msg = f"You have tried a query that is not currently handled.\n{query_and_parameters=}"
        raise ValueError(value_error_msg)

    def test_wrest_observation_info(self):
        """
        Tests the WrestObservationMetadata class' wrest_observation_info method.
        """
        mock_cursor = Mock()
        mock_cursor.fetchall.side_effect = partial(self.wrest_observation_info_fetchall_side_effect, mock_cursor)
        wrester = WrestObservationMetadata(self.create_mock_connection_object(mock_cursor), "fake-sdm-id")

        metadata = wrester.wrest_observation_info()

        assert json.loads(metadata) == {
            "spl": "uid://fake-spl",
            "bands": "fake-band",
            "array_config": "fake-array-config",
            "obs_start_time": "2016-11-21T12:19:09",
            "obs_end_time": "2016-11-21T12:20:37",
            "is_srdp": "True",
        }
