#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

""" Contacts-wrester Tests"""
import logging
import re
import sys
from functools import partial
from unittest.mock import Mock

from contacts_wrest.contacts_wrester import ProjectContactsWrester

logger = logging.getLogger("test_contacts_wrest")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


class TestContactsWrester:
    """
    Tests the methods of the ProjectContactsWrester class.
    """

    def global_ids_fetchall_side_effect(self, cursor):
        """
        :param cursor: The mocked database connection cursor for getting global IDs.
        :return: The appropriate mocked data based on the database query.
        """
        query_lower = cursor.execute.call_args[0][0].lower()
        return_val = None

        if re.search("select globalid\n +from project", query_lower):
            return_val = [tuple([229030])]

        if return_val:
            return return_val

        query_and_parameters = cursor.execute.call_args
        value_error_msg = f"You have tried a query that is not currently handled.\n{query_and_parameters=}"
        raise ValueError(value_error_msg)

    def emails_fetchall_side_effect(self, cursor):
        """
        :param cursor: The mocked database connection cursor for getting emails.
        :return: The appropriate mocked data based on the database query.
        """
        query_lower = cursor.execute.call_args[0][0].lower()
        return_val = None

        if re.search("select email\n +from email\n +where person_id in", query_lower):
            return_val = [tuple(["somebody@somewhere.edu"]), tuple(["a.coauthor@important.edu"])]

        if return_val:
            return return_val

        query_and_parameters = cursor.execute.call_args
        value_error_msg = f"You have tried a query that is not currently handled.\n{query_and_parameters=}"
        raise ValueError(value_error_msg)

    def test_retrieve_contacts(self):
        """
        Tests the retrieve_contacts method of the ProjectContactsWrester.
        """
        mock_global_ids_cursor = Mock()
        mock_global_ids_cursor.fetchall.side_effect = partial(
            self.global_ids_fetchall_side_effect, mock_global_ids_cursor
        )
        mock_emails_cursor = Mock()
        mock_emails_cursor.fetchall.side_effect = partial(self.emails_fetchall_side_effect, mock_emails_cursor)
        wrester = ProjectContactsWrester(
            project_code="fake-project-code", global_ids_cursor=mock_global_ids_cursor, emails_cursor=mock_emails_cursor
        )

        assert wrester.retrieve_contacts() == ["somebody@somewhere.edu", "a.coauthor@important.edu"]
