#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Tests methods from core_sampler.py.
"""
import re
from functools import partial
from unittest.mock import Mock

from core_sampler.core_sampler import CoreSampler


class TestCoreSampler:
    """
    Tests methods of the CoreSampler class.
    """

    def create_mock_connection_object(self, mock_cursor) -> Mock:
        """
        Creates a mock connection object to be used in the below tests.
        """
        # Testing utility method to create a reusable mock object to handle the functions of the MDDBConnector class
        mock_db_connection = Mock()

        mock_db_connection.cursor.return_value = mock_cursor
        return mock_db_connection

    def sample_project_fetchall_side_effect(self, cursor):
        """
        A side effect method for mocking the cursor.fetchall() response.
        This is used in testing of the CoreSampler class' sample_project method.
        """
        # format: "table_name": primary_key_name
        table_primary_key = {
            "ancillary_products": "ancillary_product_locator",
            "fake_foreign_table": "project_code",
            "filegroups": "filegroup_id",
            "product_groups": "product_group_id",
            "projects": "project_code",
            "science_products": "science_product_locator",
        }

        # For consistency; some queries have upper-case and lower-case names and database keywords
        query_lower = cursor.execute.call_args[0][0].lower()

        return_val = None

        try:
            val_to_search = list(cursor.execute.call_args[0][1].values())[0]
        except IndexError:  # If no argument was provided
            return_val = [{"project_code": "fake_project_code"}]

        # return_val = results based on the query
        if re.search("select c.column_name\n +from information_schema.table_constraints tc", query_lower):
            return_val = [{"column_name": table_primary_key[val_to_search]}]
        if re.search(
            r"select kcu.table_name +as foreign_table.*string_agg\(kcu.column_name, ','\) "
            r"as fk_columns.*from information_schema.table_constraints tco",
            query_lower,
            re.DOTALL,
        ):
            return_val = [{"foreign_table": "fake_foreign_table", "fk_columns": "fake_foreign_key"}]
        if re.search("select science_product_locator from science_products_projects", query_lower):
            return_val = [{"science_product_locator": "uid://fake-spl"}]
        if re.search(r"select \* from product_groups", query_lower):
            # Getting product groups data
            return_val = [
                {
                    "product_group_id": 12345,
                    "product_group_type": "fake_product_group_type",
                    "alma_ous_id": "fake_alma_ous_id",
                    "program_id": "fake_program_id",
                    "parent_product_group_id": 0,
                }
            ]
        if re.search(r"select \* from ancillary_products", query_lower):
            return_val = [
                {
                    "ancillary_product_locator": "uid://fake-ancillary-product-locator",
                    "ancillary_product_type": "fake_ancillary_product_type",
                    "product_group_id": 12345,
                    "science_product_locator": "uid://fake-spl",
                    "filegroup_id": 999_999_999,
                }
            ]
        if re.search(r"select \* from filegroups", query_lower):
            return_val = [
                {
                    "filegroup_id": 999_999_999,
                    "project_code": "fake_project_code",
                    "parent_filegroup_id": None,
                    "datasize": 316_000,
                    "type": "fake_type",
                }
            ]
        if re.search("select ancillary_product_locator from ancillary_products", query_lower):
            return_val = [{"ancillary_product_locator": "uid://fake-ancillary-product-locator"}]
        if re.search(r"select \* from science_products where science_product_locator = any", query_lower):
            return_val = [
                {
                    "science_product_locator": "uid://fake-spl",
                    "science_product_type": "Fake Type",
                    "metadata_ingestion_date": "1901-02-10 12:00:00.000000",
                    "metadata_ingestion_version": "1",
                    "filegroup_id": 999_999_999,
                    "external_system": "Fake External System",
                    "collection": "Fake Collection",
                    "allow_download": False,
                    "is_srdp": False,
                }
            ]
        if re.search(
            r"select product_group_id from science_products_product_groups.*"
            r"where science_product_locator in %\(locator_list\)s",
            query_lower,
            re.DOTALL,
        ):
            return_val = [{"product_group_id": 12345}]
        if re.search(r"select \* from projects where project_code = %\(project_code\)s", query_lower):
            return_val = [
                {
                    "project_code": "fake_project_code",
                    "legacy_id": "fake_legacy_id",
                    "total_observation_time": 0.014130208328424487,
                    "opt_project_id": "fake_opt_project_id",
                    "title": "Fake Project Title",
                    "abstract": "Fake Abstract",
                    "proprietary_expiration": "1901-01-10 12:00:00.000000",
                    "last_addition": "1900-01-10",
                    "starttime": 25282,
                    "endtime": 30000,
                    "proprietary_duration": 0,
                    "proprietary_period_start": "1901-01-10 12:00:00.000000",
                    "observing_types": "Fake,Observing,Types",
                }
            ]

        if return_val:
            return return_val

        query_and_parameters = cursor.execute.call_args
        value_error_msg = f"You have tried a query that is not currently handled.\n{query_and_parameters=}"
        raise ValueError(value_error_msg)

    def sample_eb_fetchall_side_effect(self, cursor):
        """
        A side effect method for mocking the cursor.fetchall() response.
        This is used in testing of the CoreSampler class' sample_project method.
        """
        # mock_cursor.fetchall.return_value = [
        #     {
        #         "column_name": "execution_block_id",
        #         "foreign_table": "fake_foreign_table",
        #         "fk_columns": "fake_foreign_key_column",
        #         "execution_block_id": 22903,
        #     }
        # ]
        # For consistency; some queries have upper-case and lower-case names and database keywords
        query_lower = cursor.execute.call_args[0][0].lower()
        return_val = None

        if re.search("select c.column_name\n +from information_schema.table_constraints tc", query_lower):
            return_val = [{"column_name": "execution_block_id"}]
        if re.search(
            r"select \* from execution_blocks where execution_block_id = %\(execution_block_id\)s", query_lower
        ):
            return_val = [
                {
                    "execution_block_id": 22903,
                    "ost_exec_block_id": 0,
                    "filegroup_id": 50208,
                    "calibration_level": 0,
                    "telescope": "FakeTelescope",
                    "configuration": "FakeConfiguration",
                    "scheduling_block_id": 12345678,
                    "ngas_fileset_id": "fake_ngas_fileset_id",
                    "project_code": "fake_project_code",
                    "starttime": 0,
                    "endtime": 1,
                    "calibration_status": "fake_calibration_status",
                    "scheduling_block_type": "fake_scheduling_block_type",
                    "band_code": "fake_band_code",
                    "ingestion_complete": False,
                    "segment": None,
                    "alma_ous_id": None,
                    "science_product_locator": "uid://fake-spl",
                    "type": "fake_type",
                    "num_antennas": 0,
                    "max_uv_distance": 28800,
                    "min_uv_distance": 1,
                }
            ]
        if re.search(
            r"select kcu.table_name +as foreign_table,\n +string_agg\(kcu.column_name, ','\) as fk_columns", query_lower
        ):
            return_val = [{"foreign_table": "fake_foreign_table", "fk_columns": "fake_foreign_key_column"}]
        if re.search("with recursive recursive_fake_foreign_table as ", query_lower):
            return_val = [{"execution_block_id": "22903", "foreign_table_pkey": "foreign_table_pkey_value"}]

        if return_val:
            return return_val

        query_and_parameters = cursor.execute.call_args
        value_error_msg = f"You have tried a query that is not currently handled.\n{query_and_parameters=}"
        raise ValueError(value_error_msg)

    def test_sample_project(self, capsys):
        """
        :param capsys: The built-in pytest fixture for getting data from writes to stdout and stderr.
        Tests the sample_project method of the CoreSampler class.
        """
        # Test that providing a project ID gives the correct result

        # Test setup
        mock_cursor = Mock()

        # Here, add any configuration for the cursor
        mock_cursor.fetchall.side_effect = partial(self.sample_project_fetchall_side_effect, mock_cursor)

        mock_db_connection = self.create_mock_connection_object(mock_cursor)
        sampler = CoreSampler(mock_db_connection)

        # Run sample_project
        sampler.sample_project("fake_project_code")

        # Prep expected result data for test
        expected_result = (
            "COPY projects (project_code, legacy_id, total_observation_time, opt_project_id, title, "
            "abstract, proprietary_expiration, last_addition, starttime, endtime, proprietary_duration, "
            "proprietary_period_start, observing_types) FROM stdin;\nfake_project_code\tfake_legacy_id"
            "\t0.014130208328424487\tfake_opt_project_id\tFake Project Title\tFake Abstract"
            "\t1901-01-10 12:00:00.000000\t1900-01-10\t25282\t30000\t0\t1901-01-10 12:00:00.000000"
            "\tFake,Observing,Types\n\\."
            "\nCOPY science_products (science_product_locator, science_product_type, "
            "metadata_ingestion_date, metadata_ingestion_version, filegroup_id, external_system, "
            "collection, allow_download, is_srdp) FROM stdin;\nuid://fake-spl\tFake Type"
            "\t1901-02-10 12:00:00.000000\t1\t999999999\tFake External System\tFake Collection\tFalse"
            "\tFalse\n\\."
            "\nCOPY product_groups (product_group_id, product_group_type, alma_ous_id, program_id, "
            "parent_product_group_id) FROM stdin;\n12345\tfake_product_group_type\tfake_alma_ous_id"
            "\tfake_program_id\t0\n\\.\nCOPY ancillary_products (ancillary_product_locator, "
            "ancillary_product_type, product_group_id, science_product_locator, filegroup_id) "
            "FROM stdin;\nuid://fake-ancillary-product-locator\tfake_ancillary_product_type\t12345"
            "\tuid://fake-spl\t999999999\n\\.\nCOPY filegroups (filegroup_id, project_code, "
            "parent_filegroup_id, datasize, type) FROM stdin;\n999999999\tfake_project_code\t\\N\t316000"
            "\tfake_type\n\\.\nCOPY fake_foreign_table (project_code) FROM stdin;"
            "\nfake_project_code\n\\.\n"
        )

        # Test expected result
        assert capsys.readouterr().out == expected_result

    def test_sample_eb(self, capsys):
        """
        :param capsys: The built-in pytest fixture for getting data from writes to stdout and stderr.
        Tests the sample_eb method of the CoreSampler class.
        """
        # Test that providing an SDM ID gives the correct result
        # Test setup
        mock_cursor = Mock()

        # Here, add any configuration for the cursor
        mock_cursor.fetchall.side_effect = partial(self.sample_eb_fetchall_side_effect, mock_cursor)

        # for find_eb_id to work; NRAO HQ's zip code.
        mock_cursor.fetchone.return_value = {"execution_block_id": 22903}

        mock_db_connection = self.create_mock_connection_object(mock_cursor)
        sampler = CoreSampler(mock_db_connection)

        # Run sample_eb
        sampler.sample_eb("fake_sdm_id")

        # Prep expected result data for test
        expected_result = (
            "COPY execution_blocks (execution_block_id, ost_exec_block_id, filegroup_id, "
            "calibration_level, telescope, configuration, scheduling_block_id, ngas_fileset_id, "
            "project_code, starttime, endtime, calibration_status, scheduling_block_type, band_code, "
            "ingestion_complete, segment, alma_ous_id, science_product_locator, type, num_antennas, "
            "max_uv_distance, min_uv_distance) FROM stdin;"
            "22903\t0\t50208\t0\tFakeTelescope\tFakeConfiguration\t12345678\tfake_ngas_fileset_id"
            "\tfake_project_code\t0\t1\tfake_calibration_status\tfake_scheduling_block_type"
            "\tfake_band_code\tFalse\t\\N\t\\N\tuid://fake-spl\tfake_type\t0\t28800\t1\\."
            "COPY fake_foreign_table (execution_block_id, foreign_table_pkey) FROM stdin;"
            "22903\tforeign_table_pkey_value\\."
        )

        # Test that the expected SQL was generated
        assert capsys.readouterr().out.replace("\n", "") == expected_result

    def test_sample_eb_get_correct_eb_rows(self):
        """
        Tests the sample_eb method of the CoreSampler class.
        Specifically, tests if the call to sample_eb results
        in a correct SQL call to return the execution block
        rows based on the provided SDM and execution block ID.
        """
        # Test setup
        mock_cursor = Mock()

        # Here, add any configuration for the cursor
        mock_cursor.fetchall.side_effect = partial(self.sample_eb_fetchall_side_effect, mock_cursor)

        # for find_eb_id to work; NRAO HQ's zip code.
        mock_cursor.fetchone.return_value = {"execution_block_id": 22903}

        mock_db_connection = self.create_mock_connection_object(mock_cursor)
        sampler = CoreSampler(mock_db_connection)

        # Run sample_eb
        sampler.sample_eb("fake_sdm_id")

        # Test that the mock cursor has been called with the expected SQL for fetch
        mock_cursor.execute.assert_any_call(
            "SELECT * FROM execution_blocks WHERE execution_block_id = %(execution_block_id)s",
            {"execution_block_id": 22903},
        )
