# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
"""
    Tests methods of the core sampler's database.py.
"""
from unittest.mock import Mock

from core_sampler.database import PGTable


class TestPGTable:
    """
    Tests the methods of the PGTable class.
    """

    def test_escape(self):
        """
        Tests the escape method of the PGTable class.
        """
        pg_table = PGTable(Mock(), "name")

        # Testing for string values
        assert pg_table.escape("16B-069") == "E'16B-069'"

        # Testing for None
        assert pg_table.escape(None) is None

        # Testing for a non-string value
        assert pg_table.escape(37) == "37"

    def test_primary_key_columns_execute_call(self):
        """
        Tests that the primary_key_columns method of the PGTable class
        uses the expected arguments (SQL call and a dictionary) in the execute call.
        """
        mock_cursor = Mock()
        mock_cursor.fetchall.return_value = [{"column_name": "not_examined"}]
        pg_table = PGTable(mock_cursor, "table_name")

        pg_table.primary_key_columns()

        # Test that the generated SQL is what is expected
        mock_cursor.execute.assert_called_once_with(
            """SELECT c.column_name
                 FROM information_schema.table_constraints tc
                 JOIN information_schema.constraint_column_usage AS ccu USING (constraint_schema, constraint_name)
                 JOIN information_schema.columns AS c ON c.table_schema = tc.constraint_schema AND
                                                         tc.table_name = c.table_name AND
                                                         ccu.column_name = c.column_name
                WHERE constraint_type = 'PRIMARY KEY' AND
                      tc.table_name = %(tablename)s""",
            {"tablename": "table_name"},
        )

    def test_primary_key_columns_number_of_columns(self):
        """
        Tests that the primary_key_columns method of the PGTable class
        functions as expected when there are zero, one, or more than
        one primary key.
        """
        # Test setup
        mock_cursor = Mock()
        pg_table = PGTable(mock_cursor, "table_name")

        # For one primary key
        mock_cursor.fetchall.return_value = [{"column_name": "table_primary_key"}]
        primary_key_columns = pg_table.primary_key_columns()
        assert primary_key_columns == ["table_primary_key"]

        # For 2 (1+) primary keys
        mock_cursor.fetchall.return_value = [
            {"column_name": "table_primary_key"},
            {"column_name": "table_other_primary_key"},
        ]
        primary_key_columns = pg_table.primary_key_columns()
        assert primary_key_columns == ["table_primary_key", "table_other_primary_key"]

        # For no primary keys
        mock_cursor.fetchall.return_value = []
        primary_key_columns = pg_table.primary_key_columns()
        assert primary_key_columns == []

    def test_self_join_columns(self):
        """
        Tests the self_join_columns of the PGTable class.
        """
        # Test setup
        mock_cursor = Mock()
        pg_table = PGTable(mock_cursor, "table_name")

        # Test for no self-join columns
        mock_cursor.fetchall.return_value = []
        self_join_cols = pg_table.self_join_columns()
        assert self_join_cols == []

        # Test for any self-join columns
        mock_cursor.fetchall.return_value = [{"foreign_table": "table_name", "fk_columns": "foreign_key_col_name"}]
        self_join_cols = pg_table.self_join_columns()
        assert self_join_cols == ["foreign_key_col_name"]
