# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Unit tests for enum.py module.

This test suite verifies the functionality of the enum classes in the enum.py module,
including HtcEventCodes, WorkflowStatusMessages, and DAGFinalStage.
The tests provide 100% coverage of all class methods and properties.
"""

import pytest

# Import the enum classes from the module
from wf_monitor.enum import DAGFinalStage, HtcEventCodes, WorkflowStatusMessages


class TestHtcEventCodes:
    """
    Test suite for the HtcEventCodes enum class.

    Tests all enum values, properties, and methods to ensure they
    correctly identify different types of HTCondor events.
    """

    def test_initialization(self):
        """
        Tests that HtcEventCodes initializes correctly with htc_value.

        :return: None
        """
        # Verify enums have the correct htc_value
        assert HtcEventCodes.JobSubmitted.htc_value == 0
        assert HtcEventCodes.JobExecuting.htc_value == 1
        assert HtcEventCodes.ExecutableError.htc_value == 2
        assert HtcEventCodes.JobCheckpoint.htc_value == 3
        assert HtcEventCodes.JobEvicted.htc_value == 4
        assert HtcEventCodes.JobTerminated.htc_value == 5
        assert HtcEventCodes.JobImageUpdate.htc_value == 6
        assert HtcEventCodes.ShadowException.htc_value == 7
        assert HtcEventCodes.GenericUnused.htc_value == 8
        assert HtcEventCodes.JobAborted.htc_value == 9
        assert HtcEventCodes.JobSuspended.htc_value == 10
        assert HtcEventCodes.JobUnsuspended.htc_value == 11
        assert HtcEventCodes.JobHeld.htc_value == 12
        assert HtcEventCodes.JobReleased.htc_value == 13
        assert HtcEventCodes.ParallelNodeExecuted.htc_value == 14
        assert HtcEventCodes.ParallelNodeTerminated.htc_value == 15
        assert HtcEventCodes.PostScriptTerminated.htc_value == 16

        # 17-20 Unused
        assert HtcEventCodes.RemoteError.htc_value == 21
        assert HtcEventCodes.RemoteSocketLost.htc_value == 22
        assert HtcEventCodes.RemoteSocketRegained.htc_value == 23
        assert HtcEventCodes.RemoteReconnectFailure.htc_value == 24
        assert HtcEventCodes.GridResourceReturned.htc_value == 25
        assert HtcEventCodes.GridResourceDown.htc_value == 26
        assert HtcEventCodes.GridResourceSubmission.htc_value == 27
        assert HtcEventCodes.JobAdInformationEvent.htc_value == 28
        assert HtcEventCodes.JobRemoteStatusUnknown.htc_value == 29
        assert HtcEventCodes.JobRemoteStatusReacquired.htc_value == 30
        assert HtcEventCodes.JobStageIn.htc_value == 31
        assert HtcEventCodes.JobStageOut.htc_value == 32
        assert HtcEventCodes.JobAdAttributeUpdate.htc_value == 33
        assert HtcEventCodes.PreSkipEvent.htc_value == 34
        assert HtcEventCodes.ClusterSubmit.htc_value == 35
        assert HtcEventCodes.ClusterRemove.htc_value == 36
        assert HtcEventCodes.FactoryPaused.htc_value == 37
        assert HtcEventCodes.FactoryResumed.htc_value == 38
        assert HtcEventCodes.GenericTimeout.htc_value == 39
        assert HtcEventCodes.FileTransfer.htc_value == 40

    def test_is_initialization_message(self):
        """
        Tests the is_initialization_message property.
        """
        for event in HtcEventCodes:
            if event.name in ["JobSubmitted", "ClusterSubmit"]:
                assert event.is_initialization_message
            else:
                assert not event.is_initialization_message

    def test_is_completion_message(self):
        """
        Tests the is_completion_message property.
        """
        for event in HtcEventCodes:
            if event.name in ["JobTerminated", "ClusterRemove"]:
                assert event.is_completion_message
            else:
                assert not event.is_completion_message

    def test_is_informational_message(self):
        """
        Tests the is_informational_message property.
        """
        for event in HtcEventCodes:
            if event.name in [
                "JobExecuting",
                "JobCheckpoint",
                "JobExecuting",
                "JobCheckpoint",
                "JobImageUpdate",
                "ParallelNodeExecuted",
                "ParallelNodeTerminated",
                "PostScriptTerminated",
                "RemoteSocketLost",
                "RemoteSocketRegained",
                "GridResourceReturned",
                "GridResourceDown",
                "GridResourceSubmission",
                "JobAdInformationEvent",
                "JobRemoteStatusUnknown",
                "JobRemoteStatusReacquired",
                "JobStageIn",
                "JobStageOut",
                "JobAdAttributeUpdate",
                "PreSkipEvent",
                "FactoryPaused",
                "FactoryResumed",
                "FileTransfer",
            ]:
                assert event.is_informational_message
            else:
                assert not event.is_informational_message

    def test_is_error_message(self):
        """
        Tests the is_error_message property.
        """
        for event in HtcEventCodes:
            if event.name in ["ExecutableError", "JobHeld", "RemoteError"]:
                assert event.is_error_message
            else:
                assert not event.is_error_message

    def test_is_resubmitting_message(self):
        """
        Tests the is_resubmitting_message property.
        """
        for event in HtcEventCodes:
            if event.name == "RemoteReconnectFailure":
                assert event.is_resubmitting_message
            else:
                assert not event.is_resubmitting_message

    def test_is_pause_message(self):
        """
        Tests the is_pause_message property.
        """
        for event in HtcEventCodes:
            if event.name in ["ShadowException", "JobSuspended"]:
                assert event.is_pause_message
            else:
                assert not event.is_pause_message

    def test_is_unpause_message(self):
        """
        Tests the is_unpause_message property.
        """
        for event in HtcEventCodes:
            if event.name in ["JobUnsuspended", "JobReleased"]:
                assert event.is_unpause_message
            else:
                assert not event.is_unpause_message

    def test_is_aborted_message(self):
        """
        Tests the is_aborted_message property.
        """
        for event in HtcEventCodes:
            if event.name in ["JobEvicted", "JobAborted"]:
                assert event.is_aborted_message
            else:
                assert not event.is_aborted_message

    def test_is_chirp_message(self):
        """
        Tests the is_chirp_message property.
        """
        for event in HtcEventCodes:
            if event.name == "GenericUnused":
                assert event.is_chirp_message
            else:
                assert not event.is_chirp_message


# pylint:disable=too-few-public-methods
class TestWorkflowStatusMessages:
    """
    Test suite for the WorkflowStatusMessages enum class.

    Tests all enum values to ensure they have the correct string representations.
    """

    def test_enum_values(self):
        """
        Test that WorkflowStatusMessages enum values are correctly defined.
        """
        assert WorkflowStatusMessages.EXECUTING.value == "workflow-executing"
        assert WorkflowStatusMessages.CONTINUE.value == "workflow-continuing"
        assert WorkflowStatusMessages.FAILED.value == "workflow-failed"
        assert WorkflowStatusMessages.TERMINATED.value == "workflow-terminated"
        assert WorkflowStatusMessages.ABORTED.value == "workflow-aborted"
        assert WorkflowStatusMessages.PAUSED.value == "workflow-paused"
        assert WorkflowStatusMessages.RESTARTED.value == "workflow-restarted"
        assert WorkflowStatusMessages.UPDATED.value == "workflow-updated"
        assert WorkflowStatusMessages.UNEXPECTED.value == "unexpected-message"
        assert WorkflowStatusMessages.COMPLETE.value == "workflow-complete"


class TestDAGFinalStage:
    """
    Test suite for the DAGFinalStage enum class.

    Tests all enum values and class methods to ensure they correctly
    identify final DAG stages.
    """

    def test_values_method(self):
        """
        Tests the values class method.
        """
        values_dict = DAGFinalStage.values()

        # Check that it's a dictionary
        assert isinstance(values_dict, dict)

        # Check that all enum values are keys in the dictionary
        assert "CONVEY" in values_dict
        assert "DELIVER" in values_dict
        assert "FINISH" in values_dict
        assert "POST" in values_dict

        # Check that the values in the dictionary are the correct enum members
        assert values_dict["CONVEY"] == DAGFinalStage.CONVEY
        assert values_dict["DELIVER"] == DAGFinalStage.DELIVER
        assert values_dict["FINISH"] == DAGFinalStage.FINISH
        assert values_dict["POST"] == DAGFinalStage.POST

    def test_is_final_stage(self):
        """
        Tests the is_final_stage class method.
        """
        # Strings that should be identified as final stages
        assert DAGFinalStage.is_final_stage("CONVEY") is True
        assert DAGFinalStage.is_final_stage("DELIVER") is True
        assert DAGFinalStage.is_final_stage("FINISH") is True
        assert DAGFinalStage.is_final_stage("POST") is True

        # Strings that should not be identified as final stages
        assert DAGFinalStage.is_final_stage("START") is False
        assert DAGFinalStage.is_final_stage("MIDDLE") is False
        assert DAGFinalStage.is_final_stage("convey") is False  # Test for case-sensitivity
        assert DAGFinalStage.is_final_stage("PROCESS") is False


if __name__ == "__main__":
    pytest.main(["-xvs", __file__, "--cov=enum", "--cov-report=term-missing"])
