# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Unit tests for the WorkflowMonitor class and related functions in the wf_monitor.monitor module.
This module provides comprehensive test coverage for log parsing, event handling,
message formatting, and other functionality of the wf_monitor tool.
"""
from copy import deepcopy
from pathlib import Path
from threading import Thread
from time import sleep
from tracemalloc import Frame
from unittest import mock

import pytest
from wf_monitor.enum import WorkflowStatusMessages
from wf_monitor.monitor import (
    WorkflowMonitor,
    fmt_timestamp,
    get_chirp_type,
    get_retval,
    is_final_stage,
    log_decorator_factory,
    main,
    make_arg_parser,
    send_message,
    timeout_handler,
)

DAG_NODE_PROCESS = "DAG Node: PROCESS"
DAG_NODE_FINISH = "DAG Node: FINISH"


class TestHelperFunctions:
    """Tests for helper functions in the monitor module outside the WorkflowMonitor class."""

    def test_fmt_timestamp(self):
        """
        Tests that the fmt_timestamp function formats timestamps correctly.
        """
        # Test with a valid timestamp
        timestamp = "2023-01-15 14:30:45"
        expected = "2023-01-15T14:30:45+07:00"
        assert fmt_timestamp(timestamp) == expected

    def test_get_retval(self):
        """
        Tests that the get_retval function extracts return values correctly.
        """
        # Test with a return value of 0
        body = "Job executing on host: example.host.com\n\treturn value 0"
        assert get_retval(body) == 0

        # Test with a return value of 1
        body = "Job executing on host: example.host.com\n\treturn value 1"
        assert get_retval(body) == 1

        # Test with a 2-digit return value
        body = "Job executing on host: example.host.com\n\treturn value 22"
        assert get_retval(body) == 22

        # Test with no return value
        body = "Job executing on host: example.host.com"
        assert get_retval(body) is None

    def test_get_chirp_type(self):
        """
        Tests that the get_chirp_type function correctly identifies chirp message types.
        """
        # Test with a SCG message
        chirp_msg = "Chirp message containing nraorsync info"
        assert get_chirp_type(chirp_msg) == "nraorsync"

        # Test with a non-SCG message
        chirp_msg = "Regular chirp message"
        assert get_chirp_type(chirp_msg) == "stage_update"

    def test_is_final_stage(self):
        """
        Tests that the is_final_stage function correctly identifies final DAG stages.
        """
        # Test with a final stage
        body = DAG_NODE_FINISH
        assert is_final_stage(body) is True

        # Test with a non-final stage
        body = DAG_NODE_PROCESS
        assert is_final_stage(body) is False

    def test_log_decorator_factory(self):
        """
        Tests that the log_decorator_factory correctly creates decorators that log messages.
        """
        # Create a test function
        @log_decorator_factory("Test message")
        def test_func(arg1, arg2=None):
            return arg1, arg2

        # Call the decorated function and verify it returns the correct results
        with mock.patch("logging.Logger.info") as mock_info:
            result = test_func("test", arg2="value")
            assert result == ("test", "value")
            mock_info.assert_called_once_with("Test message")

    def test_send_message(self, mock_message_sender: mock.MagicMock):
        """
        Tests that send_message correctly sends message via the messenger.
        """
        message = {"service": "workflow", "subject": {"workflow_request_id": 123}, "type": "workflow-executing"}

        with mock.patch("logging.Logger.debug") as mock_log:
            send_message(message, mock_message_sender)
            mock_log.assert_called_once()
            mock_message_sender.send_message.assert_called_once()

    def test_timeout_handler(self):
        """
        Tests that the timeout_handler raises TimeoutError with correct message.
        """
        with pytest.raises(TimeoutError) as excinfo:
            # Creating a fake Frame instance so the method is getting what it normally does
            timeout_handler(-1, Frame(tuple(("fake-file-name", 2))))
        assert "No HTCondor log file found in timeout period" in str(excinfo.value)


@pytest.fixture(name="mock_message_sender")
def get_mock_message_sender() -> mock.MagicMock:
    """
    Fixture providing a mocked MessageSender.
    """
    with mock.patch("wf_monitor.monitor.MessageSender") as mock_sender_class:
        mock_sender = mock.MagicMock()
        mock_sender_class.return_value = mock_sender
        yield mock_sender


# This is so we can have all these methods in this class
# pylint:disable=too-many-public-methods
class TestWorkflowMonitor:
    """Tests for WorkflowMonitor class methods."""

    # The numbers at the beginning (e.g.) make it so certain code does or does not execute, and they are HTC event codes
    EVENT_STR_INFORMATIONAL_MESSAGE = (
        "001 (123.000.000) 2023-01-15 14:30:45 Job executing on host: example.com\n\tSome info\n...\n"
    )

    PARSE_MESSAGE_INFORMATIONAL_EXPECTED = {
        "service": "workflow",
        "subject": {
            "type": "WorkflowRequest",
            "workflow_request_id": 123,
        },
        "routing_key": "workflow",
        "type": WorkflowStatusMessages.CONTINUE.value,
        "condor_metadata": {
            "event_type": WorkflowStatusMessages.CONTINUE.value,
            "condor_job_id": "123",
            "retval": None,
            "timestamp": fmt_timestamp("2023-01-15 14:30:45"),
            "log": EVENT_STR_INFORMATIONAL_MESSAGE,
        },
    }

    # nraorysnc is so certain code executes
    EVENT_STR_CHIRP_MESSAGE = (
        "008 (123.000.000) 2023-01-15 14:30:45 Job executing on host: nraorsync\n\tSome info\n...\n"
    )

    PARSE_MESSAGE_CHIRP_EXPECTED = {
        "service": "workflow",
        "subject": {
            "type": "WorkflowRequest",
            "workflow_request_id": 123,
        },
        "routing_key": "workflow",
        "type": WorkflowStatusMessages.FAILED.value,
        "condor_metadata": {
            "event_type": WorkflowStatusMessages.FAILED.value,
            "condor_job_id": "123",
            "retval": None,
            "timestamp": fmt_timestamp("2023-01-15 14:30:45"),
            "log": EVENT_STR_CHIRP_MESSAGE,
        },
    }

    EVENT_STR_JOB_TERMINATED_MESSAGE = (
        "005 (123.000.000) 2023-01-15 14:30:45 Job executing on host: example.com\n\tSome info\n...\n"
    )

    PARSE_MESSAGE_JOB_TERMINATED_EXPECTED = deepcopy(PARSE_MESSAGE_CHIRP_EXPECTED)
    PARSE_MESSAGE_JOB_TERMINATED_EXPECTED["condor_metadata"]["log"] = EVENT_STR_JOB_TERMINATED_MESSAGE

    SAMPLE_HTC_WORKFLOW_EXECUTING_EVENT = (
        "000 (123.000.000) 2023-01-15 14:30:45 Job submitted from host: example.com\n"
        f"\t{DAG_NODE_PROCESS}\n\tSome additional information\n...\n"
    )

    SAMPLE_HTC_WORKFLOW_EXECUTING_EVENT_HTC_LOG_DAG_JOB = [
        "000 (123.000.000) ",  # First part of event
        "2023-01-15 14:30:45 Job submitted from host: example.com\n",  # Second part of event
        f"\t{DAG_NODE_PROCESS}\n",  # DAG info
        "\tSome additional information\n",  # More info
        "",  # To test readline returning this
        "...\n",  # End of first event
    ]

    SAMPLE_HTC_JOB_CONTINUING_EVENT = (
        "001 (123.000.000) 2023-01-15 14:30:45 Job executing on host: example.com\n"
        "\tSome additional information\n...\n"
    )

    PARSE_MESSAGE_JOB_CONTINUING_EVENT_EXPECTED = {
        "service": "workflow",
        "subject": {
            "type": "WorkflowRequest",
            "workflow_request_id": 123,
        },
        "routing_key": "workflow",
        "type": WorkflowStatusMessages.CONTINUE.value,
        "condor_metadata": {
            "event_type": WorkflowStatusMessages.CONTINUE.value,
            "condor_job_id": "123",
            "retval": None,
            "timestamp": fmt_timestamp("2023-01-15 14:30:45"),
            "log": SAMPLE_HTC_JOB_CONTINUING_EVENT,
        },
    }

    SAMPLE_HTC_TERMINATION_EVENT = (
        "005 (123.000.000) 2023-01-15 14:35:45 Job terminated.\n\tSome information\n\treturn value 0\n...\n"
    )

    SAMPLE_HTC_TERMINATION_EVENT_HTC_LOG_DAG_JOB = [
        "\n",  # To test a newline being in a log
        "005 (123.000.000) ",  # Start of termination event
        "2023-01-15 14:35:45 Job terminated.\n",  # Part of termination
        "\tSome information\n",  # More of termination
        "\treturn value 0\n",  # More of termination
        "...\n",  # End of event
    ]

    SAMPLE_DAG_EXECUTION_EVENT = (
        "001 (123.000.000) 2023-01-15 14:30:45 Job executing on host: example.com\n"
        f"\t{DAG_NODE_PROCESS}\n\tSome additional information\n...\n"
    )

    SAMPLE_FINAL_DAG_EVENT = (
        "000 (123.000.000) 2023-01-15 14:30:45 Job submitted from host: example.com\n"
        f"\t{DAG_NODE_FINISH}\n\tSome additional information\n...\n"
    )

    @staticmethod
    def write_contents_to_logfile(logfile_path: Path, contents: str):
        """
        Writes the specified log contents to the provided log file.
        :param logfile_path: The path to the testing log file.
        :param contents: The contents to write to the log file.
        """
        with open(logfile_path, "w", encoding="utf8") as logfile:
            logfile.write(contents + "")

    @pytest.fixture
    def log_file_path(self, tmp_path: Path) -> Path:
        """
        Fixture providing a temporary log file path.

        :param tmp_path: Pytest temporary directory fixture
        :return: Path to the temporary log file
        """
        return tmp_path / "test_condor.log"

    @pytest.fixture
    def dag_log_file_path(self, tmp_path: Path) -> Path:
        """
        Fixture providing a temporary DAG log file path.

        :param tmp_path: Pytest temporary directory fixture
        :return: Path to the temporary DAG log file
        """
        return tmp_path / "dag.log"

    def test_init(self, log_file_path: Path):
        """
        Tests the initialization of WorkflowMonitor.
        """
        with mock.patch("wf_monitor.monitor.WorkflowMonitor.determine_monitor_type"), mock.patch(
            "messaging.messenger.MessageSender.__init__"
        ) as mock_message_sender_init:
            mock_message_sender_init.return_value = None  # To avoid a TypeError
            monitor = WorkflowMonitor(str(log_file_path), 123, False)

            assert monitor.logfile_path == log_file_path
            assert monitor.workflow_request_id == 123
            assert monitor.is_dag is False
            assert not monitor.events
            assert monitor.log == ""

            assert mock_message_sender_init.call_count == 1  # Make sure it tried to create a MessageSender

    def test_determine_monitor_type_not_dag(self, log_file_path: Path):
        """
        Tests determine_monitor_type when not monitoring a DAG log.
        """
        with mock.patch("wf_monitor.monitor.WorkflowMonitor.read_htcondor_log") as mock_read, mock.patch(
            "messaging.messenger.MessageSender.__init__"
        ) as mock_message_sender_init:
            mock_message_sender_init.return_value = None  # To avoid a TypeError
            _ = WorkflowMonitor(str(log_file_path), 123, False, 1)
            mock_read.assert_called_once_with(log_file_path, 1)

    def test_determine_monitor_type_dag(self, dag_log_file_path: Path):
        """
        Tests determine_monitor_type when monitoring a DAG log.
        """
        with mock.patch("wf_monitor.monitor.Process") as mock_process, mock.patch(
            "messaging.messenger.MessageSender.__init__"
        ) as mock_message_sender_init:
            mock_message_sender_init.return_value = None  # To avoid a TypeError
            instance = mock_process.return_value

            _ = WorkflowMonitor(str(dag_log_file_path), 123, True, 1)

            # Verify two processes were started (one for DAG log, one for condor log)
            assert instance.start.call_count == 2
            assert instance.join.call_count == 2

    def test_log_exists_success(self, log_file_path: Path):
        """
        Tests log_exists when the log file is found.
        """
        # Create the log file
        log_file_path.touch()

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            result = monitor.log_exists(log_file_path, 10)
            assert result is True

    def test_log_exists_timeout(self, log_file_path: Path):
        """
        Tests log_exists when the log file is not found within the timeout.
        """
        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)

            with mock.patch("pathlib.Path.is_file") as mock_is_file, pytest.raises(TimeoutError):
                # So the 'while' loop runs for if the file doesn't exist
                # and then a TimeoutError happens at 4 seconds
                mock_is_file.side_effect = [
                    False,  # At around 0 seconds
                    False,  # at around 3 seconds (the time it sleeps between checks in log_exists)
                ]
                monitor.log_exists(log_file_path, 4)

    @pytest.mark.parametrize(
        "event, expected_result",
        [
            (SAMPLE_HTC_JOB_CONTINUING_EVENT, PARSE_MESSAGE_JOB_CONTINUING_EVENT_EXPECTED),
            (EVENT_STR_INFORMATIONAL_MESSAGE, PARSE_MESSAGE_INFORMATIONAL_EXPECTED),
            (EVENT_STR_CHIRP_MESSAGE, PARSE_MESSAGE_CHIRP_EXPECTED),
            (EVENT_STR_JOB_TERMINATED_MESSAGE, PARSE_MESSAGE_JOB_TERMINATED_EXPECTED),
        ],
    )
    def test_parse_message_valid(self, event, expected_result):
        """
        Tests parse_message with a valid event.
        """
        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            actual_result = monitor.parse_message(event)

            assert actual_result == expected_result

    def test_parse_message_invalid(self):
        """
        Tests parse_message with an invalid event.
        """
        invalid_event = "This is not a valid HTCondor event"

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)

            with pytest.raises(ValueError):
                monitor.parse_message(invalid_event)

    # pylint:disable=protected-access # This is so we can test protected methods

    def test_parse_event_type(self):
        """
        Tests that _parse_event_type correctly converts HTCondor event numbers to WorkflowStatusMessages enum values.
        """
        assert WorkflowMonitor._parse_event_type(0) == WorkflowStatusMessages.EXECUTING.value  # Initialization message
        assert WorkflowMonitor._parse_event_type(12) == WorkflowStatusMessages.FAILED.value  # Error message
        assert WorkflowMonitor._parse_event_type(24) == WorkflowStatusMessages.RESTARTED.value  # Re-submitting message
        assert WorkflowMonitor._parse_event_type(5) == WorkflowStatusMessages.TERMINATED.value  # Completion message
        assert WorkflowMonitor._parse_event_type(9) == WorkflowStatusMessages.ABORTED.value  # Aborted message
        assert WorkflowMonitor._parse_event_type(7) == WorkflowStatusMessages.PAUSED.value  # Pause message
        assert WorkflowMonitor._parse_event_type(11) == WorkflowStatusMessages.CONTINUE.value  # Unpause message
        assert WorkflowMonitor._parse_event_type(16) == WorkflowStatusMessages.CONTINUE.value  # Informational message
        assert WorkflowMonitor._parse_event_type(8) == WorkflowStatusMessages.UPDATED.value  # Chirp message
        assert WorkflowMonitor._parse_event_type(39) == WorkflowStatusMessages.UNEXPECTED.value  # Generic timeout msg

    def test_get_job_result(self):
        """
        Tests that _get_job_result returns the correct status based on job's return value.
        """
        assert WorkflowMonitor._get_job_result(0) == WorkflowStatusMessages.COMPLETE.value
        assert WorkflowMonitor._get_job_result(1) == WorkflowStatusMessages.FAILED.value
        assert WorkflowMonitor._get_job_result(123) == WorkflowStatusMessages.FAILED.value

    # pylint:enable=protected-access

    def test_close(self, mock_message_sender: mock.MagicMock):
        """
        Tests that the close method properly closes the message router.
        """
        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.message_router = mock_message_sender
            monitor.workflow_request_id = 123

            with mock.patch("time.sleep") as mock_sleep:
                monitor.close()
                mock_sleep.assert_called_once_with(3)
                mock_message_sender.close.assert_called_once()

    def test_read_htcondor_log_file_not_found(self, log_file_path: Path):
        """
        Tests the read_htcondor_log method for when the log file is not found within the timeout.
        """
        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists") as mock_log_exists:
                mock_log_exists.side_effect = TimeoutError("Test timeout")

                with mock.patch("logging.Logger.error") as mock_error, pytest.raises(SystemExit):
                    monitor.read_htcondor_log(log_file_path, 4)
                mock_error.assert_called_once()
                mock_log_exists.assert_called_once()

    def test_read_htcondor_log_standard_job(
        self,
        log_file_path: Path,
        mock_message_sender: mock.MagicMock,
    ):
        """
        Tests the read_htcondor_log method for a standard (non-DAG) job.
        """
        TestWorkflowMonitor.write_contents_to_logfile(
            log_file_path,
            TestWorkflowMonitor.SAMPLE_HTC_JOB_CONTINUING_EVENT + TestWorkflowMonitor.SAMPLE_HTC_TERMINATION_EVENT,
        )

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.is_dag = False
            monitor.message_router = mock_message_sender
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists", return_value=True), mock.patch(
                "wf_monitor.monitor.WorkflowMonitor.parse_message"
            ) as mock_parse, mock.patch("wf_monitor.monitor.send_message") as mock_send:
                # Set up mock parse_message to return appropriate events
                # We are mocking this to test how many times the method was called
                mock_parse.side_effect = [
                    {"type": "workflow-continuing", "condor_metadata": {"log": self.SAMPLE_HTC_JOB_CONTINUING_EVENT}},
                    {"type": "workflow-terminated", "condor_metadata": {"log": self.SAMPLE_HTC_TERMINATION_EVENT}},
                ]

                monitor.read_htcondor_log(log_file_path, 5)

                # Verify the events were processed and messages sent
                assert mock_parse.call_count == 2
                assert mock_send.call_count == 2
                assert len(monitor.events) == 2
                assert monitor.log == TestWorkflowMonitor.SAMPLE_HTC_JOB_CONTINUING_EVENT

    def test_read_htcondor_log_dag_job(
        self,
        log_file_path: Path,
        mock_message_sender: mock.MagicMock,
    ):
        """
        Tests the read_htcondor_log method for a DAG job with regular execution events.
        """
        TestWorkflowMonitor.write_contents_to_logfile(
            log_file_path,
            "".join(
                TestWorkflowMonitor.SAMPLE_HTC_WORKFLOW_EXECUTING_EVENT_HTC_LOG_DAG_JOB
                + TestWorkflowMonitor.SAMPLE_HTC_TERMINATION_EVENT_HTC_LOG_DAG_JOB
            ),
        )

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.is_dag = True
            monitor.message_router = mock_message_sender
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists", return_value=True), mock.patch(
                "wf_monitor.monitor.WorkflowMonitor.parse_message"
            ) as mock_parse, mock.patch("wf_monitor.monitor.send_message") as mock_send, mock.patch(
                "wf_monitor.monitor.is_final_stage"
            ) as mock_is_final:
                # Set up mock parse_message to return appropriate events
                # We are mocking this to test how many times the method was called
                mock_parse.side_effect = [
                    {"type": "workflow-executing", "condor_metadata": {"log": DAG_NODE_FINISH}},
                    {
                        "type": "workflow-terminated",
                        "condor_metadata": {"log": self.SAMPLE_HTC_TERMINATION_EVENT_HTC_LOG_DAG_JOB},
                    },
                ]

                # Set up is_final_stage to return True (it would only be set for the first event)
                # This is so we can test how many times it was set
                mock_is_final.return_value = True

                monitor.read_htcondor_log(log_file_path, 5)

                # Verify the events were processed
                assert mock_parse.call_count == 2
                assert mock_send.call_count == 1  # Only called for the workflow-executing event
                assert mock_is_final.call_count == 1  # Only set for the workflow-executing event
                assert len(monitor.events) == 2
                assert monitor.log == "".join(TestWorkflowMonitor.SAMPLE_HTC_WORKFLOW_EXECUTING_EVENT_HTC_LOG_DAG_JOB)

    def test_read_htcondor_log_dag_job_final_stage_with_preceding_non_final_stage(
        self,
        log_file_path: Path,
        mock_message_sender: mock.MagicMock,
    ):
        """
        Test the read_htcondor_log method for a DAG job with an execution event that is
        the final stage of the DAG log, sandwiched in between two termination events,
        the first of which isn't for a final stage.
        """
        # Drop the new line at the beginning
        termination_event_modified = TestWorkflowMonitor.SAMPLE_HTC_TERMINATION_EVENT_HTC_LOG_DAG_JOB[1:]

        # Switch from "DAG Node: Process" to "DAG Node: FINISH" so certain code executes
        workflow_executing_event_modified = deepcopy(
            TestWorkflowMonitor.SAMPLE_HTC_WORKFLOW_EXECUTING_EVENT_HTC_LOG_DAG_JOB
        )
        workflow_executing_event_modified[2] = workflow_executing_event_modified[2].replace(
            DAG_NODE_PROCESS, DAG_NODE_FINISH
        )

        TestWorkflowMonitor.write_contents_to_logfile(
            log_file_path,
            "".join(termination_event_modified + workflow_executing_event_modified + termination_event_modified),
        )

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.is_dag = True
            monitor.message_router = mock_message_sender
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists", return_value=True), mock.patch(
                "wf_monitor.monitor.WorkflowMonitor.parse_message"
            ) as mock_parse, mock.patch("wf_monitor.monitor.send_message") as mock_send, mock.patch(
                "wf_monitor.monitor.is_final_stage"
            ) as mock_is_final:
                # Set up mock parse_message to return appropriate events
                # We are mocking this to test how many times the method was called
                mock_parse.side_effect = [
                    {"type": "workflow-terminated", "condor_metadata": {"log": termination_event_modified}},
                    {"type": "workflow-executing", "condor_metadata": {"log": workflow_executing_event_modified}},
                    {"type": "workflow-terminated", "condor_metadata": {"log": termination_event_modified}},
                ]

                # Set up is_final_stage to be set in the processing of the executing message
                mock_is_final.return_value = True

                monitor.read_htcondor_log(log_file_path, 5)

                # Verify the events were processed
                assert mock_parse.call_count == 3
                assert mock_send.call_count == 1
                assert mock_is_final.call_count == 1  # Only set for the executing event
                assert len(monitor.events) == 3
                assert monitor.log == "".join(termination_event_modified + workflow_executing_event_modified)

    def test_read_htcondor_log_dag_job_final_stage(
        self,
        log_file_path: Path,
        mock_message_sender: mock.MagicMock,
    ):
        """
        Tests the read_htcondor_log method for a DAG job with final stage execution.
        """
        # Make it a list and not a string
        modified_sample_final_dag_event = TestWorkflowMonitor.SAMPLE_FINAL_DAG_EVENT.splitlines(keepends=True)

        TestWorkflowMonitor.write_contents_to_logfile(
            log_file_path,
            "".join(modified_sample_final_dag_event + TestWorkflowMonitor.SAMPLE_HTC_TERMINATION_EVENT_HTC_LOG_DAG_JOB),
        )

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.is_dag = True
            monitor.message_router = mock_message_sender
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists", return_value=True), mock.patch(
                "wf_monitor.monitor.WorkflowMonitor.parse_message"
            ) as mock_parse, mock.patch("wf_monitor.monitor.send_message") as mock_send, mock.patch(
                "wf_monitor.monitor.is_final_stage"
            ) as mock_is_final:
                # Set up mock parse_message to return appropriate events
                # We are mocking this to test how many times the method was called
                mock_parse.side_effect = [
                    {"type": "workflow-executing", "condor_metadata": {"log": modified_sample_final_dag_event}},
                    {
                        "type": "workflow-terminated",
                        "condor_metadata": {"log": TestWorkflowMonitor.SAMPLE_HTC_TERMINATION_EVENT_HTC_LOG_DAG_JOB},
                    },
                ]

                # Set up is_final_stage to return True for the final stage event
                mock_is_final.return_value = True

                monitor.read_htcondor_log(log_file_path, 5)

                # Verify the events were processed
                assert mock_parse.call_count == 2
                assert mock_send.call_count == 1
                assert mock_is_final.call_count == 1  # See if it was set once (for the final stage event)
                assert len(monitor.events) == 2
                assert monitor.log == "".join(modified_sample_final_dag_event)

    def test_read_htcondor_log_malformed_event(
        self,
        log_file_path: Path,
        mock_message_sender: mock.MagicMock,
    ):
        """
        Tests the read_htcondor_log method for when a malformed event is encountered.
        """
        TestWorkflowMonitor.write_contents_to_logfile(
            log_file_path,
            "".join(
                [
                    "This is a malformed event\n",  # Malformed event
                    "...\n",  # End of malformed event,
                    TestWorkflowMonitor.SAMPLE_HTC_TERMINATION_EVENT,
                ]
            ),
        )

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.is_dag = False
            monitor.message_router = mock_message_sender
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists", return_value=True), mock.patch(
                "wf_monitor.monitor.WorkflowMonitor.parse_message"
            ) as mock_parse, mock.patch("wf_monitor.monitor.send_message") as mock_send, mock.patch(
                "logging.Logger.error"
            ) as mock_error:
                # First call raises ValueError, second returns a valid event
                # We are mocking this to test how many times the method was called
                mock_parse.side_effect = [
                    ValueError("Malformed event"),
                    {
                        "type": "workflow-terminated",
                        "condor_metadata": {"log": TestWorkflowMonitor.SAMPLE_HTC_TERMINATION_EVENT},
                    },
                ]

                monitor.read_htcondor_log(log_file_path, 5)

                # Verify that the error was logged and we continued processing
                mock_error.assert_called_once()
                assert mock_parse.call_count == 2
                assert mock_send.call_count == 1
                assert len(monitor.events) == 1
                assert monitor.log == ""

    def test_read_htcondor_log_readline_empty(
        self,
        log_file_path: Path,
        mock_message_sender: mock.MagicMock,
    ):
        """
        Tests read_htcondor_log for when readline returns "".
        """

        # Set up a thread to write to the file while it's being read
        def writer_thread():
            with open(log_file_path, "a", encoding="utf8") as f:
                f.write("Event line 1\n")
                f.flush()
                sleep(0.5)
                f.write("Event line 2\n")
                f.flush()
                sleep(0.5)
                f.write("...\n")  # End of event marker
                f.flush()
                sleep(1)
                f.write("Final event line\n")
                f.write("...\n")  # Final event marker
                f.flush()

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.is_dag = False
            monitor.message_router = mock_message_sender
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists", return_value=True), mock.patch(
                "wf_monitor.monitor.WorkflowMonitor.parse_message"
            ) as mock_parse, mock.patch("wf_monitor.monitor.send_message") as mock_send, mock.patch(
                "logging.Logger.error"
            ) as mock_error:
                # First call raises ValueError, second returns a valid event
                # We are mocking this to test how many times the method was called
                mock_parse.side_effect = [
                    ValueError("Malformed event"),
                    {"type": "workflow-terminated", "condor_metadata": {"log": "valid event"}},
                ]
                # Start the writer thread
                writer = Thread(target=writer_thread)
                writer.start()

                monitor.read_htcondor_log(log_file_path, 5)

                # Verify error was logged and we continued processing
                mock_error.assert_called_once()
                assert mock_parse.call_count == 2
                assert mock_send.call_count == 1
                assert len(monitor.events) == 1
                assert monitor.log == ""

    def test_read_dag_log(
        self,
        dag_log_file_path: Path,
        mock_message_sender: mock.MagicMock,
    ):
        """
        Tests read_dag_log for successful execution.
        """
        TestWorkflowMonitor.write_contents_to_logfile(
            dag_log_file_path,
            TestWorkflowMonitor.SAMPLE_DAG_EXECUTION_EVENT + TestWorkflowMonitor.SAMPLE_HTC_TERMINATION_EVENT,
        )

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.message_router = mock_message_sender
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists", return_value=True), mock.patch(
                "wf_monitor.monitor.WorkflowMonitor.parse_message"
            ) as mock_parse, mock.patch("wf_monitor.monitor.send_message") as mock_send:
                # Set up mock parse_message to return appropriate events
                # We are mocking this to test how many times the method was called
                mock_parse.side_effect = [
                    {"type": "workflow-continuing", "condor_metadata": {"log": self.SAMPLE_DAG_EXECUTION_EVENT}},
                    {"type": "workflow-terminated", "condor_metadata": {"log": self.SAMPLE_HTC_TERMINATION_EVENT}},
                ]

                monitor.read_dag_log(dag_log_file_path, 5)

                # Verify the events were processed and messages sent
                assert mock_parse.call_count == 2
                assert mock_send.call_count == 2
                assert len(monitor.events) == 2
                assert monitor.log == "".join(TestWorkflowMonitor.SAMPLE_DAG_EXECUTION_EVENT)

    def test_read_dag_log_parse_message_malformed_event(
        self,
        dag_log_file_path: Path,
        mock_message_sender: mock.MagicMock,
    ):
        """
        Tests read_dag_log for a log with a malformed message.
        """
        TestWorkflowMonitor.write_contents_to_logfile(
            dag_log_file_path,
            "".join(
                [
                    TestWorkflowMonitor.SAMPLE_DAG_EXECUTION_EVENT,
                    "\n",  # To test for when there is a newline in between events
                    TestWorkflowMonitor.SAMPLE_HTC_TERMINATION_EVENT,
                ]
            ),
        )

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.message_router = mock_message_sender
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists", return_value=True), mock.patch(
                "wf_monitor.monitor.WorkflowMonitor.parse_message"
            ) as mock_parse, mock.patch("wf_monitor.monitor.send_message") as mock_send, mock.patch(
                "logging.Logger.error"
            ) as mock_error:
                # First call raises ValueError, second returns a valid event
                # We are mocking this to test how many times the method was called
                mock_parse.side_effect = [
                    ValueError("Malformed event"),
                    {"type": "workflow-terminated", "condor_metadata": {"log": "valid event"}},
                ]

                monitor.read_dag_log(dag_log_file_path, 5)

                # Verify error was logged and we continued processing
                mock_error.assert_called_once()
                assert mock_parse.call_count == 2
                assert mock_send.call_count == 1
                assert len(monitor.events) == 1
                assert monitor.log == ""

    def test_read_dag_log_readline_empty(
        self,
        dag_log_file_path: Path,
        mock_message_sender: mock.MagicMock,
    ):
        """
        Tests read_dag_log for when readline returns "".
        """

        # Set up a thread to write to the file while it's being read
        def writer_thread():
            with open(dag_log_file_path, "a", encoding="utf8") as f:
                f.write("Event line 1\n")
                f.flush()
                sleep(0.5)
                f.write("Event line 2\n")
                f.flush()
                sleep(0.5)
                f.write("...\n")  # End of event marker
                f.flush()
                sleep(1)
                f.write("Final event line\n")
                f.write("...\n")  # Final event marker
                f.flush()

        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.message_router = mock_message_sender
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists", return_value=True), mock.patch(
                "wf_monitor.monitor.WorkflowMonitor.parse_message"
            ) as mock_parse, mock.patch("wf_monitor.monitor.send_message") as mock_send, mock.patch(
                "logging.Logger.error"
            ) as mock_error:
                # First call raises ValueError, second returns a valid event
                # We are mocking this to test how many times the method was called
                mock_parse.side_effect = [
                    ValueError("Malformed event"),
                    {"type": "workflow-terminated", "condor_metadata": {"log": "valid event"}},
                ]
                # Start the writer thread
                writer = Thread(target=writer_thread)
                writer.start()

                monitor.read_dag_log(dag_log_file_path, 5)

                # Verify error was logged and we continued processing
                mock_error.assert_called_once()
                assert mock_parse.call_count == 2
                assert mock_send.call_count == 1
                assert len(monitor.events) == 1
                assert monitor.log == ""

    def test_read_dag_log_system_exit(
        self,
        dag_log_file_path: Path,
    ):
        """
        Tests the read_dag_log method for when a TimeoutError is raised in log_exists.

        There is no need to create a mock instance of the MessageSender class
        because an instance of it is never accessed in this test.
        """
        with mock.patch("wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None):
            monitor = WorkflowMonitor(None, None, None)
            monitor.workflow_request_id = 123
            monitor.events = []
            monitor.log = ""

            with mock.patch("wf_monitor.monitor.WorkflowMonitor.log_exists") as mock_log_exists:
                mock_log_exists.side_effect = TimeoutError("Test timeout")

                with mock.patch("logging.Logger.error") as mock_error, pytest.raises(SystemExit):
                    monitor.read_dag_log(dag_log_file_path, 5)

                mock_error.assert_called_once()

    def test_str_and_repr(self, log_file_path: Path):
        """
        Tests the __str__ and __repr__ methods.
        """
        with mock.patch("wf_monitor.monitor.WorkflowMonitor.determine_monitor_type"), mock.patch(
            "wf_monitor.monitor.WorkflowMonitor.__init__", return_value=None
        ):
            monitor = WorkflowMonitor(None, None, False)
            monitor.workflow_request_id = 123
            monitor.logfile_path = log_file_path
            monitor.events = []
            monitor.log = ""

            # Test __str__
            str_result = str(monitor)
            assert "WorkflowMonitor" in str_result
            assert str(log_file_path) in str_result

            # Test __repr__
            repr_result = repr(monitor)
            assert "WorkflowMonitor" in repr_result
            assert str(log_file_path) in repr_result


# pylint:enable=too-many-public-methods


class TestCommandLine:
    """
    Tests for command-line functionality.
    """

    def test_make_arg_parser(self):
        """
        Tests that make_arg_parser creates a properly configured parser.
        """
        parser = make_arg_parser()

        # Test default values
        args = parser.parse_args(["/path/to/log", "123", "True"])
        assert args.log_path == "/path/to/log"
        assert args.workflow_request_id == "123"
        assert args.is_dag == "True"
        assert args.profile == "dsoc-dev"

        # Test with specifying the profile
        args = parser.parse_args(["-P", "test", "/path/to/log", "123", "True"])
        assert args.log_path == "/path/to/log"
        assert args.workflow_request_id == "123"
        assert args.is_dag == "True"
        assert args.profile == "test"

    def test_main(self):
        """
        Tests that the main function correctly sets up and runs the monitor.
        """
        with mock.patch("sys.argv", ["wf_monitor", "/path/to/log", "123", "False"]), mock.patch(
            "wf_monitor.monitor.WorkflowMonitor"
        ) as mock_monitor_class:
            mock_monitor = mock.MagicMock()
            mock_monitor_class.return_value = mock_monitor

            with mock.patch("logging.Logger.info"):
                main()

                # Verify WorkflowMonitor was created with correct args and close was called
                mock_monitor_class.assert_called_once_with("/path/to/log", 123, False)
                mock_monitor.close.assert_called_once()
