#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Reads in an HTCondor log file and parses the relevant data from it, splitting it into events
and formatting those into messages that are sent to the workspaces.workflow-service.workflow-status
RabbitMQ exchange. Also input is a workflow request ID that is sent alongside the event data.
"""

import argparse
import functools
import logging
import re
import signal
import time
import tracemalloc
from multiprocessing import Process
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple, Union

from messaging.messenger import MessageSender

from . import __version__ as version
from .enum import DAGFinalStage, HtcEventCodes, WorkflowStatusMessages

# pylint: disable=E0401, E0402, W1203

logger = logging.getLogger("wf_monitor")

SCG_MSG = "nraorsync"


def log_decorator_factory(msg: str):
    """
    Factory that produces log decorators with a given message

    :param msg: Message to be logged
    :return: Log decorator
    """

    def log_decorator(func: Callable) -> Callable:
        """
        Decorator that logs a message before and after a function is called

        :param func: Function to be decorated
        :return: Wrapper function
        """

        @functools.wraps(func)
        def wrapper(*args: Tuple, **kwargs: Dict) -> Any:
            """
            Decorator boilerplate: wrapper function for log decorator

            :param args: args for function f
            :param kwargs: kwargs for function f
            :return: Return value of the function
            """
            logger.info(msg)
            result = func(*args, **kwargs)
            return result

        return wrapper

    return log_decorator


def timeout_handler(signum: int, frame: tracemalloc.Frame) -> None:
    """
    Signal handler function for handling a SIGALRM that arrives when WorkflowMonitor.read_log()
    times out.

    :param signum: Signal number (would be signal.SIGALRM in this case)
    :param frame: Stack frame the signal was sent from
    """
    msg = (
        "No HTCondor log file found in timeout period. The job either did "
        "not log any output or the path to the log was not entered correctly."
    )
    raise TimeoutError(msg)


# Timeout handling solution found here:
# https://stackoverflow.com/questions/492519/timeout-on-a-function-call
signal.signal(signal.SIGALRM, timeout_handler)


def fmt_timestamp(timestamp: str) -> str:
    """
    Format timestamp string into ISO 8601 format
    2020-08-27T21:45:09+07:00

    :param timestamp: Unformatted timestamp string
    :return: Formatted timestamp string
    """
    r_yr = r"[0-9]{4}"
    r_nums = r"[0-9]{2}"
    r_timestamp = rf"(?P<yr>{r_yr})-(?P<mon>{r_nums})-(?P<d>{r_nums}) (?P<h>{r_nums}):(?P<m>{r_nums}):(?P<s>{r_nums})"
    match = re.search(r_timestamp, timestamp)

    return "{year}-{mon}-{day}T{h}:{m}:{s}+07:00".format(
        year=match.group("yr"),
        mon=match.group("mon"),
        day=match.group("d"),
        h=match.group("h"),
        m=match.group("m"),
        s=match.group("s"),
    )


def get_retval(body: str) -> Union[int, None]:
    """
    Parse return value from job terminated event

    :param body: Body of event log entry
    :return: Return value as an int
    """
    r_retval = r"return value (?P<retval>[0-9]+)"

    retval = re.search(r_retval, body)
    if retval:
        return int(retval.group("retval"))
    return None


def get_chirp_type(chirp_msg: str) -> str:
    """
    Parse chirp message type from event

    :param chirp_msg: Body of event log entry
    :return: Chirp message type (str)
    """
    return SCG_MSG if (SCG_MSG in chirp_msg) else "stage_update"


def is_final_stage(body: str) -> bool:
    """
    Parse current DAG stage from job executing event

    :param body: Body of event log entry
    :return: boolean for final stage
    """

    dagnode = re.search(r"DAG Node: (?P<dagnode>[A-Z]+)", body)
    if dagnode:
        logger.info(f"current DAG stage: {dagnode.group('dagnode')}")
        if DAGFinalStage.is_final_stage(dagnode.group("dagnode")):
            return True
    return False


def get_termination_types() -> List[str]:
    """
    Return list of termination types
    :return:
    """
    return [
        WorkflowStatusMessages.TERMINATED.value,
        WorkflowStatusMessages.COMPLETE.value,
        WorkflowStatusMessages.FAILED.value,
        WorkflowStatusMessages.ABORTED.value,
    ]


class WorkflowMonitor:
    """
    Class that monitors the events in a given workflow via the log file produced by it
    """

    def __init__(self, logfile_path: str, workflow_request_id: int, is_dag: bool, timeout: int = 60):
        self.logfile_path = Path(logfile_path)
        self.workflow_request_id = workflow_request_id
        self.is_dag = is_dag
        self.message_router = MessageSender("workflow")
        # Leaving the list of events in for now; used for testing; might remove later
        self.events = []
        # Leaving the log in for now; used for testing; might remove later
        self.log = ""

        self.determine_monitor_type(timeout)

        logger.info(f"Finished monitoring {self.logfile_path}.")

    def determine_monitor_type(self, timeout: int):
        """
        DAGs must be monitored differently from standard condor jobs.

        :param timeout: timeout for finding a usable log
        :return:
        """
        if not self.is_dag:
            self.read_htcondor_log(self.logfile_path, timeout)
        else:
            base_path = self.logfile_path.parent
            condor_log_path = base_path / "condor.log"

            dag_proc = Process(
                target=self.read_dag_log,
                args=(
                    self.logfile_path,
                    timeout,
                ),
            )
            condor_proc = Process(
                target=self.read_htcondor_log,
                args=(
                    condor_log_path,
                    timeout,
                ),
            )

            dag_proc.start()
            condor_proc.start()

            condor_proc.join()
            dag_proc.join()

    @log_decorator_factory("Looking for HTCondor log file...")
    def log_exists(self, log_path: Path, timeout: int) -> bool:
        """
        Checks file location repeatedly and returns true if log file is found.

        :param log_path: path to log file of interest
        :param timeout: Amount of time in seconds to wait for the log to appear.
        :raises TimeoutError: If log file not found within timeout seconds
        :return: True if log file found; else raises TimeoutError
        """
        time_remaining = timeout
        time_between_attempts = 3  # seconds
        signal.alarm(timeout)
        while not log_path.is_file():
            logger.info(f"Attempting to read log file {log_path}... {time_remaining} seconds left...")
            time.sleep(time_between_attempts)
            time_remaining -= time_between_attempts
        # Log file found, cancel alarm
        signal.alarm(0)
        return True

    @log_decorator_factory("Reading HTCondor log...")
    def read_htcondor_log(self, log_path: Path, timeout: int):
        """
        Reads HTCondor log as it's being written. When a full event is read in,
        the event is parsed and sent to the workflow events AMQP exchange

        :param log_path: path to log file of interest (might be different from provided for DAGs)
        :param timeout: Time in seconds to wait for the job to write to the log
        """
        logger.info(f"Attempting to read {log_path}")
        exists = False
        is_final_node = False
        job_count = 0

        try:
            exists = self.log_exists(log_path, timeout)
        except TimeoutError as err:
            # log_exists() timed out
            logger.error(err)
            quit(-1)

        if exists:
            logger.info(f"Log file {log_path} found.")
            event_str = ""
            with open(log_path, "r") as infile:
                # Read while file is being written to
                while True:
                    where = infile.tell()
                    line = infile.readline()
                    if not line:
                        time.sleep(1)
                        infile.seek(where)
                    else:
                        if line == "...\n":
                            # Line is at the end of an event
                            event_str += line
                            # Parse and send event
                            try:
                                event = self.parse_message(event_str)
                            except ValueError as err:
                                # Event not well-formed
                                logger.error(err)
                                continue
                            self.events.append(event["condor_metadata"])

                            if event["type"] in get_termination_types():
                                # ignore terminations from condor.log file when run a DAG
                                if self.is_dag:
                                    # unless it's the last termination, then just exit
                                    if is_final_node:
                                        logger.info("Final DAG stage. Exiting condor.log...")
                                        break

                                    logger.info(
                                        f"event type: {event['type']}, Running a DAG, stage over. continuing..."
                                    )
                                    # Empty contents to prepare for next event
                                    self.log += event_str
                                    event_str = ""
                                    continue
                                else:
                                    # if not a DAG, send message and terminate when found
                                    send_message(event, self.message_router)
                                    break
                            else:
                                # determine current DAG stage from event, if final, set exit condition for DAG
                                if self.is_dag and event["type"] == WorkflowStatusMessages.EXECUTING.value:
                                    job_count += 1
                                    # DAG is still running, change message to 'continue'
                                    event["type"] = WorkflowStatusMessages.CONTINUE.value
                                    # set is final node flag if necessary
                                    is_final_node = is_final_stage(event["condor_metadata"]["log"])

                                send_message(event, self.message_router)
                                # Empty contents to prepare for next event
                                self.log += event_str
                                event_str = ""
                        elif line == "\n":
                            # Line empty
                            time.sleep(1)
                        else:
                            # Line is anywhere else
                            event_str += line

    @log_decorator_factory("Reading HTCondor DAG log...")
    def read_dag_log(self, log_path: Path, timeout: int):
        """
        Reads HTCondor DAG log as it's being written. When a full event is read in,
        the event is parsed and sent to the workflow events AMQP exchange

        :param log_path: path to log file of interest
        :param timeout: Time in seconds to wait for the job to write to the log
        """
        logger.info(f"Attempting to read {log_path}")
        exists = False
        try:
            exists = self.log_exists(log_path, timeout)
        except TimeoutError as err:
            # log_exists() timed out
            logger.error(err)
            quit(-1)

        if exists:
            logger.info(f"Log file {log_path} found.")
            event_str = ""
            with open(log_path, "r") as infile:
                # Read while file is being written to
                while True:
                    where = infile.tell()
                    line = infile.readline()
                    if not line:
                        time.sleep(1)
                        infile.seek(where)
                    else:
                        if line == "...\n":
                            # Line is at the end of an event
                            event_str += line
                            # Parse and send event
                            try:
                                event = self.parse_message(event_str)
                            except ValueError as err:
                                # Event not well-formed
                                logger.error(err)
                                continue
                            self.events.append(event["condor_metadata"])
                            send_message(event, self.message_router)
                            if event["type"] in get_termination_types():
                                break
                            # Empty contents to prepare for next event
                            self.log += event_str
                            event_str = ""
                        elif line == "\n":
                            # Line empty
                            time.sleep(1)
                        else:
                            # Line is anywhere else
                            event_str += line

    @log_decorator_factory("Parsing HTCondor log...")
    def parse_message(self, event_str: str) -> Dict[str, str]:
        """
        Parse event for relevant details:
        - HTCondor event number
        - HTCondor job number
        - Timestamp of event
        - Description of event
        - Body with additional information

        :param event_str: String representation of an HTCondor event
        :return: Message in dictionary form that can be sent over AMQP
        """
        # Regex that matches an event entry in the HTCondor log file.
        r_condor_event = (
            r"(?P<eventnum>[0-9]{3}) \((?P<jobnum>[0-9]+)\.[0-9]+\.[0-9]+\) "
            r"(?P<timestamp>[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}) "
            r"(?P<desc>[a-zA-Z0-9 @?&#%:=<>\"\"\(\)\/\-.,!_]+)\n"
            r"(?P<body>[\t ]+(?:(?!\.\.\.)[\s\S])*)*"
            r"(?P<terminator>\.\.\.\n)"
        )
        match = re.match(r_condor_event, event_str)
        if match:
            # Event is well-formed
            message = self._event_to_message(match)

            return message

        # Regex didn't return any matches (event not well-formed)
        raise ValueError(f"HTCondor event not well-formatted. Event contents:\n{event_str}")

    def _event_to_message(self, event_match: re.Match) -> Dict[str, str]:
        """
        Translate a well-formed HTCondor event into a message in dict format
        that can be sent over AMQP

        :param event_match: Regex match object for well-formed event
        :return: Message in dictionary form that can be sent over AMQP
        """
        job_id = event_match.group("jobnum")

        event_type = self._parse_event_type(int(event_match.group("eventnum")))

        if event_type == "workflow-updated":
            chirp_type = get_chirp_type(event_match.group("desc"))
            if chirp_type == SCG_MSG:
                event_type = "workflow-failed"

        if event_type == "workflow-terminated":
            retval = get_retval(event_match.group("body"))
            event_type = self._get_job_result(retval)
        else:
            retval = None

        return {
            "service": "workflow",
            "subject": {
                "type": "WorkflowRequest",
                "workflow_request_id": self.workflow_request_id,
            },
            "routing_key": "workflow",
            "type": event_type,
            "condor_metadata": {
                "event_type": event_type,
                "condor_job_id": job_id,
                "retval": retval,
                "timestamp": fmt_timestamp(event_match.group("timestamp")),
                "log": event_match.group(0),
            },
        }

    @staticmethod
    def _parse_event_type(condor_event_num: int) -> str:
        """
        Translate HTCondor event number into a readable str that can be sent in
        an AMQP-friendly message

        :param condor_event_num: Event number that HTCondor spits out to its log
        :return: String representation of the meaning of the given HTCondor
            event number (if supported); else "unknown"
        """

        # event_meanings = {
        #     0: "workflow-executing",
        #     1: "workflow-continuing",
        #     2: "workflow-failed",
        #     3: "workflow-continuing",
        #     4: "workflow-continuing",  # "workflow-paused",
        #     5: "workflow-terminated",
        #     6: "workflow-continuing",
        #     7: "workflow-continuing",  # "workflow-paused",
        #     8: "unknown-unused",
        #     9: "workflow-aborted",
        #     10: "workflow-continuing",  # "workflow-paused",
        #     11: "workflow-continuing",
        #     12: "workflow-failed",
        #     13: "workflow-continuing",
        #     14: "workflow-continuing",
        #     15: "workflow-continuing",
        #     16: "workflow-continuing",
        #     # skipped values 17-20
        #     21: "workflow-failed",
        #     22: "workflow-continuing",  # "communications-paused",
        #     23: "workflow-continuing",
        #     24: "workflow-failed",
        #     25: "workflow-continuing",
        #     26: "workflow-continuing",
        #     27: "workflow-continuing",
        #     28: "workflow-continuing",  # "workflow-configuration-updated",
        #     29: "unknown-contact-lost",
        #     30: "workflow-continuing",
        #     31: "workflow-continuing",
        #     32: "workflow-continuing",
        #     33: "workflow-continuing",  # "workflow-configuration-updated",
        #     34: "workflow-continuing",
        #     35: "workflow-executing",
        #     36: "workflow-terminated",
        #     37: "workflow-continuing",  # "workflow-paused",
        #     38: "workflow-continuing",
        #     39: "workflow-continuing",  # "communications-error",
        #     40: "workflow-continuing",
        # }
        #
        # return event_meanings.get(condor_event_num, "unknown")

        condor_code = HtcEventCodes(condor_event_num)

        if condor_code.is_initialization_message:
            workflow_message = WorkflowStatusMessages.EXECUTING
        elif condor_code.is_error_message:
            workflow_message = WorkflowStatusMessages.FAILED
        elif condor_code.is_resubmitting_message:
            workflow_message = WorkflowStatusMessages.RESTARTED
        elif condor_code.is_completion_message:
            workflow_message = WorkflowStatusMessages.TERMINATED
        elif condor_code.is_aborted_message:
            workflow_message = WorkflowStatusMessages.ABORTED
        elif condor_code.is_pause_message:
            workflow_message = WorkflowStatusMessages.PAUSED
        elif condor_code.is_unpause_message or condor_code.is_informational_message:
            workflow_message = WorkflowStatusMessages.CONTINUE
        elif condor_code.is_chirp_message:
            workflow_message = WorkflowStatusMessages.UPDATED
        else:
            workflow_message = WorkflowStatusMessages.UNEXPECTED

        return workflow_message.value

    @staticmethod
    def _get_job_result(return_value: int) -> str:
        """
        Determine the result of a terminated job

        :param return_value: Return value of the terminated job
        :return: String representation of the termination result
        """
        return WorkflowStatusMessages.COMPLETE.value if return_value == 0 else WorkflowStatusMessages.FAILED.value

    def close(self):
        logger.info("Waiting 3 seconds for final messages to clear....")
        time.sleep(3)
        self.message_router.close()
        logger.info(f"Monitor closed for {self.workflow_request_id}")

    def __str__(self):
        return f"{self.__class__.__name__}, monitoring {self.logfile_path} that has events {self.events}"

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.logfile_path}, {self.events})"


_DESCRIPTION = "Workspaces workflow monitor, version {}. Monitor execution of a workflow from the command line."


def send_message(message: Dict[str, str], messenger: MessageSender):
    """
    Send message with HTCondor event details to AMQP exchange

    :param messenger:
    :param message: Message with event details
    """
    decorated_send = log_decorator_factory("Sending event...")(messenger.send_message)
    logger.debug(f"wf_monitor > Sending message: {message}")
    decorated_send(**message)


def make_arg_parser() -> argparse.ArgumentParser:
    """
    Initialize argument parser for command-line arguments

    :return: Argument parser
    """
    parser = argparse.ArgumentParser(
        description=_DESCRIPTION.format(version),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-P",
        "--profile",
        action="store",
        required=False,
        default="dsoc-dev",
        help="profile name to use; e.g. test, production; DEFAULT: dsoc-dev",
    )
    parser.add_argument(
        "log_path",
        action="store",
        help="path to the workflow's HTCondor log",
    )
    parser.add_argument(
        "workflow_request_id",
        action="store",
        help="ID of the workflow request that ran the workflow",
    )
    parser.add_argument(
        "is_dag",
        action="store",
        help="is the workflow to be monitored a DAG workflow?",
    )

    return parser


def main():
    """
    Parse command-line args and launch the monitor

    :return:
    """

    args = make_arg_parser().parse_args()
    wf_monitor = WorkflowMonitor(args.log_path.rstrip(), int(args.workflow_request_id), eval(args.is_dag))
    logger.info(f"Closing wf_monitor for workflow request #{args.workflow_request_id}")
    wf_monitor.close()
