/*
 * Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 */
import { Component, OnInit, Input, Output, EventEmitter } from '@angular/core';
import {FormArray, FormGroup, FormControl, FormBuilder} from "@angular/forms";
import { ValidateDate } from 'src/app/shared/date-validator';

export interface Filter {
  data: Array<string>
  filter: string
  event: any
}

@Component({
  selector: 'app-filter-menu',
  templateUrl: './filter-menu.component.html',
  styleUrls: ['./filter-menu.component.scss']
})

export class FilterMenuComponent implements OnInit {
  @Input() state: any;
  @Input() exec_status: any;
  @Input() download_status: any;
  @Input() stage1QaStaff: any;
  @Input() stage2QaStaff: any;
  @Input() srdpStatus: any;
  @Input() filters: any;

  @Output() filterMenuEventEmitter = new EventEmitter<Filter>();

  form: FormGroup;
  public dateGroup: FormGroup;

  constructor() {}

  ngOnInit(): void {

    // maxdate and mindate are the real fields that get submitted as filter
    // maxDate and minDate are just there for the UI controls
    // they are different because angular bootstraps deals with the control values as objects, not strings
    // the two sets of fields are to switch from objects to strings and back
    // we will filter out the fake ones on form submit
    this.dateGroup = new FormGroup({
      maxDate: new FormControl('', { validators: ValidateDate, updateOn: "blur"}),
      minDate: new FormControl('', { validators: ValidateDate, updateOn: "blur"}),
      maxdate: new FormControl(''),
      mindate: new FormControl(''),
    });

    if (this.filters['obs_min']) {
      this.dateGroup.patchValue({
        'minDate': this.createNgDate(this.filters['obs_min']),
        'mindate': this.createNgDate(this.filters['obs_min'])
      })
    }
    if (this.filters['obs_max']) {
      this.dateGroup.patchValue({
        'maxDate': this.createNgDate(this.filters['obs_max']),
        'maxdate': this.createNgDate(this.filters['obs_max'])
      })
    }

    this.onChanges();
  }

  onChanges(): void {
    this.dateGroup.get('maxDate').valueChanges.subscribe(val => {
      if(this.dateGroup.get('maxDate').valid && val !== null) {
        this.dateGroup.get('maxdate').setValue(val.year + '-' + (val.month < 10 ? '0' : '') + val.month + '-' + (val.day < 10 ? '0' : '') + val.day);
        this.addFilter(this.dateGroup.get('maxdate').value, "obs_max", true)
      } else if (val === null) {
        this.addFilter(this.dateGroup.get('maxdate').value, "obs_max", false)
      }
    });
    this.dateGroup.get('minDate').valueChanges.subscribe(val => {
      if(this.dateGroup.get('minDate').valid && val !== null) {
        this.dateGroup.get('mindate').setValue(val.year + '-' + (val.month < 10 ? '0' : '') + val.month + '-' + (val.day < 10 ? '0' : '') + val.day);
        this.addFilter(this.dateGroup.get('mindate').value, "obs_min", true)
      } else if (val === null) {
        this.addFilter(this.dateGroup.get('mindate').value, "obs_min", false)
      }
    });
  }

  public addFilter(d: string, s: string, e?: any) {
    let filter: Array<string> = []
    filter.push(d)
    this.filterMenuEventEmitter.emit({'data': filter, 'filter': s,'event': e})
  }

  createNgDate(dateString: string): any {
    let dateArrary = dateString.split('-')
    return {
      'year': parseInt(dateArrary[0]),
      'month': parseInt(dateArrary[1]),
      'day': parseInt(dateArrary[2])
    }
  }

}
