#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

import time

import sqlalchemy.orm
import transaction
import zope.sqlalchemy
from prometheus_client import Summary
from prometheus_client.core import REGISTRY, CounterMetricFamily
from pyramid.config import Configurator
from pyramid.events import NewRequest
from pyramid.renderers import JSONP
from pyramid.request import Request
from pyramid.response import Response
from pyramid_beaker import BeakerSessionFactoryConfig, session_factory_from_settings

from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.capability_service import (
    CapabilityCanceller,
    CapabilityLauncher,
    CapabilityQueueStarter,
    CapabilityService,
)
from workspaces.capability.services.state_machine_info import StateMachineInfo
from workspaces.system.schema import get_engine, get_session_factory
from workspaces.system.services.archive_service import ArchiveService
from workspaces.system.services.casa_matrix_service import CasaMatrixService
from workspaces.workflow.services.workflow_service import WorkflowServiceRESTClient

from .utility.notificationsender import TransactionalNotificationSender


# Copied from here: https://stackoverflow.com/questions/21107057/pyramid-cors-for-ajax-requests
def add_cors_headers_response_callback(event: NewRequest):
    """
    Event handler that adds CORS HTTP headers to responses from this server; allows external servers
    (the front end) to send requests and not have them bounce off

    More information about CORS: https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS

    :param event: Server event
    """

    def cors_headers(request: Request, response: Response):
        """
        Callback function that adds CORS headers to a response

        :param response: Server response
        """
        response.headers.update(
            {
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Methods": "POST,GET,DELETE,PUT,OPTIONS",
                "Access-Control-Allow-Headers": "Origin, Content-Type, Accept, Authorization",
                "Access-Control-Allow-Credentials": "true",
                "Access-Control-Max-Age": "1728000",
                "Content-Type": "application/json",
            }
        )

    event.request.add_response_callback(cors_headers)


# ---------------------------------------------------------
#
#    S E R V I C E S
#
# ---------------------------------------------------------


def add_services(config: Configurator, session_factory: BeakerSessionFactoryConfig) -> Configurator:
    """
    Add capability info, capability service, and workflow service to Pyramid request configuration

    :param config: Pyramid Configurator object
    :param session_factory: Pyramid Beaker session factory
    :return: Updated Configurator
    """

    # # these objects should exist across request boundaries
    workflow_rest_client = WorkflowServiceRESTClient()
    notification_rest_client = TransactionalNotificationSender(transaction.manager)
    capability_info = CapabilityInfo(get_tm_session(session_factory, transaction.manager))
    state_machine_info = StateMachineInfo(get_tm_session(session_factory, transaction.manager))
    capability_service = CapabilityService(
        capability_info,
        state_machine_info,
        workflow_rest_client,
        notification_rest_client,
    )
    archive_service = ArchiveService(session_factory)
    casa_matrix_service = CasaMatrixService(get_tm_session(session_factory, transaction.manager))

    # hook up the queue report collector for Prometheus
    REGISTRY.register(QueueReportCollector(capability_info))

    def create_capability_info(request):
        return CapabilityInfo(get_tm_session(session_factory, transaction.manager))

    # it turns out to be necessary to share the message router here, or we have trouble with deliveries
    # being round-robined to request-based routers that have no listeners
    def create_capability_launcher(request):
        return CapabilityLauncher(request.capability_info, capability_service.messenger)

    def create_queue_restarter(request):
        return CapabilityQueueStarter(capability_service.messenger)

    def create_capability_canceller(request):
        return CapabilityCanceller(capability_service.messenger)

    # make capability_info available for use in Pyramid
    config.add_request_method(
        # r.tm is the transaction manager used by pyramid_tm
        create_capability_info,
        "capability_info",
        reify=True,
    )

    # make capability launcher available for use in Pyramid
    config.add_request_method(create_capability_launcher, "capability_launcher", reify=True)

    # make capability queue starter available for use in Pyramid
    config.add_request_method(create_queue_restarter, "capability_queue_restarter", reify=True)

    # make capability canceller available for use in Pyramid
    config.add_request_method(create_capability_canceller, "capability_canceller", reify=True)

    # make workflow_info available for use in Pyramid
    config.add_request_method(
        # r.tm is the transaction manager used by pyramid_tm
        lambda request: workflow_rest_client,
        "workflow_service",
        reify=True,
    )

    config.add_request_method(lambda request: notification_rest_client, "notification_service", reify=True)

    # make archive_service available for use in Pyramid
    config.add_request_method(lambda request: archive_service, "archive_service", reify=True)

    # make casa_matrix_service available for use in Pyramid
    config.add_request_method(lambda request: casa_matrix_service, "casa_matrix_service", reify=True)

    return config


def lookup_request(request):
    return request.capability_info.lookup_capability_request(request.matchdict["request_id"])


# ---------------------------------------------------------
#
#    F R A M E W O R K    S E T U P
#
# ---------------------------------------------------------


def get_tm_session(
    session_factory: BeakerSessionFactoryConfig,
    transaction_manager: transaction.TransactionManager,
) -> sqlalchemy.orm.Session:
    """
    Enable Zope's transaction manager on our session
    :param session_factory:
    :param transaction_manager:
    :return:
    """
    dbsession = session_factory()
    zope.sqlalchemy.register(dbsession, transaction_manager=transaction_manager)
    return dbsession


# ---------------------------------------------------------
#
#    P R O M E T H E U S   S E T U P
#
# ---------------------------------------------------------

# Prometheus logging
http_requests = Summary("http_request_timing", "HTTP Requests", ["status_code", "route_name"])


def prometheus_route_timing_factory(handler, registry):
    # if timing support is enabled, return a wrapper
    def prometheus_route_timer(request):
        start = time.time()
        response = handler(request)
        end = time.time()
        if request.matched_route:
            http_requests.labels(status_code=response.status_code, route_name=request.matched_route.name).observe(
                end - start
            )
        return response

    return prometheus_route_timer


# Let's export the queue report too, just for fun
class QueueReportCollector:
    """
    Collects the capability queue report and exposes it via Prometheus
    """

    def __init__(self, capability_info: CapabilityInfo):
        self.capability_info = capability_info

    def collect(self):
        c = CounterMetricFamily(
            "capability_queue",
            "Queue status for each capability",
            labels=["capability", "status"],
        )

        for report in self.capability_info.report():
            c.add_metric([report.capability, "waiting"], report.waiting)
            c.add_metric([report.capability, "executing"], report.executing)
        yield c


# ---------------------------------------------------------
#
#    M A I N    E N T R Y P O I N T
#
# ---------------------------------------------------------


def main(global_config, **settings):
    with Configurator(settings=settings) as config:
        # Helpers
        config.add_subscriber(add_cors_headers_response_callback, NewRequest)

        session_factory = session_factory_from_settings(settings)
        config.set_session_factory(session_factory)
        config.add_renderer("jsonp", JSONP(param_name="callback"))

        settings["tm.manager_hook"] = "pyramid_tm.explicit_manager"

        # use pyramid_tm to hook the transaction lifecycle to the request
        config.include("pyramid_tm")

        # use pyramid_retry to retry a request when transient exceptions occur
        config.include("pyramid_retry")

        session_factory = get_session_factory(get_engine())
        config.registry["dbsession_factory"] = session_factory

        # Add services to config
        config = add_services(config, session_factory)

        config.include("pyramid_beaker")
        # Include routes from routes file
        config.include(".routes")
        # config.scan(".views")
        config.scan(".")

        # add the prometheus timing metric
        config.add_tween("capability.server.prometheus_route_timing_factory")
        return config.make_wsgi_app()
