#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
File containing definitions for the capability version routes of the Workspaces REST API
"""

import copy
import http
import json
import logging
import subprocess

from pycapo import CapoConfig
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound, HTTPPreconditionFailed
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config

from workspaces.capability.schema import CapabilityVersion

logger = logging.getLogger(__name__)

NOTIF_SETTINGS_KEY = "edu.nrao.workspaces.NotificationSettings"

from workspaces.capability.enums import CapabilityVersionState


@view_config(route_name="view_all_versions", renderer="json")
def view_all_versions(request: Request):
    """
    Pyramid view that accepts a request to view all versions of a capability request
    URL: capability/request/{capability_request_id}/versions

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["capability_request_id"])
    if capability_request:
        if capability_request.versions:
            return capability_request.versions
        else:
            no_versions_msg = (
                f"Capability request with ID {request.matchdict['capability_request_id']} has no versions."
            )
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['capability_request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_latest_version", renderer="json")
def view_latest_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the latest version of a capability request
    URL: capability/request/{capability_request_id}/version/latest

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["capability_request_id"])
    if capability_request:
        if capability_request.current_version:
            version_json = capability_request.current_version.__json__()
            version_json["files"] = []

            for file in capability_request.current_version.files:
                version_json["files"].append(file.__json__())

            return Response(json_body=version_json)
        else:
            no_versions_msg = (
                f"Capability request with ID {request.matchdict['capability_request_id']} has no versions."
            )
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['capability_request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_specific_version", renderer="json")
def view_specific_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if capability_request:
        if version := capability_request.versions[int(version_id) - 1]:
            return version
        else:
            no_versions_msg = f"Capability request with ID {capability_request_id} has no version {version_id}"
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = f"Capability request with ID {capability_request_id} not found."
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_internal_notes", renderer="json")
def view_internal_notes(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get notes for a specific version
    URL: capability/request/{capability_request_id}/version/{version_id}/notes

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of version notes
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if capability_request:
        if version := capability_request.versions[int(version_id) - 1]:
            return Response(status_int=http.HTTPStatus.OK, json_body={"resp": f"{version.internal_notes}"})
        else:
            no_versions_msg = f"Capability request with ID {capability_request_id} has no version {version_id}"
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = f"Capability request with ID {capability_request_id} not found."
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="update_internal_notes", renderer="json")
def update_internal_notes(request: Request) -> Response:
    """
    Pyramid view that accepts a request to update notes for a specific version
    URL: capability/request/{capability_request_id}/version/{version_id}/notes

    :param request: POST request
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability request with given id does not exist
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]
    params = request.json_body
    # sanitize input before writing/persisting
    # \\u0000 is an invalid character that is incompatible with postgres json columns
    # from StackOverflow: https://stackoverflow.com/questions/63092267/how-to-handle-api-responsesjson-containing-x00-or-u0000-in-its-data-and-s
    new_notes = json.loads(json.dumps(params["edits"])).replace("\\u0000", "")

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if capability_request:
        if version := capability_request.versions[int(version_id) - 1]:
            version.internal_notes = new_notes
            return Response(
                body=f"Successfully updated internal_notes for capability request #{capability_request_id} v{version_id}."
            )
        else:
            no_versions_msg = f"Capability request with ID {capability_request_id} has no version {version_id}"
            # Need to raise on failure when a transaction is in-progress so that it rolls back instead of committing
            raise HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = f"Capability request with ID {capability_request_id} not found."
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="create_capability_version", renderer="json")
def create_capability_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to create a capability version
    URL: capability/request/{capability_request_id}/version/create

    :param request: POST request, expecting JSON parameter "parameters"
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability request with given id does not exist
    """
    expected_params = ["parameters"]
    capability_request_id = request.matchdict["capability_request_id"]
    params = request.json_body

    if not all([expected in params for expected in expected_params]):
        # JSON params do not contain all expected params
        params_not_given_msg = f"Expected JSON parameters {expected_params}. Received only {params}."
        return HTTPBadRequest(detail=params_not_given_msg)
    elif not request.capability_info.lookup_capability_request(capability_request_id):
        # CapabilityRequest with given id does not exist; can't create a new version for it
        does_not_exist_msg = (f"Capability request #{capability_request_id} does not exist. Cannot create new version",)
        return HTTPPreconditionFailed(detail=does_not_exist_msg)
    else:
        new_capability_version = request.capability_info.create_capability_version(
            capability_request_id, parameters=params["parameters"]
        )
        return Response(json_body=new_capability_version.__json__())


@view_config(request_method="PUT", route_name="update_parameters")
def update_parameters(request: Request) -> Response:
    """
    Note: this is only currently set up to handle top-level (not nested) parameters.

    :param request: a PUT request whose body contains a dict under the key "parameters",
                    among other things. This should contain the parameters and values
                    to be set for the latest version of the capability request whose
                    value is specified in the URL.
    :return: 200 OK response if the parameters were all able to be added,
             or a 400 Request if the user supplied a bad parameter and/or value.

    """
    json_body = request.json_body

    # Dictionary in the form of "param": [acceptable values]
    approved_params_vals = {"is_partial_pass": [True, False]}

    not_added_params = dict()

    params_to_set = json_body["parameters"]
    logger.info(f"In capability views' capability_version.py; received these parameters: {params_to_set}")

    capability_request_id = request.matchdict["capability_request_id"]
    capability_version_number = request.matchdict["version_id"]

    version: CapabilityVersion = request.capability_info.lookup_version(
        capability_request_id, capability_version_number
    )
    updated_version_params = copy.deepcopy(version.parameters)
    logger.info(f"Before adding, {updated_version_params=}")

    for param, param_val in params_to_set.items():
        if param in approved_params_vals and param_val in approved_params_vals[param]:
            # Update parameter
            updated_version_params[param] = param_val
            logger.info(f"Added {param}: {param_val}")
        else:
            not_added_params[param] = param_val
            logger.info(f"Could not updated (not-allowed parameter {param} with value {param_val}")

    # Set the updated parameters
    version.parameters = updated_version_params

    logger.info(f"Updated params: {updated_version_params}\n\nAbout to save:")

    # Persist version (with updated parameters) to the database
    request.capability_info.save_version(version)

    # Return the response
    return Response(status_code=200 if len(not_added_params) == 0 else 400, json_body=not_added_params)


@view_config(route_name="submit_capability_version", renderer="json")
def submit_capability_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to submit a capability version for execution
    URL: capability/request/{capability_request_id}/version/{version_id}/submit

    :param request: POST request, expecting JSON parameter "parameters"
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability request with given id does not exist
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_number = request.matchdict["version_number"]

    version = request.capability_info.lookup_version(capability_request_id, version_number)

    if not version_number:
        # Capability request not found
        does_not_exist_msg = (
            f"Capability request #{capability_request_id} v{version_number} does not exist; cannot submit it."
        )
        return HTTPPreconditionFailed(detail=does_not_exist_msg)

    execution = request.capability_launcher.launch_capability(version)
    return Response(json_body=execution.__json__())


@view_config(route_name="cancel_capability_version", renderer="json")
def cancel_capability_version(request: Request) -> Response:
    """
    Accepts a capability request id and version id and cancels it
    URL: capability/request/{capability_request_id}/version/{version_id}/cancel

    :return: 200 OK response
        or 412 response (HTTPPreconditionFailed) if capability execution with the given ids does not exist
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_number = request.matchdict["version_number"]

    version = request.capability_info.lookup_version(capability_request_id, version_number)
    execution = request.capability_info.lookup_execution_by_request_version(capability_request_id, version_number)
    if execution:
        request.capability_canceller.cancel_capability_execution(execution)
        return Response(body=f"Capability request #{capability_request_id} v{version_number} successfully cancelled.")
    elif version:
        if version.state == "Created":
            cap_request = request.capability_info.lookup_capability_request(version.capability_request_id)
            # Set version to Cancelled state
            version.state = CapabilityVersionState.Cancelled.name

            # Seal both version and requests
            version.sealed = True
            cap_request.sealed = True

            # Save
            request.capability_info.save_version(version)
            request.capability_info.save_entity(cap_request)
            return Response(
                body=f"Capability request #{capability_request_id} v{version_number} successfully cancelled."
            )
    else:
        does_not_exist_msg = (
            f"Capability request #{capability_request_id} v{version_number} has not been executed; cannot cancel it."
        )
        return HTTPPreconditionFailed(detail=does_not_exist_msg)


@view_config(route_name="update_file_for_latest", renderer="json")
def add_or_update_file(request: Request) -> Response:
    """
    Pyramid view that accepts a request to edit an existing file to the latest version of a capability request
    URL: capability/request/{capability_request_id}/version/latest/files/{filename}

    :param request: PUT request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["capability_request_id"])

    if capability_request:
        if capability_request.current_version:
            # Preventing empty request bodies that were clearing out existing file content
            if request.body.strip():
                versionFile = request.capability_info.lookup_version_file(
                    request.matchdict["capability_request_id"],
                    capability_request.current_version.version_number,
                    request.matchdict["filename"],
                )
                if versionFile:
                    versionFile.content = request.body
                    file = request.capability_info.save_edited_version_file(versionFile)
                    return file
                else:
                    # Add the file if it doesn't exist:
                    return request.capability_info.save_version_file(
                        capability_request.current_version, request.matchdict["filename"], request.body
                    )
            else:
                no_content_msg = (
                    f"No content to save in file {request.matchdict['filename']} "
                    f"for capability request with ID {request.matchdict['capability_request_id']}."
                )
            return HTTPPreconditionFailed(no_content_msg)
        else:
            no_versions_msg = (
                f"Capability request with ID {request.matchdict['capability_request_id']} has no versions."
            )
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['capability_request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_version_metadata", renderer="json")
def view_version_metadata(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the workflow metadata of a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}/workflow_metadata

    :param request: GET request
    :return: 200 OK response with JSON metadata of specified version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if version := request.capability_info.lookup_version(capability_request_id, version_id):
        return version.workflow_metadata
    else:
        no_versions_msg = f"Capability request {capability_request_id} version {version_id} not found."
        return HTTPNotFound(detail=no_versions_msg)


@view_config(route_name="edit_version_metadata", renderer="json")
def edit_version_metadata(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the workflow metadata of a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}/workflow_metadata

    :param request: GET request
    :return: 200 OK response with JSON metadata of specified version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    if version := request.capability_info.lookup_version(capability_request_id, version_id):
        # sqlalchemy does not detect in-place mutations of JSON
        # https://docs.sqlalchemy.org/en/14/core/type_basics.html#sqlalchemy.types.JSON
        # to get past this we use a deepcopy of version.workflow_metadata to create a "new" JSON obj
        version.workflow_metadata = copy.deepcopy(version.workflow_metadata)
        if version.workflow_metadata is None:
            version.workflow_metadata = {}
        for key in request.json_body:
            version.workflow_metadata[key] = request.json_body[key]
        request.capability_info.save_version(version)
        return request.json_body
    else:
        no_versions_msg = f"Capability request {capability_request_id} version {version_id} not found."
        return HTTPNotFound(detail=no_versions_msg)


@view_config(route_name="get_user_email", renderer="json")
def get_user_email(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the email contact information for a capability request in order to
    prevent issues with empty user_email parameters causing QA controls to fail in the UI
    URL: capability/request/{capability_request_id}/version/{version_id}/user_email

    :param request: GET request
    :return: 200 OK response with JSON-formatted user_email field
        or 404 response (HTTPNotFound) if the email information couldn't be found
    """
    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    # We only send emails to the PI/CoI in production!
    email_pi = CapoConfig().getboolean("edu.nrao.workspaces.NotificationSettings.sendPIEmail")
    if not email_pi:
        user_email = CapoConfig().settings(NOTIF_SETTINGS_KEY).analystEmail
    else:
        if version := request.capability_info.lookup_version(capability_request_id, version_id):
            if not version.parameters:
                not_found_msg = f"No parameters found for version {version_id} of request with ID {capability_request_id}. Couldn't find user_email information."
                return HTTPNotFound(detail=not_found_msg)
            elif (
                "user_email" in version.parameters
                and version.parameters["user_email"]
                and "workspaces-analysts" not in version.parameters["user_email"]
            ):
                # exclude the case of follow-on requests when the default user_email is the analyst list,
                # otherwise send to specified address
                return Response(
                    status_int=http.HTTPStatus.OK, json_body={"resp": f"{version.parameters['user_email']}"}
                )
            # If user_email not populated (or is an analyst list), search for the PI/CoI emails with the contacts_wrester
            elif "product_locator" in version.parameters and version.parameters["product_locator"]:
                process_call = ["aat_wrest", "-sc", version.parameters["product_locator"]]
            elif "sdmId" in version.parameters and version.parameters["sdmId"]:
                process_call = ["aat_wrest", "-si", version.parameters["sdmId"]]
            elif (
                "metadata" in version.parameters
                and version.parameters.metadata
                and "sdm_id" in version.parameters.metadata
                and version.parameters.metadata["sdm_id"]
            ):
                process_call = ["aat_wrest", "-si", version.parameters["sdmId"]]
            else:
                not_found_msg = f"No user email, product locator, or sdm ID found in version {version_id} of request with ID {capability_request_id}. Cannot find user_email information."
                return HTTPNotFound(detail=not_found_msg)
        else:
            not_found_msg = f"Capability request with ID {capability_request_id} not found."
            return HTTPNotFound(detail=not_found_msg)

        with subprocess.Popen(
            process_call,
            stdout=subprocess.PIPE,
        ) as project_data:
            data = project_data.communicate()[0]
            project = json.loads(data)
            project_code = project["projectCode"]

            with subprocess.Popen(
                ["contacts_wrest", project_code],
                stdout=subprocess.PIPE,
            ) as wrested_contacts:
                contacts = wrested_contacts.communicate()[0]
                if "ERROR" in contacts.decode():
                    err_msg = f"Contact wrester failed for project {project_code}"
                    return HTTPPreconditionFailed(err_msg)

        email_list = json.loads(contacts.decode())
        user_email = email_list
        if isinstance(email_list, list):
            user_email = ", ".join(email_list)

    logger.info(f"get_user_email endpoint retrieved: {user_email}")

    return Response(status_int=http.HTTPStatus.OK, json_body={"resp": f"{user_email}"})
