"""
    Unit tests for the new 'std_calibration_and_imaging' workflow.
    Run these inside the workflow container.
"""
#
# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

import json
import pytest
import requests
from requests import Response

TIMEOUT = 30
SERVICE_URL = "http://workflow:3456" # From docker.properties
WORKFLOW = "std_calibration_and_imaging"

# Set to False if you're testing this locally, which you should do from inside
# the workflow container. Always make sure this is True when committing it.
TESTING_IN_PIPELINE = True

def mocked_requests(url: str) -> Response:
    """
    :param url: The URL for the mocked request.
    :return: Response. The mocked response for the mocked request.
    """
    response = Response()
    match url:
        case "http://workflow:3456/workflows/std_calibration_and_imaging/requests/create":
            response.status_code = 200
            response._content = (b'{"type": "WorkflowRequest", "workflow_request_id": 1,'
                                 b'"workflow_name": "std_calibration_and_imaging", "argument":'
                                 b'{"metadata": {}, "user_email": "dlopezsa@nrao.edu",'
                                 b'"provided_files": [],'
                                 b'"product_locator": "uid://evla/execblock/'
                                 b'dd42ec9d-1a6c-4441-b25a-fd0dd35d3a82",'
                                 b'"files_to_transfer": "", "need_project_metadata": true},'
                                 b'"state": "Created", "results_dir": "", "files": [],'
                                 b'"created_at": "2024-09-19T16:45:25.970049+00:00",'
                                 b'"updated_at": "2024-09-19T16:45:25.970049+00:00",'
                                 b'"num_to_clean": 0, "num_cleaned": 0, "controller": "WS", '
                                 b'"progress": []}')


        case "http://workflow:3456/workflows/std_calibration_and_imaging/requests/1/files/PPR.xml":
            response.status_code = 200
        case "http://workflow:3456/workflows/std_calibration_and_imaging/requests/1/submit":
            response.status_code = 200
        case _:
            response.status_code = 404
            response._content = (b'{"error": "You have requested a page'
                                b'that has not been set up to be mocked".}')
    return response

# So far, we only need this for the module, so this has a scope of 'module'.
# This ensures the setup/teardown is only done once for this module.
@pytest.fixture(scope='module', name='wf_request')
def create_std_calibration_and_imaging_wf_request() -> Response:
    """
        Creates a standard calibration and imaging workflow request.
        :return: Response. The response from the server.
    """
    url = f"{SERVICE_URL}/workflows/{WORKFLOW}/requests/create"

    if TESTING_IN_PIPELINE is False:
        product_locator = "uid://evla/execblock/dd42ec9d-1a6c-4441-b25a-fd0dd35d3a82"
        wf_request_params = {"metadata": {},
                             "user_email": "workspaces-analysts-test@listmgr.nrao.edu",
                             "provided_files": [], "product_locator": product_locator,
                             "files_to_transfer": "", "need_project_metadata": True}

        # Try to create WF request
        response = requests.post(url, json=wf_request_params, timeout=TIMEOUT)
    else:
        response = mocked_requests(url)

    # yield is used rather than return so we can have set up
    # and teardown (before and after the yield line, respectively)
    yield response

@pytest.fixture(name='wf_request_id', scope='module')
def std_calibration_and_imaging_wf_request_id(wf_request) -> int:
    """
    :param wf_request: The return value (requests.Response)
                       from the wf_request fixture.
    :return: An int. The ID of the newly-created workflow request.
    """
    return json.loads(wf_request.content.decode())["workflow_request_id"]

def test_create_wf_request(wf_request):
    """
        Tests the correctness of the creating a new
        standard calibration and imaging workflow.
    """
    assert wf_request.status_code == 200

def test_add_ppr_to_wf_request(wf_request_id):
    """
        Tests the correctness of adding a PPR to the
        database for a standard calibration and imaging workflow.
    """
    url = f"{SERVICE_URL}/workflows/{WORKFLOW}/requests/{wf_request_id}/files/PPR.xml"
    if TESTING_IN_PIPELINE is False:
        with open("../std_calibration_and_imaging_PPR.xml", "r", encoding="utf8") as ppr_file:
            ppr_content = ppr_file.read()
        response = requests.put(url, data=ppr_content, timeout=TIMEOUT)
    else:
        response = mocked_requests(url)
    assert response.status_code == 200

def test_submit_wf_request(wf_request_id):
    """
        Tests the correctness of adding a PPR to the
        database for a standard calibration and imaging workflow.
    """
    url = f"{SERVICE_URL}/workflows/{WORKFLOW}/requests/{wf_request_id}/submit"
    if TESTING_IN_PIPELINE is False:
        response = requests.post(url, timeout=TIMEOUT)
    else:
        response = mocked_requests(url)

    assert response.status_code == 200
