#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
The entities in this module are responsible for creating the low-level connections to the AMQP service and routing
messages from Python-land into AMQP and back into Python.
"""


import logging
import time
from abc import ABC, ABCMeta, abstractmethod
from threading import Lock
from typing import Dict, List

import kombu.log
from kombu import Connection, Exchange, Message, Queue
from kombu.mixins import ConsumerMixin
from messaging.retry import retry
from pycapo import CapoConfig

logger = logging.getLogger()


kombu.log.setup_logging("DEBUG")


class MessageFormatException(Exception):
    pass


class AMQPServerInitializer:
    """
    Object responsible for creating a connection to the correct AMQP server and establishing exchanges and queues for
    that server
    """

    def __init__(self, service_name: str):
        self.service_name = service_name
        self.connection = self._setup_connection()

    @staticmethod
    def _setup_connection() -> Connection:
        """
        Setup connection to the Workspaces AMQP server

        :return: AMQP connection
        """
        config = CapoConfig().settings("edu.nrao.archive.configuration.AmqpServer")
        logger.debug("Setting up AMQP connection")

        return Connection(
            hostname=config.hostname,
            port=int(config.port),
            userid=config.username,
            password=config.password,
        )

    def create_exchange(self) -> Exchange:
        """
        Create AMQP topic exchange given a name

        :return: Newly created exchange
        """
        if self.service_name == "archive":
            return Exchange("archive.events", type="topic")

        return Exchange("workspaces", type="topic")

    def listen_to_exchange(self, exchange: str, exchange_type="topic", queue_name="", routing_key="#"):
        """
        Start listening for events from a certain exchange.

        :param exchange:    the name of the exchange
        :param exchange_type:  the type of exchange, defaults to "topic"
        :param queue_name:  if supplied, create a persistent queue with this name. Otherwise generate a temporary queue
        :param routing_key: supply a routing key pattern if you want to be choosy about which messages you catch
        """
        exchange = Exchange(exchange, type=exchange_type)
        queue = Queue(name=queue_name, exchange=exchange, routing_key=routing_key)
        queue.maybe_bind(self.connection)
        queue.declare()

    def create_queues(self, exchange: Exchange) -> List[Queue]:
        """
        Create correct queues for exchange given a service name and set them up for use
        TODO: Refactor to only create a "persistent queue for the user"

        :param exchange: Name of the AMQP exchange
        :return: List of newly created and set-up queues
        """
        q_name = self.service_name + "_queue"

        if self.service_name == "capability":
            cp_queue = Queue(q_name, exchange, routing_key=self.service_name)
            cp_wf_queue = Queue("cp_wf_queue", exchange, routing_key="workflow")

            queues = [cp_queue, cp_wf_queue]
        elif self.service_name == "workflow":
            queues = [Queue(q_name, exchange, routing_key=self.service_name)]
        elif self.service_name == "archive":
            # Exchange used to communicate with the AAT system
            # Queue that will accept ingestion-complete events
            archive_ingestion_queue = Queue(
                "ws_ingestion_complete_queue",
                exchange,
                routing_key="ingestion-complete.#",
            )
            # Queue that will accept URLs for active CARTA instances
            archive_ws_workflow_events_queue = Queue(
                "archive.ws-workflow.events",
                exchange,
                routing_key="ws-workflow.carta-instance-ready.#",
            )
            queues = [archive_ingestion_queue, archive_ws_workflow_events_queue]
        else:
            return []

        self.setup_queues(queues)
        return queues

    @retry(logger=logger)
    def setup_queues(self, queues: List[Queue]):
        """
        Perform setup for AMQP queues

        :param queues: List of queues to set up
        """
        for queue in queues:
            queue.maybe_bind(self.connection)
            queue.declare()


class MessageReceiverIF(ABC):
    """
    Abstract interface for a message-receiving object
    """

    @abstractmethod
    def receive_message(self, message):
        pass


class ReceiverIF(metaclass=ABCMeta):
    @abstractmethod
    def receive_message(self, message):
        """
        Callback invoked upon consuming of an AMQP message; sends consumed message to a MessageReceiver and
        acknowledges message's receipt

        :param message: AMQP message
        """


class SenderIF(metaclass=ABCMeta):
    @abstractmethod
    def send_message(self, **message: Dict):
        """
        Send message to AMQP server

        :param message: Body of AMQP message
        """


class MessageSender(SenderIF):
    """
    The MessageSender is responsible for:

    1. Maintaining a connection to the Workspaces AMQP server
    2. Sending messages to the server
    """

    def __init__(self, service_name: str):
        logger.debug("Creating MessageSender for service %s", service_name)
        server_initializer = AMQPServerInitializer(service_name)
        self.connection = server_initializer.connection
        self.exchange = server_initializer.create_exchange()
        self.queues = server_initializer.create_queues(self.exchange)
        self.sending_lock = Lock()
        self.producer = self.connection.Producer()

    def close(self):
        logger.debug("Closing connection in MessageSender")
        self.producer.close()
        self.connection.release()

    def send_message(self, **message: Dict):
        """
        Send message to AMQP server

        :param message: Body of AMQP message
        """
        with self.sending_lock:
            if "routing_key" not in message.keys():
                if "service" in message.keys():
                    message["routing_key"] = message["service"]
                else:
                    raise MessageFormatException("Message has no routing key. Keys are: " + ", ".join(message.keys()))

            self.producer.publish(
                message,
                routing_key=message["routing_key"],
                exchange=self.exchange,
                retry=True,
            )
        logger.debug(
            f"Sent message to exchange '{self.exchange.name}' with routing_key '{message['routing_key']}' and body '{message}'"
        )


class MessageReceiver(ReceiverIF, ConsumerMixin):
    """
    The MessageReceiver is responsible for:

    1. Maintaining a connection to the Workspaces AMQP server
    2. Consuming messages from the server and transferring them to a destination for further processing
    """

    def __init__(self, destination: MessageReceiverIF, service_name: str):
        logger.debug("Creating MessageReceiver for service %s", service_name)
        server_initializer = AMQPServerInitializer(service_name)
        self.connection = server_initializer.connection
        self.exchange = server_initializer.create_exchange()
        self.queues = server_initializer.create_queues(self.exchange)
        self.destination = destination

    def close(self):
        logger.debug("Closing connection in messenger")
        self.connection.release()

    def receive_message(self, message: Message):
        """
        Callback invoked upon consuming of an AMQP message; sends consumed message to a MessageReceiver and
        acknowledges message's receipt

        :param message: AMQP message
        """
        started = time.perf_counter()
        try:
            self.destination.receive_message(message.body)
        finally:
            elapsed = time.perf_counter() - started
            logger.debug("Message handled in %s seconds", elapsed)
            message.ack()

    def get_consumers(self, consumer, channel):
        """
        Implementation of abstract method from superclass ConsumerProducerMixin; returns a list of queues that this
        object will consume from
        """
        return [
            consumer(
                queues=queue,
                on_message=self.receive_message,
                prefetch_count=10,
            )
            for queue in self.queues
        ]

    def on_connection_revived(self):
        logger.warning("AMQP connection was revived")

    def on_decode_error(self, message, exc):
        logger.exception("Unable to decode message: %s", message, exc_info=exc)

    def on_connection_error(self, exc, interval):
        logger.exception("AMQP connection error, sleeping %s seconds", interval, exc_info=exc)
