#!/bin/bash
# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#

set -o errexit

ROOT_DIR="$1"
CASA_DIR="${ROOT_DIR}/home/casa/packages"

# Do nothing if legit CASA packages already exist
if [ -d "$CASA_DIR" ]; then
    exit 0
fi

# Sample recipes
STD_CAL_RECIPE=$(cat <<-END
<ProcessingProcedure>

    <ProcedureTitle>hifv_cal</ProcedureTitle>

    <ProcessingCommand>
        <Command xmlns="">hifv_importdata</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_hanning</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_flagdata</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">hm_tbuff</Keyword>
                <Value xmlns="">1.5int</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">fracspw</Keyword>
                <Value xmlns="">0.01</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">intents</Keyword>
                <Value xmlns="">*POINTING*,*FOCUS*,*ATMOSPHERE*,*SIDEBAND_RATIO*, *UNKNOWN*, *SYSTEM_CONFIGURATION*, *UNSPECIFIED#UNSPECIFIED*</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_vlasetjy</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_priorcals</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_syspower</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_testBPdcals</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_checkflag</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">checkflagmode</Keyword>
                <Value xmlns="">bpd-vla</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_semiFinalBPdcals</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_checkflag</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">checkflagmode</Keyword>
                <Value xmlns="">allcals-vla</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_solint</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_fluxboot</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_finalcals</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_applycals</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_checkflag</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">checkflagmode</Keyword>
                <Value xmlns="">target-vla</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_statwt</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_plotsummary</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hif_makeimlist</Command>
        <ParameterSet>
           <Parameter>
               <Keyword xmlns="">intent</Keyword>
               <Value xmlns="">PHASE,BANDPASS</Value>
           </Parameter>
           <Parameter>
               <Keyword xmlns="">specmode</Keyword>
               <Value xmlns="">cont</Value>
           </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hif_makeimages</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">hm_masking</Keyword>
                <Value xmlns="">centralregion</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_exportdata</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

</ProcessingProcedure>
END
)

RESTORE_VLA_RECIPE=$(cat <<-END
<ProcessingProcedure>

    <ProcedureTitle>hifv_deliver_ms</ProcedureTitle>

    <!-- import MS: if importonly=True, no calibration shall be performed -->
    {{#importonly}}
    <ProcessingCommand>
        <Command xmlns="">hifv_importdata</Command>
        <ParameterSet/>
    </ProcessingCommand>

    <!-- If onlineflag=True, online flag shall be applied to the imported MSes. -->
    {{#onlineflag}}
    <ProcessingCommand>
        <Command xmlns="">hifv_flagdata</Command>
        <ParameterSet/>
    </ProcessingCommand>
    {{/onlineflag}}
    {{/importonly}}
    {{^importonly}}
    <ProcessingCommand>
        <Command xmlns="">hifv_restoredata</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">copytoraw</Keyword>
                <Value xmlns="">False</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>
    {{/importonly}}

    <!-- if mstransform=True, data selection, time and/or spectral averaging shall be performed -->
    {{#mstransform}}
    <ProcessingCommand>
        <Command xmlns="">hif_mstransform</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">chanbin</Keyword>
                <Value xmlns="">{{chanbin}}</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">timebin</Keyword>
                <Value xmlns="">{{timebin}}</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">intent</Keyword>
                <Value xmlns="">{{intent}}</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">field</Keyword>
                <Value xmlns="">{{field}}</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">spw</Keyword>
                <Value xmlns="">{{spw}}</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>
    {{/mstransform}}

    <ProcessingCommand>
        <Command xmlns="">hifv_exportdata</Command>
        <ParameterSet/>
    </ProcessingCommand>

</ProcessingProcedure>
END
)

RESTORE_ALMA_RECIPE=$(cat <<-END
<ProcessingProcedure>

    <ProcedureTitle>hifa_deliver_ms</ProcedureTitle>

    <!-- import MS: if importonly=True, no calibration shall be performed -->
    {{#importonly}}
    <ProcessingCommand>
        <Command xmlns="">hifa_importdata</Command>
        <ParameterSet/>
    </ProcessingCommand>

    <!-- If onlineflag=True, online flag shall be applied to the imported MSes. -->
    {{#onlineflag}}
    <ProcessingCommand>
        <Command xmlns="">hifa_flagdata</Command>
        <ParameterSet/>
    </ProcessingCommand>
    {{/onlineflag}}
    {{/importonly}}
    {{^importonly}}
    <ProcessingCommand>
        <Command xmlns="">hifa_restoredata</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">copytoraw</Keyword>
                <Value xmlns="">False</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>
    {{/importonly}}

    <!-- if mstransform=True, data selection, time and/or spectral averaging shall be performed -->
    {{#mstransform}}
    <ProcessingCommand>
        <Command xmlns="">hif_mstransform</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">chanbin</Keyword>
                <Value xmlns="">{{chanbin}}</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">timebin</Keyword>
                <Value xmlns="">{{timebin}}</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">intent</Keyword>
                <Value xmlns="">{{intent}}</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">field</Keyword>
                <Value xmlns="">{{field}}</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">spw</Keyword>
                <Value xmlns="">{{spw}}</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>
    {{/mstransform}}

    <ProcessingCommand>
        <Command xmlns="">hifa_exportdata</Command>
        <ParameterSet/>
    </ProcessingCommand>

</ProcessingProcedure>
END
)

IMG_RECIPE=$(cat <<-END
<ProcessingProcedure>

    <ProcedureTitle>hifv_contimage_selfcal</ProcedureTitle>

    <ProcessingCommand>
        <Command xmlns="">hifv_importdata</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_flagtargetsdata</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hif_mstransform</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hif_checkproductsize</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">maximsize</Keyword>
                <Value xmlns="">16384</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hif_makeimlist</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">specmode</Keyword>
                <Value xmlns="">cont</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">datatype</Keyword>
                <Value xmlns="">regcal</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hif_makeimages</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">hm_cyclefactor</Keyword>
                <Value xmlns="">3.0</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hif_selfcal</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hif_makeimlist</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">specmode</Keyword>
                <Value xmlns="">cont</Value>
            </Parameter>
            <Parameter>
                <Keyword xmlns="">datatype</Keyword>
                <Value xmlns="">selfcal</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hif_makeimages</Command>
        <ParameterSet>
            <Parameter>
                <Keyword xmlns="">hm_cyclefactor</Keyword>
                <Value xmlns="">3.0</Value>
            </Parameter>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_pbcor</Command>
        <ParameterSet>
        </ParameterSet>
    </ProcessingCommand>

    <ProcessingCommand>
        <Command xmlns="">hifv_exportdata</Command>
        <ParameterSet>
          <Parameter>
            <Keyword xmlns="">imaging_products_only</Keyword>
            <Value xmlns="">True</Value>
          </Parameter>
        </ParameterSet>
    </ProcessingCommand>

</ProcessingProcedure>
END
)

# Separating steps for RHEL6 and RHEL7 so they can differ

# RHEL6
# Make RHEL release directories
CASA_RHEL_DIR="${CASA_DIR}/RHEL6/release"
SSA_CASA_RHEL_DIR="${ROOT_DIR}/home/ssa/casa/RHEL6/docker"
mkdir -p "${CASA_RHEL_DIR}" "${SSA_CASA_RHEL_DIR}"
cd "${CASA_RHEL_DIR}"

# Mock some CASA releases with pipeline directories
CASAS=("casa-6.1.1-13-pipeline-2020.1.0.36" "casa-pipeline-release-5.6.1-8.el6" "casa-release-5.5.0-149")
for CASA in "${CASAS[@]}"
do
    recipedir="${CASA}/pipeline/pipeline/recipes"
    mkdir -p "${recipedir}"

    # Populate some recipe files
    echo "$STD_CAL_RECIPE" | tee "${recipedir}/procedure_hifv.xml" > /dev/null
    echo "$RESTORE_VLA_RECIPE" | tee "${recipedir}/template_hifv_deliver_ms.xml" > /dev/null
    echo "$RESTORE_ALMA_RECIPE" | tee "${recipedir}/template_hifa_deliver_ms.xml" > /dev/null
    echo "$IMG_RECIPE" | tee "${recipedir}/procedure_hifv_contimage_selfcal.xml" > /dev/null

    # Symlink to where WS expects them to live
    ln -s "${CASA_RHEL_DIR}/${CASA}" "${SSA_CASA_RHEL_DIR}/${CASA}"
done


# RHEL7
# Make RHEL release directories
CASA_RHEL_DIR="${CASA_DIR}/RHEL7/release"
SSA_CASA_RHEL_DIR="${ROOT_DIR}/home/ssa/casa/RHEL7/docker"
mkdir -p "${CASA_RHEL_DIR}" "${SSA_CASA_RHEL_DIR}"
cd "${CASA_RHEL_DIR}"

# Mock some CASA releases with pipeline directories
CASAS=("casa-6.1.1-13-pipeline-2020.1.0.36" "casa-6.4.1-12-pipeline-2022.2.0.64" "casa-6.4.1-12-pipeline-2022.2.0.68" "casa-release-5.3.0-143" "casa-6.5.4-9-pipeline-2023.1.0.124")
for CASA in "${CASAS[@]}"
do
    recipedir="${CASA}/pipeline/pipeline/recipes"
    mkdir -p "${recipedir}"

    # Populate some recipe files
    echo "$STD_CAL_RECIPE" | tee "${recipedir}/procedure_hifv.xml" > /dev/null
    echo "$RESTORE_VLA_RECIPE" | tee "${recipedir}/template_hifv_deliver_ms.xml" > /dev/null
    echo "$RESTORE_ALMA_RECIPE" | tee "${recipedir}/template_hifa_deliver_ms.xml" > /dev/null
    echo "$IMG_RECIPE" | tee "${recipedir}/procedure_hifv_contimage_selfcal.xml" > /dev/null

    # Symlink to where WS expects them to live
    ln -s "${CASA_RHEL_DIR}/${CASA}" "${SSA_CASA_RHEL_DIR}/${CASA}"
done
