#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Testing capability actions
"""
# pylint: disable=C0301, E0401, E0402, E0611, W0621
import json
from typing import Dict, List, Optional
from unittest.mock import MagicMock, patch

import pytest
from hypothesis import given
from hypothesis import strategies as st

from workspaces.capability.enums import (
    CapabilityRequestState,
    CapabilityVersionState,
    QueueState,
)
from workspaces.capability.schema import (
    AnnounceQa,
    Capability,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityVersion,
    CapabilityVersionFile,
    QaPass,
    SendNotification,
)
from workspaces.capability.services.capability_info import CapabilityInfo

from ..conftest import (
    SAMPLE_CAPABILITY_NAMES,
    SQLITE_MAX_INT,
    SQLITE_MIN_INT,
    clear_test_database,
)

FAKE_WORKFLOW_URL = "http://fake-workflow.edu"
FAKE_ANALYST = "some-analyst"


# Register Capability JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(Capability))
# >>> def test(generated_capability_json: Dict):
# >>>   capability = Capability.from_json(generated_capability_json)
st.register_type_strategy(
    Capability,
    st.fixed_dictionaries(
        {
            "type": st.just("Capability"),
            "name": st.sampled_from(SAMPLE_CAPABILITY_NAMES),
            "max_jobs": st.integers(min_value=1, max_value=25),
            "enabled": st.booleans(),
            "has_image_products": st.booleans(),
            "paused": st.booleans(),
        }
    ),
)

# Register CapabilityRequest JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityRequest))
# >>> def test(generated_request_json: Dict):
# >>>   request = CapabilityRequest.from_json(generated_request_json)
st.register_type_strategy(
    CapabilityRequest,
    st.fixed_dictionaries(
        {
            "type": st.just("CapabilityRequest"),
            "id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "capability_name": st.sampled_from(SAMPLE_CAPABILITY_NAMES),
            "state": st.sampled_from([name for name, _ in CapabilityRequestState.__members__.items()]),
            "parameters": st.one_of(st.lists(st.text()), st.none()),
            "ingested": st.booleans(),
            "created_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
            "updated_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
        }
    ),
)

# Register CapabilityVersion JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityVersion))
# >>> def test(generated_version_json: Dict):
# >>>   version = CapabilityVersion.from_json(generated_version_json)
st.register_type_strategy(
    CapabilityVersion,
    st.fixed_dictionaries(
        {
            "type": st.just("CapabilityVersion"),
            "capability_request_id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "version_number": st.integers(min_value=0, max_value=50),
            "parameters": st.one_of(st.none(), st.fixed_dictionaries({"user_email": st.emails()})),
            "workflow_metadata": st.one_of(
                st.none(), st.dictionaries(st.text(min_size=1), st.text(min_size=1), min_size=1)
            ),
            "files": st.none(),
        }
    ),
)

# Register CapabilityVersionFile JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityVersionFile))
# >>> def test(generated_version_file_json: Dict):
# >>>   version_file = CapabilityVersionFile.from_json(generated_version_file_json)
st.register_type_strategy(
    CapabilityVersionFile,
    st.fixed_dictionaries({"filename": st.text(min_size=1), "content": st.binary(min_size=10).map(bytes.hex)}),
)

# Register CapabilityExecution JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityExecution))
# >>> def test(generated_execution_json: Dict):
# >>>   execution = CapabilityExecution.from_json(generated_request_json)
st.register_type_strategy(
    CapabilityExecution,
    st.fixed_dictionaries(
        {
            "type": st.just("CapabilityExecution"),
            "id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "state_name": st.sampled_from(["Start", "Queued", "Executing", "Foo", "Bar"]),
            "queue_state": st.sampled_from([name for name, _ in QueueState.__members__.items()]),
            "capability_request_id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "version_number": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "current_workflow_request_id": st.integers(min_value=1, max_value=15),
            "delivery_url": st.from_regex(r"https:\/\/dl-nrao-unittest\.aoc\.nrao\.edu(?:\/[A-z0-9]*)+"),
            "created_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
            "updated_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
        }
    ),
)


@pytest.mark.skip("Broken by version rework")
@given(
    st.one_of(st.lists(st.text()), st.none()),
    st.from_type(Capability),
    st.integers(min_value=2, max_value=10),
)
def test_qa_pass(
    mock_capability_info: CapabilityInfo,
    parameters: Optional[List[str]],
    capability_json: Dict,
    number_of_versions: int,
):
    """
    Confirm that QA Pass capability functions as intended.

    :param mock_capability_info:
    :param parameters:
    :param capability_json:
    :param number_of_versions:
    :return:
    """
    capability = Capability.from_json(capability_json)
    with clear_test_database(mock_capability_info):
        mock_capability_info.save_entity(capability)
        expected_request = mock_capability_info.create_capability_request(capability.name, parameters)

        # Create a number of new versions
        expected_versions = []
        for _ in range(number_of_versions):
            created_version = mock_capability_info.create_capability_version(expected_request.id, parameters)
            expected_versions.append(created_version)

        # Latest version is always the one to pass
        version_to_pass = expected_versions[-1]
        passing_execution = mock_capability_info.create_execution(expected_request)
        passing_execution.version = version_to_pass
        passing_execution.version_number = version_to_pass.version_number
        passing_execution.state_name = "Awaiting QA"
        version_to_fail = expected_versions[0]
        failing_execution = mock_capability_info.create_execution(expected_request)
        failing_execution.version = version_to_fail
        failing_execution.version_number = version_to_fail.version_number
        failing_execution.state_name = "Executing"
        failing_execution.current_workflow_request_id = -1

        assert len(expected_versions) == number_of_versions

        with patch("workspaces.capability.schema.requests.post") as mock_post:
            with patch("workspaces.capability.schema.CapoConfig") as mock_capo_config:
                mock_capo_config.return_value.settings.return_value.serviceUrl = FAKE_WORKFLOW_URL
                QaPass()(passing_execution)
                # Assert failing execution has its execution aborted
                mock_post.assert_called_with(
                    f"{FAKE_WORKFLOW_URL}/workflows/requests/{failing_execution.current_workflow_request_id}/abort"
                )

        # Assert all versions except passing one are failed and passing one is complete
        for expected_version in expected_versions:
            actual_version = mock_capability_info.lookup_version(expected_request.id, expected_version.version_number)
            if actual_version.version_number != version_to_pass.version_number:
                assert actual_version.state == CapabilityVersionState.Failed.name


@given(st.from_type(CapabilityExecution))
def test_announce_qa(execution_json: Dict):
    execution = CapabilityExecution.from_json(execution_json)

    with patch("workspaces.capability.schema.requests.post") as mock_post:
        with patch("workspaces.capability.schema.CapoConfig") as mock_capo_config:
            mock_capo_config.return_value.settings.return_value.serviceUrl = FAKE_WORKFLOW_URL
            AnnounceQa()(execution)
            mock_post.assert_called_with(
                f"{FAKE_WORKFLOW_URL}/workflows/requests/{execution.current_workflow_request_id}/qa/qa_ready",
                json={"capability_version": execution.version_number},
            )


@given(st.from_type(CapabilityExecution), st.from_type(CapabilityVersion))
def test_send_notification(execution, version):
    # manufacture some types
    execution = CapabilityExecution.from_json(execution)
    version = CapabilityVersion.from_json(version)
    execution.version = version
    version.request = CapabilityRequest()
    version.request.capability = Capability()
    version.request.capability.name = "test_capability"

    # fake out Capo a bit
    with patch("workspaces.capability.schema.CapoConfig") as fake_capo:
        fake_capo.return_value.settings.return_value.analystEmail = FAKE_ANALYST

        # first, let's try sending to QA
        action = SendNotification(arguments=json.dumps({"template": "template", "send_to_qa": True}))
        execution_manager = MagicMock()
        action(execution, execution_manager)
        assert execution_manager.notifier.send_email.call_args[0][0] == "template"
        assert execution_manager.notifier.send_email.call_args[0][1]["destination_email"] == FAKE_ANALYST

        # second, let's try sending to the user, if they have an email address
        execution_manager = MagicMock()
        action = SendNotification(arguments=json.dumps({"template": "template", "send_to_qa": False}))
        action(execution, execution_manager)
        if version.parameters and "user_email" in version.parameters:
            assert execution_manager.notifier.send_email.call_args[0][0] == "template"
            assert (
                execution_manager.notifier.send_email.call_args[0][1]["destination_email"]
                == version.parameters["user_email"]
            )
        else:
            assert not execution_manager.notifier.send_email.called

        # Third, let's try with a CC and a user
        execution_manager = MagicMock()
        action = SendNotification(
            arguments=json.dumps({"template": "template", "send_to_qa": False, "email_das": True})
        )
        action(execution, execution_manager)
        if version.parameters and "user_email" in version.parameters:
            assert execution_manager.notifier.send_email.call_args[0][0] == "template"
            assert (
                execution_manager.notifier.send_email.call_args[0][1]["destination_email"]
                == version.parameters["user_email"]
            )
            assert execution_manager.notifier.send_email.call_args[0][1]["cc_email"] == [FAKE_ANALYST]
        else:
            assert not execution_manager.notifier.send_email.called
