#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
from datetime import datetime
from unittest.mock import MagicMock, patch

from workspaces.system.schema import AbstractFile
from workspaces.workflow.schema import Workflow, WorkflowRequest
from workspaces.workflow.services.remote_processing_service import CapoInjector

from .test_workflow_info import FakeWorkflowInfo

"""
tests for remote_processing_service.CapoInjector
"""

expected_remote_subspace = {
    "edu.nrao.workspaces.WorkflowSettings.serviceUrl": "http://workflow:3456",
    "edu.nrao.workspaces.NotificationSettings.serviceUrl": "http://notification:3458",
    "edu.nrao.workspaces.UISettings.serviceUrl": "http://localhost:4444/workspaces",
    "edu.nrao.workspaces.ProcessingSettings.useCasa": "false",
    "edu.nrao.workspaces.ProcessingSettings.CasaVersion.vlass": "/home/casa/packages/pipeline/casa-6.1.3-3-pipeline-2021.1.1.32",
    "edu.nrao.archive.workflow.config.CasaVersions.homeForReprocessing": "/home/casa/packages/pipeline/current",
}

expected_remote_cv_subspace = {
    "edu.nrao.workspaces.WorkflowSettings.serviceUrl": "http://workflow:3456",
    "edu.nrao.workspaces.NotificationSettings.serviceUrl": "http://notification:3458",
    "edu.nrao.workspaces.UISettings.serviceUrl": "http://localhost:4444/workspaces",
    "edu.nrao.workspaces.ProductFetcherSettings.locatorServiceUrlPrefix": "https://webtest.aoc.nrao.edu/archive-service/location?locator",
    "edu.nrao.workspaces.ProductFetcherSettings.defaultThreadsPerHost": "1",
    "edu.nrao.workspaces.ProductFetcherSettings.executionSite": "local_test",
    "edu.nrao.workspaces.ProductFetcherSettings.metadataServiceUrlPrefix": "https://webtest.aoc.nrao.edu/archive-service/product/metadata",
    "almaMetadataDatabase.jdbcUsername": "docker",
    "almaMetadataDatabase.jdbcPassword": "docker",
    "almaMetadataDatabase.jdbcUrl": "alma",
    "edu.nrao.workspaces.ProcessingSettings.useCasa": "false",
    "edu.nrao.workspaces.ProcessingSettings.CasaVersion.vlass": "/home/casa/packages/pipeline/casa-6.1.3-3-pipeline-2021.1.1.32",
    "edu.nrao.archive.workflow.config.CasaVersions.homeForReprocessing": "/home/casa/packages/pipeline/current",
    "edu.nrao.workspaces.DeliverySettings.downloadDirectory": "/tmp/delivery_root",
    "edu.nrao.workspaces.DeliverySettings.downloadUrl": "http://localhost:4444/dl",
    "edu.nrao.workspaces.ProcessingSettings.rootDirectory": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool",
    "edu.nrao.workspaces.ProcessingSettings.scriptLocation": "/lustre/aoc/cluster/pipeline/docker/workspaces/sbin",
}

remote_request = WorkflowRequest(
    workflow_request_id=-3,
    workflow=Workflow(workflow_name="vlass_seci", uses_casa=True, remote_capable=True),
    workflow_name="vlass_seci",
    argument={"remote": True},
    state="Created",
    results_dir="/i/am/a/path",
    created_at=datetime(2021, 1, 1),
    updated_at=datetime(2021, 1, 1),
)

remote_cv_request = WorkflowRequest(
    workflow_request_id=-4,
    workflow=Workflow(workflow_name="restore_cms", uses_casa=True, remote_capable=True),
    workflow_name="restore_cms",
    argument={"remote-cv": True},
    state="Created",
    results_dir="/i/am/a/path",
    created_at=datetime(2021, 1, 1),
    updated_at=datetime(2021, 1, 1),
)


info = FakeWorkflowInfo()

remote_injector = CapoInjector(info, remote_request)
remote_cv_injector = CapoInjector(info, remote_cv_request)


class TestCapoInjector:
    def test_is_remote_workflow(self):
        check = remote_injector.is_remote_workflow()
        assert check is True

    def test_identify_subspace(self):
        view = remote_injector.identify_subspace()
        # Verifies keys match
        assert set(view) == set(expected_remote_subspace)

    def test_create_subspace(self):
        subspace = remote_injector.create_subspace()
        # Verifies values match
        assert set(subspace.values()) == set(expected_remote_subspace.values())

    def test_write_subspace(self):
        abs_file = remote_injector.write_subspace()
        assert isinstance(abs_file, AbstractFile)
        assert abs_file.filename == "docker.properties"
        assert (
            abs_file.content
            == """edu.nrao.workspaces.WorkflowSettings.serviceUrl = http://workflow:3456\nedu.nrao.workspaces.NotificationSettings.serviceUrl = http://notification:3458\nedu.nrao.workspaces.UISettings.serviceUrl = http://localhost:4444/workspaces\nedu.nrao.workspaces.ProcessingSettings.useCasa = false\nedu.nrao.workspaces.ProcessingSettings.CasaVersion.vlass = /home/casa/packages/pipeline/casa-6.1.3-3-pipeline-2021.1.1.32\nedu.nrao.archive.workflow.config.CasaVersions.homeForReprocessing = /home/casa/packages/pipeline/current""".encode()
        )

    def test_is_remote_cv_workflow(self):
        check = remote_cv_injector.is_remote_workflow()
        assert check is True

    def test_identify_cv_subspace(self):
        view = remote_cv_injector.identify_subspace()
        # Verifies keys match
        assert set(view) == set(expected_remote_cv_subspace)

    def test_create_cv_subspace(self):
        subspace = remote_cv_injector.create_subspace()
        # Verifies values match
        assert set(subspace.values()) == set(expected_remote_cv_subspace.values())

    def test_write_cv_subspace(self):
        abs_file = remote_cv_injector.write_subspace()
        assert isinstance(abs_file, AbstractFile)
        assert abs_file.filename == "docker.properties"
        assert (
            abs_file.content
            == """edu.nrao.workspaces.WorkflowSettings.serviceUrl = http://workflow:3456\nedu.nrao.workspaces.NotificationSettings.serviceUrl = http://notification:3458\nedu.nrao.workspaces.UISettings.serviceUrl = http://localhost:4444/workspaces\nedu.nrao.workspaces.ProcessingSettings.useCasa = false\nedu.nrao.workspaces.ProcessingSettings.CasaVersion.vlass = /home/casa/packages/pipeline/casa-6.1.3-3-pipeline-2021.1.1.32\nedu.nrao.archive.workflow.config.CasaVersions.homeForReprocessing = /home/casa/packages/pipeline/current\nedu.nrao.workspaces.ProductFetcherSettings.locatorServiceUrlPrefix = https://webtest.aoc.nrao.edu/archive-service/location?locator\nedu.nrao.workspaces.ProductFetcherSettings.defaultThreadsPerHost = 1\nedu.nrao.workspaces.ProductFetcherSettings.executionSite = local_test\nedu.nrao.workspaces.ProductFetcherSettings.metadataServiceUrlPrefix = https://webtest.aoc.nrao.edu/archive-service/product/metadata\nalmaMetadataDatabase.jdbcUsername = docker\nalmaMetadataDatabase.jdbcPassword = docker\nalmaMetadataDatabase.jdbcUrl = alma\nedu.nrao.workspaces.DeliverySettings.downloadDirectory = /tmp/delivery_root\nedu.nrao.workspaces.DeliverySettings.downloadUrl = http://localhost:4444/dl\nedu.nrao.workspaces.ProcessingSettings.rootDirectory = /lustre/aoc/cluster/pipeline/docker/workspaces/spool\nedu.nrao.workspaces.ProcessingSettings.scriptLocation = /lustre/aoc/cluster/pipeline/docker/workspaces/sbin""".encode()
        )

    @patch("pathlib.Path.write_bytes")
    def test_subspace_to_directory(self, mock_write):
        remote_injector.subspace_to_directory()
        assert mock_write.call_count == 1

    @patch("pathlib.Path.unlink")
    @patch("glob.glob")
    @patch("os.listdir")
    @patch("pathlib.Path.exists", MagicMock(return_value=True))
    def test_clear_subspace(self, mock_os, mock_glob, mock_remove):
        remote_injector.clear_subspace()
        assert mock_os.call_count == 1
        assert mock_glob.call_count == 1
