#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

import logging
from typing import Dict, List, Optional, Union

import transaction
from sqlalchemy.orm import Session, registry

from workspaces.capability.schema import (
    Action,
    DoubleQAPartialPassStateMachine,
    DoubleQAStateMachine,
    SimpleStateMachine,
    SimpleStateMachineWithAnalystSupport,
    SingleQAStateMachine,
    State,
    StateMachine,
    StateMachineRepr,
    Transition,
)

mapper_registry = registry()

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class StateMachineInfo:
    def __init__(self, session: Session):
        self.session = session

    def save_entity(self, entity: Union[State, Transition, Action]):
        """
        Persist state machine entity to the database, if it doesn't already exist

        :param entity: Entity to save
        """
        self.session.merge(entity)
        self.session.flush()

    def save_entities(self, entities: List[Union[State, Transition, Action]]):
        """
        Persist state machine entities to the database

        :param entities: Entities to save
        """
        self.session.add_all(entities)
        self.session.flush()

    def lookup_transition(
        self, capability_name: str, from_state: State, to_state: State, pattern: str
    ) -> Optional[Transition]:
        """
        Lookup capability state transition

        :param capability_name: Capability name of transition
        :param from_state: From state
        :param to_state: To state
        :param pattern: The transition's pattern
        :return: Found transition; or None if no transition is found
        """
        result = self.session.query(Transition).filter_by(
            capability_name=capability_name, from_state=from_state, to_state=to_state, pattern=pattern
        )
        return result.first()

    def persist_machine(self, machine: StateMachineRepr, metadata: Dict):
        """
        Persist state machine to database

        :param machine: State machine to persist
        :param metadata: Metadata for state machine
        """
        workflow_name = metadata["workflow_name"]

        for from_state_name, details in machine.items():
            for (to_state_name, message_type), actions in details.items():
                from_state = State(capability_name=workflow_name, state_name=from_state_name)
                to_state = State(capability_name=workflow_name, state_name=to_state_name)

                pattern = f"type == {message_type}"

                # Persist current transition
                current_transition = Transition(
                    capability_name=workflow_name,
                    from_state=from_state,
                    pattern=pattern,
                    to_state=to_state,
                )
                self.save_entity(current_transition)
                # Re-lookup transition to grab transition ID
                current_transition = self.lookup_transition(workflow_name, from_state, to_state, pattern)

                # Persist actions
                for action in actions:
                    action.transition_id = current_transition.transition_id
                    self.save_entity(action)

                transaction.commit()

    def init_state_machines(self):
        """
        Initialize database with state machine definitions
        """
        self._clear_state_machines()
        transaction.commit()

        machine_type_mapping = {
            "simple": SimpleStateMachine,
            "single_qa": SingleQAStateMachine,
            "double_qa": DoubleQAStateMachine,
            "simple_with_analyst_support": SimpleStateMachineWithAnalystSupport,
            "double_qa_partial_pass": DoubleQAPartialPassStateMachine,
        }
        state_machines: list[StateMachine] = self.session.query(StateMachine).all()
        for machine in state_machines:
            MachineType = machine_type_mapping.get(machine.machine_type)
            if MachineType is None:
                raise KeyError(
                    f"Expected state machine of type '{machine.machine_type}' from the database in the machine_type_mapping, found None"
                )
            new_machine = MachineType(
                capability_name=machine.capability_name,
                machine_type=machine.machine_type,
                associated_workflows=machine.associated_workflows,
            ).init_state_machine()
            logger.debug(
                "Setting up state machine type %s for workflows %s",
                machine.machine_type,
                machine.associated_workflows["workflow_name"],
            )
            self.persist_machine(new_machine, machine.associated_workflows)

    def _clear_state_machines(self):
        """
        Clear old state machine records from the database
        """
        self.session.query(Action).delete()
        reset_actions_counter_sql = "ALTER SEQUENCE capability_state_actions_action_id_seq RESTART WITH 1"
        self.session.execute(reset_actions_counter_sql)
        self.session.query(Transition).delete()
        reset_transitions_counter_sql = "ALTER SEQUENCE capability_state_transitions_transition_id_seq RESTART WITH 1"
        self.session.execute(reset_transitions_counter_sql)
        self.session.flush()
