#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

""" Classes for notification """

# pylint: disable=C0103, E0401, R0903, W1201

import email
import logging
import smtplib
from typing import Dict

import requests
from email_validator import EmailNotValidError, validate_email
from pycapo import CapoConfig

from workspaces.notification.services.interfaces import (
    NotificationInfoIF,
    NotificationServiceIF,
)

logger = logging.getLogger(__name__)
NOTIF_SETTINGS_KEY = "edu.nrao.workspaces.NotificationSettings"


class NotificationServiceRESTClient(NotificationServiceIF):
    """REST endpoint for NotificationService"""

    def __init__(self):
        self.url = CapoConfig().settings(NOTIF_SETTINGS_KEY).serviceUrl
        self.linkUrl = CapoConfig().settings("edu.nrao.workspaces.UISettings").serviceUrl
        self.sendPI = CapoConfig().getboolean("edu.nrao.workspaces.NotificationSettings.sendPIEmail")

    def send_email(self, notification: str, arguments: Dict):
        requests.post(f"{self.url}/notify/{notification}/send", json=arguments).raise_for_status()


class NotificationService(NotificationServiceIF):
    """Email notifier"""

    def __init__(self, info: NotificationInfoIF):
        self.info = info

    def send_email(self, template_name, parameters: dict):
        """
        Sends an email using parameters provided

        :param template_name: the email template to use
        :param parameters: any information needed to send the email
        :return:
        """
        smtp_server = "smtp.aoc.nrao.edu"
        sender_email = "do-not-reply@nrao.edu"

        if not parameters["destination_email"]:
            return "destination_email missing from parameters."

        # This gets picked up from whatever's set in the 'user_email' parameter in the capability version parameters
        if isinstance(parameters["destination_email"], list):
            receiver_emails = parameters["destination_email"]
        else:
            # sometimes the list gets passed in as a string, split it if needed and strip extra spaces
            receiver_emails = [address.strip() for address in parameters["destination_email"].split(",")]

        # Let's not send emails into the void
        if not receiver_emails:
            return "No PI email found."

        # verify email addresses are valid. Drop invalid addresses
        for address in receiver_emails:
            try:
                validate_email(address)
            except EmailNotValidError as e:
                logger.info(f"Invalid email address {address} found, removing from receivers list.")
                receiver_emails.remove(address)

        if len(receiver_emails) > 0:
            logger.info(f">> In send_email, sending email to: {receiver_emails}")
            logger.info(f">> The parameters are: {parameters}")
            template = self.info.lookup_template(template_name)

            if template is None:
                logger.info("No template '" + template_name + "' found.")

            file = template.render(parameters)

            # Load the headers and email content from the template
            msg = email.message_from_bytes(file.content)

            # Attach the 'from', 'to', and 'cc' headers if they weren't set in the template
            if msg["From"] is None:
                msg["From"] = sender_email

            if msg["To"] is None:
                msg["To"] = ", ".join(receiver_emails)

            if msg["Cc"] is None and "cc_email" in parameters and parameters["cc_email"] != [None]:
                if isinstance(parameters["cc_email"], list):
                    msg["Cc"] = ", ".join(parameters["cc_email"])
                else:
                    msg["Cc"] = parameters["cc_email"]

            with smtplib.SMTP(smtp_server) as server:
                r = server.send_message(msg)

            return r
        else:
            logger.info("No valid email addresses provided. Skipping sending mail.")
