# CASA Version Matrix Service
The CASA version matrix is responsible for returning valid CASA + pipeline versions for processing as well as
maintaining our lists of installed and allowed versions for various capabilities.

## Getting valid versions
Each capability has a list of CASA versions that are tested and verified for use. These allowed versions are
cross-referenced against the installed versions in `/home/ssa/casa/` to get a full list of versions that are allowed for
processing with the given capability.

Examples of using the service to get a single version or a list of versions (the version name(s) and full path(s) are
in the response body):
```python
import requests
prefix = 'http://capability:3457/casa_matrix/'

# Single version (telescope optional), gives the preferred version from CAPO
r = requests.get(prefix + 'version', json={'capability': 'restore_cms', 'telescope': 'alma'})
print(r.json())
# {'version': '6.4.1-12|2022.2.0.64', 'path': '/home/ssa/casa/RHEL7/docker/casa-6.4.1-12-pipeline-2022.2.0.64'}

# Single version, returns the given version if it's valid
r = requests.get(prefix + 'version', json={'version': '6.4.1-12|2022.2.0.68'})
print(r.json())
# {'version': '6.4.1-12|2022.2.0.68', 'path': '/home/ssa/casa/RHEL7/docker/casa-6.4.1-12-pipeline-2022.2.0.68'}

# List of versions with the preferred version first in the list, multiple capabilities can be given
r = requests.get(prefix + 'versions', json={'capabilities': ['restore_cms']})
print(r.json())
# {'restore_cms': [{'version': '6.4.1-12|2022.2.0.64', 'path': '/home/ssa/casa/RHEL7/docker/casa-6.4.1-12-pipeline-2022.2.0.64'}, {'version': '6.4.1-12|2022.2.0.68', 'path': '/home/ssa/casa/RHEL7/docker/casa-6.4.1-12-pipeline-2022.2.0.68'}, {'version': '5.6.1-8|default', 'path': '/home/ssa/casa/RHEL6/docker/casa-pipeline-release-5.6.1-8.el6'}]}
```

## Getting CASA recipes
Each capability has a `casa_recipe` field pointing to a recipe XML file in a CASA installation. This service can be used
to return the contents of the recipe file for the correct CASA version.

Note, to override searching for a valid CASA version, pass the full path to a CASA version in the `version_path`
parameter.

Example:
```python
import requests
prefix = 'http://capability:3457/casa_matrix/'

# Returns the contents of template_hifv_deliver_ms.xml for the default CASA version
r = requests.get(prefix + 'recipe', json={'capability': 'restore_cms'})
print(r.json())
# <ProcessingProcedure>...</ProcessingProcedure>

# Returns the contents of template_hifa_deliver_ms.xml for the given CASA version if it's valid
r = requests.get(prefix + 'recipe', json={'version': '6.4.1-12|2022.2.0.68', 'capability': 'restore_cms', 'telescope': 'alma'})
print(r.json())
# <ProcessingProcedure>...</ProcessingProcedure>

# Returns the contents of procedure_hifv.xml for the given CASA version, regardless if it's valid
r = requests.get(prefix + 'recipe', json={'capability': 'std_calibration', 'version_path': '/home/casa/packages/RHEL7/release/casa-pipeline-release-5.6.1-8.el7'})
print(r.json())
# <ProcessingProcedure>...</ProcessingProcedure>
```


## Refreshing installed versions
We maintain a list of symlinks to usable CASA + pipeline versions to filter out CASA packages that do not have a
corresponding pipeline. Our symlinks are in `/home/ssa/casa/` and link to CASA's managed installations in
`/home/casa/packages`. It is recommended to point all WS activities (e.g. workflows, capabilities) to our managed
symlinks.

Example of using the service to refresh the installed version links:
```python
import requests
requests.post('http://capability:3457/casa_matrix/links')
```

## Managing allowed versions
Each capability has a list of CASA versions that are allowed for processing. These are maintained in the `casa_matrix_*`
DB tables.

Examples of using the service to add/update/delete versions from the tables:

```python
import requests
url = 'http://capability:3457/casa_matrix/db/version'

# Add
requests.post(url, json={'version': '2.2.2', 'is_cluster_compatible': False, 'capabilities': ['curator', 'download']})

# Update
requests.put(url, json={'version': '2.2.2', 'is_cluster_compatible': True, 'capabilities': ['restore_cms', 'std_cms_imaging']})

# Delete
requests.delete(url, json={'version': '2.2.2'})
```
