#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import json
import logging

import requests
from lxml import etree

from pycapo import CapoConfig

logger = logging.getLogger(__name__)


# TODO: implement PPR generator
def add_ppr_parameter_tags(xml_to_change, command_to_params):
    """
    Adds parameters to the Parameter Set block for a casa recipe. Parameter blocks have the following structure:
        <ParameterSet>
            <Parameter>
                <Keyword xmnls="">parameter_name</Keyword>
                <Value xmnls="">paramater_value</Value>
            </Parameter>
        </ParameterSet>

    :param xml_to_change: The XML string to add parameter tags to
    :param command_to_params: A dictionary pairing Command names to a set of parameter names and values
    :return: the passed in xml string plus the additional parameter tags
    """
    root = etree.fromstring(xml_to_change)
    for command_name in command_to_params:
        # Find the <Command> tag with the appropriate command text, then grab the <ParameterSet> tag
        command_element = root.xpath(f'.//Command[text()="{command_name}"]')[0]
        param_set_element = command_element.getnext()

        # Add each parameter specified in the input
        command_param_dict = command_to_params[command_name]
        for param_to_add in command_param_dict.keys():
            logger.info(f"Adding parameter {param_to_add} with value {command_param_dict[param_to_add]}")

            # Create the <Parameter> tag
            param_element = etree.SubElement(param_set_element, "Parameter")

            # Add the <Keyword> tag
            param_name_element = etree.SubElement(param_element, "Keyword")
            param_name_element.text = param_to_add
            param_name_element.attrib["xmlns"] = ""

            # Add the <Value> tag
            param_value_element = etree.SubElement(param_element, "Value")
            param_value_element.text = command_param_dict[param_to_add]
            param_value_element.attrib["xmlns"] = ""

    # Output the finalized XML string, ensuring it is encoded correctly as a string
    etree.indent(root)
    return etree.tostring(root, encoding=str)


class CasaMatrixClient:
    """
    Methods for interacting with the CASA version matrix.
    """

    @staticmethod
    def casa_version(capability: str, version: str | None = None, telescope: str | None = None) -> dict[str, str, str]:
        """
        Fetches a valid CASA version from the CASA matrix. Will return the provided one if it's valid.

        :param capability: Capability name to filter versions
        :param version: Desired version
        :param telescope: Telescope to filter versions
        :return: A dict containing the version name and full path
        """
        ret_version = {}
        matrix_url = (
            CapoConfig().get("edu.nrao.workspaces.CapabilitySettings.externalServiceUrl") + "/casa_matrix/version"
        )

        params = {"capability": capability, "version": version, "telescope": telescope}

        response = requests.get(matrix_url, json=params)

        if response.ok:
            ret_version = json.loads(response.text)
        else:
            logger.critical(f"Unable to fetch a CASA version from the metrix service with params: {params}")

        return ret_version

    @staticmethod
    def casa_recipe(
        capability: str,
        version: str | None = None,
        telescope: str | None = None,
        version_path: str | None = None,
        params_to_add: dict = None,
    ) -> dict[str, str, str]:
        """
        Fetches a CASA recipe that can be injected into a PPR template. Will use the provided version's recipe if it's
        valid.

        :param capability: Capability name to filter versions
        :param version: Desired version
        :param telescope: Telescope to filter versions
        :param version_path: Full path to a CASA version, providing this overrides the matrix version filtering
        :param params_to_add: Dictionary pairing Command names to a set of parameter names and values
        :return: A dict containing the version's recipe, name, and full path
        """
        recipe = {}
        matrix_url = (
            CapoConfig().get("edu.nrao.workspaces.CapabilitySettings.externalServiceUrl") + "/casa_matrix/recipe"
        )

        params = {"capability": capability, "version": version, "telescope": telescope, "version_path": version_path}

        if params_to_add:
            params["params_to_add"] = params_to_add

        response = requests.get(matrix_url, json=params)

        if response.ok:
            recipe = json.loads(response.text)
        else:
            logger.critical(f"Unable to fetch a CASA recipe from the metrix service with params: {params}")

        return recipe
