#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import datetime
import json
import logging
from typing import Callable, Dict, List

import arrow
import chevron
import sqlalchemy as sa
from sqlalchemy.orm import registry, relationship

from workspaces.system.schema import (
    AbstractFile,
    AbstractRequest,
    JSONSerializable,
    mapper_registry,
)

logger = logging.getLogger(__name__)


@mapper_registry.mapped
class Workflow(JSONSerializable):
    """
    A Workflow is a suite of tasks defined by templates that must be
    executed together with some user-submitted files.
    """

    workflow_name: str
    uses_casa: bool
    requires_lustre: bool
    remote_capable: bool
    external_wf: bool
    templates: List["WorkflowTemplate"]
    requests: List["WorkflowRequest"]

    __tablename__ = "workflows"
    workflow_name = sa.Column("workflow_name", sa.String, primary_key=True)
    uses_casa = sa.Column("uses_casa", sa.Boolean, default=False, server_default="false", nullable=False)
    requires_lustre = sa.Column(
        "requires_lustre",
        sa.Boolean,
        server_default="False",
        nullable=False,
        comment="does this entire workflow require lustre access? (downloads, ingestion, etc.)",
    )
    remote_capable = sa.Column(
        "remote_capable",
        sa.Boolean,
        server_default="True",
        nullable=False,
        comment="can this workflow (or part of this workflow) run remote?",
    )
    external_wf = sa.Column(
        "external_wf",
        sa.Boolean,
        server_default="False",
        nullable=False,
        comment="is this workflow initiated outside of Workspaces or does it need to report "
        "to another system? (Archive, VLASS, etc.)",
    )
    templates = relationship("WorkflowTemplate", backref="workflows", lazy="dynamic")
    requests = relationship("WorkflowRequest", back_populates="workflow")

    def __repr__(self):
        return f"<Workflow workflow_name={self.workflow_name}>"

    def render_templates(self, argument: Dict) -> List[AbstractFile]:
        """
        Render the templates associated with this workflow

        :param argument: the workflow argument JSON
        :return:         a list of rendered template files
        """
        rendered_files = []

        """

        Original:
          - All the user's input files are collected at submission time
          - This method combines user input files with rendered template results
          - They wind up mixed together in the list on this request

        New:
          - User's input files are collected prior to submission
          - They already exist in the list of files on this request
          - At submission time, we only need to render the templates and append them to the request
        """
        for template in self.templates:
            rendered_files.append(template.render(argument))

        return rendered_files

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "workflow_name": self.workflow_name,
            "uses_casa": self.uses_casa,
            "requires_lustre": self.requires_lustre,
            "remote_capable": self.remote_capable,
            "external_wf": self.external_wf,
            "templates": list(self.templates),  # <- the list() here is necessitated by lazy loading
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        return Workflow(
            json["workflow_name"],
            json["uses_casa"],
            json["requires_lustre"],
            json["remote_capable"],
            json["external_wf"],
            json["templates"],
        )


@mapper_registry.mapped
class WorkflowTemplate(JSONSerializable):
    """
    Workflow Templates are Mustache-formatted files associated with a given Workflow
    """

    filename: str
    content: bytes

    __tablename__ = "workflow_templates"
    filename = sa.Column("filename", sa.String, primary_key=True)
    content = sa.Column("content", sa.LargeBinary, nullable=False)
    workflow_name = sa.Column(
        "workflow_name",
        sa.String,
        sa.ForeignKey("workflows.workflow_name"),
        primary_key=True,
    )

    def render(self, argument: Dict) -> AbstractFile:
        """
        Render a given Mustache-formatted Workflow Template with the specified arguments.

        :param argument:    argument(s) needed for template render
        :return:            an AbstractFile rendered from mustache template
        """
        # add some rendering helpers to the argument
        render_helpers = {"make_condor_jobname": self.make_condor_jobname}
        if "splits" in argument:
            argument["splits_str"] = json.dumps(argument["splits"])

        argument = {**render_helpers, **argument}

        # render
        contents = chevron.render(self.content.decode("utf8"), argument)
        logger.debug(
            "rendering %s with argument %s to %s",
            self.content.decode("utf8"),
            argument,
            contents,
        )
        return AbstractFile(self.filename, contents.encode("utf8"))

    @property
    def template_file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @template_file.setter
    def template_file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        # Defer to the actual template file's contents for JSON conversion
        return {
            "filename": self.filename,
            "content": self.content.decode("utf8"),
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        return WorkflowTemplate(json["filename"], json["content"])

    @staticmethod
    def make_condor_jobname(content: str, renderer: Callable[[str], str]):
        # get the rendered content
        rendered = renderer(content)
        return rendered.replace(".", "-").replace("+", "-")

    def __repr__(self):
        return f"<WorkflowTemplate filename={self.filename} for workflow={self.workflow_name}>"


@mapper_registry.mapped
class WorkflowRequest(JSONSerializable):
    """
    Workflow Requests invoke a particular workflow with a particular argument and some user-supplied files.
    """

    workflow_request_id: int
    workflow: Workflow
    workflow_name: str
    argument: dict  # JSON-formatted
    num_to_clean: int
    state: str
    results_dir: str
    files: List["WorkflowRequestFile"]
    job_id: str
    __tablename__ = "workflow_requests"
    workflow_request_id = sa.Column("workflow_request_id", sa.Integer, primary_key=True, autoincrement=True)
    htcondor_job_id = sa.Column("htcondor_job_id", sa.Integer, default=None, server_default="null")
    workflow_name = sa.Column("workflow_name", sa.String, sa.ForeignKey("workflows.workflow_name"))
    argument = sa.Column("argument", sa.JSON)
    state = sa.Column("state", sa.String)
    results_dir = sa.Column("results_dir", sa.String)
    files = relationship("WorkflowRequestFile", backref="request")
    workflow = relationship("Workflow", back_populates="requests")
    progress = relationship("WorkflowProgress", backref="request")
    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
    )
    updated_at = sa.Column(
        "updated_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        server_onupdate=sa.func.now(),
        nullable=False,
    )
    num_to_clean = sa.Column(
        "num_to_clean",
        sa.Integer,
        server_default="0",
        comment="Number of areas to annihilate to be considered cleaned",
        nullable=False,
    )
    num_cleaned = sa.Column(
        "num_cleaned",
        sa.Integer,
        server_default="0",
        comment="Number of areas annihilated",
        nullable=False,
    )
    # this property should not ever be altered by code as it's a calculated column, however it needs to be
    # accessible for annihilator queries
    cleaned = sa.Column(
        "cleaned",
        sa.Boolean,
        sa.Computed("num_to_clean <= num_cleaned"),
        comment="Has this request been cleaned from lustre?",
        nullable=False,
    )
    htcondor_iterations = sa.Column(
        "htcondor_iterations",
        sa.Integer,
        default=0,
        nullable=False,
    )
    controller = sa.Column(
        "controller",
        sa.String,
        sa.ForeignKey("processing_controllers.control_system"),
        comment="the system that initiated the workflow request, defaults to Workspaces (WS)",
    )

    @property
    def created_at_ago(self) -> str:
        return arrow.get(self.created_at).humanize()

    @property
    def updated_at_ago(self) -> str:
        return arrow.get(self.updated_at).humanize()

    @property
    def request(self) -> AbstractRequest:
        return AbstractRequest(self.workflow_name, self.argument)

    def update_status(self, status: str):
        self.state = status

    def add_num_to_clean(self, num_to_clean: int = 1):
        self.num_to_clean += num_to_clean

    def update_num_cleaned(self, num_cleaned: int):
        self.num_cleaned = num_cleaned

    def __getitem__(self, item):
        return self.files

    def __repr__(self):
        return f"<WorkflowRequest workflow_request_id={self.workflow_request_id}>"

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "workflow_request_id": self.workflow_request_id,
            "workflow_name": self.workflow_name,
            "argument": self.argument,
            "state": self.state,
            "results_dir": self.results_dir,
            "files": self.files,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
            "num_to_clean": self.num_to_clean,
            "num_cleaned": self.num_cleaned,
            "controller": self.controller,
            "progress": self.progress,
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        # clean up the dictionary a bit
        copy = dict(json)

        # type will confuse the constructor
        del copy["type"]

        # fix the types of the datetimes
        copy["created_at"] = datetime.datetime.fromisoformat(copy["created_at"])
        copy["updated_at"] = datetime.datetime.fromisoformat(copy["updated_at"])

        # now return an instance
        return WorkflowRequest(**copy)


@mapper_registry.mapped
class WorkflowRequestFile(JSONSerializable):
    """
    A Workflow Request File is a file supplied by the user and attached to the request they have submitted.
    """

    filename: str
    content: bytes

    def abstract_file(self):
        return AbstractFile(self.filename, self.content)

    __tablename__ = "workflow_request_files"
    workflow_request_id = sa.Column(
        "workflow_request_id",
        sa.Integer,
        sa.ForeignKey("workflow_requests.workflow_request_id"),
        primary_key=True,
    )
    filename = sa.Column("filename", sa.String, primary_key=True)
    content = sa.Column("content", sa.LargeBinary, nullable=False)

    @property
    def file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @file.setter
    def file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __repr__(self):
        return f"<WorkflowRequestFile filename={self.filename}>"

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "workflow_request_id": self.workflow_request_id,
            "filename": self.filename,
            "content": self.content.hex(),
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        return WorkflowRequestFile(json["filename"], bytes.fromhex(json["content"]))


@mapper_registry.mapped
class WorkflowProgress(JSONSerializable):
    """
    Track workflow progress through stage start and end times
    """

    __tablename__ = "workflow_progress"
    workflow_request_id = sa.Column(
        "workflow_request_id",
        sa.Integer,
        sa.ForeignKey("workflow_requests.workflow_request_id"),
        primary_key=True,
        comment="The workflow request of interest",
    )
    stage_name = sa.Column(
        "stage_name",
        sa.String,
        nullable=False,
        primary_key=True,
        comment="The workflow stage to track progress for",
    )
    start = sa.Column(
        "start_time",
        sa.DateTime(timezone=True),
        comment="The start time of this stage",
    )
    end = sa.Column(
        "end_time",
        sa.DateTime(timezone=True),
        comment="The end time of this stage",
    )

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "workflow_request_id": self.workflow_request_id,
            "stage_name": self.stage_name,
            "start": self.start.isoformat(),
            "end": self.end.isoformat() if self.end is not None else None,
        }


@mapper_registry.mapped
class ProcessingSite(JSONSerializable):
    """
    A processing site is a location that Workflow Requests can be submitted to for execution
    """

    processing_site: str
    template_key: str
    data_center: bool

    __tablename__ = "processing_sites"
    processing_site = sa.Column(
        "processing_site",
        sa.String,
        primary_key=True,
        comment="the abbreviation of the processing site (NMT, NAASC, etc.)",
    )
    template_key = sa.Column(
        "template_key",
        sa.String,
        nullable=True,
        comment="the key used by WS template rendering to ensure correct site setup",
    )
    data_center = sa.Column(
        "data_center",
        sa.Boolean,
        server_default="False",
        nullable=False,
        comment="is this site also a data retrieval site? (DSOC, NAASC, etc.)",
    )

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "processing_site": self.processing_site,
            "template_key": self.template_key,
            "data_center": self.data_center,
        }
