#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Workspaces Workflow Service - Workflow Recovery Facility
"""
import logging
import os
from pathlib import Path
from typing import List

import arrow
from sqlalchemy import text

from workspaces.workflow.schema import WorkflowRequest
from workspaces.workflow.services.interfaces import WorkflowInfoIF
from workspaces.workflow.services.monitor import WorkflowMonitor

logger = logging.getLogger(__name__)


def _ensure_initial_startup() -> bool:
    """
    Docker containers are recreated whenever software is deployed or containers are restarted.
    Check if initialization file is present in container to ensure that workflow reacquisition
    only runs on container restart.

    :return: True if file doesn't exist, False otherwise
    """

    ensure_initial_file = "reacquire.txt"

    if Path(ensure_initial_file).exists():
        return False

    content = f"The Workflow Reacquire system was started within this container at {arrow.utcnow()}\n"

    eif = open(ensure_initial_file, "w")
    eif.write(content)
    eif.close()

    logger.info("Workflow reacquisition beginning...")
    return True


def get_environment() -> str:
    """
    Determine which environment the system is running on. Important for making sure only
    requests belonging to the current environment reacquire monitor processes.

    :return: string representing the environment - i.e. dev, test, or prod
    """
    capo_profile = os.environ["CAPO_PROFILE"]

    if capo_profile != "docker":
        environment = capo_profile.split("-")
        return environment[1]

    return capo_profile


class MonitorRecover:
    def __init__(self, info: WorkflowInfoIF):
        self.info = info

    def _on_restart(self) -> List[WorkflowRequest]:
        """
        Mine workflow_requests table for in-flight requests

        :return: List of in-flight requests eligible for reacquisition
        """
        inflights = []
        profile = get_environment()

        for row in (
            self.info.session.get_bind()
            .execute(
                text(
                    f"""
                SELECT workflow_request_id as wf_id
                FROM workflow_requests
                WHERE state not in ('Created', 'Complete', 'Failed', 'Error')
                AND results_dir like ('%{profile}%')
                AND cleaned = 'False'
                """
                )
            )
            .fetchall()
        ):
            inflights.append(self.info.lookup_workflow_request(request_id=row[0]))

        return inflights

    def _reacquire_logfile(self, wf_request: WorkflowRequest) -> Path:
        """
        For an In-flight WorkflowRequest, reacquire the condor log file

        :return: Path to the HTCondor logfile
        """
        results_dir = Path(wf_request.results_dir)

        if results_dir.is_dir():
            if results_dir.resolve() == Path.cwd():
                logger.info(f"Results directory for Request #{wf_request.workflow_request_id} not found. Skipping.")
                return Path.cwd()
            else:
                logger.info(f"Searching for log file for Request #{wf_request.workflow_request_id}...")
                # Start by looking for a dag log.
                # Most workflows are DAGs and monitoring will terminate prematurely if condor.log is used
                dag_log = list(results_dir.glob("*.dag.dagman.log"))
                if dag_log:
                    logger.info("Found dagman logfile!")
                    return dag_log[0]

                # No dag log was found, fall back to condor.log
                condor_log = list(results_dir.glob("condor.log"))
                if condor_log:
                    logger.info("Found condor logfile!")
                    return condor_log[0]
                else:
                    # No usable log was found
                    logger.info(f"No log file found for Request #{wf_request.workflow_request_id}. Skipping.")
                    return Path.cwd()

        else:
            logger.info(
                f"Results directory {results_dir} for request #{wf_request.workflow_request_id} "
                f"is not a directory! Cleaning request record."
            )
            # directory has already been removed, clear id
            wf_request.htcondor_job_id = None
            # trigger the cleaned flag to update
            wf_request.update_num_cleaned(1)
            self.info.save_request(wf_request)
            # signal not a directory with current location
            return Path.cwd()

    @staticmethod
    def _restart_monitor(logfile: Path, wf_request: WorkflowRequest):
        """
        Restart a wf_monitor process for the provided WorkflowRequest

        :param logfile: The HTCondor logfile to monitor
        :param wf_request: The WorkflowRequest to update
        :return:
        """
        if logfile is None or logfile.resolve() == Path.cwd():
            return

        logger.info("Running wf_monitor.")
        is_dag = True if str(logfile).endswith(".dagman.log") else False

        WorkflowMonitor(logfile, wf_request, is_dag).start()
        logger.info(f"Successfully reacquired monitor for request #{wf_request.workflow_request_id}. Continuing...")

    def reacquire_inflight_requests(self):
        """
        Run reacquire for all in-flight requests

        :return:
        """

        if _ensure_initial_startup():
            inflights = self._on_restart()
            if len(inflights) == 0:
                logger.info("No in-flight workflows found, skipping workflow reacquisition.")

            for request in inflights:
                logger.info(f"Reacquiring monitor for Request #{request.workflow_request_id}...")
                self._restart_monitor(self._reacquire_logfile(request), request)

            logger.info("Workflow reacquisition complete.")
        else:
            logger.info("Not initial container startup, skipping workflow reacquisition.")
